# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Site formerly occupied by region of interest (ROI) module

We removed this module because it needed some work and was in a place obvious
enough to be confusing.

The last commit where you can find the full module ready for more work is
``0831bf1``

You might want to have a look at :mod:`nipy.labs.spatial_models`. Please see
:mod:``nipy.labs.spatial_models.discrete_domain`,
:mod:``nipy.labs.spatial_models.mroi` and the example in
``examples/labs/need_data/demo_roi.py``.
"""
