# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import sys
sys.path.insert(0, "..")

import model

'''
The following line does not match the ``model`` module - is it outdated?
'''
#model.run()
## from io import data_path

## study = Study(root=io.data_path)
## subject = Subject(subj, study=study)
## runmodel = Run(subject, run)


                   
