# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import nipy.algorithms.statistics.tests.test_rft
import nipy.algorithms.statistics.tests.test_intrinsic_volumes
