# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
3D plotting of neuroimaging volumes.
"""
from .viz_tools.maps_3d import affine_img_src, plot_map_3d, plot_anat_3d,\
    demo_plot_map_3d

