# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Package contains generic functions for data input/output. This includes
methods for accessing file systems and network resources.
"""
__docformat__ = 'restructuredtext'
