/* Generated by Cython 0.17.4 on Sat Jan 12 17:27:31 2013 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02040000
    #error Cython requires Python 2.4+.
#else
#include <stddef.h> /* For offsetof */
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if PY_VERSION_HEX < 0x02050000
  typedef int Py_ssize_t;
  #define PY_SSIZE_T_MAX INT_MAX
  #define PY_SSIZE_T_MIN INT_MIN
  #define PY_FORMAT_SIZE_T ""
  #define CYTHON_FORMAT_SSIZE_T ""
  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)
  #define PyInt_AsSsize_t(o)   __Pyx_PyInt_AsInt(o)
  #define PyNumber_Index(o)    ((PyNumber_Check(o) && !PyFloat_Check(o)) ? PyNumber_Int(o) : \
                                (PyErr_Format(PyExc_TypeError, \
                                              "expected index value, got %.200s", Py_TYPE(o)->tp_name), \
                                 (PyObject*)0))
  #define __Pyx_PyIndex_Check(o) (PyNumber_Check(o) && !PyFloat_Check(o) && \
                                  !PyComplex_Check(o))
  #define PyIndex_Check __Pyx_PyIndex_Check
  #define PyErr_WarnEx(category, message, stacklevel) PyErr_Warn(category, message)
  #define __PYX_BUILD_PY_SSIZE_T "i"
#else
  #define __PYX_BUILD_PY_SSIZE_T "n"
  #define CYTHON_FORMAT_SSIZE_T "z"
  #define __Pyx_PyIndex_Check PyIndex_Check
#endif
#if PY_VERSION_HEX < 0x02060000
  #define Py_REFCNT(ob) (((PyObject*)(ob))->ob_refcnt)
  #define Py_TYPE(ob)   (((PyObject*)(ob))->ob_type)
  #define Py_SIZE(ob)   (((PyVarObject*)(ob))->ob_size)
  #define PyVarObject_HEAD_INIT(type, size) \
          PyObject_HEAD_INIT(type) size,
  #define PyType_Modified(t)
  typedef struct {
     void *buf;
     PyObject *obj;
     Py_ssize_t len;
     Py_ssize_t itemsize;
     int readonly;
     int ndim;
     char *format;
     Py_ssize_t *shape;
     Py_ssize_t *strides;
     Py_ssize_t *suboffsets;
     void *internal;
  } Py_buffer;
  #define PyBUF_SIMPLE 0
  #define PyBUF_WRITABLE 0x0001
  #define PyBUF_FORMAT 0x0004
  #define PyBUF_ND 0x0008
  #define PyBUF_STRIDES (0x0010 | PyBUF_ND)
  #define PyBUF_C_CONTIGUOUS (0x0020 | PyBUF_STRIDES)
  #define PyBUF_F_CONTIGUOUS (0x0040 | PyBUF_STRIDES)
  #define PyBUF_ANY_CONTIGUOUS (0x0080 | PyBUF_STRIDES)
  #define PyBUF_INDIRECT (0x0100 | PyBUF_STRIDES)
  #define PyBUF_RECORDS (PyBUF_STRIDES | PyBUF_FORMAT | PyBUF_WRITABLE)
  #define PyBUF_FULL (PyBUF_INDIRECT | PyBUF_FORMAT | PyBUF_WRITABLE)
  typedef int (*getbufferproc)(PyObject *, Py_buffer *, int);
  typedef void (*releasebufferproc)(PyObject *, Py_buffer *);
#endif
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos) \
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#if PY_MAJOR_VERSION < 3 && PY_MINOR_VERSION < 6
  #define PyUnicode_FromString(s) PyUnicode_Decode(s, strlen(s), "UTF-8", "strict")
#endif
#if PY_MAJOR_VERSION >= 3
  #define Py_TPFLAGS_CHECKTYPES 0
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#if (PY_VERSION_HEX < 0x02060000) || (PY_MAJOR_VERSION >= 3)
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ? \
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((k=k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PyBytesObject                PyStringObject
  #define PyBytes_Type                 PyString_Type
  #define PyBytes_Check                PyString_Check
  #define PyBytes_CheckExact           PyString_CheckExact
  #define PyBytes_FromString           PyString_FromString
  #define PyBytes_FromStringAndSize    PyString_FromStringAndSize
  #define PyBytes_FromFormat           PyString_FromFormat
  #define PyBytes_DecodeEscape         PyString_DecodeEscape
  #define PyBytes_AsString             PyString_AsString
  #define PyBytes_AsStringAndSize      PyString_AsStringAndSize
  #define PyBytes_Size                 PyString_Size
  #define PyBytes_AS_STRING            PyString_AS_STRING
  #define PyBytes_GET_SIZE             PyString_GET_SIZE
  #define PyBytes_Repr                 PyString_Repr
  #define PyBytes_Concat               PyString_Concat
  #define PyBytes_ConcatAndDel         PyString_ConcatAndDel
#endif
#if PY_VERSION_HEX < 0x02060000
  #define PySet_Check(obj)             PyObject_TypeCheck(obj, &PySet_Type)
  #define PyFrozenSet_Check(obj)       PyObject_TypeCheck(obj, &PyFrozenSet_Type)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_VERSION_HEX < 0x03020000
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if (PY_MAJOR_VERSION < 3) || (PY_VERSION_HEX >= 0x03010300)
  #define __Pyx_PySequence_GetSlice(obj, a, b) PySequence_GetSlice(obj, a, b)
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) PySequence_SetSlice(obj, a, b, value)
  #define __Pyx_PySequence_DelSlice(obj, a, b) PySequence_DelSlice(obj, a, b)
#else
  #define __Pyx_PySequence_GetSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), (PyObject*)0) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_GetSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object is unsliceable", (obj)->ob_type->tp_name), (PyObject*)0)))
  #define __Pyx_PySequence_SetSlice(obj, a, b, value) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_SetSlice(obj, a, b, value)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice assignment", (obj)->ob_type->tp_name), -1)))
  #define __Pyx_PySequence_DelSlice(obj, a, b) (unlikely(!(obj)) ? \
        (PyErr_SetString(PyExc_SystemError, "null argument to internal routine"), -1) : \
        (likely((obj)->ob_type->tp_as_mapping) ? (PySequence_DelSlice(obj, a, b)) : \
            (PyErr_Format(PyExc_TypeError, "'%.200s' object doesn't support slice deletion", (obj)->ob_type->tp_name), -1)))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),((char *)(n)))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),((char *)(n)),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),((char *)(n)))
#else
  #define __Pyx_GetAttrString(o,n)   PyObject_GetAttrString((o),(n))
  #define __Pyx_SetAttrString(o,n,a) PyObject_SetAttrString((o),(n),(a))
  #define __Pyx_DelAttrString(o,n)   PyObject_DelAttrString((o),(n))
#endif
#if PY_VERSION_HEX < 0x02050000
  #define __Pyx_NAMESTR(n) ((char *)(n))
  #define __Pyx_DOCSTR(n)  ((char *)(n))
#else
  #define __Pyx_NAMESTR(n) (n)
  #define __Pyx_DOCSTR(n)  (n)
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
#define _USE_MATH_DEFINES
#endif
#include <math.h>
#define __PYX_HAVE__nipy__algorithms__statistics__intvol
#define __PYX_HAVE_API__nipy__algorithms__statistics__intvol
#include "stdio.h"
#include "stdlib.h"
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include "math.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif


/* inline attribute */
#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* unused attribute */
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif

typedef struct {PyObject **p; char *s; const long n; const char* encoding; const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry; /*proto*/


/* Type Conversion Predeclarations */

#define __Pyx_PyBytes_FromUString(s) PyBytes_FromString((char*)s)
#define __Pyx_PyBytes_AsUString(s)   ((unsigned char*) PyBytes_AsString(s))

#define __Pyx_Owned_Py_None(b) (Py_INCREF(Py_None), Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);

static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject*);

#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))

#ifdef __GNUC__
  /* Test for GCC > 2.95 */
  #if __GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95))
    #define likely(x)   __builtin_expect(!!(x), 1)
    #define unlikely(x) __builtin_expect(!!(x), 0)
  #else /* __GNUC__ > 2 ... */
    #define likely(x)   (x)
    #define unlikely(x) (x)
  #endif /* __GNUC__ > 2 ... */
#else /* __GNUC__ */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
    
static PyObject *__pyx_m;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "intvol.pyx",
  "numpy.pxd",
  "type.pxd",
};
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name; /* for error messages only */
  struct __Pyx_StructField_* fields;
  size_t size;     /* sizeof(type) */
  size_t arraysize[8]; /* length of array in each dimension */
  int ndim;
  char typegroup; /* _R_eal, _C_omplex, Signed _I_nt, _U_nsigned int, _S_truct, _P_ointer, _O_bject, c_H_ar */
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "numpy.pxd":723
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "numpy.pxd":724
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "numpy.pxd":725
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "numpy.pxd":726
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "numpy.pxd":730
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "numpy.pxd":731
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "numpy.pxd":732
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "numpy.pxd":733
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "numpy.pxd":737
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "numpy.pxd":738
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "numpy.pxd":747
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "numpy.pxd":748
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "numpy.pxd":749
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "numpy.pxd":751
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "numpy.pxd":752
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "numpy.pxd":753
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "numpy.pxd":755
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "numpy.pxd":756
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "numpy.pxd":758
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "numpy.pxd":759
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "numpy.pxd":760
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif


/*--- Type declarations ---*/

/* "numpy.pxd":762
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "numpy.pxd":763
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "numpy.pxd":764
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "numpy.pxd":766
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname); /*proto*/
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          if (acquire_gil) { \
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure(); \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
              PyGILState_Release(__pyx_gilstate_save); \
          } else { \
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__); \
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil) \
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext() \
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif /* CYTHON_REFNANNY */
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name); /*proto*/

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found); /*proto*/

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name); /*proto*/

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[], \
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args, \
    const char* function_name); /*proto*/

#if PY_VERSION_HEX < 0x02050000
#ifndef PyAnySet_CheckExact
#define PyAnySet_CheckExact(ob) \
    ((ob)->ob_type == &PySet_Type || \
     (ob)->ob_type == &PyFrozenSet_Type)
#define PySet_New(iterable) \
    PyObject_CallFunctionObjArgs((PyObject *)&PySet_Type, (iterable), NULL)
#define Pyx_PyFrozenSet_New(iterable) \
    PyObject_CallFunctionObjArgs((PyObject *)&PyFrozenSet_Type, (iterable), NULL)
#define PySet_Size(anyset) \
    PyObject_Size((anyset))
#define PySet_Contains(anyset, key) \
    PySequence_Contains((anyset), (key))
#define PySet_Pop(set) \
    PyObject_CallMethod(set, (char *)"pop", NULL)
static CYTHON_INLINE int PySet_Clear(PyObject *set) {
    PyObject *ret = PyObject_CallMethod(set, (char *)"clear", NULL);
    if (!ret) return -1;
    Py_DECREF(ret); return 0;
}
static CYTHON_INLINE int PySet_Discard(PyObject *set, PyObject *key) {
    PyObject *ret = PyObject_CallMethod(set, (char *)"discard", (char *)"O", key);
    if (!ret) return -1;
    Py_DECREF(ret); return 0;
}
static CYTHON_INLINE int PySet_Add(PyObject *set, PyObject *key) {
    PyObject *ret = PyObject_CallMethod(set, (char *)"add", (char *)"O", key);
    if (!ret) return -1;
    Py_DECREF(ret); return 0;
}
#endif /* PyAnySet_CheckExact (<= Py2.4) */
#endif /* < Py2.5  */

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb); /*proto*/
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb); /*proto*/

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause); /*proto*/

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type); /*proto*/

static CYTHON_INLINE int  __Pyx_GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);

static void __Pyx_RaiseBufferFallbackError(void); /*proto*/

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
#define __Pyx_GetItemInt_List(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_List_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyList_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyList_GET_ITEM(o, PyList_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt_Tuple(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Tuple_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    else if ((-PyTuple_GET_SIZE(o) <= i) & (i < 0)) {
        PyObject *r = PyTuple_GET_ITEM(o, PyTuple_GET_SIZE(o) + i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
#define __Pyx_GetItemInt(o, i, size, to_py_func) (((size) <= sizeof(Py_ssize_t)) ? \
                                                    __Pyx_GetItemInt_Fast(o, i) : \
                                                    __Pyx_GetItemInt_Generic(o, to_py_func(i)))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (PyList_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyList_GET_SIZE(o)))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = (likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if (likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {  /* inlined PySequence_GetItem() */
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (unlikely(l < 0)) return NULL;
                i += l;
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

static void __Pyx_RaiseBufferIndexError(int axis); /*proto*/

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

static CYTHON_INLINE int __Pyx_mod_int(int, int); /* proto */

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact); /*proto*/

static CYTHON_INLINE long __Pyx_mod_long(long, long); /* proto */

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

static CYTHON_INLINE int __Pyx_IterFinish(void); /*proto*/

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected); /*proto*/

typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


static Py_ssize_t __Pyx_zeros[] = {0, 0, 0, 0, 0, 0, 0, 0};
static Py_ssize_t __Pyx_minusones[] = {-1, -1, -1, -1, -1, -1, -1, -1};

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, long level); /*proto*/

static CYTHON_INLINE void __Pyx_RaiseImportError(PyObject *name);

static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_from_py_Py_intptr_t(PyObject *);

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t);

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(_WIN32) && defined(__cplusplus) && CYTHON_CCOMPLEX
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eqf(a, b)   ((a)==(b))
    #define __Pyx_c_sumf(a, b)  ((a)+(b))
    #define __Pyx_c_difff(a, b) ((a)-(b))
    #define __Pyx_c_prodf(a, b) ((a)*(b))
    #define __Pyx_c_quotf(a, b) ((a)/(b))
    #define __Pyx_c_negf(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zerof(z) ((z)==(float)0)
    #define __Pyx_c_conjf(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_absf(z)     (::std::abs(z))
        #define __Pyx_c_powf(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zerof(z) ((z)==0)
    #define __Pyx_c_conjf(z)    (conjf(z))
    #if 1
        #define __Pyx_c_absf(z)     (cabsf(z))
        #define __Pyx_c_powf(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq(a, b)   ((a)==(b))
    #define __Pyx_c_sum(a, b)  ((a)+(b))
    #define __Pyx_c_diff(a, b) ((a)-(b))
    #define __Pyx_c_prod(a, b) ((a)*(b))
    #define __Pyx_c_quot(a, b) ((a)/(b))
    #define __Pyx_c_neg(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero(z) ((z)==(double)0)
    #define __Pyx_c_conj(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (::std::abs(z))
        #define __Pyx_c_pow(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero(z) ((z)==0)
    #define __Pyx_c_conj(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs(z)     (cabs(z))
        #define __Pyx_c_pow(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject *);

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject *);

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject *);

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject *);

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject *);

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject *);

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject *);

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject *);

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject *);

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject *);

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject *);

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject *);

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name); /*proto*/

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);  /*proto*/

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename); /*proto*/

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t); /*proto*/


/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'nipy.algorithms.statistics.intvol' */
static double __pyx_v_4nipy_10algorithms_10statistics_6intvol_PI;
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu3_tet(double, double, double, double, double, double, double, double, double, double, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tet(double, double, double, double, double, double, double, double, double, double, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tet(double, double, double, double, double, double, double, double, double, double, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE double __pyx_f_4nipy_10algorithms_10statistics_6intvol_limited_acos(double); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(double, double, double, double, double, double, double, double, double, double, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(double, double, double, double, double, double, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tri(double, double, double, double, double, double, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(double, double, double, int __pyx_skip_dispatch); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t = { "intp_t", NULL, sizeof(__pyx_t_5numpy_intp_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_intp_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_intp_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float_t = { "float_t", NULL, sizeof(__pyx_t_5numpy_float_t), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "nipy.algorithms.statistics.intvol"
int __pyx_module_is_main_nipy__algorithms__statistics__intvol = 0;

/* Implementation of 'nipy.algorithms.statistics.intvol' */
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_mu3_tet(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_2mu2_tet(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_4mu1_tet(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_6_mu1_tetface(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_Ds0s0, double __pyx_v_Ds0s1, double __pyx_v_Ds1s1, double __pyx_v_Ds0t0, double __pyx_v_Ds0t1, double __pyx_v_Ds1t0, double __pyx_v_Ds1t1, double __pyx_v_Dt0t0, double __pyx_v_Dt0t1, double __pyx_v_Dt1t1); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_8mu2_tri(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D22); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_10mu1_tri(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D22); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_12mu1_edge(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D11); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_14EC3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_16Lips3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords, PyObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_18_convert_stride3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v, PyObject *__pyx_v_stride1, PyObject *__pyx_v_stride2); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_20_convert_stride2(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v, PyObject *__pyx_v_stride1, PyObject *__pyx_v_stride2); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_22_convert_stride1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v, PyObject *__pyx_v_stride1, PyObject *__pyx_v_stride2); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_24Lips2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords, PyObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_26EC2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_28Lips1d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_coords, PyArrayObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_30EC1d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mask); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static char __pyx_k_1[] = "mask should be filled with 0/1 values, but be of type np.int";
static char __pyx_k_8[] = "cube_with_strides_center";
static char __pyx_k_17[] = "shape of mask does not match coordinates";
static char __pyx_k_73[] = "ndarray is not C contiguous";
static char __pyx_k_75[] = "ndarray is not Fortran contiguous";
static char __pyx_k_77[] = "Non-native byte order not supported";
static char __pyx_k_79[] = "unknown dtype code in numpy.pxd (%d)";
static char __pyx_k_80[] = "Format string allocated too short, see comment in numpy.pxd";
static char __pyx_k_83[] = "Format string allocated too short.";
static char __pyx_k_85[] = "\nThe estimators for the intrinsic volumes appearing in this module\nwere partially supported by NSF grant DMS-0405970.\n\nTaylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n   with an application to brain mapping.\"\n   Journal of the American Statistical Association, 102(479):913-928.\n\n";
static char __pyx_k_86[] = "scipy.sparse";
static char __pyx_k_87[] = "nipy.utils.arrays";
static char __pyx_k_90[] = "/Users/mb312/dev_trees/nipy/nipy/algorithms/statistics/intvol.pyx";
static char __pyx_k_91[] = "nipy.algorithms.statistics.intvol";
static char __pyx_k__B[] = "B";
static char __pyx_k__D[] = "D";
static char __pyx_k__H[] = "H";
static char __pyx_k__I[] = "I";
static char __pyx_k__L[] = "L";
static char __pyx_k__O[] = "O";
static char __pyx_k__Q[] = "Q";
static char __pyx_k__b[] = "b";
static char __pyx_k__c[] = "c";
static char __pyx_k__d[] = "d";
static char __pyx_k__f[] = "f";
static char __pyx_k__g[] = "g";
static char __pyx_k__h[] = "h";
static char __pyx_k__i[] = "i";
static char __pyx_k__j[] = "j";
static char __pyx_k__k[] = "k";
static char __pyx_k__l[] = "l";
static char __pyx_k__m[] = "m";
static char __pyx_k__q[] = "q";
static char __pyx_k__r[] = "r";
static char __pyx_k__s[] = "s";
static char __pyx_k__v[] = "v";
static char __pyx_k__Zd[] = "Zd";
static char __pyx_k__Zf[] = "Zf";
static char __pyx_k__Zg[] = "Zg";
static char __pyx_k__d2[] = "d2";
static char __pyx_k__d3[] = "d3";
static char __pyx_k__d4[] = "d4";
static char __pyx_k__l0[] = "l0";
static char __pyx_k__l1[] = "l1";
static char __pyx_k__l2[] = "l2";
static char __pyx_k__l3[] = "l3";
static char __pyx_k__m2[] = "m2";
static char __pyx_k__m3[] = "m3";
static char __pyx_k__m4[] = "m4";
static char __pyx_k__mr[] = "mr";
static char __pyx_k__ms[] = "ms";
static char __pyx_k__np[] = "np";
static char __pyx_k__pi[] = "pi";
static char __pyx_k__rr[] = "rr";
static char __pyx_k__s0[] = "s0";
static char __pyx_k__s1[] = "s1";
static char __pyx_k__s2[] = "s2";
static char __pyx_k__ss[] = "ss";
static char __pyx_k__v0[] = "v0";
static char __pyx_k__v1[] = "v1";
static char __pyx_k__v2[] = "v2";
static char __pyx_k__v3[] = "v3";
static char __pyx_k__w0[] = "w0";
static char __pyx_k__w1[] = "w1";
static char __pyx_k__w2[] = "w2";
static char __pyx_k__w3[] = "w3";
static char __pyx_k__D00[] = "D00";
static char __pyx_k__D01[] = "D01";
static char __pyx_k__D02[] = "D02";
static char __pyx_k__D03[] = "D03";
static char __pyx_k__D11[] = "D11";
static char __pyx_k__D12[] = "D12";
static char __pyx_k__D13[] = "D13";
static char __pyx_k__D22[] = "D22";
static char __pyx_k__D23[] = "D23";
static char __pyx_k__D33[] = "D33";
static char __pyx_k__ds2[] = "ds2";
static char __pyx_k__ds3[] = "ds3";
static char __pyx_k__ds4[] = "ds4";
static char __pyx_k__int[] = "int";
static char __pyx_k__res[] = "res";
static char __pyx_k__ss0[] = "ss0";
static char __pyx_k__ss1[] = "ss1";
static char __pyx_k__ss2[] = "ss2";
static char __pyx_k__sum[] = "sum";
static char __pyx_k__EC1d[] = "EC1d";
static char __pyx_k__EC2d[] = "EC2d";
static char __pyx_k__EC3d[] = "EC3d";
static char __pyx_k__bool[] = "bool";
static char __pyx_k__intp[] = "intp";
static char __pyx_k__mask[] = "mask";
static char __pyx_k__ndim[] = "ndim";
static char __pyx_k__npix[] = "npix";
static char __pyx_k__nvox[] = "nvox";
static char __pyx_k__size[] = "size";
static char __pyx_k__ss0d[] = "ss0d";
static char __pyx_k__ss1d[] = "ss1d";
static char __pyx_k__ss2d[] = "ss2d";
static char __pyx_k__Ds0s0[] = "Ds0s0";
static char __pyx_k__Ds0s1[] = "Ds0s1";
static char __pyx_k__Ds0t0[] = "Ds0t0";
static char __pyx_k__Ds0t1[] = "Ds0t1";
static char __pyx_k__Ds1s1[] = "Ds1s1";
static char __pyx_k__Ds1t0[] = "Ds1t0";
static char __pyx_k__Ds1t1[] = "Ds1t1";
static char __pyx_k__Dt0t0[] = "Dt0t0";
static char __pyx_k__Dt0t1[] = "Dt0t1";
static char __pyx_k__Dt1t1[] = "Dt1t1";
static char __pyx_k__array[] = "array";
static char __pyx_k__dtype[] = "dtype";
static char __pyx_k__fmask[] = "fmask";
static char __pyx_k__index[] = "index";
static char __pyx_k__numpy[] = "numpy";
static char __pyx_k__pmask[] = "pmask";
static char __pyx_k__range[] = "range";
static char __pyx_k__shape[] = "shape";
static char __pyx_k__union[] = "union";
static char __pyx_k__utils[] = "utils";
static char __pyx_k__value[] = "value";
static char __pyx_k__verts[] = "verts";
static char __pyx_k__zeros[] = "zeros";
static char __pyx_k__Lips1d[] = "Lips1d";
static char __pyx_k__Lips2d[] = "Lips2d";
static char __pyx_k__Lips3d[] = "Lips3d";
static char __pyx_k__coords[] = "coords";
static char __pyx_k__fpmask[] = "fpmask";
static char __pyx_k__hstack[] = "hstack";
static char __pyx_k__mask_c[] = "mask_c";
static char __pyx_k__pindex[] = "pindex";
static char __pyx_k__sorted[] = "sorted";
static char __pyx_k__unique[] = "unique";
static char __pyx_k__fcoords[] = "fcoords";
static char __pyx_k__reshape[] = "reshape";
static char __pyx_k__squeeze[] = "squeeze";
static char __pyx_k__stride1[] = "stride1";
static char __pyx_k__stride2[] = "stride2";
static char __pyx_k__strides[] = "strides";
static char __pyx_k____main__[] = "__main__";
static char __pyx_k____test__[] = "__test__";
static char __pyx_k__coords_c[] = "coords_c";
static char __pyx_k__dstrides[] = "dstrides";
static char __pyx_k__issubset[] = "issubset";
static char __pyx_k__cvertices[] = "cvertices";
static char __pyx_k__ValueError[] = "ValueError";
static char __pyx_k__difference[] = "difference";
static char __pyx_k__dok_matrix[] = "dok_matrix";
static char __pyx_k__pmask_shape[] = "pmask_shape";
static char __pyx_k__RuntimeError[] = "RuntimeError";
static char __pyx_k__strides_from[] = "strides_from";
static char __pyx_k__join_complexes[] = "join_complexes";
static char __pyx_k___convert_stride1[] = "_convert_stride1";
static char __pyx_k___convert_stride2[] = "_convert_stride2";
static char __pyx_k___convert_stride3[] = "_convert_stride3";
static PyObject *__pyx_kp_s_1;
static PyObject *__pyx_kp_s_17;
static PyObject *__pyx_kp_u_73;
static PyObject *__pyx_kp_u_75;
static PyObject *__pyx_kp_u_77;
static PyObject *__pyx_kp_u_79;
static PyObject *__pyx_n_s_8;
static PyObject *__pyx_kp_u_80;
static PyObject *__pyx_kp_u_83;
static PyObject *__pyx_n_s_86;
static PyObject *__pyx_n_s_87;
static PyObject *__pyx_kp_s_90;
static PyObject *__pyx_n_s_91;
static PyObject *__pyx_n_s__D;
static PyObject *__pyx_n_s__D00;
static PyObject *__pyx_n_s__D01;
static PyObject *__pyx_n_s__D02;
static PyObject *__pyx_n_s__D03;
static PyObject *__pyx_n_s__D11;
static PyObject *__pyx_n_s__D12;
static PyObject *__pyx_n_s__D13;
static PyObject *__pyx_n_s__D22;
static PyObject *__pyx_n_s__D23;
static PyObject *__pyx_n_s__D33;
static PyObject *__pyx_n_s__Ds0s0;
static PyObject *__pyx_n_s__Ds0s1;
static PyObject *__pyx_n_s__Ds0t0;
static PyObject *__pyx_n_s__Ds0t1;
static PyObject *__pyx_n_s__Ds1s1;
static PyObject *__pyx_n_s__Ds1t0;
static PyObject *__pyx_n_s__Ds1t1;
static PyObject *__pyx_n_s__Dt0t0;
static PyObject *__pyx_n_s__Dt0t1;
static PyObject *__pyx_n_s__Dt1t1;
static PyObject *__pyx_n_s__EC1d;
static PyObject *__pyx_n_s__EC2d;
static PyObject *__pyx_n_s__EC3d;
static PyObject *__pyx_n_s__Lips1d;
static PyObject *__pyx_n_s__Lips2d;
static PyObject *__pyx_n_s__Lips3d;
static PyObject *__pyx_n_s__RuntimeError;
static PyObject *__pyx_n_s__ValueError;
static PyObject *__pyx_n_s____main__;
static PyObject *__pyx_n_s____test__;
static PyObject *__pyx_n_s___convert_stride1;
static PyObject *__pyx_n_s___convert_stride2;
static PyObject *__pyx_n_s___convert_stride3;
static PyObject *__pyx_n_s__array;
static PyObject *__pyx_n_s__bool;
static PyObject *__pyx_n_s__c;
static PyObject *__pyx_n_s__coords;
static PyObject *__pyx_n_s__coords_c;
static PyObject *__pyx_n_s__cvertices;
static PyObject *__pyx_n_s__d2;
static PyObject *__pyx_n_s__d3;
static PyObject *__pyx_n_s__d4;
static PyObject *__pyx_n_s__difference;
static PyObject *__pyx_n_s__dok_matrix;
static PyObject *__pyx_n_s__ds2;
static PyObject *__pyx_n_s__ds3;
static PyObject *__pyx_n_s__ds4;
static PyObject *__pyx_n_s__dstrides;
static PyObject *__pyx_n_s__dtype;
static PyObject *__pyx_n_s__fcoords;
static PyObject *__pyx_n_s__fmask;
static PyObject *__pyx_n_s__fpmask;
static PyObject *__pyx_n_s__hstack;
static PyObject *__pyx_n_s__i;
static PyObject *__pyx_n_s__index;
static PyObject *__pyx_n_s__int;
static PyObject *__pyx_n_s__intp;
static PyObject *__pyx_n_s__issubset;
static PyObject *__pyx_n_s__j;
static PyObject *__pyx_n_s__join_complexes;
static PyObject *__pyx_n_s__k;
static PyObject *__pyx_n_s__l;
static PyObject *__pyx_n_s__l0;
static PyObject *__pyx_n_s__l1;
static PyObject *__pyx_n_s__l2;
static PyObject *__pyx_n_s__l3;
static PyObject *__pyx_n_s__m;
static PyObject *__pyx_n_s__m2;
static PyObject *__pyx_n_s__m3;
static PyObject *__pyx_n_s__m4;
static PyObject *__pyx_n_s__mask;
static PyObject *__pyx_n_s__mask_c;
static PyObject *__pyx_n_s__mr;
static PyObject *__pyx_n_s__ms;
static PyObject *__pyx_n_s__ndim;
static PyObject *__pyx_n_s__np;
static PyObject *__pyx_n_s__npix;
static PyObject *__pyx_n_s__numpy;
static PyObject *__pyx_n_s__nvox;
static PyObject *__pyx_n_s__pi;
static PyObject *__pyx_n_s__pindex;
static PyObject *__pyx_n_s__pmask;
static PyObject *__pyx_n_s__pmask_shape;
static PyObject *__pyx_n_s__r;
static PyObject *__pyx_n_s__range;
static PyObject *__pyx_n_s__res;
static PyObject *__pyx_n_s__reshape;
static PyObject *__pyx_n_s__rr;
static PyObject *__pyx_n_s__s;
static PyObject *__pyx_n_s__s0;
static PyObject *__pyx_n_s__s1;
static PyObject *__pyx_n_s__s2;
static PyObject *__pyx_n_s__shape;
static PyObject *__pyx_n_s__size;
static PyObject *__pyx_n_s__sorted;
static PyObject *__pyx_n_s__squeeze;
static PyObject *__pyx_n_s__ss;
static PyObject *__pyx_n_s__ss0;
static PyObject *__pyx_n_s__ss0d;
static PyObject *__pyx_n_s__ss1;
static PyObject *__pyx_n_s__ss1d;
static PyObject *__pyx_n_s__ss2;
static PyObject *__pyx_n_s__ss2d;
static PyObject *__pyx_n_s__stride1;
static PyObject *__pyx_n_s__stride2;
static PyObject *__pyx_n_s__strides;
static PyObject *__pyx_n_s__strides_from;
static PyObject *__pyx_n_s__sum;
static PyObject *__pyx_n_s__union;
static PyObject *__pyx_n_s__unique;
static PyObject *__pyx_n_s__utils;
static PyObject *__pyx_n_s__v;
static PyObject *__pyx_n_s__v0;
static PyObject *__pyx_n_s__v1;
static PyObject *__pyx_n_s__v2;
static PyObject *__pyx_n_s__v3;
static PyObject *__pyx_n_s__value;
static PyObject *__pyx_n_s__verts;
static PyObject *__pyx_n_s__w0;
static PyObject *__pyx_n_s__w1;
static PyObject *__pyx_n_s__w2;
static PyObject *__pyx_n_s__w3;
static PyObject *__pyx_n_s__zeros;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_int_15;
static PyObject *__pyx_k_slice_3;
static PyObject *__pyx_k_slice_4;
static PyObject *__pyx_k_slice_5;
static PyObject *__pyx_k_tuple_2;
static PyObject *__pyx_k_tuple_6;
static PyObject *__pyx_k_tuple_7;
static PyObject *__pyx_k_tuple_9;
static PyObject *__pyx_k_slice_21;
static PyObject *__pyx_k_slice_22;
static PyObject *__pyx_k_slice_23;
static PyObject *__pyx_k_slice_43;
static PyObject *__pyx_k_slice_44;
static PyObject *__pyx_k_slice_57;
static PyObject *__pyx_k_slice_58;
static PyObject *__pyx_k_tuple_10;
static PyObject *__pyx_k_tuple_11;
static PyObject *__pyx_k_tuple_12;
static PyObject *__pyx_k_tuple_13;
static PyObject *__pyx_k_tuple_14;
static PyObject *__pyx_k_tuple_15;
static PyObject *__pyx_k_tuple_16;
static PyObject *__pyx_k_tuple_18;
static PyObject *__pyx_k_tuple_19;
static PyObject *__pyx_k_tuple_20;
static PyObject *__pyx_k_tuple_24;
static PyObject *__pyx_k_tuple_25;
static PyObject *__pyx_k_tuple_26;
static PyObject *__pyx_k_tuple_27;
static PyObject *__pyx_k_tuple_28;
static PyObject *__pyx_k_tuple_29;
static PyObject *__pyx_k_tuple_30;
static PyObject *__pyx_k_tuple_31;
static PyObject *__pyx_k_tuple_32;
static PyObject *__pyx_k_tuple_33;
static PyObject *__pyx_k_tuple_34;
static PyObject *__pyx_k_tuple_35;
static PyObject *__pyx_k_tuple_36;
static PyObject *__pyx_k_tuple_37;
static PyObject *__pyx_k_tuple_38;
static PyObject *__pyx_k_tuple_39;
static PyObject *__pyx_k_tuple_40;
static PyObject *__pyx_k_tuple_41;
static PyObject *__pyx_k_tuple_42;
static PyObject *__pyx_k_tuple_45;
static PyObject *__pyx_k_tuple_46;
static PyObject *__pyx_k_tuple_47;
static PyObject *__pyx_k_tuple_48;
static PyObject *__pyx_k_tuple_49;
static PyObject *__pyx_k_tuple_50;
static PyObject *__pyx_k_tuple_51;
static PyObject *__pyx_k_tuple_52;
static PyObject *__pyx_k_tuple_53;
static PyObject *__pyx_k_tuple_54;
static PyObject *__pyx_k_tuple_55;
static PyObject *__pyx_k_tuple_56;
static PyObject *__pyx_k_tuple_59;
static PyObject *__pyx_k_tuple_60;
static PyObject *__pyx_k_tuple_61;
static PyObject *__pyx_k_tuple_62;
static PyObject *__pyx_k_tuple_63;
static PyObject *__pyx_k_tuple_64;
static PyObject *__pyx_k_tuple_65;
static PyObject *__pyx_k_tuple_66;
static PyObject *__pyx_k_tuple_67;
static PyObject *__pyx_k_tuple_68;
static PyObject *__pyx_k_tuple_69;
static PyObject *__pyx_k_tuple_70;
static PyObject *__pyx_k_tuple_71;
static PyObject *__pyx_k_tuple_72;
static PyObject *__pyx_k_tuple_74;
static PyObject *__pyx_k_tuple_76;
static PyObject *__pyx_k_tuple_78;
static PyObject *__pyx_k_tuple_81;
static PyObject *__pyx_k_tuple_82;
static PyObject *__pyx_k_tuple_84;
static PyObject *__pyx_k_tuple_88;
static PyObject *__pyx_k_tuple_92;
static PyObject *__pyx_k_tuple_94;
static PyObject *__pyx_k_tuple_96;
static PyObject *__pyx_k_tuple_98;
static PyObject *__pyx_k_tuple_100;
static PyObject *__pyx_k_tuple_102;
static PyObject *__pyx_k_tuple_104;
static PyObject *__pyx_k_tuple_106;
static PyObject *__pyx_k_codeobj_89;
static PyObject *__pyx_k_codeobj_93;
static PyObject *__pyx_k_codeobj_95;
static PyObject *__pyx_k_codeobj_97;
static PyObject *__pyx_k_codeobj_99;
static PyObject *__pyx_k_codeobj_101;
static PyObject *__pyx_k_codeobj_103;
static PyObject *__pyx_k_codeobj_105;
static PyObject *__pyx_k_codeobj_107;

/* "nipy/algorithms/statistics/intvol.pyx":36
 * 
 * 
 * cpdef double mu3_tet(double D00, double D01, double D02, double D03,             # <<<<<<<<<<<<<<
 *                      double D11, double D12, double D13,
 *                      double D22, double D23,
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_1mu3_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu3_tet(double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_C00;
  double __pyx_v_C01;
  double __pyx_v_C02;
  double __pyx_v_C11;
  double __pyx_v_C12;
  double __pyx_v_C22;
  double __pyx_v_v2;
  double __pyx_r;
  int __pyx_t_1;

  /* "nipy/algorithms/statistics/intvol.pyx":76
 *   """
 *   cdef double C00, C01, C02, C11, C12, C22, v2
 *   C00 = D00 - 2*D03 + D33             # <<<<<<<<<<<<<<
 *   C01 = D01 - D13 - D03 + D33
 *   C02 = D02 - D23 - D03 + D33
 */
  __pyx_v_C00 = ((__pyx_v_D00 - (2.0 * __pyx_v_D03)) + __pyx_v_D33);

  /* "nipy/algorithms/statistics/intvol.pyx":77
 *   cdef double C00, C01, C02, C11, C12, C22, v2
 *   C00 = D00 - 2*D03 + D33
 *   C01 = D01 - D13 - D03 + D33             # <<<<<<<<<<<<<<
 *   C02 = D02 - D23 - D03 + D33
 *   C11 = D11 - 2*D13 + D33
 */
  __pyx_v_C01 = (((__pyx_v_D01 - __pyx_v_D13) - __pyx_v_D03) + __pyx_v_D33);

  /* "nipy/algorithms/statistics/intvol.pyx":78
 *   C00 = D00 - 2*D03 + D33
 *   C01 = D01 - D13 - D03 + D33
 *   C02 = D02 - D23 - D03 + D33             # <<<<<<<<<<<<<<
 *   C11 = D11 - 2*D13 + D33
 *   C12 = D12 - D13 - D23 + D33
 */
  __pyx_v_C02 = (((__pyx_v_D02 - __pyx_v_D23) - __pyx_v_D03) + __pyx_v_D33);

  /* "nipy/algorithms/statistics/intvol.pyx":79
 *   C01 = D01 - D13 - D03 + D33
 *   C02 = D02 - D23 - D03 + D33
 *   C11 = D11 - 2*D13 + D33             # <<<<<<<<<<<<<<
 *   C12 = D12 - D13 - D23 + D33
 *   C22 = D22 - 2*D23 + D33
 */
  __pyx_v_C11 = ((__pyx_v_D11 - (2.0 * __pyx_v_D13)) + __pyx_v_D33);

  /* "nipy/algorithms/statistics/intvol.pyx":80
 *   C02 = D02 - D23 - D03 + D33
 *   C11 = D11 - 2*D13 + D33
 *   C12 = D12 - D13 - D23 + D33             # <<<<<<<<<<<<<<
 *   C22 = D22 - 2*D23 + D33
 *   v2 = (C00 * (C11 * C22 - C12 * C12) -
 */
  __pyx_v_C12 = (((__pyx_v_D12 - __pyx_v_D13) - __pyx_v_D23) + __pyx_v_D33);

  /* "nipy/algorithms/statistics/intvol.pyx":81
 *   C11 = D11 - 2*D13 + D33
 *   C12 = D12 - D13 - D23 + D33
 *   C22 = D22 - 2*D23 + D33             # <<<<<<<<<<<<<<
 *   v2 = (C00 * (C11 * C22 - C12 * C12) -
 *         C01 * (C01 * C22 - C02 * C12) +
 */
  __pyx_v_C22 = ((__pyx_v_D22 - (2.0 * __pyx_v_D23)) + __pyx_v_D33);

  /* "nipy/algorithms/statistics/intvol.pyx":84
 *   v2 = (C00 * (C11 * C22 - C12 * C12) -
 *         C01 * (C01 * C22 - C02 * C12) +
 *         C02 * (C01 * C12 - C11 * C02))             # <<<<<<<<<<<<<<
 *   # Rounding errors near 0 cause NaNs
 *   if v2 <= 0:
 */
  __pyx_v_v2 = (((__pyx_v_C00 * ((__pyx_v_C11 * __pyx_v_C22) - (__pyx_v_C12 * __pyx_v_C12))) - (__pyx_v_C01 * ((__pyx_v_C01 * __pyx_v_C22) - (__pyx_v_C02 * __pyx_v_C12)))) + (__pyx_v_C02 * ((__pyx_v_C01 * __pyx_v_C12) - (__pyx_v_C11 * __pyx_v_C02))));

  /* "nipy/algorithms/statistics/intvol.pyx":86
 *         C02 * (C01 * C12 - C11 * C02))
 *   # Rounding errors near 0 cause NaNs
 *   if v2 <= 0:             # <<<<<<<<<<<<<<
 *       return 0
 *   return sqrt(v2) / 6.
 */
  __pyx_t_1 = (__pyx_v_v2 <= 0.0);
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":87
 *   # Rounding errors near 0 cause NaNs
 *   if v2 <= 0:
 *       return 0             # <<<<<<<<<<<<<<
 *   return sqrt(v2) / 6.
 * 
 */
    __pyx_r = 0.0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":88
 *   if v2 <= 0:
 *       return 0
 *   return sqrt(v2) / 6.             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (sqrt(__pyx_v_v2) / 6.);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_1mu3_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_mu3_tet[] = " Compute the 3rd intrinsic volume of a tetrahedron.\n\n  3rd intrinsic volume (just volume in this case) of a tetrahedron with\n  coordinates implied by dot products below.\n\n  Parameters\n  ----------\n  D00 : float\n    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is\n    ``cv0.dot(cv0)``\n  D01 : float\n    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second\n    vertex.\n  D02 : float\n    ``cv0.dot(cv2)``\n  D03 : float\n    ``cv0.dot(cv3)``\n  D11 : float\n    ``cv1.dot(cv1)``\n  D12 : float\n    ``cv1.dot(cv2)``\n  D13 : float\n    ``cv1.dot(cv3)``\n  D22 : float\n    ``cv2.dot(cv2)``\n  D23 : float\n    ``cv2.dot(cv2)``\n  D33 : float\n    ``cv3.dot(cv3)``\n\n  Returns\n  -------\n  mu3 : float\n    volume of tetrahedron\n  ";
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_1mu3_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_D00;
  double __pyx_v_D01;
  double __pyx_v_D02;
  double __pyx_v_D03;
  double __pyx_v_D11;
  double __pyx_v_D12;
  double __pyx_v_D13;
  double __pyx_v_D22;
  double __pyx_v_D23;
  double __pyx_v_D33;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mu3_tet (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__D00,&__pyx_n_s__D01,&__pyx_n_s__D02,&__pyx_n_s__D03,&__pyx_n_s__D11,&__pyx_n_s__D12,&__pyx_n_s__D13,&__pyx_n_s__D22,&__pyx_n_s__D23,&__pyx_n_s__D33,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D00)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D01)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D02)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D03)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D11)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D13)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D22)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D23)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D33)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, 9); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mu3_tet") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_D00 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_D00 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D01 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_D01 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D02 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_D02 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D03 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_D03 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D11 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_D11 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D12 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_D12 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D13 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_D13 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 37; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D22 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_D22 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D23 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_D23 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 38; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D33 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_D33 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 39; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mu3_tet", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu3_tet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_mu3_tet(__pyx_self, __pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D03, __pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":36
 * 
 * 
 * cpdef double mu3_tet(double D00, double D01, double D02, double D03,             # <<<<<<<<<<<<<<
 *                      double D11, double D12, double D13,
 *                      double D22, double D23,
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_mu3_tet(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mu3_tet", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu3_tet(__pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D03, __pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 36; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu3_tet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":91
 * 
 * 
 * cpdef double mu2_tet(double D00, double D01, double D02, double D03,             # <<<<<<<<<<<<<<
 *                      double D11, double D12, double D13,
 *                      double D22, double D23,
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_3mu2_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tet(double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_mu;
  double __pyx_r;

  /* "nipy/algorithms/statistics/intvol.pyx":130
 *     Half tetrahedron surface area
 *   """
 *   cdef double mu = 0             # <<<<<<<<<<<<<<
 *   mu += mu2_tri(D00, D01, D02, D11, D12, D22)
 *   mu += mu2_tri(D00, D02, D03, D22, D23, D33)
 */
  __pyx_v_mu = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":131
 *   """
 *   cdef double mu = 0
 *   mu += mu2_tri(D00, D01, D02, D11, D12, D22)             # <<<<<<<<<<<<<<
 *   mu += mu2_tri(D00, D02, D03, D22, D23, D33)
 *   mu += mu2_tri(D11, D12, D13, D22, D23, D33)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(__pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D11, __pyx_v_D12, __pyx_v_D22, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":132
 *   cdef double mu = 0
 *   mu += mu2_tri(D00, D01, D02, D11, D12, D22)
 *   mu += mu2_tri(D00, D02, D03, D22, D23, D33)             # <<<<<<<<<<<<<<
 *   mu += mu2_tri(D11, D12, D13, D22, D23, D33)
 *   mu += mu2_tri(D00, D01, D03, D11, D13, D33)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(__pyx_v_D00, __pyx_v_D02, __pyx_v_D03, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":133
 *   mu += mu2_tri(D00, D01, D02, D11, D12, D22)
 *   mu += mu2_tri(D00, D02, D03, D22, D23, D33)
 *   mu += mu2_tri(D11, D12, D13, D22, D23, D33)             # <<<<<<<<<<<<<<
 *   mu += mu2_tri(D00, D01, D03, D11, D13, D33)
 *   return mu * 0.5
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(__pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":134
 *   mu += mu2_tri(D00, D02, D03, D22, D23, D33)
 *   mu += mu2_tri(D11, D12, D13, D22, D23, D33)
 *   mu += mu2_tri(D00, D01, D03, D11, D13, D33)             # <<<<<<<<<<<<<<
 *   return mu * 0.5
 * 
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(__pyx_v_D00, __pyx_v_D01, __pyx_v_D03, __pyx_v_D11, __pyx_v_D13, __pyx_v_D33, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":135
 *   mu += mu2_tri(D11, D12, D13, D22, D23, D33)
 *   mu += mu2_tri(D00, D01, D03, D11, D13, D33)
 *   return mu * 0.5             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_mu * 0.5);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_3mu2_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_2mu2_tet[] = " Compute the 2nd intrinsic volume of tetrahedron\n\n  2nd intrinsic volume (half the surface area) of a tetrahedron with coordinates\n  implied by dot products below.\n\n  Parameters\n  ----------\n  D00 : float\n    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is\n    ``cv0.dot(cv0)``\n  D01 : float\n    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second\n    vertex.\n  D02 : float\n    ``cv0.dot(cv2)``\n  D03 : float\n    ``cv0.dot(cv3)``\n  D11 : float\n    ``cv1.dot(cv1)``\n  D12 : float\n    ``cv1.dot(cv2)``\n  D13 : float\n    ``cv1.dot(cv3)``\n  D22 : float\n    ``cv2.dot(cv2)``\n  D23 : float\n    ``cv2.dot(cv2)``\n  D33 : float\n    ``cv3.dot(cv3)``\n\n  Returns\n  -------\n  mu2 : float\n    Half tetrahedron surface area\n  ";
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_3mu2_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_D00;
  double __pyx_v_D01;
  double __pyx_v_D02;
  double __pyx_v_D03;
  double __pyx_v_D11;
  double __pyx_v_D12;
  double __pyx_v_D13;
  double __pyx_v_D22;
  double __pyx_v_D23;
  double __pyx_v_D33;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mu2_tet (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__D00,&__pyx_n_s__D01,&__pyx_n_s__D02,&__pyx_n_s__D03,&__pyx_n_s__D11,&__pyx_n_s__D12,&__pyx_n_s__D13,&__pyx_n_s__D22,&__pyx_n_s__D23,&__pyx_n_s__D33,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D00)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D01)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D02)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D03)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D11)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D13)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D22)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D23)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D33)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, 9); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mu2_tet") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_D00 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_D00 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D01 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_D01 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D02 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_D02 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D03 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_D03 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D11 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_D11 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D12 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_D12 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D13 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_D13 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D22 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_D22 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 93; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D23 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_D23 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 93; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D33 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_D33 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 94; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mu2_tet", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu2_tet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_2mu2_tet(__pyx_self, __pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D03, __pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":91
 * 
 * 
 * cpdef double mu2_tet(double D00, double D01, double D02, double D03,             # <<<<<<<<<<<<<<
 *                      double D11, double D12, double D13,
 *                      double D22, double D23,
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_2mu2_tet(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mu2_tet", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tet(__pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D03, __pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 91; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu2_tet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":138
 * 
 * 
 * cpdef double mu1_tet(double D00, double D01, double D02, double D03,             # <<<<<<<<<<<<<<
 *                      double D11, double D12, double D13,
 *                      double D22, double D23,
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_5mu1_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tet(double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_mu;
  double __pyx_r;

  /* "nipy/algorithms/statistics/intvol.pyx":179
 *   """
 *   cdef double mu
 *   mu = 0             # <<<<<<<<<<<<<<
 *   mu += _mu1_tetface(D00, D01, D11, D02, D03, D12, D13, D22, D23, D33)
 *   mu += _mu1_tetface(D00, D02, D22, D01, D03, D12, D23, D11, D13, D33)
 */
  __pyx_v_mu = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":180
 *   cdef double mu
 *   mu = 0
 *   mu += _mu1_tetface(D00, D01, D11, D02, D03, D12, D13, D22, D23, D33)             # <<<<<<<<<<<<<<
 *   mu += _mu1_tetface(D00, D02, D22, D01, D03, D12, D23, D11, D13, D33)
 *   mu += _mu1_tetface(D00, D03, D33, D01, D02, D13, D23, D11, D12, D22)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_D00, __pyx_v_D01, __pyx_v_D11, __pyx_v_D02, __pyx_v_D03, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":181
 *   mu = 0
 *   mu += _mu1_tetface(D00, D01, D11, D02, D03, D12, D13, D22, D23, D33)
 *   mu += _mu1_tetface(D00, D02, D22, D01, D03, D12, D23, D11, D13, D33)             # <<<<<<<<<<<<<<
 *   mu += _mu1_tetface(D00, D03, D33, D01, D02, D13, D23, D11, D12, D22)
 *   mu += _mu1_tetface(D11, D12, D22, D01, D13, D02, D23, D00, D03, D33)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_D00, __pyx_v_D02, __pyx_v_D22, __pyx_v_D01, __pyx_v_D03, __pyx_v_D12, __pyx_v_D23, __pyx_v_D11, __pyx_v_D13, __pyx_v_D33, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":182
 *   mu += _mu1_tetface(D00, D01, D11, D02, D03, D12, D13, D22, D23, D33)
 *   mu += _mu1_tetface(D00, D02, D22, D01, D03, D12, D23, D11, D13, D33)
 *   mu += _mu1_tetface(D00, D03, D33, D01, D02, D13, D23, D11, D12, D22)             # <<<<<<<<<<<<<<
 *   mu += _mu1_tetface(D11, D12, D22, D01, D13, D02, D23, D00, D03, D33)
 *   mu += _mu1_tetface(D11, D13, D33, D01, D12, D03, D23, D00, D02, D22)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_D00, __pyx_v_D03, __pyx_v_D33, __pyx_v_D01, __pyx_v_D02, __pyx_v_D13, __pyx_v_D23, __pyx_v_D11, __pyx_v_D12, __pyx_v_D22, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":183
 *   mu += _mu1_tetface(D00, D02, D22, D01, D03, D12, D23, D11, D13, D33)
 *   mu += _mu1_tetface(D00, D03, D33, D01, D02, D13, D23, D11, D12, D22)
 *   mu += _mu1_tetface(D11, D12, D22, D01, D13, D02, D23, D00, D03, D33)             # <<<<<<<<<<<<<<
 *   mu += _mu1_tetface(D11, D13, D33, D01, D12, D03, D23, D00, D02, D22)
 *   mu += _mu1_tetface(D22, D23, D33, D02, D12, D03, D13, D00, D01, D11)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_D11, __pyx_v_D12, __pyx_v_D22, __pyx_v_D01, __pyx_v_D13, __pyx_v_D02, __pyx_v_D23, __pyx_v_D00, __pyx_v_D03, __pyx_v_D33, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":184
 *   mu += _mu1_tetface(D00, D03, D33, D01, D02, D13, D23, D11, D12, D22)
 *   mu += _mu1_tetface(D11, D12, D22, D01, D13, D02, D23, D00, D03, D33)
 *   mu += _mu1_tetface(D11, D13, D33, D01, D12, D03, D23, D00, D02, D22)             # <<<<<<<<<<<<<<
 *   mu += _mu1_tetface(D22, D23, D33, D02, D12, D03, D13, D00, D01, D11)
 *   return mu
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_D11, __pyx_v_D13, __pyx_v_D33, __pyx_v_D01, __pyx_v_D12, __pyx_v_D03, __pyx_v_D23, __pyx_v_D00, __pyx_v_D02, __pyx_v_D22, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":185
 *   mu += _mu1_tetface(D11, D12, D22, D01, D13, D02, D23, D00, D03, D33)
 *   mu += _mu1_tetface(D11, D13, D33, D01, D12, D03, D23, D00, D02, D22)
 *   mu += _mu1_tetface(D22, D23, D33, D02, D12, D03, D13, D00, D01, D11)             # <<<<<<<<<<<<<<
 *   return mu
 * 
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_D22, __pyx_v_D23, __pyx_v_D33, __pyx_v_D02, __pyx_v_D12, __pyx_v_D03, __pyx_v_D13, __pyx_v_D00, __pyx_v_D01, __pyx_v_D11, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":186
 *   mu += _mu1_tetface(D11, D13, D33, D01, D12, D03, D23, D00, D02, D22)
 *   mu += _mu1_tetface(D22, D23, D33, D02, D12, D03, D13, D00, D01, D11)
 *   return mu             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_mu;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_5mu1_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_4mu1_tet[] = " Return 3rd intrinsic volume of tetrahedron\n\n  Compute the 3rd intrinsic volume (sum of external angles * edge\n  lengths) of a tetrahedron for which the input arguments represent the\n  coordinate dot products of the vertices.\n\n  Parameters\n  ----------\n  D00 : float\n    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is\n    ``cv0.dot(cv0)``\n  D01 : float\n    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second\n    vertex.\n  D02 : float\n    ``cv0.dot(cv2)``\n  D03 : float\n    ``cv0.dot(cv3)``\n  D11 : float\n    ``cv1.dot(cv1)``\n  D12 : float\n    ``cv1.dot(cv2)``\n  D13 : float\n    ``cv1.dot(cv3)``\n  D22 : float\n    ``cv2.dot(cv2)``\n  D23 : float\n    ``cv2.dot(cv2)``\n  D33 : float\n    ``cv3.dot(cv3)``\n\n  Returns\n  -------\n  mu1 : float\n    3rd intrinsic volume of tetrahedron\n  ";
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_5mu1_tet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_D00;
  double __pyx_v_D01;
  double __pyx_v_D02;
  double __pyx_v_D03;
  double __pyx_v_D11;
  double __pyx_v_D12;
  double __pyx_v_D13;
  double __pyx_v_D22;
  double __pyx_v_D23;
  double __pyx_v_D33;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mu1_tet (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__D00,&__pyx_n_s__D01,&__pyx_n_s__D02,&__pyx_n_s__D03,&__pyx_n_s__D11,&__pyx_n_s__D12,&__pyx_n_s__D13,&__pyx_n_s__D22,&__pyx_n_s__D23,&__pyx_n_s__D33,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D00)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D01)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D02)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D03)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D11)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D13)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D22)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D23)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D33)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, 9); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mu1_tet") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_D00 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_D00 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D01 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_D01 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D02 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_D02 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D03 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_D03 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D11 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_D11 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 139; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D12 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_D12 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 139; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D13 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_D13 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 139; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D22 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_D22 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 140; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D23 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_D23 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 140; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D33 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_D33 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 141; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mu1_tet", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu1_tet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_4mu1_tet(__pyx_self, __pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D03, __pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":138
 * 
 * 
 * cpdef double mu1_tet(double D00, double D01, double D02, double D03,             # <<<<<<<<<<<<<<
 *                      double D11, double D12, double D13,
 *                      double D22, double D23,
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_4mu1_tet(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D03, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D13, double __pyx_v_D22, double __pyx_v_D23, double __pyx_v_D33) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mu1_tet", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tet(__pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D03, __pyx_v_D11, __pyx_v_D12, __pyx_v_D13, __pyx_v_D22, __pyx_v_D23, __pyx_v_D33, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 138; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu1_tet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":189
 * 
 * 
 * cdef inline double limited_acos(double val) nogil:             # <<<<<<<<<<<<<<
 *     """ Check for -1 <= val <= 1 before returning acos(val)
 * 
 */

static CYTHON_INLINE double __pyx_f_4nipy_10algorithms_10statistics_6intvol_limited_acos(double __pyx_v_val) {
  double __pyx_r;
  int __pyx_t_1;

  /* "nipy/algorithms/statistics/intvol.pyx":194
 *     Avoids nan values from small rounding errors
 *     """
 *     if val >= 1:             # <<<<<<<<<<<<<<
 *         return 0
 *     elif val <= -1:
 */
  __pyx_t_1 = (__pyx_v_val >= 1.0);
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":195
 *     """
 *     if val >= 1:
 *         return 0             # <<<<<<<<<<<<<<
 *     elif val <= -1:
 *         return PI
 */
    __pyx_r = 0.0;
    goto __pyx_L0;
    goto __pyx_L3;
  }

  /* "nipy/algorithms/statistics/intvol.pyx":196
 *     if val >= 1:
 *         return 0
 *     elif val <= -1:             # <<<<<<<<<<<<<<
 *         return PI
 *     return acos(val)
 */
  __pyx_t_1 = (__pyx_v_val <= -1.0);
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":197
 *         return 0
 *     elif val <= -1:
 *         return PI             # <<<<<<<<<<<<<<
 *     return acos(val)
 * 
 */
    __pyx_r = __pyx_v_4nipy_10algorithms_10statistics_6intvol_PI;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":198
 *     elif val <= -1:
 *         return PI
 *     return acos(val)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = acos(__pyx_v_val);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":202
 * 
 * @cython.cdivision(True)
 * cpdef double _mu1_tetface(double Ds0s0,             # <<<<<<<<<<<<<<
 *                           double Ds0s1,
 *                           double Ds1s1,
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_7_mu1_tetface(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(double __pyx_v_Ds0s0, double __pyx_v_Ds0s1, double __pyx_v_Ds1s1, double __pyx_v_Ds0t0, double __pyx_v_Ds0t1, double __pyx_v_Ds1t0, double __pyx_v_Ds1t1, double __pyx_v_Dt0t0, double __pyx_v_Dt0t1, double __pyx_v_Dt1t1, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_A00;
  double __pyx_v_A01;
  double __pyx_v_A02;
  double __pyx_v_A11;
  double __pyx_v_A12;
  double __pyx_v_A22;
  double __pyx_v_np_len;
  double __pyx_v_a;
  double __pyx_v_acosval;
  double __pyx_v_length;
  double __pyx_v_norm_proj0;
  double __pyx_v_norm_proj1;
  double __pyx_v_inner_prod_proj;
  double __pyx_r;
  int __pyx_t_1;

  /* "nipy/algorithms/statistics/intvol.pyx":215
 *     cdef double length, norm_proj0, norm_proj1, inner_prod_proj
 * 
 *     A00 = Ds1s1 - 2 * Ds0s1 + Ds0s0             # <<<<<<<<<<<<<<
 *     # all norms divided by this value, leading to NaN value for output, for
 *     # values <= 0
 */
  __pyx_v_A00 = ((__pyx_v_Ds1s1 - (2.0 * __pyx_v_Ds0s1)) + __pyx_v_Ds0s0);

  /* "nipy/algorithms/statistics/intvol.pyx":218
 *     # all norms divided by this value, leading to NaN value for output, for
 *     # values <= 0
 *     if A00 <= 0:             # <<<<<<<<<<<<<<
 *       return 0
 *     A11 = Dt0t0 - 2 * Ds0t0 + Ds0s0
 */
  __pyx_t_1 = (__pyx_v_A00 <= 0.0);
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":219
 *     # values <= 0
 *     if A00 <= 0:
 *       return 0             # <<<<<<<<<<<<<<
 *     A11 = Dt0t0 - 2 * Ds0t0 + Ds0s0
 *     A22 = Dt1t1 - 2 * Ds0t1 + Ds0s0
 */
    __pyx_r = 0.0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":220
 *     if A00 <= 0:
 *       return 0
 *     A11 = Dt0t0 - 2 * Ds0t0 + Ds0s0             # <<<<<<<<<<<<<<
 *     A22 = Dt1t1 - 2 * Ds0t1 + Ds0s0
 *     A01 = Ds1t0 - Ds0t0 - Ds0s1 + Ds0s0
 */
  __pyx_v_A11 = ((__pyx_v_Dt0t0 - (2.0 * __pyx_v_Ds0t0)) + __pyx_v_Ds0s0);

  /* "nipy/algorithms/statistics/intvol.pyx":221
 *       return 0
 *     A11 = Dt0t0 - 2 * Ds0t0 + Ds0s0
 *     A22 = Dt1t1 - 2 * Ds0t1 + Ds0s0             # <<<<<<<<<<<<<<
 *     A01 = Ds1t0 - Ds0t0 - Ds0s1 + Ds0s0
 *     A02 = Ds1t1 - Ds0t1 - Ds0s1 + Ds0s0
 */
  __pyx_v_A22 = ((__pyx_v_Dt1t1 - (2.0 * __pyx_v_Ds0t1)) + __pyx_v_Ds0s0);

  /* "nipy/algorithms/statistics/intvol.pyx":222
 *     A11 = Dt0t0 - 2 * Ds0t0 + Ds0s0
 *     A22 = Dt1t1 - 2 * Ds0t1 + Ds0s0
 *     A01 = Ds1t0 - Ds0t0 - Ds0s1 + Ds0s0             # <<<<<<<<<<<<<<
 *     A02 = Ds1t1 - Ds0t1 - Ds0s1 + Ds0s0
 *     A12 = Dt0t1 - Ds0t0 - Ds0t1 + Ds0s0
 */
  __pyx_v_A01 = (((__pyx_v_Ds1t0 - __pyx_v_Ds0t0) - __pyx_v_Ds0s1) + __pyx_v_Ds0s0);

  /* "nipy/algorithms/statistics/intvol.pyx":223
 *     A22 = Dt1t1 - 2 * Ds0t1 + Ds0s0
 *     A01 = Ds1t0 - Ds0t0 - Ds0s1 + Ds0s0
 *     A02 = Ds1t1 - Ds0t1 - Ds0s1 + Ds0s0             # <<<<<<<<<<<<<<
 *     A12 = Dt0t1 - Ds0t0 - Ds0t1 + Ds0s0
 *     length = sqrt(A00)
 */
  __pyx_v_A02 = (((__pyx_v_Ds1t1 - __pyx_v_Ds0t1) - __pyx_v_Ds0s1) + __pyx_v_Ds0s0);

  /* "nipy/algorithms/statistics/intvol.pyx":224
 *     A01 = Ds1t0 - Ds0t0 - Ds0s1 + Ds0s0
 *     A02 = Ds1t1 - Ds0t1 - Ds0s1 + Ds0s0
 *     A12 = Dt0t1 - Ds0t0 - Ds0t1 + Ds0s0             # <<<<<<<<<<<<<<
 *     length = sqrt(A00)
 *     norm_proj0 = A11 - A01 * A01 / A00
 */
  __pyx_v_A12 = (((__pyx_v_Dt0t1 - __pyx_v_Ds0t0) - __pyx_v_Ds0t1) + __pyx_v_Ds0s0);

  /* "nipy/algorithms/statistics/intvol.pyx":225
 *     A02 = Ds1t1 - Ds0t1 - Ds0s1 + Ds0s0
 *     A12 = Dt0t1 - Ds0t0 - Ds0t1 + Ds0s0
 *     length = sqrt(A00)             # <<<<<<<<<<<<<<
 *     norm_proj0 = A11 - A01 * A01 / A00
 *     norm_proj1 = A22 - A02 * A02 / A00
 */
  __pyx_v_length = sqrt(__pyx_v_A00);

  /* "nipy/algorithms/statistics/intvol.pyx":226
 *     A12 = Dt0t1 - Ds0t0 - Ds0t1 + Ds0s0
 *     length = sqrt(A00)
 *     norm_proj0 = A11 - A01 * A01 / A00             # <<<<<<<<<<<<<<
 *     norm_proj1 = A22 - A02 * A02 / A00
 *     inner_prod_proj = A12 - A01 * A02 / A00
 */
  __pyx_v_norm_proj0 = (__pyx_v_A11 - ((__pyx_v_A01 * __pyx_v_A01) / __pyx_v_A00));

  /* "nipy/algorithms/statistics/intvol.pyx":227
 *     length = sqrt(A00)
 *     norm_proj0 = A11 - A01 * A01 / A00
 *     norm_proj1 = A22 - A02 * A02 / A00             # <<<<<<<<<<<<<<
 *     inner_prod_proj = A12 - A01 * A02 / A00
 *     np_len = norm_proj0 * norm_proj1
 */
  __pyx_v_norm_proj1 = (__pyx_v_A22 - ((__pyx_v_A02 * __pyx_v_A02) / __pyx_v_A00));

  /* "nipy/algorithms/statistics/intvol.pyx":228
 *     norm_proj0 = A11 - A01 * A01 / A00
 *     norm_proj1 = A22 - A02 * A02 / A00
 *     inner_prod_proj = A12 - A01 * A02 / A00             # <<<<<<<<<<<<<<
 *     np_len = norm_proj0 * norm_proj1
 *     if np_len <= 0: # would otherwise lead to NaN return value
 */
  __pyx_v_inner_prod_proj = (__pyx_v_A12 - ((__pyx_v_A01 * __pyx_v_A02) / __pyx_v_A00));

  /* "nipy/algorithms/statistics/intvol.pyx":229
 *     norm_proj1 = A22 - A02 * A02 / A00
 *     inner_prod_proj = A12 - A01 * A02 / A00
 *     np_len = norm_proj0 * norm_proj1             # <<<<<<<<<<<<<<
 *     if np_len <= 0: # would otherwise lead to NaN return value
 *         return 0
 */
  __pyx_v_np_len = (__pyx_v_norm_proj0 * __pyx_v_norm_proj1);

  /* "nipy/algorithms/statistics/intvol.pyx":230
 *     inner_prod_proj = A12 - A01 * A02 / A00
 *     np_len = norm_proj0 * norm_proj1
 *     if np_len <= 0: # would otherwise lead to NaN return value             # <<<<<<<<<<<<<<
 *         return 0
 *     # hedge for small rounding errors above 1 and below -1
 */
  __pyx_t_1 = (__pyx_v_np_len <= 0.0);
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":231
 *     np_len = norm_proj0 * norm_proj1
 *     if np_len <= 0: # would otherwise lead to NaN return value
 *         return 0             # <<<<<<<<<<<<<<
 *     # hedge for small rounding errors above 1 and below -1
 *     acosval = limited_acos(inner_prod_proj / sqrt(np_len))
 */
    __pyx_r = 0.0;
    goto __pyx_L0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "nipy/algorithms/statistics/intvol.pyx":233
 *         return 0
 *     # hedge for small rounding errors above 1 and below -1
 *     acosval = limited_acos(inner_prod_proj / sqrt(np_len))             # <<<<<<<<<<<<<<
 *     a = (PI - acosval) * length / (2 * PI)
 *     return a
 */
  __pyx_v_acosval = __pyx_f_4nipy_10algorithms_10statistics_6intvol_limited_acos((__pyx_v_inner_prod_proj / sqrt(__pyx_v_np_len)));

  /* "nipy/algorithms/statistics/intvol.pyx":234
 *     # hedge for small rounding errors above 1 and below -1
 *     acosval = limited_acos(inner_prod_proj / sqrt(np_len))
 *     a = (PI - acosval) * length / (2 * PI)             # <<<<<<<<<<<<<<
 *     return a
 * 
 */
  __pyx_v_a = (((__pyx_v_4nipy_10algorithms_10statistics_6intvol_PI - __pyx_v_acosval) * __pyx_v_length) / (2.0 * __pyx_v_4nipy_10algorithms_10statistics_6intvol_PI));

  /* "nipy/algorithms/statistics/intvol.pyx":235
 *     acosval = limited_acos(inner_prod_proj / sqrt(np_len))
 *     a = (PI - acosval) * length / (2 * PI)
 *     return a             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_a;
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_7_mu1_tetface(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_7_mu1_tetface(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_Ds0s0;
  double __pyx_v_Ds0s1;
  double __pyx_v_Ds1s1;
  double __pyx_v_Ds0t0;
  double __pyx_v_Ds0t1;
  double __pyx_v_Ds1t0;
  double __pyx_v_Ds1t1;
  double __pyx_v_Dt0t0;
  double __pyx_v_Dt0t1;
  double __pyx_v_Dt1t1;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_mu1_tetface (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__Ds0s0,&__pyx_n_s__Ds0s1,&__pyx_n_s__Ds1s1,&__pyx_n_s__Ds0t0,&__pyx_n_s__Ds0t1,&__pyx_n_s__Ds1t0,&__pyx_n_s__Ds1t1,&__pyx_n_s__Dt0t0,&__pyx_n_s__Dt0t1,&__pyx_n_s__Dt1t1,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds0s0)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds0s1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds1s1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds0t0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds0t1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds1t0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  6:
        if (likely((values[6] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Ds1t1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 6); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  7:
        if (likely((values[7] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Dt0t0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 7); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  8:
        if (likely((values[8] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Dt0t1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 8); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  9:
        if (likely((values[9] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__Dt1t1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, 9); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_mu1_tetface") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_Ds0s0 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_Ds0s0 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Ds0s1 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_Ds0s1 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 203; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Ds1s1 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_Ds1s1 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 204; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Ds0t0 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_Ds0t0 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 205; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Ds0t1 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_Ds0t1 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 206; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Ds1t0 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_Ds1t0 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 207; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Ds1t1 = __pyx_PyFloat_AsDouble(values[6]); if (unlikely((__pyx_v_Ds1t1 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 208; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Dt0t0 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_Dt0t0 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 209; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Dt0t1 = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_Dt0t1 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 210; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_Dt1t1 = __pyx_PyFloat_AsDouble(values[9]); if (unlikely((__pyx_v_Dt1t1 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_mu1_tetface", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._mu1_tetface", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_6_mu1_tetface(__pyx_self, __pyx_v_Ds0s0, __pyx_v_Ds0s1, __pyx_v_Ds1s1, __pyx_v_Ds0t0, __pyx_v_Ds0t1, __pyx_v_Ds1t0, __pyx_v_Ds1t1, __pyx_v_Dt0t0, __pyx_v_Dt0t1, __pyx_v_Dt1t1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":202
 * 
 * @cython.cdivision(True)
 * cpdef double _mu1_tetface(double Ds0s0,             # <<<<<<<<<<<<<<
 *                           double Ds0s1,
 *                           double Ds1s1,
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_6_mu1_tetface(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_Ds0s0, double __pyx_v_Ds0s1, double __pyx_v_Ds1s1, double __pyx_v_Ds0t0, double __pyx_v_Ds0t1, double __pyx_v_Ds1t0, double __pyx_v_Ds1t1, double __pyx_v_Dt0t0, double __pyx_v_Dt0t1, double __pyx_v_Dt1t1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_mu1_tetface", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol__mu1_tetface(__pyx_v_Ds0s0, __pyx_v_Ds0s1, __pyx_v_Ds1s1, __pyx_v_Ds0t0, __pyx_v_Ds0t1, __pyx_v_Ds1t0, __pyx_v_Ds1t1, __pyx_v_Dt0t0, __pyx_v_Dt0t1, __pyx_v_Dt1t1, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 202; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._mu1_tetface", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":238
 * 
 * 
 * cpdef double mu2_tri(double D00, double D01, double D02,             # <<<<<<<<<<<<<<
 *                      double D11, double D12,
 *                      double D22) nogil:
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_9mu2_tri(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D22, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_C00;
  double __pyx_v_C01;
  double __pyx_v_C11;
  double __pyx_v_L;
  double __pyx_r;
  int __pyx_t_1;

  /* "nipy/algorithms/statistics/intvol.pyx":269
 *   """
 *   cdef double C00, C01, C11, L
 *   C00 = D11 - 2*D01 + D00             # <<<<<<<<<<<<<<
 *   C01 = D12 - D01 - D02 + D00
 *   C11 = D22 - 2*D02 + D00
 */
  __pyx_v_C00 = ((__pyx_v_D11 - (2.0 * __pyx_v_D01)) + __pyx_v_D00);

  /* "nipy/algorithms/statistics/intvol.pyx":270
 *   cdef double C00, C01, C11, L
 *   C00 = D11 - 2*D01 + D00
 *   C01 = D12 - D01 - D02 + D00             # <<<<<<<<<<<<<<
 *   C11 = D22 - 2*D02 + D00
 *   L = C00 * C11 - C01 * C01
 */
  __pyx_v_C01 = (((__pyx_v_D12 - __pyx_v_D01) - __pyx_v_D02) + __pyx_v_D00);

  /* "nipy/algorithms/statistics/intvol.pyx":271
 *   C00 = D11 - 2*D01 + D00
 *   C01 = D12 - D01 - D02 + D00
 *   C11 = D22 - 2*D02 + D00             # <<<<<<<<<<<<<<
 *   L = C00 * C11 - C01 * C01
 *   # Negative area appeared to result from floating point errors on PPC
 */
  __pyx_v_C11 = ((__pyx_v_D22 - (2.0 * __pyx_v_D02)) + __pyx_v_D00);

  /* "nipy/algorithms/statistics/intvol.pyx":272
 *   C01 = D12 - D01 - D02 + D00
 *   C11 = D22 - 2*D02 + D00
 *   L = C00 * C11 - C01 * C01             # <<<<<<<<<<<<<<
 *   # Negative area appeared to result from floating point errors on PPC
 *   if L < 0:
 */
  __pyx_v_L = ((__pyx_v_C00 * __pyx_v_C11) - (__pyx_v_C01 * __pyx_v_C01));

  /* "nipy/algorithms/statistics/intvol.pyx":274
 *   L = C00 * C11 - C01 * C01
 *   # Negative area appeared to result from floating point errors on PPC
 *   if L < 0:             # <<<<<<<<<<<<<<
 *       return 0.0
 *   return sqrt(L) * 0.5
 */
  __pyx_t_1 = (__pyx_v_L < 0.0);
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":275
 *   # Negative area appeared to result from floating point errors on PPC
 *   if L < 0:
 *       return 0.0             # <<<<<<<<<<<<<<
 *   return sqrt(L) * 0.5
 * 
 */
    __pyx_r = 0.0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":276
 *   if L < 0:
 *       return 0.0
 *   return sqrt(L) * 0.5             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (sqrt(__pyx_v_L) * 0.5);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_9mu2_tri(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_8mu2_tri[] = " Compute the 2nd intrinsic volume of triangle\n\n  2nd intrinsic volume (just area in this case) of a triangle with coordinates\n  implied by the dot products below.\n\n  Parameters\n  ----------\n  D00 : float\n    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is\n    ``cv0.dot(cv0)``\n  D01 : float\n    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second\n    vertex.\n  D02 : float\n    ``cv0.dot(cv2)``\n  D11 : float\n    ``cv1.dot(cv1)``\n  D12 : float\n    ``cv1.dot(cv2)``\n  D22 : float\n    ``cv2.dot(cv2)``\n\n  Returns\n  -------\n  mu2 : float\n    area of triangle\n  ";
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_9mu2_tri(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_D00;
  double __pyx_v_D01;
  double __pyx_v_D02;
  double __pyx_v_D11;
  double __pyx_v_D12;
  double __pyx_v_D22;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mu2_tri (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__D00,&__pyx_n_s__D01,&__pyx_n_s__D02,&__pyx_n_s__D11,&__pyx_n_s__D12,&__pyx_n_s__D22,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D00)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D01)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tri", 1, 6, 6, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D02)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tri", 1, 6, 6, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D11)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tri", 1, 6, 6, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tri", 1, 6, 6, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D22)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu2_tri", 1, 6, 6, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mu2_tri") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_D00 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_D00 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D01 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_D01 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D02 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_D02 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D11 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_D11 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D12 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_D12 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D22 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_D22 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 240; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mu2_tri", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu2_tri", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_8mu2_tri(__pyx_self, __pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D11, __pyx_v_D12, __pyx_v_D22);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":238
 * 
 * 
 * cpdef double mu2_tri(double D00, double D01, double D02,             # <<<<<<<<<<<<<<
 *                      double D11, double D12,
 *                      double D22) nogil:
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_8mu2_tri(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D22) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mu2_tri", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(__pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D11, __pyx_v_D12, __pyx_v_D22, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 238; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu2_tri", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":279
 * 
 * 
 * cpdef double mu1_tri(double D00, double D01, double D02,             # <<<<<<<<<<<<<<
 *                      double D11, double D12,
 *                      double D22) nogil:
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_11mu1_tri(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tri(double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D22, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_mu;
  double __pyx_r;

  /* "nipy/algorithms/statistics/intvol.pyx":309
 *     1/2 perimeter of triangle
 *   """
 *   cdef double mu = 0             # <<<<<<<<<<<<<<
 *   mu += mu1_edge(D00, D01, D11)
 *   mu += mu1_edge(D00, D02, D22)
 */
  __pyx_v_mu = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":310
 *   """
 *   cdef double mu = 0
 *   mu += mu1_edge(D00, D01, D11)             # <<<<<<<<<<<<<<
 *   mu += mu1_edge(D00, D02, D22)
 *   mu += mu1_edge(D11, D12, D22)
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(__pyx_v_D00, __pyx_v_D01, __pyx_v_D11, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":311
 *   cdef double mu = 0
 *   mu += mu1_edge(D00, D01, D11)
 *   mu += mu1_edge(D00, D02, D22)             # <<<<<<<<<<<<<<
 *   mu += mu1_edge(D11, D12, D22)
 *   return mu * 0.5
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(__pyx_v_D00, __pyx_v_D02, __pyx_v_D22, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":312
 *   mu += mu1_edge(D00, D01, D11)
 *   mu += mu1_edge(D00, D02, D22)
 *   mu += mu1_edge(D11, D12, D22)             # <<<<<<<<<<<<<<
 *   return mu * 0.5
 * 
 */
  __pyx_v_mu = (__pyx_v_mu + __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(__pyx_v_D11, __pyx_v_D12, __pyx_v_D22, 0));

  /* "nipy/algorithms/statistics/intvol.pyx":313
 *   mu += mu1_edge(D00, D02, D22)
 *   mu += mu1_edge(D11, D12, D22)
 *   return mu * 0.5             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_mu * 0.5);
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_11mu1_tri(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_10mu1_tri[] = " Compute the 1st intrinsic volume of triangle\n\n  1st intrinsic volume (1/2 the perimeter) of a triangle with coordinates\n  implied by the dot products below.\n\n  Parameters\n  ----------\n  D00 : float\n    If ``cv0`` is a 3-vector of coordinates for the first vertex, `D00` is\n    ``cv0.dot(cv0)``\n  D01 : float\n    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the second\n    vertex.\n  D02 : float\n    ``cv0.dot(cv2)``\n  D11 : float\n    ``cv1.dot(cv1)``\n  D12 : float\n    ``cv1.dot(cv2)``\n  D22 : float\n    ``cv2.dot(cv2)``\n\n  Returns\n  -------\n  mu1 : float\n    1/2 perimeter of triangle\n  ";
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_11mu1_tri(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_D00;
  double __pyx_v_D01;
  double __pyx_v_D02;
  double __pyx_v_D11;
  double __pyx_v_D12;
  double __pyx_v_D22;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mu1_tri (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__D00,&__pyx_n_s__D01,&__pyx_n_s__D02,&__pyx_n_s__D11,&__pyx_n_s__D12,&__pyx_n_s__D22,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D00)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D01)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tri", 1, 6, 6, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D02)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tri", 1, 6, 6, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D11)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tri", 1, 6, 6, 3); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D12)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tri", 1, 6, 6, 4); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  5:
        if (likely((values[5] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D22)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_tri", 1, 6, 6, 5); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mu1_tri") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_D00 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_D00 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D01 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_D01 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D02 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_D02 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D11 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_D11 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D12 = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_D12 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 280; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D22 = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_D22 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 281; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mu1_tri", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu1_tri", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_10mu1_tri(__pyx_self, __pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D11, __pyx_v_D12, __pyx_v_D22);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":279
 * 
 * 
 * cpdef double mu1_tri(double D00, double D01, double D02,             # <<<<<<<<<<<<<<
 *                      double D11, double D12,
 *                      double D22) nogil:
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_10mu1_tri(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D02, double __pyx_v_D11, double __pyx_v_D12, double __pyx_v_D22) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mu1_tri", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tri(__pyx_v_D00, __pyx_v_D01, __pyx_v_D02, __pyx_v_D11, __pyx_v_D12, __pyx_v_D22, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 279; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu1_tri", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":316
 * 
 * 
 * cpdef double mu1_edge(double D00, double D01, double D11) nogil:             # <<<<<<<<<<<<<<
 *   """ Compute the 1st intrinsic volume (length) of line segment
 * 
 */

static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_13mu1_edge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D11, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_r;

  /* "nipy/algorithms/statistics/intvol.pyx":337
 *     length of line segment
 *   """
 *   return sqrt(D00 - 2*D01 + D11)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = sqrt(((__pyx_v_D00 - (2.0 * __pyx_v_D01)) + __pyx_v_D11));
  goto __pyx_L0;

  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_13mu1_edge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_12mu1_edge[] = " Compute the 1st intrinsic volume (length) of line segment\n\n  Length of a line segment with vertex coordinates implied by dot products\n  below.\n\n  Parameters\n  ----------\n  D00 : float\n    If ``cv0`` is a 3-vector of coordinates for the line start, `D00` is\n    ``cv0.dot(cv0)``\n  D01 : float\n    ``cv0.dot(cv1)`` where ``cv1`` is the coordinates for the line end.\n  D11 : float\n    ``cv1.dot(cv1)``\n\n  Returns\n  -------\n  mu0 : float\n    length of line segment\n  ";
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_13mu1_edge(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_D00;
  double __pyx_v_D01;
  double __pyx_v_D11;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mu1_edge (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__D00,&__pyx_n_s__D01,&__pyx_n_s__D11,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D00)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D01)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_edge", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__D11)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mu1_edge", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mu1_edge") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_D00 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_D00 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D01 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_D01 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    __pyx_v_D11 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_D11 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mu1_edge", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu1_edge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_12mu1_edge(__pyx_self, __pyx_v_D00, __pyx_v_D01, __pyx_v_D11);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":316
 * 
 * 
 * cpdef double mu1_edge(double D00, double D01, double D11) nogil:             # <<<<<<<<<<<<<<
 *   """ Compute the 1st intrinsic volume (length) of line segment
 * 
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_12mu1_edge(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_D00, double __pyx_v_D01, double __pyx_v_D11) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mu1_edge", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(__pyx_v_D00, __pyx_v_D01, __pyx_v_D11, 0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 316; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.mu1_edge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_15EC3d(PyObject *__pyx_self, PyObject *__pyx_v_mask); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_14EC3d[] = " Compute Euler characteristic of region within `mask`\n\n    Given a 3d `mask`, compute the 0th intrinsic volume (Euler characteristic)\n    of the masked region. The region is broken up into tetrahedra / triangles /\n    edges / vertices, which are included based on whether all voxels in the\n    tetrahedron / triangle / edge / vertex are in the mask or not.\n\n    Parameters\n    ----------\n    mask : ndarray((i,j,k), np.int)\n         Binary mask determining whether or not a voxel is in the mask.\n\n    Returns\n    -------\n    mu0 : int\n        Euler characteristic\n\n    Notes\n    -----\n    The array mask is assumed to be binary. At the time of writing, it is not\n    clear how to get cython to use np.bool arrays.\n\n    The 3d cubes are triangulated into 6 tetrahedra of equal volume, as\n    described in the reference below.\n\n    References\n    ----------\n    Taylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n      with an application to brain mapping.\"\n      Journal of the American Statistical Association, 102(479):913-928.\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_15EC3d = {__Pyx_NAMESTR("EC3d"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_15EC3d, METH_O, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_14EC3d)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_15EC3d(PyObject *__pyx_self, PyObject *__pyx_v_mask) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EC3d (wrapper)", 0);
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_14EC3d(__pyx_self, ((PyObject *)__pyx_v_mask));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":340
 * 
 * 
 * def EC3d(mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic of region within `mask`
 * 
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_14EC3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mask) {
  PyArrayObject *__pyx_v_mask_c = 0;
  PyArrayObject *__pyx_v_fpmask = 0;
  PyArrayObject *__pyx_v_d2 = 0;
  PyArrayObject *__pyx_v_d3 = 0;
  PyArrayObject *__pyx_v_d4 = 0;
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_j;
  npy_intp __pyx_v_k;
  npy_intp __pyx_v_l;
  npy_intp __pyx_v_s0;
  npy_intp __pyx_v_s1;
  npy_intp __pyx_v_s2;
  npy_intp __pyx_v_ds2;
  npy_intp __pyx_v_ds3;
  npy_intp __pyx_v_ds4;
  npy_intp __pyx_v_index;
  npy_intp __pyx_v_m;
  CYTHON_UNUSED npy_intp __pyx_v_nvox;
  npy_intp __pyx_v_ss0;
  npy_intp __pyx_v_ss1;
  npy_intp __pyx_v_ss2;
  npy_intp __pyx_v_v0;
  npy_intp __pyx_v_v1;
  npy_intp __pyx_v_v2;
  npy_intp __pyx_v_v3;
  npy_intp __pyx_v_l0;
  PyObject *__pyx_v_pmask_shape = NULL;
  PyObject *__pyx_v_pmask = NULL;
  PyArrayObject *__pyx_v_strides = 0;
  PyObject *__pyx_v_union = NULL;
  PyObject *__pyx_v_c = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d2;
  __Pyx_Buffer __pyx_pybuffer_d2;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d3;
  __Pyx_Buffer __pyx_pybuffer_d3;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d4;
  __Pyx_Buffer __pyx_pybuffer_d4;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fpmask;
  __Pyx_Buffer __pyx_pybuffer_fpmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask_c;
  __Pyx_Buffer __pyx_pybuffer_mask_c;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_strides;
  __Pyx_Buffer __pyx_pybuffer_strides;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  npy_intp __pyx_t_11;
  npy_intp __pyx_t_12;
  npy_intp __pyx_t_13;
  PyArrayObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyArrayObject *__pyx_t_23 = NULL;
  PyArrayObject *__pyx_t_24 = NULL;
  PyArrayObject *__pyx_t_25 = NULL;
  long __pyx_t_26;
  long __pyx_t_27;
  long __pyx_t_28;
  long __pyx_t_29;
  long __pyx_t_30;
  long __pyx_t_31;
  npy_intp __pyx_t_32;
  npy_intp __pyx_t_33;
  npy_intp __pyx_t_34;
  long __pyx_t_35;
  npy_intp __pyx_t_36;
  npy_intp __pyx_t_37;
  long __pyx_t_38;
  npy_intp __pyx_t_39;
  long __pyx_t_40;
  npy_intp __pyx_t_41;
  long __pyx_t_42;
  npy_intp __pyx_t_43;
  npy_intp __pyx_t_44;
  npy_intp __pyx_t_45;
  npy_intp __pyx_t_46;
  long __pyx_t_47;
  npy_intp __pyx_t_48;
  npy_intp __pyx_t_49;
  long __pyx_t_50;
  npy_intp __pyx_t_51;
  long __pyx_t_52;
  npy_intp __pyx_t_53;
  npy_intp __pyx_t_54;
  npy_intp __pyx_t_55;
  long __pyx_t_56;
  npy_intp __pyx_t_57;
  npy_intp __pyx_t_58;
  long __pyx_t_59;
  npy_intp __pyx_t_60;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("EC3d", 0);
  __pyx_pybuffer_mask_c.pybuffer.buf = NULL;
  __pyx_pybuffer_mask_c.refcount = 0;
  __pyx_pybuffernd_mask_c.data = NULL;
  __pyx_pybuffernd_mask_c.rcbuffer = &__pyx_pybuffer_mask_c;
  __pyx_pybuffer_fpmask.pybuffer.buf = NULL;
  __pyx_pybuffer_fpmask.refcount = 0;
  __pyx_pybuffernd_fpmask.data = NULL;
  __pyx_pybuffernd_fpmask.rcbuffer = &__pyx_pybuffer_fpmask;
  __pyx_pybuffer_d2.pybuffer.buf = NULL;
  __pyx_pybuffer_d2.refcount = 0;
  __pyx_pybuffernd_d2.data = NULL;
  __pyx_pybuffernd_d2.rcbuffer = &__pyx_pybuffer_d2;
  __pyx_pybuffer_d3.pybuffer.buf = NULL;
  __pyx_pybuffer_d3.refcount = 0;
  __pyx_pybuffernd_d3.data = NULL;
  __pyx_pybuffernd_d3.rcbuffer = &__pyx_pybuffer_d3;
  __pyx_pybuffer_d4.pybuffer.buf = NULL;
  __pyx_pybuffer_d4.refcount = 0;
  __pyx_pybuffernd_d4.data = NULL;
  __pyx_pybuffernd_d4.rcbuffer = &__pyx_pybuffer_d4;
  __pyx_pybuffer_strides.pybuffer.buf = NULL;
  __pyx_pybuffer_strides.refcount = 0;
  __pyx_pybuffernd_strides.data = NULL;
  __pyx_pybuffernd_strides.rcbuffer = &__pyx_pybuffer_strides;

  /* "nipy/algorithms/statistics/intvol.pyx":372
 *       Journal of the American Statistical Association, 102(479):913-928.
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):             # <<<<<<<<<<<<<<
 *       raise ValueError('mask should be filled with 0/1 '
 *                        'values, but be of type np.int')
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__unique); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(((PyObject *)((PyObject*)(&PySet_Type))), ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__issubset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 372; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  if (__pyx_t_5) {

    /* "nipy/algorithms/statistics/intvol.pyx":373
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":389
 *         np.npy_intp ss0, ss1, ss2 # strides
 *         np.npy_intp v0, v1, v2, v3 # vertices
 *         np.npy_intp l0 = 0             # <<<<<<<<<<<<<<
 * 
 *     mask_c = mask
 */
  __pyx_v_l0 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":391
 *         np.npy_intp l0 = 0
 * 
 *     mask_c = mask             # <<<<<<<<<<<<<<
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 */
  if (!(likely(((__pyx_v_mask) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_mask, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = ((PyArrayObject *)__pyx_v_mask);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask_c, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    __pyx_pybuffernd_mask_c.diminfo[0].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask_c.diminfo[0].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask_c.diminfo[1].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask_c.diminfo[1].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask_c.diminfo[2].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask_c.diminfo[2].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_6 = 0;
  __Pyx_INCREF(__pyx_v_mask);
  __pyx_v_mask_c = ((PyArrayObject *)__pyx_v_mask);

  /* "nipy/algorithms/statistics/intvol.pyx":393
 *     mask_c = mask
 * 
 *     pmask_shape = np.array(mask.shape) + 1             # <<<<<<<<<<<<<<
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1,:-1] = mask_c
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_int_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 393; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_pmask_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":394
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)             # <<<<<<<<<<<<<<
 *     pmask[:-1,:-1,:-1] = mask_c
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 394; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 394; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 394; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__int); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 394; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 394; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 394; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_v_pmask = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":395
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 */
  if (PyObject_SetItem(__pyx_v_pmask, ((PyObject *)__pyx_k_tuple_6), ((PyObject *)__pyx_v_mask_c)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 395; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":397
 *     pmask[:-1,:-1,:-1] = mask_c
 * 
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])             # <<<<<<<<<<<<<<
 * 
 *     fpmask = pmask.reshape(-1)
 */
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_11 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__shape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_12 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_13 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_13 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 397; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s0 = __pyx_t_11;
  __pyx_v_s1 = __pyx_t_12;
  __pyx_v_s2 = __pyx_t_13;

  /* "nipy/algorithms/statistics/intvol.pyx":399
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[np.intp_t, ndim=1] strides
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__reshape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_k_tuple_7), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_14 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_fpmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
    }
    __pyx_pybuffernd_fpmask.diminfo[0].strides = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fpmask.diminfo[0].shape = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_14 = 0;
  __pyx_v_fpmask = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":402
 *     cdef:
 *         np.ndarray[np.intp_t, ndim=1] strides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)             # <<<<<<<<<<<<<<
 * 
 *     # First do the interior contributions.
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__strides_from); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__bool); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_15));
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_16 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__intp); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_15, ((PyObject *)__pyx_n_s__dtype), __pyx_t_16) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_15)); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_15)); __pyx_t_15 = 0;
  if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_16);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_v_strides, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    __pyx_pybuffernd_strides.diminfo[0].strides = __pyx_pybuffernd_strides.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_strides.diminfo[0].shape = __pyx_pybuffernd_strides.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 402; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_17 = 0;
  __pyx_v_strides = ((PyArrayObject *)__pyx_t_16);
  __pyx_t_16 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":407
 *     # We first figure out which vertices, edges, triangles, tetrahedra
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 */
  __pyx_t_16 = __Pyx_GetName(__pyx_m, __pyx_n_s__join_complexes); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_9));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_9));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_9));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_1 = PyObject_Call(__pyx_t_15, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":408
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 408; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 408; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_10));
  PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)__pyx_k_tuple_10));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_10));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_15, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_15), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 408; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_15)); __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":409
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 */
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_11));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_11));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_11));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_18 = PyObject_Call(__pyx_t_15, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":410
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 410; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 410; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_12));
  PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)__pyx_k_tuple_12));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_12));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_15, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_19 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_15), NULL); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 410; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_15)); __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":411
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 */
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_13));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_13));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_13));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_20 = PyObject_Call(__pyx_t_15, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":412
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_14));
  PyTuple_SET_ITEM(__pyx_t_15, 0, ((PyObject *)__pyx_k_tuple_14));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_14));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_15, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_21 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_15), NULL); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_15)); __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":413
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])             # <<<<<<<<<<<<<<
 *     c = cube_with_strides_center((0,0,0), strides)
 * 
 */
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_15));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_15));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_15));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_22 = PyObject_Call(__pyx_t_15, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(7); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_18);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_19);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_22);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_18 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  __pyx_t_22 = PySequence_Tuple(((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_22));
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_16, ((PyObject *)__pyx_t_22), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  __pyx_v_union = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":414
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)             # <<<<<<<<<<<<<<
 * 
 *     d4 = np.array(list(c[4].difference(union[4])))
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 414; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 414; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_16));
  PyTuple_SET_ITEM(__pyx_t_22, 0, ((PyObject *)__pyx_k_tuple_16));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_16));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_22, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_16 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_22), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 414; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  __pyx_v_c = __pyx_t_16;
  __pyx_t_16 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":416
 *     c = cube_with_strides_center((0,0,0), strides)
 * 
 *     d4 = np.array(list(c[4].difference(union[4])))             # <<<<<<<<<<<<<<
 *     d3 = np.array(list(c[3].difference(union[3])))
 *     d2 = np.array(list(c[2].difference(union[2])))
 */
  __pyx_t_16 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyObject_GetAttr(__pyx_t_16, __pyx_n_s__array); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_c, 4, sizeof(long), PyInt_FromLong); if (!__pyx_t_16) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_16, __pyx_n_s__difference); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_union, 4, sizeof(long), PyInt_FromLong); if (!__pyx_t_16) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_21 = PyTuple_New(1); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_21), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_21)); __pyx_t_21 = 0;
  __pyx_t_21 = PyTuple_New(1); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_21), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(((PyObject *)__pyx_t_21)); __pyx_t_21 = 0;
  __pyx_t_21 = PyTuple_New(1); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_22, ((PyObject *)__pyx_t_21), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_21)); __pyx_t_21 = 0;
  if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_23 = ((PyArrayObject *)__pyx_t_16);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d4.rcbuffer->pybuffer, (PyObject*)__pyx_t_23, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d4.rcbuffer->pybuffer, (PyObject*)__pyx_v_d4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
    }
    __pyx_pybuffernd_d4.diminfo[0].strides = __pyx_pybuffernd_d4.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d4.diminfo[0].shape = __pyx_pybuffernd_d4.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d4.diminfo[1].strides = __pyx_pybuffernd_d4.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d4.diminfo[1].shape = __pyx_pybuffernd_d4.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 416; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_23 = 0;
  __pyx_v_d4 = ((PyArrayObject *)__pyx_t_16);
  __pyx_t_16 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":417
 * 
 *     d4 = np.array(list(c[4].difference(union[4])))
 *     d3 = np.array(list(c[3].difference(union[3])))             # <<<<<<<<<<<<<<
 *     d2 = np.array(list(c[2].difference(union[2])))
 * 
 */
  __pyx_t_16 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_21 = PyObject_GetAttr(__pyx_t_16, __pyx_n_s__array); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_c, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_16) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyObject_GetAttr(__pyx_t_16, __pyx_n_s__difference); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_union, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_16) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_22, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_21, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_24 = ((PyArrayObject *)__pyx_t_16);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_t_24, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_v_d3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    __pyx_pybuffernd_d3.diminfo[0].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d3.diminfo[0].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d3.diminfo[1].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d3.diminfo[1].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 417; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_24 = 0;
  __pyx_v_d3 = ((PyArrayObject *)__pyx_t_16);
  __pyx_t_16 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":418
 *     d4 = np.array(list(c[4].difference(union[4])))
 *     d3 = np.array(list(c[3].difference(union[3])))
 *     d2 = np.array(list(c[2].difference(union[2])))             # <<<<<<<<<<<<<<
 * 
 *     ds2 = d2.shape[0]
 */
  __pyx_t_16 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_16, __pyx_n_s__array); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_c, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_16) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_21 = PyObject_GetAttr(__pyx_t_16, __pyx_n_s__difference); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_union, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_16) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyTuple_New(1); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_21, ((PyObject *)__pyx_t_22), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  __pyx_t_22 = PyTuple_New(1); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_22), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  __pyx_t_22 = PyTuple_New(1); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_22), NULL); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_22)); __pyx_t_22 = 0;
  if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_25 = ((PyArrayObject *)__pyx_t_16);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_t_25, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_v_d2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
    }
    __pyx_pybuffernd_d2.diminfo[0].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d2.diminfo[0].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d2.diminfo[1].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d2.diminfo[1].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_25 = 0;
  __pyx_v_d2 = ((PyArrayObject *)__pyx_t_16);
  __pyx_t_16 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":420
 *     d2 = np.array(list(c[2].difference(union[2])))
 * 
 *     ds2 = d2.shape[0]             # <<<<<<<<<<<<<<
 *     ds3 = d3.shape[0]
 *     ds4 = d4.shape[0]
 */
  __pyx_v_ds2 = (__pyx_v_d2->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":421
 * 
 *     ds2 = d2.shape[0]
 *     ds3 = d3.shape[0]             # <<<<<<<<<<<<<<
 *     ds4 = d4.shape[0]
 * 
 */
  __pyx_v_ds3 = (__pyx_v_d3->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":422
 *     ds2 = d2.shape[0]
 *     ds3 = d3.shape[0]
 *     ds4 = d4.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     ss0 = strides[0]
 */
  __pyx_v_ds4 = (__pyx_v_d4->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":424
 *     ds4 = d4.shape[0]
 * 
 *     ss0 = strides[0]             # <<<<<<<<<<<<<<
 *     ss1 = strides[1]
 *     ss2 = strides[2]
 */
  __pyx_t_26 = 0;
  __pyx_t_7 = -1;
  if (__pyx_t_26 < 0) {
    __pyx_t_26 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_26 < 0)) __pyx_t_7 = 0;
  } else if (unlikely(__pyx_t_26 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_7 = 0;
  if (unlikely(__pyx_t_7 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_7);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 424; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_v_ss0 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_strides.diminfo[0].strides));

  /* "nipy/algorithms/statistics/intvol.pyx":425
 * 
 *     ss0 = strides[0]
 *     ss1 = strides[1]             # <<<<<<<<<<<<<<
 *     ss2 = strides[2]
 * 
 */
  __pyx_t_27 = 1;
  __pyx_t_7 = -1;
  if (__pyx_t_27 < 0) {
    __pyx_t_27 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_27 < 0)) __pyx_t_7 = 0;
  } else if (unlikely(__pyx_t_27 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_7 = 0;
  if (unlikely(__pyx_t_7 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_7);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 425; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_v_ss1 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_strides.diminfo[0].strides));

  /* "nipy/algorithms/statistics/intvol.pyx":426
 *     ss0 = strides[0]
 *     ss1 = strides[1]
 *     ss2 = strides[2]             # <<<<<<<<<<<<<<
 * 
 *     nvox = mask.size
 */
  __pyx_t_28 = 2;
  __pyx_t_7 = -1;
  if (__pyx_t_28 < 0) {
    __pyx_t_28 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_28 < 0)) __pyx_t_7 = 0;
  } else if (unlikely(__pyx_t_28 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_7 = 0;
  if (unlikely(__pyx_t_7 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_7);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 426; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_v_ss2 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_strides.diminfo[0].strides));

  /* "nipy/algorithms/statistics/intvol.pyx":428
 *     ss2 = strides[2]
 * 
 *     nvox = mask.size             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0-1):
 */
  __pyx_t_16 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__size); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 428; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_13 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_16); if (unlikely((__pyx_t_13 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 428; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_v_nvox = __pyx_t_13;

  /* "nipy/algorithms/statistics/intvol.pyx":430
 *     nvox = mask.size
 * 
 *     for i in range(s0-1):             # <<<<<<<<<<<<<<
 *         for j in range(s1-1):
 *             for k in range(s2-1):
 */
  __pyx_t_29 = (__pyx_v_s0 - 1);
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_29; __pyx_t_13+=1) {
    __pyx_v_i = __pyx_t_13;

    /* "nipy/algorithms/statistics/intvol.pyx":431
 * 
 *     for i in range(s0-1):
 *         for j in range(s1-1):             # <<<<<<<<<<<<<<
 *             for k in range(s2-1):
 *                 index = i*ss0+j*ss1+k*ss2
 */
    __pyx_t_30 = (__pyx_v_s1 - 1);
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_30; __pyx_t_12+=1) {
      __pyx_v_j = __pyx_t_12;

      /* "nipy/algorithms/statistics/intvol.pyx":432
 *     for i in range(s0-1):
 *         for j in range(s1-1):
 *             for k in range(s2-1):             # <<<<<<<<<<<<<<
 *                 index = i*ss0+j*ss1+k*ss2
 *                 for l in range(ds4):
 */
      __pyx_t_31 = (__pyx_v_s2 - 1);
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_31; __pyx_t_11+=1) {
        __pyx_v_k = __pyx_t_11;

        /* "nipy/algorithms/statistics/intvol.pyx":433
 *         for j in range(s1-1):
 *             for k in range(s2-1):
 *                 index = i*ss0+j*ss1+k*ss2             # <<<<<<<<<<<<<<
 *                 for l in range(ds4):
 *                     v0 = index + d4[l,0]
 */
        __pyx_v_index = (((__pyx_v_i * __pyx_v_ss0) + (__pyx_v_j * __pyx_v_ss1)) + (__pyx_v_k * __pyx_v_ss2));

        /* "nipy/algorithms/statistics/intvol.pyx":434
 *             for k in range(s2-1):
 *                 index = i*ss0+j*ss1+k*ss2
 *                 for l in range(ds4):             # <<<<<<<<<<<<<<
 *                     v0 = index + d4[l,0]
 *                     m = fpmask[v0]
 */
        __pyx_t_32 = __pyx_v_ds4;
        for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
          __pyx_v_l = __pyx_t_33;

          /* "nipy/algorithms/statistics/intvol.pyx":435
 *                 index = i*ss0+j*ss1+k*ss2
 *                 for l in range(ds4):
 *                     v0 = index + d4[l,0]             # <<<<<<<<<<<<<<
 *                     m = fpmask[v0]
 *                     if m:
 */
          __pyx_t_34 = __pyx_v_l;
          __pyx_t_35 = 0;
          __pyx_t_7 = -1;
          if (__pyx_t_34 < 0) {
            __pyx_t_34 += __pyx_pybuffernd_d4.diminfo[0].shape;
            if (unlikely(__pyx_t_34 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_34 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_7 = 0;
          if (__pyx_t_35 < 0) {
            __pyx_t_35 += __pyx_pybuffernd_d4.diminfo[1].shape;
            if (unlikely(__pyx_t_35 < 0)) __pyx_t_7 = 1;
          } else if (unlikely(__pyx_t_35 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_7 = 1;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 435; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v0 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_35, __pyx_pybuffernd_d4.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":436
 *                 for l in range(ds4):
 *                     v0 = index + d4[l,0]
 *                     m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                     if m:
 *                         v1 = index + d4[l,1]
 */
          __pyx_t_36 = __pyx_v_v0;
          __pyx_t_7 = -1;
          if (__pyx_t_36 < 0) {
            __pyx_t_36 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_36 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_36 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 436; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_fpmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":437
 *                     v0 = index + d4[l,0]
 *                     m = fpmask[v0]
 *                     if m:             # <<<<<<<<<<<<<<
 *                         v1 = index + d4[l,1]
 *                         v2 = index + d4[l,2]
 */
          if (__pyx_v_m) {

            /* "nipy/algorithms/statistics/intvol.pyx":438
 *                     m = fpmask[v0]
 *                     if m:
 *                         v1 = index + d4[l,1]             # <<<<<<<<<<<<<<
 *                         v2 = index + d4[l,2]
 *                         v3 = index + d4[l,3]
 */
            __pyx_t_37 = __pyx_v_l;
            __pyx_t_38 = 1;
            __pyx_t_7 = -1;
            if (__pyx_t_37 < 0) {
              __pyx_t_37 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_37 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_37 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_7 = 0;
            if (__pyx_t_38 < 0) {
              __pyx_t_38 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_38 < 0)) __pyx_t_7 = 1;
            } else if (unlikely(__pyx_t_38 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_7 = 1;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 438; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v1 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_38, __pyx_pybuffernd_d4.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":439
 *                     if m:
 *                         v1 = index + d4[l,1]
 *                         v2 = index + d4[l,2]             # <<<<<<<<<<<<<<
 *                         v3 = index + d4[l,3]
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]
 */
            __pyx_t_39 = __pyx_v_l;
            __pyx_t_40 = 2;
            __pyx_t_7 = -1;
            if (__pyx_t_39 < 0) {
              __pyx_t_39 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_39 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_39 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_7 = 0;
            if (__pyx_t_40 < 0) {
              __pyx_t_40 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_40 < 0)) __pyx_t_7 = 1;
            } else if (unlikely(__pyx_t_40 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_7 = 1;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 439; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v2 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_39, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_40, __pyx_pybuffernd_d4.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":440
 *                         v1 = index + d4[l,1]
 *                         v2 = index + d4[l,2]
 *                         v3 = index + d4[l,3]             # <<<<<<<<<<<<<<
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]
 *                         l0 = l0 - m
 */
            __pyx_t_41 = __pyx_v_l;
            __pyx_t_42 = 3;
            __pyx_t_7 = -1;
            if (__pyx_t_41 < 0) {
              __pyx_t_41 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_41 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_41 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_7 = 0;
            if (__pyx_t_42 < 0) {
              __pyx_t_42 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_42 < 0)) __pyx_t_7 = 1;
            } else if (unlikely(__pyx_t_42 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_7 = 1;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 440; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v3 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_d4.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":441
 *                         v2 = index + d4[l,2]
 *                         v3 = index + d4[l,3]
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]             # <<<<<<<<<<<<<<
 *                         l0 = l0 - m
 * 
 */
            __pyx_t_43 = __pyx_v_v1;
            __pyx_t_7 = -1;
            if (__pyx_t_43 < 0) {
              __pyx_t_43 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_43 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_43 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_44 = __pyx_v_v2;
            __pyx_t_7 = -1;
            if (__pyx_t_44 < 0) {
              __pyx_t_44 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_44 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_44 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_45 = __pyx_v_v3;
            __pyx_t_7 = -1;
            if (__pyx_t_45 < 0) {
              __pyx_t_45 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_45 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_45 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 441; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_m = (((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_43, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_44, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_45, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":442
 *                         v3 = index + d4[l,3]
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]
 *                         l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *                 for l in range(ds3):
 */
            __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
            goto __pyx_L12;
          }
          __pyx_L12:;
        }

        /* "nipy/algorithms/statistics/intvol.pyx":444
 *                         l0 = l0 - m
 * 
 *                 for l in range(ds3):             # <<<<<<<<<<<<<<
 *                     v0 = index + d3[l,0]
 *                     m = fpmask[v0]
 */
        __pyx_t_32 = __pyx_v_ds3;
        for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
          __pyx_v_l = __pyx_t_33;

          /* "nipy/algorithms/statistics/intvol.pyx":445
 * 
 *                 for l in range(ds3):
 *                     v0 = index + d3[l,0]             # <<<<<<<<<<<<<<
 *                     m = fpmask[v0]
 *                     if m:
 */
          __pyx_t_46 = __pyx_v_l;
          __pyx_t_47 = 0;
          __pyx_t_7 = -1;
          if (__pyx_t_46 < 0) {
            __pyx_t_46 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_46 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_46 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_7 = 0;
          if (__pyx_t_47 < 0) {
            __pyx_t_47 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_47 < 0)) __pyx_t_7 = 1;
          } else if (unlikely(__pyx_t_47 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_7 = 1;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 445; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v0 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_46, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_47, __pyx_pybuffernd_d3.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":446
 *                 for l in range(ds3):
 *                     v0 = index + d3[l,0]
 *                     m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                     if m:
 *                         v1 = index + d3[l,1]
 */
          __pyx_t_48 = __pyx_v_v0;
          __pyx_t_7 = -1;
          if (__pyx_t_48 < 0) {
            __pyx_t_48 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_48 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_48 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_48, __pyx_pybuffernd_fpmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":447
 *                     v0 = index + d3[l,0]
 *                     m = fpmask[v0]
 *                     if m:             # <<<<<<<<<<<<<<
 *                         v1 = index + d3[l,1]
 *                         v2 = index + d3[l,2]
 */
          if (__pyx_v_m) {

            /* "nipy/algorithms/statistics/intvol.pyx":448
 *                     m = fpmask[v0]
 *                     if m:
 *                         v1 = index + d3[l,1]             # <<<<<<<<<<<<<<
 *                         v2 = index + d3[l,2]
 *                         m = m * fpmask[v1] * fpmask[v2]
 */
            __pyx_t_49 = __pyx_v_l;
            __pyx_t_50 = 1;
            __pyx_t_7 = -1;
            if (__pyx_t_49 < 0) {
              __pyx_t_49 += __pyx_pybuffernd_d3.diminfo[0].shape;
              if (unlikely(__pyx_t_49 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_49 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_7 = 0;
            if (__pyx_t_50 < 0) {
              __pyx_t_50 += __pyx_pybuffernd_d3.diminfo[1].shape;
              if (unlikely(__pyx_t_50 < 0)) __pyx_t_7 = 1;
            } else if (unlikely(__pyx_t_50 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_7 = 1;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 448; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v1 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_49, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_50, __pyx_pybuffernd_d3.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":449
 *                     if m:
 *                         v1 = index + d3[l,1]
 *                         v2 = index + d3[l,2]             # <<<<<<<<<<<<<<
 *                         m = m * fpmask[v1] * fpmask[v2]
 *                         l0 = l0 + m
 */
            __pyx_t_51 = __pyx_v_l;
            __pyx_t_52 = 2;
            __pyx_t_7 = -1;
            if (__pyx_t_51 < 0) {
              __pyx_t_51 += __pyx_pybuffernd_d3.diminfo[0].shape;
              if (unlikely(__pyx_t_51 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_51 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_7 = 0;
            if (__pyx_t_52 < 0) {
              __pyx_t_52 += __pyx_pybuffernd_d3.diminfo[1].shape;
              if (unlikely(__pyx_t_52 < 0)) __pyx_t_7 = 1;
            } else if (unlikely(__pyx_t_52 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_7 = 1;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 449; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v2 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_51, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_52, __pyx_pybuffernd_d3.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":450
 *                         v1 = index + d3[l,1]
 *                         v2 = index + d3[l,2]
 *                         m = m * fpmask[v1] * fpmask[v2]             # <<<<<<<<<<<<<<
 *                         l0 = l0 + m
 * 
 */
            __pyx_t_53 = __pyx_v_v1;
            __pyx_t_7 = -1;
            if (__pyx_t_53 < 0) {
              __pyx_t_53 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_53 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_53 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 450; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_54 = __pyx_v_v2;
            __pyx_t_7 = -1;
            if (__pyx_t_54 < 0) {
              __pyx_t_54 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_54 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_54 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 450; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_m = ((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_54, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":451
 *                         v2 = index + d3[l,2]
 *                         m = m * fpmask[v1] * fpmask[v2]
 *                         l0 = l0 + m             # <<<<<<<<<<<<<<
 * 
 *                 for l in range(ds2):
 */
            __pyx_v_l0 = (__pyx_v_l0 + __pyx_v_m);
            goto __pyx_L15;
          }
          __pyx_L15:;
        }

        /* "nipy/algorithms/statistics/intvol.pyx":453
 *                         l0 = l0 + m
 * 
 *                 for l in range(ds2):             # <<<<<<<<<<<<<<
 *                     v0 = index + d2[l,0]
 *                     m = fpmask[v0]
 */
        __pyx_t_32 = __pyx_v_ds2;
        for (__pyx_t_33 = 0; __pyx_t_33 < __pyx_t_32; __pyx_t_33+=1) {
          __pyx_v_l = __pyx_t_33;

          /* "nipy/algorithms/statistics/intvol.pyx":454
 * 
 *                 for l in range(ds2):
 *                     v0 = index + d2[l,0]             # <<<<<<<<<<<<<<
 *                     m = fpmask[v0]
 *                     if m:
 */
          __pyx_t_55 = __pyx_v_l;
          __pyx_t_56 = 0;
          __pyx_t_7 = -1;
          if (__pyx_t_55 < 0) {
            __pyx_t_55 += __pyx_pybuffernd_d2.diminfo[0].shape;
            if (unlikely(__pyx_t_55 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_55 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_7 = 0;
          if (__pyx_t_56 < 0) {
            __pyx_t_56 += __pyx_pybuffernd_d2.diminfo[1].shape;
            if (unlikely(__pyx_t_56 < 0)) __pyx_t_7 = 1;
          } else if (unlikely(__pyx_t_56 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_7 = 1;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 454; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v0 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_55, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_56, __pyx_pybuffernd_d2.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":455
 *                 for l in range(ds2):
 *                     v0 = index + d2[l,0]
 *                     m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                     if m:
 *                         v1 = index + d2[l,1]
 */
          __pyx_t_57 = __pyx_v_v0;
          __pyx_t_7 = -1;
          if (__pyx_t_57 < 0) {
            __pyx_t_57 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_57 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_57 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 455; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_57, __pyx_pybuffernd_fpmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":456
 *                     v0 = index + d2[l,0]
 *                     m = fpmask[v0]
 *                     if m:             # <<<<<<<<<<<<<<
 *                         v1 = index + d2[l,1]
 *                         m = m * fpmask[v1]
 */
          if (__pyx_v_m) {

            /* "nipy/algorithms/statistics/intvol.pyx":457
 *                     m = fpmask[v0]
 *                     if m:
 *                         v1 = index + d2[l,1]             # <<<<<<<<<<<<<<
 *                         m = m * fpmask[v1]
 *                         l0 = l0 - m
 */
            __pyx_t_58 = __pyx_v_l;
            __pyx_t_59 = 1;
            __pyx_t_7 = -1;
            if (__pyx_t_58 < 0) {
              __pyx_t_58 += __pyx_pybuffernd_d2.diminfo[0].shape;
              if (unlikely(__pyx_t_58 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_58 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_7 = 0;
            if (__pyx_t_59 < 0) {
              __pyx_t_59 += __pyx_pybuffernd_d2.diminfo[1].shape;
              if (unlikely(__pyx_t_59 < 0)) __pyx_t_7 = 1;
            } else if (unlikely(__pyx_t_59 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_7 = 1;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 457; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v1 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_58, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_59, __pyx_pybuffernd_d2.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":458
 *                     if m:
 *                         v1 = index + d2[l,1]
 *                         m = m * fpmask[v1]             # <<<<<<<<<<<<<<
 *                         l0 = l0 - m
 *     l0 += mask.sum()
 */
            __pyx_t_60 = __pyx_v_v1;
            __pyx_t_7 = -1;
            if (__pyx_t_60 < 0) {
              __pyx_t_60 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_60 < 0)) __pyx_t_7 = 0;
            } else if (unlikely(__pyx_t_60 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
            if (unlikely(__pyx_t_7 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_7);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 458; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_m = (__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_60, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":459
 *                         v1 = index + d2[l,1]
 *                         m = m * fpmask[v1]
 *                         l0 = l0 - m             # <<<<<<<<<<<<<<
 *     l0 += mask.sum()
 *     return l0
 */
            __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
            goto __pyx_L18;
          }
          __pyx_L18:;
        }
      }
    }
  }

  /* "nipy/algorithms/statistics/intvol.pyx":460
 *                         m = m * fpmask[v1]
 *                         l0 = l0 - m
 *     l0 += mask.sum()             # <<<<<<<<<<<<<<
 *     return l0
 * 
 */
  __pyx_t_16 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_l0); if (unlikely(!__pyx_t_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 460; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__sum); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 460; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_3 = PyObject_Call(__pyx_t_22, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 460; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = PyNumber_InPlaceAdd(__pyx_t_16, __pyx_t_3); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 460; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_13 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_22); if (unlikely((__pyx_t_13 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 460; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_v_l0 = __pyx_t_13;

  /* "nipy/algorithms/statistics/intvol.pyx":461
 *                         l0 = l0 - m
 *     l0 += mask.sum()
 *     return l0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_22 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_l0); if (unlikely(!__pyx_t_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 461; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_r = __pyx_t_22;
  __pyx_t_22 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.EC3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask_c);
  __Pyx_XDECREF((PyObject *)__pyx_v_fpmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_d2);
  __Pyx_XDECREF((PyObject *)__pyx_v_d3);
  __Pyx_XDECREF((PyObject *)__pyx_v_d4);
  __Pyx_XDECREF(__pyx_v_pmask_shape);
  __Pyx_XDECREF(__pyx_v_pmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_strides);
  __Pyx_XDECREF(__pyx_v_union);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_17Lips3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_16Lips3d[] = " Estimated intrinsic volumes within masked region given coordinates\n\n    Given a 3d `mask` and coordinates `coords`, estimate the intrinsic volumes\n    of the masked region. The region is broken up into tetrahedra / triangles /\n    edges / vertices, which are included based on whether all voxels in the\n    tetrahedron / triangle / edge / vertex are in the mask or not.\n\n    Parameters\n    ----------\n    coords : ndarray((N,i,j,k))\n         Coordinates for the voxels in the mask. ``N`` will often be 3 (for 3\n         dimensional coordinates, but can be any integer > 0\n    mask : ndarray((i,j,k), np.int)\n         Binary mask determining whether or not\n         a voxel is in the mask.\n\n    Returns\n    -------\n    mu : ndarray\n        Array of intrinsic volumes [mu0, mu1, mu2, mu3], being, respectively:\n        #. Euler characteristic\n        #. 2 * mean caliper diameter\n        #. 0.5 * surface area\n        #. Volume.\n\n    Notes\n    -----\n    The array mask is assumed to be binary. At the time of writing, it\n    is not clear how to get cython to use np.bool arrays.\n\n    The 3d cubes are triangulated into 6 tetrahedra of equal volume, as\n    described in the reference below.\n\n    References\n    ----------\n    Taylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n      with an application to brain mapping.\"\n      Journal of the American Statistical Association, 102(479):913-928.\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_17Lips3d = {__Pyx_NAMESTR("Lips3d"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_17Lips3d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_16Lips3d)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_17Lips3d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_coords = 0;
  PyObject *__pyx_v_mask = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Lips3d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__coords,&__pyx_n_s__mask,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__coords)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("Lips3d", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "Lips3d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coords = values[0];
    __pyx_v_mask = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Lips3d", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.Lips3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_16Lips3d(__pyx_self, __pyx_v_coords, __pyx_v_mask);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":464
 * 
 * 
 * def Lips3d(coords, mask):             # <<<<<<<<<<<<<<
 *     """ Estimated intrinsic volumes within masked region given coordinates
 * 
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_16Lips3d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords, PyObject *__pyx_v_mask) {
  PyObject *__pyx_v_value = NULL;
  PyArrayObject *__pyx_v_coords_c = 0;
  PyArrayObject *__pyx_v_mask_c = 0;
  PyArrayObject *__pyx_v_fcoords = 0;
  PyArrayObject *__pyx_v_D = 0;
  PyArrayObject *__pyx_v_fmask = 0;
  PyArrayObject *__pyx_v_fpmask = 0;
  PyArrayObject *__pyx_v_pmask = 0;
  PyArrayObject *__pyx_v_d4 = 0;
  PyArrayObject *__pyx_v_m4 = 0;
  PyArrayObject *__pyx_v_d3 = 0;
  PyArrayObject *__pyx_v_m3 = 0;
  PyArrayObject *__pyx_v_d2 = 0;
  PyArrayObject *__pyx_v_m2 = 0;
  PyArrayObject *__pyx_v_cvertices = 0;
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_j;
  npy_intp __pyx_v_k;
  npy_intp __pyx_v_l;
  npy_intp __pyx_v_s0;
  npy_intp __pyx_v_s1;
  npy_intp __pyx_v_s2;
  npy_intp __pyx_v_ds4;
  npy_intp __pyx_v_ds3;
  npy_intp __pyx_v_ds2;
  npy_intp __pyx_v_index;
  npy_intp __pyx_v_pindex;
  npy_intp __pyx_v_m;
  npy_intp __pyx_v_nvox;
  npy_intp __pyx_v_r;
  npy_intp __pyx_v_s;
  npy_intp __pyx_v_rr;
  npy_intp __pyx_v_ss;
  npy_intp __pyx_v_mr;
  npy_intp __pyx_v_ms;
  npy_intp __pyx_v_ss0;
  npy_intp __pyx_v_ss1;
  npy_intp __pyx_v_ss2;
  npy_intp __pyx_v_v0;
  npy_intp __pyx_v_v1;
  npy_intp __pyx_v_v2;
  npy_intp __pyx_v_v3;
  npy_intp __pyx_v_w0;
  npy_intp __pyx_v_w1;
  npy_intp __pyx_v_w2;
  npy_intp __pyx_v_w3;
  double __pyx_v_l0;
  double __pyx_v_l1;
  double __pyx_v_l2;
  double __pyx_v_l3;
  double __pyx_v_res;
  PyObject *__pyx_v_pmask_shape = NULL;
  PyArrayObject *__pyx_v_strides = 0;
  PyArrayObject *__pyx_v_dstrides = 0;
  PyObject *__pyx_v_ss0d = NULL;
  PyObject *__pyx_v_ss1d = NULL;
  PyObject *__pyx_v_ss2d = NULL;
  PyObject *__pyx_v_verts = NULL;
  PyObject *__pyx_v_union = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_v = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_D;
  __Pyx_Buffer __pyx_pybuffer_D;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_coords_c;
  __Pyx_Buffer __pyx_pybuffer_coords_c;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_cvertices;
  __Pyx_Buffer __pyx_pybuffer_cvertices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d2;
  __Pyx_Buffer __pyx_pybuffer_d2;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d3;
  __Pyx_Buffer __pyx_pybuffer_d3;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d4;
  __Pyx_Buffer __pyx_pybuffer_d4;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dstrides;
  __Pyx_Buffer __pyx_pybuffer_dstrides;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fcoords;
  __Pyx_Buffer __pyx_pybuffer_fcoords;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fmask;
  __Pyx_Buffer __pyx_pybuffer_fmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fpmask;
  __Pyx_Buffer __pyx_pybuffer_fpmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_m2;
  __Pyx_Buffer __pyx_pybuffer_m2;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_m3;
  __Pyx_Buffer __pyx_pybuffer_m3;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_m4;
  __Pyx_Buffer __pyx_pybuffer_m4;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask_c;
  __Pyx_Buffer __pyx_pybuffer_mask_c;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pmask;
  __Pyx_Buffer __pyx_pybuffer_pmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_strides;
  __Pyx_Buffer __pyx_pybuffer_strides;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyArrayObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  PyArrayObject *__pyx_t_13 = NULL;
  npy_intp __pyx_t_14;
  npy_intp __pyx_t_15;
  npy_intp __pyx_t_16;
  PyArrayObject *__pyx_t_17 = NULL;
  PyArrayObject *__pyx_t_18 = NULL;
  PyArrayObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyArrayObject *__pyx_t_21 = NULL;
  PyArrayObject *__pyx_t_22 = NULL;
  long __pyx_t_23;
  __pyx_t_5numpy_intp_t __pyx_t_24;
  long __pyx_t_25;
  __pyx_t_5numpy_intp_t __pyx_t_26;
  long __pyx_t_27;
  __pyx_t_5numpy_intp_t __pyx_t_28;
  long __pyx_t_29;
  long __pyx_t_30;
  long __pyx_t_31;
  int __pyx_t_32;
  PyArrayObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  PyObject *__pyx_t_36 = NULL;
  PyObject *__pyx_t_37 = NULL;
  PyObject *__pyx_t_38 = NULL;
  PyArrayObject *__pyx_t_39 = NULL;
  PyArrayObject *__pyx_t_40 = NULL;
  PyArrayObject *__pyx_t_41 = NULL;
  Py_ssize_t __pyx_t_42;
  PyObject *(*__pyx_t_43)(PyObject *);
  PyArrayObject *__pyx_t_44 = NULL;
  PyArrayObject *__pyx_t_45 = NULL;
  PyArrayObject *__pyx_t_46 = NULL;
  PyArrayObject *__pyx_t_47 = NULL;
  long __pyx_t_48;
  long __pyx_t_49;
  long __pyx_t_50;
  npy_intp __pyx_t_51;
  npy_intp __pyx_t_52;
  npy_intp __pyx_t_53;
  long __pyx_t_54;
  npy_intp __pyx_t_55;
  npy_intp __pyx_t_56;
  npy_intp __pyx_t_57;
  npy_intp __pyx_t_58;
  npy_intp __pyx_t_59;
  npy_intp __pyx_t_60;
  npy_intp __pyx_t_61;
  npy_intp __pyx_t_62;
  npy_intp __pyx_t_63;
  npy_intp __pyx_t_64;
  npy_intp __pyx_t_65;
  npy_intp __pyx_t_66;
  npy_intp __pyx_t_67;
  npy_intp __pyx_t_68;
  npy_intp __pyx_t_69;
  npy_intp __pyx_t_70;
  npy_intp __pyx_t_71;
  long __pyx_t_72;
  npy_intp __pyx_t_73;
  npy_intp __pyx_t_74;
  long __pyx_t_75;
  npy_intp __pyx_t_76;
  long __pyx_t_77;
  npy_intp __pyx_t_78;
  long __pyx_t_79;
  npy_intp __pyx_t_80;
  long __pyx_t_81;
  npy_intp __pyx_t_82;
  long __pyx_t_83;
  npy_intp __pyx_t_84;
  long __pyx_t_85;
  npy_intp __pyx_t_86;
  npy_intp __pyx_t_87;
  npy_intp __pyx_t_88;
  npy_intp __pyx_t_89;
  npy_intp __pyx_t_90;
  npy_intp __pyx_t_91;
  npy_intp __pyx_t_92;
  npy_intp __pyx_t_93;
  npy_intp __pyx_t_94;
  npy_intp __pyx_t_95;
  npy_intp __pyx_t_96;
  npy_intp __pyx_t_97;
  npy_intp __pyx_t_98;
  npy_intp __pyx_t_99;
  npy_intp __pyx_t_100;
  npy_intp __pyx_t_101;
  npy_intp __pyx_t_102;
  npy_intp __pyx_t_103;
  npy_intp __pyx_t_104;
  npy_intp __pyx_t_105;
  npy_intp __pyx_t_106;
  npy_intp __pyx_t_107;
  npy_intp __pyx_t_108;
  npy_intp __pyx_t_109;
  npy_intp __pyx_t_110;
  npy_intp __pyx_t_111;
  npy_intp __pyx_t_112;
  npy_intp __pyx_t_113;
  npy_intp __pyx_t_114;
  npy_intp __pyx_t_115;
  npy_intp __pyx_t_116;
  npy_intp __pyx_t_117;
  npy_intp __pyx_t_118;
  npy_intp __pyx_t_119;
  npy_intp __pyx_t_120;
  npy_intp __pyx_t_121;
  npy_intp __pyx_t_122;
  npy_intp __pyx_t_123;
  npy_intp __pyx_t_124;
  npy_intp __pyx_t_125;
  npy_intp __pyx_t_126;
  npy_intp __pyx_t_127;
  npy_intp __pyx_t_128;
  npy_intp __pyx_t_129;
  npy_intp __pyx_t_130;
  npy_intp __pyx_t_131;
  npy_intp __pyx_t_132;
  npy_intp __pyx_t_133;
  npy_intp __pyx_t_134;
  npy_intp __pyx_t_135;
  npy_intp __pyx_t_136;
  npy_intp __pyx_t_137;
  npy_intp __pyx_t_138;
  npy_intp __pyx_t_139;
  npy_intp __pyx_t_140;
  npy_intp __pyx_t_141;
  npy_intp __pyx_t_142;
  npy_intp __pyx_t_143;
  npy_intp __pyx_t_144;
  npy_intp __pyx_t_145;
  npy_intp __pyx_t_146;
  npy_intp __pyx_t_147;
  npy_intp __pyx_t_148;
  npy_intp __pyx_t_149;
  long __pyx_t_150;
  npy_intp __pyx_t_151;
  long __pyx_t_152;
  npy_intp __pyx_t_153;
  npy_intp __pyx_t_154;
  long __pyx_t_155;
  npy_intp __pyx_t_156;
  long __pyx_t_157;
  npy_intp __pyx_t_158;
  long __pyx_t_159;
  npy_intp __pyx_t_160;
  long __pyx_t_161;
  npy_intp __pyx_t_162;
  npy_intp __pyx_t_163;
  npy_intp __pyx_t_164;
  npy_intp __pyx_t_165;
  npy_intp __pyx_t_166;
  npy_intp __pyx_t_167;
  npy_intp __pyx_t_168;
  npy_intp __pyx_t_169;
  npy_intp __pyx_t_170;
  npy_intp __pyx_t_171;
  npy_intp __pyx_t_172;
  npy_intp __pyx_t_173;
  npy_intp __pyx_t_174;
  npy_intp __pyx_t_175;
  npy_intp __pyx_t_176;
  npy_intp __pyx_t_177;
  npy_intp __pyx_t_178;
  npy_intp __pyx_t_179;
  npy_intp __pyx_t_180;
  npy_intp __pyx_t_181;
  npy_intp __pyx_t_182;
  npy_intp __pyx_t_183;
  npy_intp __pyx_t_184;
  npy_intp __pyx_t_185;
  npy_intp __pyx_t_186;
  npy_intp __pyx_t_187;
  npy_intp __pyx_t_188;
  long __pyx_t_189;
  npy_intp __pyx_t_190;
  long __pyx_t_191;
  npy_intp __pyx_t_192;
  npy_intp __pyx_t_193;
  long __pyx_t_194;
  npy_intp __pyx_t_195;
  long __pyx_t_196;
  npy_intp __pyx_t_197;
  npy_intp __pyx_t_198;
  npy_intp __pyx_t_199;
  npy_intp __pyx_t_200;
  npy_intp __pyx_t_201;
  npy_intp __pyx_t_202;
  npy_intp __pyx_t_203;
  double __pyx_t_204;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Lips3d", 0);
  __Pyx_INCREF(__pyx_v_coords);
  __Pyx_INCREF(__pyx_v_mask);
  __pyx_pybuffer_coords_c.pybuffer.buf = NULL;
  __pyx_pybuffer_coords_c.refcount = 0;
  __pyx_pybuffernd_coords_c.data = NULL;
  __pyx_pybuffernd_coords_c.rcbuffer = &__pyx_pybuffer_coords_c;
  __pyx_pybuffer_mask_c.pybuffer.buf = NULL;
  __pyx_pybuffer_mask_c.refcount = 0;
  __pyx_pybuffernd_mask_c.data = NULL;
  __pyx_pybuffernd_mask_c.rcbuffer = &__pyx_pybuffer_mask_c;
  __pyx_pybuffer_fcoords.pybuffer.buf = NULL;
  __pyx_pybuffer_fcoords.refcount = 0;
  __pyx_pybuffernd_fcoords.data = NULL;
  __pyx_pybuffernd_fcoords.rcbuffer = &__pyx_pybuffer_fcoords;
  __pyx_pybuffer_D.pybuffer.buf = NULL;
  __pyx_pybuffer_D.refcount = 0;
  __pyx_pybuffernd_D.data = NULL;
  __pyx_pybuffernd_D.rcbuffer = &__pyx_pybuffer_D;
  __pyx_pybuffer_fmask.pybuffer.buf = NULL;
  __pyx_pybuffer_fmask.refcount = 0;
  __pyx_pybuffernd_fmask.data = NULL;
  __pyx_pybuffernd_fmask.rcbuffer = &__pyx_pybuffer_fmask;
  __pyx_pybuffer_fpmask.pybuffer.buf = NULL;
  __pyx_pybuffer_fpmask.refcount = 0;
  __pyx_pybuffernd_fpmask.data = NULL;
  __pyx_pybuffernd_fpmask.rcbuffer = &__pyx_pybuffer_fpmask;
  __pyx_pybuffer_pmask.pybuffer.buf = NULL;
  __pyx_pybuffer_pmask.refcount = 0;
  __pyx_pybuffernd_pmask.data = NULL;
  __pyx_pybuffernd_pmask.rcbuffer = &__pyx_pybuffer_pmask;
  __pyx_pybuffer_d4.pybuffer.buf = NULL;
  __pyx_pybuffer_d4.refcount = 0;
  __pyx_pybuffernd_d4.data = NULL;
  __pyx_pybuffernd_d4.rcbuffer = &__pyx_pybuffer_d4;
  __pyx_pybuffer_m4.pybuffer.buf = NULL;
  __pyx_pybuffer_m4.refcount = 0;
  __pyx_pybuffernd_m4.data = NULL;
  __pyx_pybuffernd_m4.rcbuffer = &__pyx_pybuffer_m4;
  __pyx_pybuffer_d3.pybuffer.buf = NULL;
  __pyx_pybuffer_d3.refcount = 0;
  __pyx_pybuffernd_d3.data = NULL;
  __pyx_pybuffernd_d3.rcbuffer = &__pyx_pybuffer_d3;
  __pyx_pybuffer_m3.pybuffer.buf = NULL;
  __pyx_pybuffer_m3.refcount = 0;
  __pyx_pybuffernd_m3.data = NULL;
  __pyx_pybuffernd_m3.rcbuffer = &__pyx_pybuffer_m3;
  __pyx_pybuffer_d2.pybuffer.buf = NULL;
  __pyx_pybuffer_d2.refcount = 0;
  __pyx_pybuffernd_d2.data = NULL;
  __pyx_pybuffernd_d2.rcbuffer = &__pyx_pybuffer_d2;
  __pyx_pybuffer_m2.pybuffer.buf = NULL;
  __pyx_pybuffer_m2.refcount = 0;
  __pyx_pybuffernd_m2.data = NULL;
  __pyx_pybuffernd_m2.rcbuffer = &__pyx_pybuffer_m2;
  __pyx_pybuffer_cvertices.pybuffer.buf = NULL;
  __pyx_pybuffer_cvertices.refcount = 0;
  __pyx_pybuffernd_cvertices.data = NULL;
  __pyx_pybuffernd_cvertices.rcbuffer = &__pyx_pybuffer_cvertices;
  __pyx_pybuffer_strides.pybuffer.buf = NULL;
  __pyx_pybuffer_strides.refcount = 0;
  __pyx_pybuffernd_strides.data = NULL;
  __pyx_pybuffernd_strides.rcbuffer = &__pyx_pybuffer_strides;
  __pyx_pybuffer_dstrides.pybuffer.buf = NULL;
  __pyx_pybuffer_dstrides.refcount = 0;
  __pyx_pybuffernd_dstrides.data = NULL;
  __pyx_pybuffernd_dstrides.rcbuffer = &__pyx_pybuffer_dstrides;

  /* "nipy/algorithms/statistics/intvol.pyx":504
 *       Journal of the American Statistical Association, 102(479):913-928.
 *     """
 *     if mask.shape != coords.shape[1:]:             # <<<<<<<<<<<<<<
 *         raise ValueError('shape of mask does not match coordinates')
 *     # if the data can be squeezed, we must use the lower dimensional function
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 504; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 504; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PySequence_GetSlice(__pyx_t_2, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 504; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 504; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 504; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "nipy/algorithms/statistics/intvol.pyx":505
 *     """
 *     if mask.shape != coords.shape[1:]:
 *         raise ValueError('shape of mask does not match coordinates')             # <<<<<<<<<<<<<<
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_18), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 505; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 505; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":507
 *         raise ValueError('shape of mask does not match coordinates')
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)             # <<<<<<<<<<<<<<
 *     if mask.ndim < 3:
 *         value = np.zeros(4)
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 507; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__squeeze); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 507; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 507; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  __pyx_t_1 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 507; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_v_mask);
  __pyx_v_mask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":508
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)
 *     if mask.ndim < 3:             # <<<<<<<<<<<<<<
 *         value = np.zeros(4)
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__ndim); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 508; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_3, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 508; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 508; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "nipy/algorithms/statistics/intvol.pyx":509
 *     mask = np.squeeze(mask)
 *     if mask.ndim < 3:
 *         value = np.zeros(4)             # <<<<<<<<<<<<<<
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         if mask.ndim == 2:
 */
    __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 509; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__zeros); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 509; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_k_tuple_19), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 509; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_value = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "nipy/algorithms/statistics/intvol.pyx":510
 *     if mask.ndim < 3:
 *         value = np.zeros(4)
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)             # <<<<<<<<<<<<<<
 *         if mask.ndim == 2:
 *             value[:3] = Lips2d(coords, mask)
 */
    __pyx_t_2 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__shape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_3) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Add(((PyObject *)__pyx_t_1), __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 510; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_v_coords);
    __pyx_v_coords = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "nipy/algorithms/statistics/intvol.pyx":511
 *         value = np.zeros(4)
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         if mask.ndim == 2:             # <<<<<<<<<<<<<<
 *             value[:3] = Lips2d(coords, mask)
 *         elif mask.ndim == 1:
 */
    __pyx_t_5 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__ndim); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 511; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_int_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 511; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 511; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_4) {

      /* "nipy/algorithms/statistics/intvol.pyx":512
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         if mask.ndim == 2:
 *             value[:3] = Lips2d(coords, mask)             # <<<<<<<<<<<<<<
 *         elif mask.ndim == 1:
 *             value[:2] = Lips1d(coords, mask)
 */
      __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__Lips2d); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 512; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 512; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_coords);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_coords);
      __Pyx_GIVEREF(__pyx_v_coords);
      __Pyx_INCREF(__pyx_v_mask);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_mask);
      __Pyx_GIVEREF(__pyx_v_mask);
      __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 512; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
      if (__Pyx_PySequence_SetSlice(__pyx_v_value, 0, 3, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 512; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5;
    }

    /* "nipy/algorithms/statistics/intvol.pyx":513
 *         if mask.ndim == 2:
 *             value[:3] = Lips2d(coords, mask)
 *         elif mask.ndim == 1:             # <<<<<<<<<<<<<<
 *             value[:2] = Lips1d(coords, mask)
 *         return value
 */
    __pyx_t_2 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__ndim); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 513; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_2, __pyx_int_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 513; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 513; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_4) {

      /* "nipy/algorithms/statistics/intvol.pyx":514
 *             value[:3] = Lips2d(coords, mask)
 *         elif mask.ndim == 1:
 *             value[:2] = Lips1d(coords, mask)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
      __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__Lips1d); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 514; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 514; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_coords);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_coords);
      __Pyx_GIVEREF(__pyx_v_coords);
      __Pyx_INCREF(__pyx_v_mask);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_mask);
      __Pyx_GIVEREF(__pyx_v_mask);
      __pyx_t_3 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 514; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
      if (__Pyx_PySequence_SetSlice(__pyx_v_value, 0, 2, __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 514; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L5;
    }
    __pyx_L5:;

    /* "nipy/algorithms/statistics/intvol.pyx":515
 *         elif mask.ndim == 1:
 *             value[:2] = Lips1d(coords, mask)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "nipy/algorithms/statistics/intvol.pyx":517
 *         return value
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):             # <<<<<<<<<<<<<<
 *       raise ValueError('mask should be filled with 0/1 '
 *                        'values, but be of type np.int')
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__unique); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(((PyObject *)((PyObject*)(&PySet_Type))), ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__issubset); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_5));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 517; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = (!__pyx_t_4);
  if (__pyx_t_6) {

    /* "nipy/algorithms/statistics/intvol.pyx":518
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_20), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 518; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 518; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "nipy/algorithms/statistics/intvol.pyx":549
 *         double res
 * 
 *     coords_c = coords             # <<<<<<<<<<<<<<
 *     mask_c = mask
 *     l0 = 0; l1 = 0; l2 = 0; l3 = 0
 */
  if (!(likely(((__pyx_v_coords) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_coords, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 549; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_7 = ((PyArrayObject *)__pyx_v_coords);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 4, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer, (PyObject*)__pyx_v_coords_c, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 4, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_coords_c.diminfo[0].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_coords_c.diminfo[0].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_coords_c.diminfo[1].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_coords_c.diminfo[1].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_coords_c.diminfo[2].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_coords_c.diminfo[2].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[2]; __pyx_pybuffernd_coords_c.diminfo[3].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[3]; __pyx_pybuffernd_coords_c.diminfo[3].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[3];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 549; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_7 = 0;
  __Pyx_INCREF(__pyx_v_coords);
  __pyx_v_coords_c = ((PyArrayObject *)__pyx_v_coords);

  /* "nipy/algorithms/statistics/intvol.pyx":550
 * 
 *     coords_c = coords
 *     mask_c = mask             # <<<<<<<<<<<<<<
 *     l0 = 0; l1 = 0; l2 = 0; l3 = 0
 * 
 */
  if (!(likely(((__pyx_v_mask) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_mask, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 550; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_12 = ((PyArrayObject *)__pyx_v_mask);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask_c, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_mask_c.diminfo[0].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask_c.diminfo[0].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask_c.diminfo[1].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask_c.diminfo[1].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_mask_c.diminfo[2].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_mask_c.diminfo[2].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 550; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_12 = 0;
  __Pyx_INCREF(__pyx_v_mask);
  __pyx_v_mask_c = ((PyArrayObject *)__pyx_v_mask);

  /* "nipy/algorithms/statistics/intvol.pyx":551
 *     coords_c = coords
 *     mask_c = mask
 *     l0 = 0; l1 = 0; l2 = 0; l3 = 0             # <<<<<<<<<<<<<<
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 */
  __pyx_v_l0 = 0.0;
  __pyx_v_l1 = 0.0;
  __pyx_v_l2 = 0.0;
  __pyx_v_l3 = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":553
 *     l0 = 0; l1 = 0; l2 = 0; l3 = 0
 * 
 *     pmask_shape = np.array(mask.shape) + 1             # <<<<<<<<<<<<<<
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1,:-1] = mask_c
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 553; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 553; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 553; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 553; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 553; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_5, __pyx_int_1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 553; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_pmask_shape = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":554
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)             # <<<<<<<<<<<<<<
 *     pmask[:-1,:-1,:-1] = mask_c
 * 
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__zeros); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__int); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_pmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_pmask.diminfo[0].strides = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pmask.diminfo[0].shape = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pmask.diminfo[1].strides = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pmask.diminfo[1].shape = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_pmask.diminfo[2].strides = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_pmask.diminfo[2].shape = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_13 = 0;
  __pyx_v_pmask = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":555
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 */
  if (PyObject_SetItem(((PyObject *)__pyx_v_pmask), ((PyObject *)__pyx_k_tuple_24), ((PyObject *)__pyx_v_mask_c)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 555; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":557
 *     pmask[:-1,:-1,:-1] = mask_c
 * 
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])             # <<<<<<<<<<<<<<
 * 
 *     fpmask = pmask.reshape(-1)
 */
  __pyx_t_14 = (__pyx_v_pmask->dimensions[0]);
  __pyx_t_15 = (__pyx_v_pmask->dimensions[1]);
  __pyx_t_16 = (__pyx_v_pmask->dimensions[2]);
  __pyx_v_s0 = __pyx_t_14;
  __pyx_v_s1 = __pyx_t_15;
  __pyx_v_s2 = __pyx_t_16;

  /* "nipy/algorithms/statistics/intvol.pyx":559
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 *     fmask = mask_c.reshape(-1)
 *     fcoords = coords_c.reshape((coords_c.shape[0], -1))
 */
  __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_v_pmask), __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 559; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_25), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 559; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 559; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_fpmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_fpmask.diminfo[0].strides = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fpmask.diminfo[0].shape = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 559; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_17 = 0;
  __pyx_v_fpmask = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":560
 * 
 *     fpmask = pmask.reshape(-1)
 *     fmask = mask_c.reshape(-1)             # <<<<<<<<<<<<<<
 *     fcoords = coords_c.reshape((coords_c.shape[0], -1))
 * 
 */
  __pyx_t_3 = PyObject_GetAttr(((PyObject *)__pyx_v_mask_c), __pyx_n_s__reshape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 560; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_k_tuple_26), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 560; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 560; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_18 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_18, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_fmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_fmask.diminfo[0].strides = __pyx_pybuffernd_fmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fmask.diminfo[0].shape = __pyx_pybuffernd_fmask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 560; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_18 = 0;
  __pyx_v_fmask = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":561
 *     fpmask = pmask.reshape(-1)
 *     fmask = mask_c.reshape(-1)
 *     fcoords = coords_c.reshape((coords_c.shape[0], -1))             # <<<<<<<<<<<<<<
 * 
 *     # First do the interior contributions.
 */
  __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_v_coords_c), __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t((__pyx_v_coords_c->dimensions[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_19 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer, (PyObject*)__pyx_t_19, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer, (PyObject*)__pyx_v_fcoords, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_fcoords.diminfo[0].strides = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fcoords.diminfo[0].shape = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_fcoords.diminfo[1].strides = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_fcoords.diminfo[1].shape = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 561; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_19 = 0;
  __pyx_v_fcoords = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":572
 *         np.ndarray[np.intp_t, ndim=1] strides
 *         np.ndarray[np.intp_t, ndim=1] dstrides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)             # <<<<<<<<<<<<<<
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
 *     ss0, ss1, ss2 = strides[0], strides[1], strides[2]
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__array); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__strides_from); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__bool); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_20 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__intp); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_20) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_20) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_20, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_21 = ((PyArrayObject *)__pyx_t_20);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_v_strides, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_strides.diminfo[0].strides = __pyx_pybuffernd_strides.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_strides.diminfo[0].shape = __pyx_pybuffernd_strides.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 572; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_21 = 0;
  __pyx_v_strides = ((PyArrayObject *)__pyx_t_20);
  __pyx_t_20 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":573
 *         np.ndarray[np.intp_t, ndim=1] dstrides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)             # <<<<<<<<<<<<<<
 *     ss0, ss1, ss2 = strides[0], strides[1], strides[2]
 *     ss0d, ss1d, ss2d = dstrides[0], dstrides[1], dstrides[2]
 */
  __pyx_t_20 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_20, __pyx_n_s__array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_GetName(__pyx_m, __pyx_n_s__strides_from); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__bool); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_20, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_5));
  __pyx_t_20 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_20, __pyx_n_s__intp); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem(__pyx_t_5, ((PyObject *)__pyx_n_s__dtype), __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_5)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_22 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer, (PyObject*)__pyx_t_22, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer, (PyObject*)__pyx_v_dstrides, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_dstrides.diminfo[0].strides = __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dstrides.diminfo[0].shape = __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 573; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_22 = 0;
  __pyx_v_dstrides = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":574
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
 *     ss0, ss1, ss2 = strides[0], strides[1], strides[2]             # <<<<<<<<<<<<<<
 *     ss0d, ss1d, ss2d = dstrides[0], dstrides[1], dstrides[2]
 *     verts = []
 */
  __pyx_t_23 = 0;
  __pyx_t_8 = -1;
  if (__pyx_t_23 < 0) {
    __pyx_t_23 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_23 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_23 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 574; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_24 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_23, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_t_25 = 1;
  __pyx_t_8 = -1;
  if (__pyx_t_25 < 0) {
    __pyx_t_25 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_25 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_25 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 574; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_26 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_t_27 = 2;
  __pyx_t_8 = -1;
  if (__pyx_t_27 < 0) {
    __pyx_t_27 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_27 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_27 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 574; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_28 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_v_ss0 = __pyx_t_24;
  __pyx_v_ss1 = __pyx_t_26;
  __pyx_v_ss2 = __pyx_t_28;

  /* "nipy/algorithms/statistics/intvol.pyx":575
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
 *     ss0, ss1, ss2 = strides[0], strides[1], strides[2]
 *     ss0d, ss1d, ss2d = dstrides[0], dstrides[1], dstrides[2]             # <<<<<<<<<<<<<<
 *     verts = []
 *     for i in range(2):
 */
  __pyx_t_29 = 0;
  __pyx_t_8 = -1;
  if (__pyx_t_29 < 0) {
    __pyx_t_29 += __pyx_pybuffernd_dstrides.diminfo[0].shape;
    if (unlikely(__pyx_t_29 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_29 >= __pyx_pybuffernd_dstrides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 575; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_dstrides.diminfo[0].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 575; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_30 = 1;
  __pyx_t_8 = -1;
  if (__pyx_t_30 < 0) {
    __pyx_t_30 += __pyx_pybuffernd_dstrides.diminfo[0].shape;
    if (unlikely(__pyx_t_30 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_30 >= __pyx_pybuffernd_dstrides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 575; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_5 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_dstrides.diminfo[0].strides))); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 575; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_31 = 2;
  __pyx_t_8 = -1;
  if (__pyx_t_31 < 0) {
    __pyx_t_31 += __pyx_pybuffernd_dstrides.diminfo[0].shape;
    if (unlikely(__pyx_t_31 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_31 >= __pyx_pybuffernd_dstrides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 575; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_dstrides.diminfo[0].strides))); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 575; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_ss0d = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_ss1d = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_ss2d = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":576
 *     ss0, ss1, ss2 = strides[0], strides[1], strides[2]
 *     ss0d, ss1d, ss2d = dstrides[0], dstrides[1], dstrides[2]
 *     verts = []             # <<<<<<<<<<<<<<
 *     for i in range(2):
 *         for j in range(2):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 576; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_verts = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":577
 *     ss0d, ss1d, ss2d = dstrides[0], dstrides[1], dstrides[2]
 *     verts = []
 *     for i in range(2):             # <<<<<<<<<<<<<<
 *         for j in range(2):
 *             for k in range(2):
 */
  for (__pyx_t_16 = 0; __pyx_t_16 < 2; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "nipy/algorithms/statistics/intvol.pyx":578
 *     verts = []
 *     for i in range(2):
 *         for j in range(2):             # <<<<<<<<<<<<<<
 *             for k in range(2):
 *                 verts.append(ss0d * i + ss1d * j + ss2d * k)
 */
    for (__pyx_t_15 = 0; __pyx_t_15 < 2; __pyx_t_15+=1) {
      __pyx_v_j = __pyx_t_15;

      /* "nipy/algorithms/statistics/intvol.pyx":579
 *     for i in range(2):
 *         for j in range(2):
 *             for k in range(2):             # <<<<<<<<<<<<<<
 *                 verts.append(ss0d * i + ss1d * j + ss2d * k)
 *     cvertices = np.array(sorted(verts), np.intp)
 */
      for (__pyx_t_14 = 0; __pyx_t_14 < 2; __pyx_t_14+=1) {
        __pyx_v_k = __pyx_t_14;

        /* "nipy/algorithms/statistics/intvol.pyx":580
 *         for j in range(2):
 *             for k in range(2):
 *                 verts.append(ss0d * i + ss1d * j + ss2d * k)             # <<<<<<<<<<<<<<
 *     cvertices = np.array(sorted(verts), np.intp)
 * 
 */
        __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_i); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = PyNumber_Multiply(__pyx_v_ss0d, __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_j); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyNumber_Multiply(__pyx_v_ss1d, __pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = PyNumber_Add(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_k); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = PyNumber_Multiply(__pyx_v_ss2d, __pyx_t_2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_32 = PyList_Append(__pyx_v_verts, __pyx_t_2); if (unlikely(__pyx_t_32 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
    }
  }

  /* "nipy/algorithms/statistics/intvol.pyx":581
 *             for k in range(2):
 *                 verts.append(ss0d * i + ss1d * j + ss2d * k)
 *     cvertices = np.array(sorted(verts), np.intp)             # <<<<<<<<<<<<<<
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__array); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_verts));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_verts));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_verts));
  __pyx_t_3 = PyObject_Call(__pyx_builtin_sorted, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__intp); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_33 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer, (PyObject*)__pyx_t_33, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer, (PyObject*)__pyx_v_cvertices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_cvertices.diminfo[0].strides = __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_cvertices.diminfo[0].shape = __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_33 = 0;
  __pyx_v_cvertices = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":583
 *     cvertices = np.array(sorted(verts), np.intp)
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__join_complexes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_27));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_k_tuple_27));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_27));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":584
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_28));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_k_tuple_28));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_28));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_20 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":585
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 585; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 585; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_29));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_k_tuple_29));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_29));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_34 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 585; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":586
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_30));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_k_tuple_30));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_30));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_35 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":587
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_31));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_k_tuple_31));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_31));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_36 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":588
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 588; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 588; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_32));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_k_tuple_32));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_32));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_37 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 588; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":589
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])             # <<<<<<<<<<<<<<
 *     c = cube_with_strides_center((0,0,0), strides)
 *     m4 = np.array(list(c[4].difference(union[4])))
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 589; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 589; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_33));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_k_tuple_33));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_33));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_38 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 589; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(7); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_20);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  PyList_SET_ITEM(__pyx_t_5, 3, __pyx_t_35);
  __Pyx_GIVEREF(__pyx_t_35);
  PyList_SET_ITEM(__pyx_t_5, 4, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  PyList_SET_ITEM(__pyx_t_5, 5, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyList_SET_ITEM(__pyx_t_5, 6, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  __pyx_t_3 = 0;
  __pyx_t_20 = 0;
  __pyx_t_34 = 0;
  __pyx_t_35 = 0;
  __pyx_t_36 = 0;
  __pyx_t_37 = 0;
  __pyx_t_38 = 0;
  __pyx_t_38 = PySequence_Tuple(((PyObject *)__pyx_t_5)); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_38));
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  __pyx_v_union = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":590
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)             # <<<<<<<<<<<<<<
 *     m4 = np.array(list(c[4].difference(union[4])))
 *     m3 = np.array(list(c[3].difference(union[3])))
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_38 = PyTuple_New(2); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_34));
  PyTuple_SET_ITEM(__pyx_t_38, 0, ((PyObject *)__pyx_k_tuple_34));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_34));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_38, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  __pyx_v_c = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":591
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)
 *     m4 = np.array(list(c[4].difference(union[4])))             # <<<<<<<<<<<<<<
 *     m3 = np.array(list(c[3].difference(union[3])))
 *     m2 = np.array(list(c[2].difference(union[2])))
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_c, 4, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__difference); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_union, 4, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyTuple_New(1); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_37), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_37)); __pyx_t_37 = 0;
  __pyx_t_37 = PyTuple_New(1); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_37), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_37)); __pyx_t_37 = 0;
  __pyx_t_37 = PyTuple_New(1); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  PyTuple_SET_ITEM(__pyx_t_37, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_t_37), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_37)); __pyx_t_37 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_39 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m4.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_m4.rcbuffer->pybuffer, (PyObject*)__pyx_t_39, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_m4.rcbuffer->pybuffer, (PyObject*)__pyx_v_m4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_m4.diminfo[0].strides = __pyx_pybuffernd_m4.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_m4.diminfo[0].shape = __pyx_pybuffernd_m4.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_m4.diminfo[1].strides = __pyx_pybuffernd_m4.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_m4.diminfo[1].shape = __pyx_pybuffernd_m4.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 591; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_39 = 0;
  __pyx_v_m4 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":592
 *     c = cube_with_strides_center((0,0,0), strides)
 *     m4 = np.array(list(c[4].difference(union[4])))
 *     m3 = np.array(list(c[3].difference(union[3])))             # <<<<<<<<<<<<<<
 *     m2 = np.array(list(c[2].difference(union[2])))
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_c, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__difference); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_union, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_37, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_40 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m3.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_m3.rcbuffer->pybuffer, (PyObject*)__pyx_t_40, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_m3.rcbuffer->pybuffer, (PyObject*)__pyx_v_m3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_m3.diminfo[0].strides = __pyx_pybuffernd_m3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_m3.diminfo[0].shape = __pyx_pybuffernd_m3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_m3.diminfo[1].strides = __pyx_pybuffernd_m3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_m3.diminfo[1].shape = __pyx_pybuffernd_m3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 592; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_40 = 0;
  __pyx_v_m3 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":593
 *     m4 = np.array(list(c[4].difference(union[4])))
 *     m3 = np.array(list(c[3].difference(union[3])))
 *     m2 = np.array(list(c[2].difference(union[2])))             # <<<<<<<<<<<<<<
 * 
 *     d4 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m4[i]] for i in range(m4.shape[0])])
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_c, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_37 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__difference); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_union, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyTuple_New(1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_37, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  __pyx_t_38 = PyTuple_New(1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  __pyx_t_38 = PyTuple_New(1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_38, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_41 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m2.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_m2.rcbuffer->pybuffer, (PyObject*)__pyx_t_41, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_m2.rcbuffer->pybuffer, (PyObject*)__pyx_v_m2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_m2.diminfo[0].strides = __pyx_pybuffernd_m2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_m2.diminfo[0].shape = __pyx_pybuffernd_m2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_m2.diminfo[1].strides = __pyx_pybuffernd_m2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_m2.diminfo[1].shape = __pyx_pybuffernd_m2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 593; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_41 = 0;
  __pyx_v_m2 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":595
 *     m2 = np.array(list(c[2].difference(union[2])))
 * 
 *     d4 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m4[i]] for i in range(m4.shape[0])])             # <<<<<<<<<<<<<<
 *     d4 = np.hstack([m4, d4])
 *     ds4 = d4.shape[0]
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = (__pyx_v_m4->dimensions[0]);
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_16; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_37 = __Pyx_GetItemInt(((PyObject *)__pyx_v_m4), __pyx_v_i, sizeof(npy_intp), __Pyx_PyInt_to_py_Py_intptr_t); if (!__pyx_t_37) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_37);
    if (PyList_CheckExact(__pyx_t_37) || PyTuple_CheckExact(__pyx_t_37)) {
      __pyx_t_36 = __pyx_t_37; __Pyx_INCREF(__pyx_t_36); __pyx_t_42 = 0;
      __pyx_t_43 = NULL;
    } else {
      __pyx_t_42 = -1; __pyx_t_36 = PyObject_GetIter(__pyx_t_37); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_36);
      __pyx_t_43 = Py_TYPE(__pyx_t_36)->tp_iternext;
    }
    __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
    for (;;) {
      if (!__pyx_t_43 && PyList_CheckExact(__pyx_t_36)) {
        if (__pyx_t_42 >= PyList_GET_SIZE(__pyx_t_36)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_37 = PyList_GET_ITEM(__pyx_t_36, __pyx_t_42); __Pyx_INCREF(__pyx_t_37); __pyx_t_42++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_37 = PySequence_ITEM(__pyx_t_36, __pyx_t_42); __pyx_t_42++; if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_43 && PyTuple_CheckExact(__pyx_t_36)) {
        if (__pyx_t_42 >= PyTuple_GET_SIZE(__pyx_t_36)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_37 = PyTuple_GET_ITEM(__pyx_t_36, __pyx_t_42); __Pyx_INCREF(__pyx_t_37); __pyx_t_42++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_37 = PySequence_ITEM(__pyx_t_36, __pyx_t_42); __pyx_t_42++; if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_37 = __pyx_t_43(__pyx_t_36);
        if (unlikely(!__pyx_t_37)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_37);
      }
      __Pyx_XDECREF(__pyx_v_v);
      __pyx_v_v = __pyx_t_37;
      __pyx_t_37 = 0;
      __pyx_t_37 = __Pyx_GetName(__pyx_m, __pyx_n_s___convert_stride3); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_37);
      __pyx_t_35 = PyTuple_New(3); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_35);
      __Pyx_INCREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      __Pyx_INCREF(((PyObject *)__pyx_v_strides));
      PyTuple_SET_ITEM(__pyx_t_35, 1, ((PyObject *)__pyx_v_strides));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
      __Pyx_INCREF(((PyObject *)__pyx_k_tuple_35));
      PyTuple_SET_ITEM(__pyx_t_35, 2, ((PyObject *)__pyx_k_tuple_35));
      __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_35));
      __pyx_t_34 = PyObject_Call(__pyx_t_37, ((PyObject *)__pyx_t_35), NULL); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_34);
      __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_35)); __pyx_t_35 = 0;
      if (unlikely(__Pyx_PyList_Append(__pyx_t_5, (PyObject*)__pyx_t_34))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    }
    __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
    if (unlikely(__Pyx_PyList_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_44 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d4.rcbuffer->pybuffer, (PyObject*)__pyx_t_44, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d4.rcbuffer->pybuffer, (PyObject*)__pyx_v_d4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_d4.diminfo[0].strides = __pyx_pybuffernd_d4.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d4.diminfo[0].shape = __pyx_pybuffernd_d4.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d4.diminfo[1].strides = __pyx_pybuffernd_d4.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d4.diminfo[1].shape = __pyx_pybuffernd_d4.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_44 = 0;
  __pyx_v_d4 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":596
 * 
 *     d4 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m4[i]] for i in range(m4.shape[0])])
 *     d4 = np.hstack([m4, d4])             # <<<<<<<<<<<<<<
 *     ds4 = d4.shape[0]
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__hstack); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_m4));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_m4));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_m4));
  __Pyx_INCREF(((PyObject *)__pyx_v_d4));
  PyList_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_d4));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_d4));
  __pyx_t_38 = PyTuple_New(1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_38, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_44 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d4.rcbuffer->pybuffer, (PyObject*)__pyx_t_44, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d4.rcbuffer->pybuffer, (PyObject*)__pyx_v_d4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_d4.diminfo[0].strides = __pyx_pybuffernd_d4.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d4.diminfo[0].shape = __pyx_pybuffernd_d4.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d4.diminfo[1].strides = __pyx_pybuffernd_d4.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d4.diminfo[1].shape = __pyx_pybuffernd_d4.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_44 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_v_d4));
  __pyx_v_d4 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":597
 *     d4 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m4[i]] for i in range(m4.shape[0])])
 *     d4 = np.hstack([m4, d4])
 *     ds4 = d4.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     d3 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m3[i]] for i in range(m3.shape[0])])
 */
  __pyx_v_ds4 = (__pyx_v_d4->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":599
 *     ds4 = d4.shape[0]
 * 
 *     d3 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m3[i]] for i in range(m3.shape[0])])             # <<<<<<<<<<<<<<
 *     d3 = np.hstack([m3, d3])
 *     ds3 = d3.shape[0]
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = (__pyx_v_m3->dimensions[0]);
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_16; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_36 = __Pyx_GetItemInt(((PyObject *)__pyx_v_m3), __pyx_v_i, sizeof(npy_intp), __Pyx_PyInt_to_py_Py_intptr_t); if (!__pyx_t_36) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_36);
    if (PyList_CheckExact(__pyx_t_36) || PyTuple_CheckExact(__pyx_t_36)) {
      __pyx_t_34 = __pyx_t_36; __Pyx_INCREF(__pyx_t_34); __pyx_t_42 = 0;
      __pyx_t_43 = NULL;
    } else {
      __pyx_t_42 = -1; __pyx_t_34 = PyObject_GetIter(__pyx_t_36); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_34);
      __pyx_t_43 = Py_TYPE(__pyx_t_34)->tp_iternext;
    }
    __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
    for (;;) {
      if (!__pyx_t_43 && PyList_CheckExact(__pyx_t_34)) {
        if (__pyx_t_42 >= PyList_GET_SIZE(__pyx_t_34)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_36 = PyList_GET_ITEM(__pyx_t_34, __pyx_t_42); __Pyx_INCREF(__pyx_t_36); __pyx_t_42++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_36 = PySequence_ITEM(__pyx_t_34, __pyx_t_42); __pyx_t_42++; if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_43 && PyTuple_CheckExact(__pyx_t_34)) {
        if (__pyx_t_42 >= PyTuple_GET_SIZE(__pyx_t_34)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_36 = PyTuple_GET_ITEM(__pyx_t_34, __pyx_t_42); __Pyx_INCREF(__pyx_t_36); __pyx_t_42++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_36 = PySequence_ITEM(__pyx_t_34, __pyx_t_42); __pyx_t_42++; if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_36 = __pyx_t_43(__pyx_t_34);
        if (unlikely(!__pyx_t_36)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_36);
      }
      __Pyx_XDECREF(__pyx_v_v);
      __pyx_v_v = __pyx_t_36;
      __pyx_t_36 = 0;
      __pyx_t_36 = __Pyx_GetName(__pyx_m, __pyx_n_s___convert_stride3); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_36);
      __pyx_t_35 = PyTuple_New(3); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_35);
      __Pyx_INCREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      __Pyx_INCREF(((PyObject *)__pyx_v_strides));
      PyTuple_SET_ITEM(__pyx_t_35, 1, ((PyObject *)__pyx_v_strides));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
      __Pyx_INCREF(((PyObject *)__pyx_k_tuple_36));
      PyTuple_SET_ITEM(__pyx_t_35, 2, ((PyObject *)__pyx_k_tuple_36));
      __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_36));
      __pyx_t_37 = PyObject_Call(__pyx_t_36, ((PyObject *)__pyx_t_35), NULL); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_37);
      __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_35)); __pyx_t_35 = 0;
      if (unlikely(__Pyx_PyList_Append(__pyx_t_5, (PyObject*)__pyx_t_37))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
    }
    __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    if (unlikely(__Pyx_PyList_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_45 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_t_45, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_v_d3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_d3.diminfo[0].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d3.diminfo[0].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d3.diminfo[1].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d3.diminfo[1].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_45 = 0;
  __pyx_v_d3 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":600
 * 
 *     d3 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m3[i]] for i in range(m3.shape[0])])
 *     d3 = np.hstack([m3, d3])             # <<<<<<<<<<<<<<
 *     ds3 = d3.shape[0]
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__hstack); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_m3));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_m3));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_m3));
  __Pyx_INCREF(((PyObject *)__pyx_v_d3));
  PyList_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_d3));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_d3));
  __pyx_t_38 = PyTuple_New(1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_38, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_45 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_t_45, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_v_d3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_d3.diminfo[0].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d3.diminfo[0].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d3.diminfo[1].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d3.diminfo[1].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_45 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_v_d3));
  __pyx_v_d3 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":601
 *     d3 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m3[i]] for i in range(m3.shape[0])])
 *     d3 = np.hstack([m3, d3])
 *     ds3 = d3.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     d2 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m2[i]] for i in range(m2.shape[0])])
 */
  __pyx_v_ds3 = (__pyx_v_d3->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":603
 *     ds3 = d3.shape[0]
 * 
 *     d2 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m2[i]] for i in range(m2.shape[0])])             # <<<<<<<<<<<<<<
 *     d2 = np.hstack([m2, d2])
 *     ds2 = d2.shape[0]
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = (__pyx_v_m2->dimensions[0]);
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_16; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_34 = __Pyx_GetItemInt(((PyObject *)__pyx_v_m2), __pyx_v_i, sizeof(npy_intp), __Pyx_PyInt_to_py_Py_intptr_t); if (!__pyx_t_34) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_34);
    if (PyList_CheckExact(__pyx_t_34) || PyTuple_CheckExact(__pyx_t_34)) {
      __pyx_t_37 = __pyx_t_34; __Pyx_INCREF(__pyx_t_37); __pyx_t_42 = 0;
      __pyx_t_43 = NULL;
    } else {
      __pyx_t_42 = -1; __pyx_t_37 = PyObject_GetIter(__pyx_t_34); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_37);
      __pyx_t_43 = Py_TYPE(__pyx_t_37)->tp_iternext;
    }
    __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    for (;;) {
      if (!__pyx_t_43 && PyList_CheckExact(__pyx_t_37)) {
        if (__pyx_t_42 >= PyList_GET_SIZE(__pyx_t_37)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_34 = PyList_GET_ITEM(__pyx_t_37, __pyx_t_42); __Pyx_INCREF(__pyx_t_34); __pyx_t_42++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_34 = PySequence_ITEM(__pyx_t_37, __pyx_t_42); __pyx_t_42++; if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_43 && PyTuple_CheckExact(__pyx_t_37)) {
        if (__pyx_t_42 >= PyTuple_GET_SIZE(__pyx_t_37)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_34 = PyTuple_GET_ITEM(__pyx_t_37, __pyx_t_42); __Pyx_INCREF(__pyx_t_34); __pyx_t_42++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_34 = PySequence_ITEM(__pyx_t_37, __pyx_t_42); __pyx_t_42++; if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_34 = __pyx_t_43(__pyx_t_37);
        if (unlikely(!__pyx_t_34)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_34);
      }
      __Pyx_XDECREF(__pyx_v_v);
      __pyx_v_v = __pyx_t_34;
      __pyx_t_34 = 0;
      __pyx_t_34 = __Pyx_GetName(__pyx_m, __pyx_n_s___convert_stride3); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_34);
      __pyx_t_35 = PyTuple_New(3); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_35);
      __Pyx_INCREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_35, 0, __pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      __Pyx_INCREF(((PyObject *)__pyx_v_strides));
      PyTuple_SET_ITEM(__pyx_t_35, 1, ((PyObject *)__pyx_v_strides));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
      __Pyx_INCREF(((PyObject *)__pyx_k_tuple_37));
      PyTuple_SET_ITEM(__pyx_t_35, 2, ((PyObject *)__pyx_k_tuple_37));
      __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_37));
      __pyx_t_36 = PyObject_Call(__pyx_t_34, ((PyObject *)__pyx_t_35), NULL); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_36);
      __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_35)); __pyx_t_35 = 0;
      if (unlikely(__Pyx_PyList_Append(__pyx_t_5, (PyObject*)__pyx_t_36))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_36); __pyx_t_36 = 0;
    }
    __Pyx_DECREF(__pyx_t_37); __pyx_t_37 = 0;
    if (unlikely(__Pyx_PyList_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_46 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_t_46, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_v_d2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_d2.diminfo[0].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d2.diminfo[0].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d2.diminfo[1].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d2.diminfo[1].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_46 = 0;
  __pyx_v_d2 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":604
 * 
 *     d2 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m2[i]] for i in range(m2.shape[0])])
 *     d2 = np.hstack([m2, d2])             # <<<<<<<<<<<<<<
 *     ds2 = d2.shape[0]
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__hstack); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_m2));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_m2));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_m2));
  __Pyx_INCREF(((PyObject *)__pyx_v_d2));
  PyList_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_d2));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_d2));
  __pyx_t_38 = PyTuple_New(1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  PyTuple_SET_ITEM(__pyx_t_38, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_38), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_38)); __pyx_t_38 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_46 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_t_46, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_v_d2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_d2.diminfo[0].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d2.diminfo[0].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d2.diminfo[1].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d2.diminfo[1].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 604; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_46 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_v_d2));
  __pyx_v_d2 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":605
 *     d2 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m2[i]] for i in range(m2.shape[0])])
 *     d2 = np.hstack([m2, d2])
 *     ds2 = d2.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     nvox = mask.size
 */
  __pyx_v_ds2 = (__pyx_v_d2->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":607
 *     ds2 = d2.shape[0]
 * 
 *     nvox = mask.size             # <<<<<<<<<<<<<<
 * 
 *     D = np.zeros((8,8))
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__size); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 607; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_16 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 607; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nvox = __pyx_t_16;

  /* "nipy/algorithms/statistics/intvol.pyx":609
 *     nvox = mask.size
 * 
 *     D = np.zeros((8,8))             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0-1):
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_38 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_k_tuple_39), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_47 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_D.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_D.rcbuffer->pybuffer, (PyObject*)__pyx_t_47, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_D.rcbuffer->pybuffer, (PyObject*)__pyx_v_D, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_D.diminfo[0].strides = __pyx_pybuffernd_D.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_D.diminfo[0].shape = __pyx_pybuffernd_D.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_D.diminfo[1].strides = __pyx_pybuffernd_D.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_D.diminfo[1].shape = __pyx_pybuffernd_D.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_47 = 0;
  __pyx_v_D = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":611
 *     D = np.zeros((8,8))
 * 
 *     for i in range(s0-1):             # <<<<<<<<<<<<<<
 *         for j in range(s1-1):
 *             for k in range(s2-1):
 */
  __pyx_t_48 = (__pyx_v_s0 - 1);
  for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_48; __pyx_t_16+=1) {
    __pyx_v_i = __pyx_t_16;

    /* "nipy/algorithms/statistics/intvol.pyx":612
 * 
 *     for i in range(s0-1):
 *         for j in range(s1-1):             # <<<<<<<<<<<<<<
 *             for k in range(s2-1):
 * 
 */
    __pyx_t_49 = (__pyx_v_s1 - 1);
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_49; __pyx_t_15+=1) {
      __pyx_v_j = __pyx_t_15;

      /* "nipy/algorithms/statistics/intvol.pyx":613
 *     for i in range(s0-1):
 *         for j in range(s1-1):
 *             for k in range(s2-1):             # <<<<<<<<<<<<<<
 * 
 *                 pindex = i*ss0+j*ss1+k*ss2
 */
      __pyx_t_50 = (__pyx_v_s2 - 1);
      for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_50; __pyx_t_14+=1) {
        __pyx_v_k = __pyx_t_14;

        /* "nipy/algorithms/statistics/intvol.pyx":615
 *             for k in range(s2-1):
 * 
 *                 pindex = i*ss0+j*ss1+k*ss2             # <<<<<<<<<<<<<<
 *                 index = i*ss0d+j*ss1d+k*ss2d
 *                 for r in range(8):
 */
        __pyx_v_pindex = (((__pyx_v_i * __pyx_v_ss0) + (__pyx_v_j * __pyx_v_ss1)) + (__pyx_v_k * __pyx_v_ss2));

        /* "nipy/algorithms/statistics/intvol.pyx":616
 * 
 *                 pindex = i*ss0+j*ss1+k*ss2
 *                 index = i*ss0d+j*ss1d+k*ss2d             # <<<<<<<<<<<<<<
 *                 for r in range(8):
 *                     rr = (index+cvertices[r]) % nvox
 */
        __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_i); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_38 = PyNumber_Multiply(__pyx_t_1, __pyx_v_ss0d); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_38);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_j); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PyNumber_Multiply(__pyx_t_1, __pyx_v_ss1d); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_Add(__pyx_t_38, __pyx_t_5); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_k); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_38 = PyNumber_Multiply(__pyx_t_5, __pyx_v_ss2d); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_38);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = PyNumber_Add(__pyx_t_1, __pyx_t_38); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
        __pyx_t_51 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_5); if (unlikely((__pyx_t_51 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 616; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_index = __pyx_t_51;

        /* "nipy/algorithms/statistics/intvol.pyx":617
 *                 pindex = i*ss0+j*ss1+k*ss2
 *                 index = i*ss0d+j*ss1d+k*ss2d
 *                 for r in range(8):             # <<<<<<<<<<<<<<
 *                     rr = (index+cvertices[r]) % nvox
 *                     mr = fmask[rr]
 */
        for (__pyx_t_51 = 0; __pyx_t_51 < 8; __pyx_t_51+=1) {
          __pyx_v_r = __pyx_t_51;

          /* "nipy/algorithms/statistics/intvol.pyx":618
 *                 index = i*ss0d+j*ss1d+k*ss2d
 *                 for r in range(8):
 *                     rr = (index+cvertices[r]) % nvox             # <<<<<<<<<<<<<<
 *                     mr = fmask[rr]
 *                     for s in range(r+1):
 */
          __pyx_t_52 = __pyx_v_r;
          __pyx_t_8 = -1;
          if (__pyx_t_52 < 0) {
            __pyx_t_52 += __pyx_pybuffernd_cvertices.diminfo[0].shape;
            if (unlikely(__pyx_t_52 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_52 >= __pyx_pybuffernd_cvertices.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 618; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_8 = (__pyx_v_index + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_cvertices.diminfo[0].strides)));
          if (unlikely(__pyx_v_nvox == 0)) {
            PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 618; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_rr = __Pyx_mod_int(__pyx_t_8, __pyx_v_nvox);

          /* "nipy/algorithms/statistics/intvol.pyx":619
 *                 for r in range(8):
 *                     rr = (index+cvertices[r]) % nvox
 *                     mr = fmask[rr]             # <<<<<<<<<<<<<<
 *                     for s in range(r+1):
 *                         res = 0
 */
          __pyx_t_53 = __pyx_v_rr;
          __pyx_t_8 = -1;
          if (__pyx_t_53 < 0) {
            __pyx_t_53 += __pyx_pybuffernd_fmask.diminfo[0].shape;
            if (unlikely(__pyx_t_53 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_53 >= __pyx_pybuffernd_fmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 619; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_mr = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fmask.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_fmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":620
 *                     rr = (index+cvertices[r]) % nvox
 *                     mr = fmask[rr]
 *                     for s in range(r+1):             # <<<<<<<<<<<<<<
 *                         res = 0
 *                         ss = (index+cvertices[s]) % nvox
 */
          __pyx_t_54 = (__pyx_v_r + 1);
          for (__pyx_t_55 = 0; __pyx_t_55 < __pyx_t_54; __pyx_t_55+=1) {
            __pyx_v_s = __pyx_t_55;

            /* "nipy/algorithms/statistics/intvol.pyx":621
 *                     mr = fmask[rr]
 *                     for s in range(r+1):
 *                         res = 0             # <<<<<<<<<<<<<<
 *                         ss = (index+cvertices[s]) % nvox
 *                         ms = fmask[ss]
 */
            __pyx_v_res = 0.0;

            /* "nipy/algorithms/statistics/intvol.pyx":622
 *                     for s in range(r+1):
 *                         res = 0
 *                         ss = (index+cvertices[s]) % nvox             # <<<<<<<<<<<<<<
 *                         ms = fmask[ss]
 *                         if mr * ms:
 */
            __pyx_t_56 = __pyx_v_s;
            __pyx_t_8 = -1;
            if (__pyx_t_56 < 0) {
              __pyx_t_56 += __pyx_pybuffernd_cvertices.diminfo[0].shape;
              if (unlikely(__pyx_t_56 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_56 >= __pyx_pybuffernd_cvertices.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 622; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_8 = (__pyx_v_index + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.buf, __pyx_t_56, __pyx_pybuffernd_cvertices.diminfo[0].strides)));
            if (unlikely(__pyx_v_nvox == 0)) {
              PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 622; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_ss = __Pyx_mod_int(__pyx_t_8, __pyx_v_nvox);

            /* "nipy/algorithms/statistics/intvol.pyx":623
 *                         res = 0
 *                         ss = (index+cvertices[s]) % nvox
 *                         ms = fmask[ss]             # <<<<<<<<<<<<<<
 *                         if mr * ms:
 *                             for l in range(fcoords.shape[0]):
 */
            __pyx_t_57 = __pyx_v_ss;
            __pyx_t_8 = -1;
            if (__pyx_t_57 < 0) {
              __pyx_t_57 += __pyx_pybuffernd_fmask.diminfo[0].shape;
              if (unlikely(__pyx_t_57 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_57 >= __pyx_pybuffernd_fmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 623; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_ms = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fmask.rcbuffer->pybuffer.buf, __pyx_t_57, __pyx_pybuffernd_fmask.diminfo[0].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":624
 *                         ss = (index+cvertices[s]) % nvox
 *                         ms = fmask[ss]
 *                         if mr * ms:             # <<<<<<<<<<<<<<
 *                             for l in range(fcoords.shape[0]):
 *                                 res += fcoords[l,ss] * fcoords[l,rr]
 */
            __pyx_t_8 = (__pyx_v_mr * __pyx_v_ms);
            if (__pyx_t_8) {

              /* "nipy/algorithms/statistics/intvol.pyx":625
 *                         ms = fmask[ss]
 *                         if mr * ms:
 *                             for l in range(fcoords.shape[0]):             # <<<<<<<<<<<<<<
 *                                 res += fcoords[l,ss] * fcoords[l,rr]
 *                             D[r,s] = res
 */
              __pyx_t_58 = (__pyx_v_fcoords->dimensions[0]);
              for (__pyx_t_59 = 0; __pyx_t_59 < __pyx_t_58; __pyx_t_59+=1) {
                __pyx_v_l = __pyx_t_59;

                /* "nipy/algorithms/statistics/intvol.pyx":626
 *                         if mr * ms:
 *                             for l in range(fcoords.shape[0]):
 *                                 res += fcoords[l,ss] * fcoords[l,rr]             # <<<<<<<<<<<<<<
 *                             D[r,s] = res
 *                             D[s,r] = res
 */
                __pyx_t_60 = __pyx_v_l;
                __pyx_t_61 = __pyx_v_ss;
                __pyx_t_8 = -1;
                if (__pyx_t_60 < 0) {
                  __pyx_t_60 += __pyx_pybuffernd_fcoords.diminfo[0].shape;
                  if (unlikely(__pyx_t_60 < 0)) __pyx_t_8 = 0;
                } else if (unlikely(__pyx_t_60 >= __pyx_pybuffernd_fcoords.diminfo[0].shape)) __pyx_t_8 = 0;
                if (__pyx_t_61 < 0) {
                  __pyx_t_61 += __pyx_pybuffernd_fcoords.diminfo[1].shape;
                  if (unlikely(__pyx_t_61 < 0)) __pyx_t_8 = 1;
                } else if (unlikely(__pyx_t_61 >= __pyx_pybuffernd_fcoords.diminfo[1].shape)) __pyx_t_8 = 1;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 626; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                }
                __pyx_t_62 = __pyx_v_l;
                __pyx_t_63 = __pyx_v_rr;
                __pyx_t_8 = -1;
                if (__pyx_t_62 < 0) {
                  __pyx_t_62 += __pyx_pybuffernd_fcoords.diminfo[0].shape;
                  if (unlikely(__pyx_t_62 < 0)) __pyx_t_8 = 0;
                } else if (unlikely(__pyx_t_62 >= __pyx_pybuffernd_fcoords.diminfo[0].shape)) __pyx_t_8 = 0;
                if (__pyx_t_63 < 0) {
                  __pyx_t_63 += __pyx_pybuffernd_fcoords.diminfo[1].shape;
                  if (unlikely(__pyx_t_63 < 0)) __pyx_t_8 = 1;
                } else if (unlikely(__pyx_t_63 >= __pyx_pybuffernd_fcoords.diminfo[1].shape)) __pyx_t_8 = 1;
                if (unlikely(__pyx_t_8 != -1)) {
                  __Pyx_RaiseBufferIndexError(__pyx_t_8);
                  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 626; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                }
                __pyx_v_res = (__pyx_v_res + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.buf, __pyx_t_60, __pyx_pybuffernd_fcoords.diminfo[0].strides, __pyx_t_61, __pyx_pybuffernd_fcoords.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.buf, __pyx_t_62, __pyx_pybuffernd_fcoords.diminfo[0].strides, __pyx_t_63, __pyx_pybuffernd_fcoords.diminfo[1].strides))));
              }

              /* "nipy/algorithms/statistics/intvol.pyx":627
 *                             for l in range(fcoords.shape[0]):
 *                                 res += fcoords[l,ss] * fcoords[l,rr]
 *                             D[r,s] = res             # <<<<<<<<<<<<<<
 *                             D[s,r] = res
 *                         else:
 */
              __pyx_t_58 = __pyx_v_r;
              __pyx_t_59 = __pyx_v_s;
              __pyx_t_8 = -1;
              if (__pyx_t_58 < 0) {
                __pyx_t_58 += __pyx_pybuffernd_D.diminfo[0].shape;
                if (unlikely(__pyx_t_58 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_58 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_59 < 0) {
                __pyx_t_59 += __pyx_pybuffernd_D.diminfo[1].shape;
                if (unlikely(__pyx_t_59 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_59 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 627; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_58, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_59, __pyx_pybuffernd_D.diminfo[1].strides) = __pyx_v_res;

              /* "nipy/algorithms/statistics/intvol.pyx":628
 *                                 res += fcoords[l,ss] * fcoords[l,rr]
 *                             D[r,s] = res
 *                             D[s,r] = res             # <<<<<<<<<<<<<<
 *                         else:
 *                             D[r,s] = 0
 */
              __pyx_t_64 = __pyx_v_s;
              __pyx_t_65 = __pyx_v_r;
              __pyx_t_8 = -1;
              if (__pyx_t_64 < 0) {
                __pyx_t_64 += __pyx_pybuffernd_D.diminfo[0].shape;
                if (unlikely(__pyx_t_64 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_64 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_65 < 0) {
                __pyx_t_65 += __pyx_pybuffernd_D.diminfo[1].shape;
                if (unlikely(__pyx_t_65 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_65 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 628; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_64, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_65, __pyx_pybuffernd_D.diminfo[1].strides) = __pyx_v_res;
              goto __pyx_L35;
            }
            /*else*/ {

              /* "nipy/algorithms/statistics/intvol.pyx":630
 *                             D[s,r] = res
 *                         else:
 *                             D[r,s] = 0             # <<<<<<<<<<<<<<
 *                             D[s,r] = 0
 * 
 */
              __pyx_t_66 = __pyx_v_r;
              __pyx_t_67 = __pyx_v_s;
              __pyx_t_8 = -1;
              if (__pyx_t_66 < 0) {
                __pyx_t_66 += __pyx_pybuffernd_D.diminfo[0].shape;
                if (unlikely(__pyx_t_66 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_66 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_67 < 0) {
                __pyx_t_67 += __pyx_pybuffernd_D.diminfo[1].shape;
                if (unlikely(__pyx_t_67 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_67 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 630; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_67, __pyx_pybuffernd_D.diminfo[1].strides) = 0.0;

              /* "nipy/algorithms/statistics/intvol.pyx":631
 *                         else:
 *                             D[r,s] = 0
 *                             D[s,r] = 0             # <<<<<<<<<<<<<<
 * 
 *                 for l in range(ds4):
 */
              __pyx_t_68 = __pyx_v_s;
              __pyx_t_69 = __pyx_v_r;
              __pyx_t_8 = -1;
              if (__pyx_t_68 < 0) {
                __pyx_t_68 += __pyx_pybuffernd_D.diminfo[0].shape;
                if (unlikely(__pyx_t_68 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_68 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_69 < 0) {
                __pyx_t_69 += __pyx_pybuffernd_D.diminfo[1].shape;
                if (unlikely(__pyx_t_69 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_69 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 631; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_68, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_69, __pyx_pybuffernd_D.diminfo[1].strides) = 0.0;
            }
            __pyx_L35:;
          }
        }

        /* "nipy/algorithms/statistics/intvol.pyx":633
 *                             D[s,r] = 0
 * 
 *                 for l in range(ds4):             # <<<<<<<<<<<<<<
 *                     v0 = pindex + d4[l,0]
 *                     w0 = d4[l,4]
 */
        __pyx_t_51 = __pyx_v_ds4;
        for (__pyx_t_55 = 0; __pyx_t_55 < __pyx_t_51; __pyx_t_55+=1) {
          __pyx_v_l = __pyx_t_55;

          /* "nipy/algorithms/statistics/intvol.pyx":634
 * 
 *                 for l in range(ds4):
 *                     v0 = pindex + d4[l,0]             # <<<<<<<<<<<<<<
 *                     w0 = d4[l,4]
 *                     m = fpmask[v0]
 */
          __pyx_t_70 = __pyx_v_l;
          __pyx_t_54 = 0;
          __pyx_t_8 = -1;
          if (__pyx_t_70 < 0) {
            __pyx_t_70 += __pyx_pybuffernd_d4.diminfo[0].shape;
            if (unlikely(__pyx_t_70 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_70 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_54 < 0) {
            __pyx_t_54 += __pyx_pybuffernd_d4.diminfo[1].shape;
            if (unlikely(__pyx_t_54 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_54 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 634; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v0 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_70, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_54, __pyx_pybuffernd_d4.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":635
 *                 for l in range(ds4):
 *                     v0 = pindex + d4[l,0]
 *                     w0 = d4[l,4]             # <<<<<<<<<<<<<<
 *                     m = fpmask[v0]
 *                     if m:
 */
          __pyx_t_71 = __pyx_v_l;
          __pyx_t_72 = 4;
          __pyx_t_8 = -1;
          if (__pyx_t_71 < 0) {
            __pyx_t_71 += __pyx_pybuffernd_d4.diminfo[0].shape;
            if (unlikely(__pyx_t_71 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_71 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_72 < 0) {
            __pyx_t_72 += __pyx_pybuffernd_d4.diminfo[1].shape;
            if (unlikely(__pyx_t_72 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_72 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 635; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_w0 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_72, __pyx_pybuffernd_d4.diminfo[1].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":636
 *                     v0 = pindex + d4[l,0]
 *                     w0 = d4[l,4]
 *                     m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                     if m:
 *                         v1 = pindex + d4[l,1]
 */
          __pyx_t_73 = __pyx_v_v0;
          __pyx_t_8 = -1;
          if (__pyx_t_73 < 0) {
            __pyx_t_73 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_73 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_73 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 636; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_73, __pyx_pybuffernd_fpmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":637
 *                     w0 = d4[l,4]
 *                     m = fpmask[v0]
 *                     if m:             # <<<<<<<<<<<<<<
 *                         v1 = pindex + d4[l,1]
 *                         v2 = pindex + d4[l,2]
 */
          if (__pyx_v_m) {

            /* "nipy/algorithms/statistics/intvol.pyx":638
 *                     m = fpmask[v0]
 *                     if m:
 *                         v1 = pindex + d4[l,1]             # <<<<<<<<<<<<<<
 *                         v2 = pindex + d4[l,2]
 *                         v3 = pindex + d4[l,3]
 */
            __pyx_t_74 = __pyx_v_l;
            __pyx_t_75 = 1;
            __pyx_t_8 = -1;
            if (__pyx_t_74 < 0) {
              __pyx_t_74 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_74 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_74 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_75 < 0) {
              __pyx_t_75 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_75 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_75 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 638; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v1 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_74, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_75, __pyx_pybuffernd_d4.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":639
 *                     if m:
 *                         v1 = pindex + d4[l,1]
 *                         v2 = pindex + d4[l,2]             # <<<<<<<<<<<<<<
 *                         v3 = pindex + d4[l,3]
 *                         w1 = d4[l,5]
 */
            __pyx_t_76 = __pyx_v_l;
            __pyx_t_77 = 2;
            __pyx_t_8 = -1;
            if (__pyx_t_76 < 0) {
              __pyx_t_76 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_76 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_76 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_77 < 0) {
              __pyx_t_77 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_77 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_77 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 639; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v2 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_76, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_77, __pyx_pybuffernd_d4.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":640
 *                         v1 = pindex + d4[l,1]
 *                         v2 = pindex + d4[l,2]
 *                         v3 = pindex + d4[l,3]             # <<<<<<<<<<<<<<
 *                         w1 = d4[l,5]
 *                         w2 = d4[l,6]
 */
            __pyx_t_78 = __pyx_v_l;
            __pyx_t_79 = 3;
            __pyx_t_8 = -1;
            if (__pyx_t_78 < 0) {
              __pyx_t_78 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_78 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_78 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_79 < 0) {
              __pyx_t_79 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_79 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_79 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 640; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v3 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_78, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_79, __pyx_pybuffernd_d4.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":641
 *                         v2 = pindex + d4[l,2]
 *                         v3 = pindex + d4[l,3]
 *                         w1 = d4[l,5]             # <<<<<<<<<<<<<<
 *                         w2 = d4[l,6]
 *                         w3 = d4[l,7]
 */
            __pyx_t_80 = __pyx_v_l;
            __pyx_t_81 = 5;
            __pyx_t_8 = -1;
            if (__pyx_t_80 < 0) {
              __pyx_t_80 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_80 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_80 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_81 < 0) {
              __pyx_t_81 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_81 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_81 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 641; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_w1 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_80, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_81, __pyx_pybuffernd_d4.diminfo[1].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":642
 *                         v3 = pindex + d4[l,3]
 *                         w1 = d4[l,5]
 *                         w2 = d4[l,6]             # <<<<<<<<<<<<<<
 *                         w3 = d4[l,7]
 * 
 */
            __pyx_t_82 = __pyx_v_l;
            __pyx_t_83 = 6;
            __pyx_t_8 = -1;
            if (__pyx_t_82 < 0) {
              __pyx_t_82 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_82 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_82 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_83 < 0) {
              __pyx_t_83 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_83 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_83 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 642; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_w2 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_82, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_83, __pyx_pybuffernd_d4.diminfo[1].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":643
 *                         w1 = d4[l,5]
 *                         w2 = d4[l,6]
 *                         w3 = d4[l,7]             # <<<<<<<<<<<<<<
 * 
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]
 */
            __pyx_t_84 = __pyx_v_l;
            __pyx_t_85 = 7;
            __pyx_t_8 = -1;
            if (__pyx_t_84 < 0) {
              __pyx_t_84 += __pyx_pybuffernd_d4.diminfo[0].shape;
              if (unlikely(__pyx_t_84 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_84 >= __pyx_pybuffernd_d4.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_85 < 0) {
              __pyx_t_85 += __pyx_pybuffernd_d4.diminfo[1].shape;
              if (unlikely(__pyx_t_85 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_85 >= __pyx_pybuffernd_d4.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 643; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_w3 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d4.rcbuffer->pybuffer.buf, __pyx_t_84, __pyx_pybuffernd_d4.diminfo[0].strides, __pyx_t_85, __pyx_pybuffernd_d4.diminfo[1].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":645
 *                         w3 = d4[l,7]
 * 
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]             # <<<<<<<<<<<<<<
 * 
 *                         l3 = l3 + m * mu3_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 */
            __pyx_t_86 = __pyx_v_v1;
            __pyx_t_8 = -1;
            if (__pyx_t_86 < 0) {
              __pyx_t_86 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_86 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_86 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_87 = __pyx_v_v2;
            __pyx_t_8 = -1;
            if (__pyx_t_87 < 0) {
              __pyx_t_87 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_87 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_87 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_88 = __pyx_v_v3;
            __pyx_t_8 = -1;
            if (__pyx_t_88 < 0) {
              __pyx_t_88 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_88 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_88 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_m = (((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_86, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_87, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_88, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":647
 *                         m = m * fpmask[v1] * fpmask[v2] * fpmask[v3]
 * 
 *                         l3 = l3 + m * mu3_tet(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 */
            __pyx_t_89 = __pyx_v_w0;
            __pyx_t_90 = __pyx_v_w0;
            __pyx_t_8 = -1;
            if (__pyx_t_89 < 0) {
              __pyx_t_89 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_89 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_89 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_90 < 0) {
              __pyx_t_90 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_90 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_90 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 647; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_91 = __pyx_v_w0;
            __pyx_t_92 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_91 < 0) {
              __pyx_t_91 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_91 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_91 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_92 < 0) {
              __pyx_t_92 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_92 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_92 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 647; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_93 = __pyx_v_w0;
            __pyx_t_94 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_93 < 0) {
              __pyx_t_93 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_93 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_93 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_94 < 0) {
              __pyx_t_94 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_94 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_94 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 647; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":648
 * 
 *                         l3 = l3 + m * mu3_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],             # <<<<<<<<<<<<<<
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 *                                               D[w3,w3])
 */
            __pyx_t_95 = __pyx_v_w0;
            __pyx_t_96 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_95 < 0) {
              __pyx_t_95 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_95 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_95 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_96 < 0) {
              __pyx_t_96 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_96 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_96 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 648; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_97 = __pyx_v_w1;
            __pyx_t_98 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_97 < 0) {
              __pyx_t_97 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_97 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_97 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_98 < 0) {
              __pyx_t_98 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_98 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_98 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 648; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_99 = __pyx_v_w1;
            __pyx_t_100 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_99 < 0) {
              __pyx_t_99 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_99 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_99 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_100 < 0) {
              __pyx_t_100 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_100 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_100 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 648; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":649
 *                         l3 = l3 + m * mu3_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],             # <<<<<<<<<<<<<<
 *                                               D[w3,w3])
 * 
 */
            __pyx_t_101 = __pyx_v_w1;
            __pyx_t_102 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_101 < 0) {
              __pyx_t_101 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_101 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_101 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_102 < 0) {
              __pyx_t_102 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_102 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_102 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 649; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_103 = __pyx_v_w2;
            __pyx_t_104 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_103 < 0) {
              __pyx_t_103 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_103 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_103 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_104 < 0) {
              __pyx_t_104 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_104 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_104 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 649; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_105 = __pyx_v_w2;
            __pyx_t_106 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_105 < 0) {
              __pyx_t_105 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_105 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_105 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_106 < 0) {
              __pyx_t_106 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_106 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_106 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 649; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":650
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 *                                               D[w3,w3])             # <<<<<<<<<<<<<<
 * 
 *                         l2 = l2 - m * mu2_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 */
            __pyx_t_107 = __pyx_v_w3;
            __pyx_t_108 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_107 < 0) {
              __pyx_t_107 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_107 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_107 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_108 < 0) {
              __pyx_t_108 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_108 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_108 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 650; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_l3 = (__pyx_v_l3 + (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu3_tet((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_89, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_90, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_91, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_92, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_93, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_94, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_95, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_96, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_97, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_98, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_99, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_100, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_101, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_102, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_103, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_104, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_105, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_106, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_107, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_108, __pyx_pybuffernd_D.diminfo[1].strides)), 0)));

            /* "nipy/algorithms/statistics/intvol.pyx":652
 *                                               D[w3,w3])
 * 
 *                         l2 = l2 - m * mu2_tet(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 */
            __pyx_t_109 = __pyx_v_w0;
            __pyx_t_110 = __pyx_v_w0;
            __pyx_t_8 = -1;
            if (__pyx_t_109 < 0) {
              __pyx_t_109 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_109 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_109 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_110 < 0) {
              __pyx_t_110 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_110 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_110 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 652; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_111 = __pyx_v_w0;
            __pyx_t_112 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_111 < 0) {
              __pyx_t_111 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_111 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_111 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_112 < 0) {
              __pyx_t_112 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_112 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_112 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 652; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_113 = __pyx_v_w0;
            __pyx_t_114 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_113 < 0) {
              __pyx_t_113 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_113 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_113 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_114 < 0) {
              __pyx_t_114 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_114 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_114 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 652; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":653
 * 
 *                         l2 = l2 - m * mu2_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],             # <<<<<<<<<<<<<<
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 *                                               D[w3,w3])
 */
            __pyx_t_115 = __pyx_v_w0;
            __pyx_t_116 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_115 < 0) {
              __pyx_t_115 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_115 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_115 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_116 < 0) {
              __pyx_t_116 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_116 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_116 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_117 = __pyx_v_w1;
            __pyx_t_118 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_117 < 0) {
              __pyx_t_117 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_117 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_117 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_118 < 0) {
              __pyx_t_118 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_118 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_118 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_119 = __pyx_v_w1;
            __pyx_t_120 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_119 < 0) {
              __pyx_t_119 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_119 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_119 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_120 < 0) {
              __pyx_t_120 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_120 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_120 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 653; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":654
 *                         l2 = l2 - m * mu2_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],             # <<<<<<<<<<<<<<
 *                                               D[w3,w3])
 * 
 */
            __pyx_t_121 = __pyx_v_w1;
            __pyx_t_122 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_121 < 0) {
              __pyx_t_121 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_121 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_121 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_122 < 0) {
              __pyx_t_122 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_122 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_122 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 654; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_123 = __pyx_v_w2;
            __pyx_t_124 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_123 < 0) {
              __pyx_t_123 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_123 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_123 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_124 < 0) {
              __pyx_t_124 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_124 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_124 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 654; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_125 = __pyx_v_w2;
            __pyx_t_126 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_125 < 0) {
              __pyx_t_125 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_125 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_125 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_126 < 0) {
              __pyx_t_126 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_126 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_126 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 654; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":655
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 *                                               D[w3,w3])             # <<<<<<<<<<<<<<
 * 
 *                         l1 = l1 + m * mu1_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 */
            __pyx_t_127 = __pyx_v_w3;
            __pyx_t_128 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_127 < 0) {
              __pyx_t_127 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_127 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_127 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_128 < 0) {
              __pyx_t_128 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_128 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_128 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 655; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_l2 = (__pyx_v_l2 - (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tet((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_109, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_110, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_111, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_112, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_113, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_114, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_115, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_116, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_117, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_118, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_119, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_120, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_121, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_122, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_123, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_124, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_125, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_126, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_127, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_128, __pyx_pybuffernd_D.diminfo[1].strides)), 0)));

            /* "nipy/algorithms/statistics/intvol.pyx":657
 *                                               D[w3,w3])
 * 
 *                         l1 = l1 + m * mu1_tet(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 */
            __pyx_t_129 = __pyx_v_w0;
            __pyx_t_130 = __pyx_v_w0;
            __pyx_t_8 = -1;
            if (__pyx_t_129 < 0) {
              __pyx_t_129 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_129 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_129 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_130 < 0) {
              __pyx_t_130 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_130 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_130 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_131 = __pyx_v_w0;
            __pyx_t_132 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_131 < 0) {
              __pyx_t_131 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_131 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_131 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_132 < 0) {
              __pyx_t_132 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_132 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_132 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_133 = __pyx_v_w0;
            __pyx_t_134 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_133 < 0) {
              __pyx_t_133 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_133 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_133 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_134 < 0) {
              __pyx_t_134 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_134 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_134 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 657; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":658
 * 
 *                         l1 = l1 + m * mu1_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],             # <<<<<<<<<<<<<<
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 *                                               D[w3,w3])
 */
            __pyx_t_135 = __pyx_v_w0;
            __pyx_t_136 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_135 < 0) {
              __pyx_t_135 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_135 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_135 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_136 < 0) {
              __pyx_t_136 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_136 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_136 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 658; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_137 = __pyx_v_w1;
            __pyx_t_138 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_137 < 0) {
              __pyx_t_137 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_137 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_137 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_138 < 0) {
              __pyx_t_138 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_138 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_138 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 658; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_139 = __pyx_v_w1;
            __pyx_t_140 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_139 < 0) {
              __pyx_t_139 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_139 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_139 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_140 < 0) {
              __pyx_t_140 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_140 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_140 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 658; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":659
 *                         l1 = l1 + m * mu1_tet(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],             # <<<<<<<<<<<<<<
 *                                               D[w3,w3])
 * 
 */
            __pyx_t_141 = __pyx_v_w1;
            __pyx_t_142 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_141 < 0) {
              __pyx_t_141 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_141 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_141 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_142 < 0) {
              __pyx_t_142 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_142 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_142 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 659; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_143 = __pyx_v_w2;
            __pyx_t_144 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_143 < 0) {
              __pyx_t_143 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_143 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_143 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_144 < 0) {
              __pyx_t_144 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_144 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_144 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 659; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_145 = __pyx_v_w2;
            __pyx_t_146 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_145 < 0) {
              __pyx_t_145 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_145 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_145 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_146 < 0) {
              __pyx_t_146 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_146 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_146 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 659; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":660
 *                                               D[w0,w3], D[w1,w1], D[w1,w2],
 *                                               D[w1,w3], D[w2,w2], D[w2,w3],
 *                                               D[w3,w3])             # <<<<<<<<<<<<<<
 * 
 *                         l0 = l0 - m
 */
            __pyx_t_147 = __pyx_v_w3;
            __pyx_t_148 = __pyx_v_w3;
            __pyx_t_8 = -1;
            if (__pyx_t_147 < 0) {
              __pyx_t_147 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_147 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_147 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_148 < 0) {
              __pyx_t_148 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_148 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_148 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 660; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_l1 = (__pyx_v_l1 + (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tet((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_129, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_130, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_131, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_132, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_133, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_134, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_135, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_136, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_137, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_138, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_139, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_140, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_141, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_142, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_143, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_144, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_145, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_146, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_147, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_148, __pyx_pybuffernd_D.diminfo[1].strides)), 0)));

            /* "nipy/algorithms/statistics/intvol.pyx":662
 *                                               D[w3,w3])
 * 
 *                         l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *                 for l in range(ds3):
 */
            __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
            goto __pyx_L40;
          }
          __pyx_L40:;
        }

        /* "nipy/algorithms/statistics/intvol.pyx":664
 *                         l0 = l0 - m
 * 
 *                 for l in range(ds3):             # <<<<<<<<<<<<<<
 *                     v0 = pindex + d3[l,0]
 *                     w0 = d3[l,3]
 */
        __pyx_t_51 = __pyx_v_ds3;
        for (__pyx_t_55 = 0; __pyx_t_55 < __pyx_t_51; __pyx_t_55+=1) {
          __pyx_v_l = __pyx_t_55;

          /* "nipy/algorithms/statistics/intvol.pyx":665
 * 
 *                 for l in range(ds3):
 *                     v0 = pindex + d3[l,0]             # <<<<<<<<<<<<<<
 *                     w0 = d3[l,3]
 *                     m = fpmask[v0]
 */
          __pyx_t_149 = __pyx_v_l;
          __pyx_t_150 = 0;
          __pyx_t_8 = -1;
          if (__pyx_t_149 < 0) {
            __pyx_t_149 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_149 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_149 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_150 < 0) {
            __pyx_t_150 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_150 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_150 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 665; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v0 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_149, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_150, __pyx_pybuffernd_d3.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":666
 *                 for l in range(ds3):
 *                     v0 = pindex + d3[l,0]
 *                     w0 = d3[l,3]             # <<<<<<<<<<<<<<
 *                     m = fpmask[v0]
 *                     if m:
 */
          __pyx_t_151 = __pyx_v_l;
          __pyx_t_152 = 3;
          __pyx_t_8 = -1;
          if (__pyx_t_151 < 0) {
            __pyx_t_151 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_151 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_151 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_152 < 0) {
            __pyx_t_152 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_152 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_152 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 666; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_w0 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_151, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_152, __pyx_pybuffernd_d3.diminfo[1].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":667
 *                     v0 = pindex + d3[l,0]
 *                     w0 = d3[l,3]
 *                     m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                     if m:
 *                         v1 = pindex + d3[l,1]
 */
          __pyx_t_153 = __pyx_v_v0;
          __pyx_t_8 = -1;
          if (__pyx_t_153 < 0) {
            __pyx_t_153 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_153 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_153 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 667; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_153, __pyx_pybuffernd_fpmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":668
 *                     w0 = d3[l,3]
 *                     m = fpmask[v0]
 *                     if m:             # <<<<<<<<<<<<<<
 *                         v1 = pindex + d3[l,1]
 *                         v2 = pindex + d3[l,2]
 */
          if (__pyx_v_m) {

            /* "nipy/algorithms/statistics/intvol.pyx":669
 *                     m = fpmask[v0]
 *                     if m:
 *                         v1 = pindex + d3[l,1]             # <<<<<<<<<<<<<<
 *                         v2 = pindex + d3[l,2]
 *                         w1 = d3[l,4]
 */
            __pyx_t_154 = __pyx_v_l;
            __pyx_t_155 = 1;
            __pyx_t_8 = -1;
            if (__pyx_t_154 < 0) {
              __pyx_t_154 += __pyx_pybuffernd_d3.diminfo[0].shape;
              if (unlikely(__pyx_t_154 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_154 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_155 < 0) {
              __pyx_t_155 += __pyx_pybuffernd_d3.diminfo[1].shape;
              if (unlikely(__pyx_t_155 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_155 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 669; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v1 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_154, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_155, __pyx_pybuffernd_d3.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":670
 *                     if m:
 *                         v1 = pindex + d3[l,1]
 *                         v2 = pindex + d3[l,2]             # <<<<<<<<<<<<<<
 *                         w1 = d3[l,4]
 *                         w2 = d3[l,5]
 */
            __pyx_t_156 = __pyx_v_l;
            __pyx_t_157 = 2;
            __pyx_t_8 = -1;
            if (__pyx_t_156 < 0) {
              __pyx_t_156 += __pyx_pybuffernd_d3.diminfo[0].shape;
              if (unlikely(__pyx_t_156 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_156 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_157 < 0) {
              __pyx_t_157 += __pyx_pybuffernd_d3.diminfo[1].shape;
              if (unlikely(__pyx_t_157 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_157 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 670; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v2 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_156, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_157, __pyx_pybuffernd_d3.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":671
 *                         v1 = pindex + d3[l,1]
 *                         v2 = pindex + d3[l,2]
 *                         w1 = d3[l,4]             # <<<<<<<<<<<<<<
 *                         w2 = d3[l,5]
 * 
 */
            __pyx_t_158 = __pyx_v_l;
            __pyx_t_159 = 4;
            __pyx_t_8 = -1;
            if (__pyx_t_158 < 0) {
              __pyx_t_158 += __pyx_pybuffernd_d3.diminfo[0].shape;
              if (unlikely(__pyx_t_158 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_158 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_159 < 0) {
              __pyx_t_159 += __pyx_pybuffernd_d3.diminfo[1].shape;
              if (unlikely(__pyx_t_159 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_159 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 671; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_w1 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_158, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_159, __pyx_pybuffernd_d3.diminfo[1].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":672
 *                         v2 = pindex + d3[l,2]
 *                         w1 = d3[l,4]
 *                         w2 = d3[l,5]             # <<<<<<<<<<<<<<
 * 
 *                         m = m * fpmask[v1] * fpmask[v2]
 */
            __pyx_t_160 = __pyx_v_l;
            __pyx_t_161 = 5;
            __pyx_t_8 = -1;
            if (__pyx_t_160 < 0) {
              __pyx_t_160 += __pyx_pybuffernd_d3.diminfo[0].shape;
              if (unlikely(__pyx_t_160 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_160 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_161 < 0) {
              __pyx_t_161 += __pyx_pybuffernd_d3.diminfo[1].shape;
              if (unlikely(__pyx_t_161 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_161 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 672; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_w2 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_160, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_161, __pyx_pybuffernd_d3.diminfo[1].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":674
 *                         w2 = d3[l,5]
 * 
 *                         m = m * fpmask[v1] * fpmask[v2]             # <<<<<<<<<<<<<<
 *                         l2 = l2 + m * mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])
 */
            __pyx_t_162 = __pyx_v_v1;
            __pyx_t_8 = -1;
            if (__pyx_t_162 < 0) {
              __pyx_t_162 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_162 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_162 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_163 = __pyx_v_v2;
            __pyx_t_8 = -1;
            if (__pyx_t_163 < 0) {
              __pyx_t_163 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_163 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_163 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 674; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_m = ((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_162, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_163, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":675
 * 
 *                         m = m * fpmask[v1] * fpmask[v2]
 *                         l2 = l2 + m * mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])
 * 
 */
            __pyx_t_164 = __pyx_v_w0;
            __pyx_t_165 = __pyx_v_w0;
            __pyx_t_8 = -1;
            if (__pyx_t_164 < 0) {
              __pyx_t_164 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_164 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_164 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_165 < 0) {
              __pyx_t_165 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_165 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_165 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_166 = __pyx_v_w0;
            __pyx_t_167 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_166 < 0) {
              __pyx_t_166 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_166 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_166 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_167 < 0) {
              __pyx_t_167 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_167 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_167 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_168 = __pyx_v_w0;
            __pyx_t_169 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_168 < 0) {
              __pyx_t_168 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_168 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_168 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_169 < 0) {
              __pyx_t_169 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_169 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_169 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 675; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":676
 *                         m = m * fpmask[v1] * fpmask[v2]
 *                         l2 = l2 + m * mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])             # <<<<<<<<<<<<<<
 * 
 *                         l1 = l1 - m * mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 */
            __pyx_t_170 = __pyx_v_w1;
            __pyx_t_171 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_170 < 0) {
              __pyx_t_170 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_170 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_170 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_171 < 0) {
              __pyx_t_171 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_171 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_171 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 676; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_172 = __pyx_v_w1;
            __pyx_t_173 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_172 < 0) {
              __pyx_t_172 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_172 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_172 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_173 < 0) {
              __pyx_t_173 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_173 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_173 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 676; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_174 = __pyx_v_w2;
            __pyx_t_175 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_174 < 0) {
              __pyx_t_174 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_174 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_174 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_175 < 0) {
              __pyx_t_175 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_175 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_175 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 676; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_l2 = (__pyx_v_l2 + (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_164, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_165, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_166, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_167, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_168, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_169, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_170, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_171, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_172, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_173, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_174, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_175, __pyx_pybuffernd_D.diminfo[1].strides)), 0)));

            /* "nipy/algorithms/statistics/intvol.pyx":678
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])
 * 
 *                         l1 = l1 - m * mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])
 * 
 */
            __pyx_t_176 = __pyx_v_w0;
            __pyx_t_177 = __pyx_v_w0;
            __pyx_t_8 = -1;
            if (__pyx_t_176 < 0) {
              __pyx_t_176 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_176 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_176 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_177 < 0) {
              __pyx_t_177 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_177 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_177 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 678; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_178 = __pyx_v_w0;
            __pyx_t_179 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_178 < 0) {
              __pyx_t_178 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_178 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_178 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_179 < 0) {
              __pyx_t_179 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_179 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_179 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 678; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_180 = __pyx_v_w0;
            __pyx_t_181 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_180 < 0) {
              __pyx_t_180 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_180 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_180 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_181 < 0) {
              __pyx_t_181 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_181 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_181 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 678; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }

            /* "nipy/algorithms/statistics/intvol.pyx":679
 * 
 *                         l1 = l1 - m * mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])             # <<<<<<<<<<<<<<
 * 
 *                         l0 = l0 + m
 */
            __pyx_t_182 = __pyx_v_w1;
            __pyx_t_183 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_182 < 0) {
              __pyx_t_182 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_182 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_182 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_183 < 0) {
              __pyx_t_183 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_183 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_183 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 679; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_184 = __pyx_v_w1;
            __pyx_t_185 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_184 < 0) {
              __pyx_t_184 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_184 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_184 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_185 < 0) {
              __pyx_t_185 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_185 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_185 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 679; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_186 = __pyx_v_w2;
            __pyx_t_187 = __pyx_v_w2;
            __pyx_t_8 = -1;
            if (__pyx_t_186 < 0) {
              __pyx_t_186 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_186 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_186 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_187 < 0) {
              __pyx_t_187 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_187 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_187 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 679; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_l1 = (__pyx_v_l1 - (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tri((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_176, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_177, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_178, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_179, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_180, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_181, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_182, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_183, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_184, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_185, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_186, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_187, __pyx_pybuffernd_D.diminfo[1].strides)), 0)));

            /* "nipy/algorithms/statistics/intvol.pyx":681
 *                                               D[w1,w1], D[w1,w2], D[w2,w2])
 * 
 *                         l0 = l0 + m             # <<<<<<<<<<<<<<
 * 
 *                 for l in range(ds2):
 */
            __pyx_v_l0 = (__pyx_v_l0 + __pyx_v_m);
            goto __pyx_L43;
          }
          __pyx_L43:;
        }

        /* "nipy/algorithms/statistics/intvol.pyx":683
 *                         l0 = l0 + m
 * 
 *                 for l in range(ds2):             # <<<<<<<<<<<<<<
 *                     v0 = pindex + d2[l,0]
 *                     w0 = d2[l,2]
 */
        __pyx_t_51 = __pyx_v_ds2;
        for (__pyx_t_55 = 0; __pyx_t_55 < __pyx_t_51; __pyx_t_55+=1) {
          __pyx_v_l = __pyx_t_55;

          /* "nipy/algorithms/statistics/intvol.pyx":684
 * 
 *                 for l in range(ds2):
 *                     v0 = pindex + d2[l,0]             # <<<<<<<<<<<<<<
 *                     w0 = d2[l,2]
 *                     m = fpmask[v0]
 */
          __pyx_t_188 = __pyx_v_l;
          __pyx_t_189 = 0;
          __pyx_t_8 = -1;
          if (__pyx_t_188 < 0) {
            __pyx_t_188 += __pyx_pybuffernd_d2.diminfo[0].shape;
            if (unlikely(__pyx_t_188 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_188 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_189 < 0) {
            __pyx_t_189 += __pyx_pybuffernd_d2.diminfo[1].shape;
            if (unlikely(__pyx_t_189 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_189 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 684; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v0 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_188, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_189, __pyx_pybuffernd_d2.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":685
 *                 for l in range(ds2):
 *                     v0 = pindex + d2[l,0]
 *                     w0 = d2[l,2]             # <<<<<<<<<<<<<<
 *                     m = fpmask[v0]
 *                     if m:
 */
          __pyx_t_190 = __pyx_v_l;
          __pyx_t_191 = 2;
          __pyx_t_8 = -1;
          if (__pyx_t_190 < 0) {
            __pyx_t_190 += __pyx_pybuffernd_d2.diminfo[0].shape;
            if (unlikely(__pyx_t_190 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_190 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_191 < 0) {
            __pyx_t_191 += __pyx_pybuffernd_d2.diminfo[1].shape;
            if (unlikely(__pyx_t_191 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_191 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 685; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_w0 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_190, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_191, __pyx_pybuffernd_d2.diminfo[1].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":686
 *                     v0 = pindex + d2[l,0]
 *                     w0 = d2[l,2]
 *                     m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                     if m:
 *                         v1 = pindex + d2[l,1]
 */
          __pyx_t_192 = __pyx_v_v0;
          __pyx_t_8 = -1;
          if (__pyx_t_192 < 0) {
            __pyx_t_192 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_192 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_192 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 686; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_192, __pyx_pybuffernd_fpmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":687
 *                     w0 = d2[l,2]
 *                     m = fpmask[v0]
 *                     if m:             # <<<<<<<<<<<<<<
 *                         v1 = pindex + d2[l,1]
 *                         w1 = d2[l,3]
 */
          if (__pyx_v_m) {

            /* "nipy/algorithms/statistics/intvol.pyx":688
 *                     m = fpmask[v0]
 *                     if m:
 *                         v1 = pindex + d2[l,1]             # <<<<<<<<<<<<<<
 *                         w1 = d2[l,3]
 *                         m = m * fpmask[v1]
 */
            __pyx_t_193 = __pyx_v_l;
            __pyx_t_194 = 1;
            __pyx_t_8 = -1;
            if (__pyx_t_193 < 0) {
              __pyx_t_193 += __pyx_pybuffernd_d2.diminfo[0].shape;
              if (unlikely(__pyx_t_193 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_193 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_194 < 0) {
              __pyx_t_194 += __pyx_pybuffernd_d2.diminfo[1].shape;
              if (unlikely(__pyx_t_194 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_194 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 688; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_v1 = (__pyx_v_pindex + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_193, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_194, __pyx_pybuffernd_d2.diminfo[1].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":689
 *                     if m:
 *                         v1 = pindex + d2[l,1]
 *                         w1 = d2[l,3]             # <<<<<<<<<<<<<<
 *                         m = m * fpmask[v1]
 *                         l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
 */
            __pyx_t_195 = __pyx_v_l;
            __pyx_t_196 = 3;
            __pyx_t_8 = -1;
            if (__pyx_t_195 < 0) {
              __pyx_t_195 += __pyx_pybuffernd_d2.diminfo[0].shape;
              if (unlikely(__pyx_t_195 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_195 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_196 < 0) {
              __pyx_t_196 += __pyx_pybuffernd_d2.diminfo[1].shape;
              if (unlikely(__pyx_t_196 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_196 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 689; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_w1 = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_195, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_196, __pyx_pybuffernd_d2.diminfo[1].strides));

            /* "nipy/algorithms/statistics/intvol.pyx":690
 *                         v1 = pindex + d2[l,1]
 *                         w1 = d2[l,3]
 *                         m = m * fpmask[v1]             # <<<<<<<<<<<<<<
 *                         l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
 * 
 */
            __pyx_t_197 = __pyx_v_v1;
            __pyx_t_8 = -1;
            if (__pyx_t_197 < 0) {
              __pyx_t_197 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
              if (unlikely(__pyx_t_197 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_197 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 690; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_m = (__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_197, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

            /* "nipy/algorithms/statistics/intvol.pyx":691
 *                         w1 = d2[l,3]
 *                         m = m * fpmask[v1]
 *                         l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])             # <<<<<<<<<<<<<<
 * 
 *                         l0 = l0 - m
 */
            __pyx_t_198 = __pyx_v_w0;
            __pyx_t_199 = __pyx_v_w0;
            __pyx_t_8 = -1;
            if (__pyx_t_198 < 0) {
              __pyx_t_198 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_198 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_198 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_199 < 0) {
              __pyx_t_199 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_199 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_199 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 691; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_200 = __pyx_v_w0;
            __pyx_t_201 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_200 < 0) {
              __pyx_t_200 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_200 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_200 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_201 < 0) {
              __pyx_t_201 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_201 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_201 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 691; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_202 = __pyx_v_w1;
            __pyx_t_203 = __pyx_v_w1;
            __pyx_t_8 = -1;
            if (__pyx_t_202 < 0) {
              __pyx_t_202 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_202 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_202 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_203 < 0) {
              __pyx_t_203 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_203 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_203 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 691; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_l1 = (__pyx_v_l1 + (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_198, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_199, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_200, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_201, __pyx_pybuffernd_D.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_202, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_203, __pyx_pybuffernd_D.diminfo[1].strides)), 0)));

            /* "nipy/algorithms/statistics/intvol.pyx":693
 *                         l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
 * 
 *                         l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *     l0 += mask.sum()
 */
            __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
            goto __pyx_L46;
          }
          __pyx_L46:;
        }
      }
    }
  }

  /* "nipy/algorithms/statistics/intvol.pyx":695
 *                         l0 = l0 - m
 * 
 *     l0 += mask.sum()             # <<<<<<<<<<<<<<
 *     return np.array([l0, l1, l2, l3])
 * 
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 695; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_38 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__sum); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 695; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_1 = PyObject_Call(__pyx_t_38, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 695; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __pyx_t_38 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 695; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_204 = __pyx_PyFloat_AsDouble(__pyx_t_38); if (unlikely((__pyx_t_204 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 695; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __pyx_v_l0 = __pyx_t_204;

  /* "nipy/algorithms/statistics/intvol.pyx":696
 * 
 *     l0 += mask.sum()
 *     return np.array([l0, l1, l2, l3])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_38 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_38, __pyx_n_s__array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_38); __pyx_t_38 = 0;
  __pyx_t_38 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_38);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_l1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_37 = PyFloat_FromDouble(__pyx_v_l2); if (unlikely(!__pyx_t_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_37);
  __pyx_t_36 = PyFloat_FromDouble(__pyx_v_l3); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  __pyx_t_35 = PyList_New(4); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  PyList_SET_ITEM(__pyx_t_35, 0, __pyx_t_38);
  __Pyx_GIVEREF(__pyx_t_38);
  PyList_SET_ITEM(__pyx_t_35, 1, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_35, 2, __pyx_t_37);
  __Pyx_GIVEREF(__pyx_t_37);
  PyList_SET_ITEM(__pyx_t_35, 3, __pyx_t_36);
  __Pyx_GIVEREF(__pyx_t_36);
  __pyx_t_38 = 0;
  __pyx_t_5 = 0;
  __pyx_t_37 = 0;
  __pyx_t_36 = 0;
  __pyx_t_36 = PyTuple_New(1); if (unlikely(!__pyx_t_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_36);
  PyTuple_SET_ITEM(__pyx_t_36, 0, ((PyObject *)__pyx_t_35));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_35));
  __pyx_t_35 = 0;
  __pyx_t_35 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_36), NULL); if (unlikely(!__pyx_t_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 696; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_35);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_36)); __pyx_t_36 = 0;
  __pyx_r = __pyx_t_35;
  __pyx_t_35 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_34);
  __Pyx_XDECREF(__pyx_t_35);
  __Pyx_XDECREF(__pyx_t_36);
  __Pyx_XDECREF(__pyx_t_37);
  __Pyx_XDECREF(__pyx_t_38);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_D.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m2.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m3.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m4.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.Lips3d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_D.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d4.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m2.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m3.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_m4.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_coords_c);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask_c);
  __Pyx_XDECREF((PyObject *)__pyx_v_fcoords);
  __Pyx_XDECREF((PyObject *)__pyx_v_D);
  __Pyx_XDECREF((PyObject *)__pyx_v_fmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_fpmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_pmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_d4);
  __Pyx_XDECREF((PyObject *)__pyx_v_m4);
  __Pyx_XDECREF((PyObject *)__pyx_v_d3);
  __Pyx_XDECREF((PyObject *)__pyx_v_m3);
  __Pyx_XDECREF((PyObject *)__pyx_v_d2);
  __Pyx_XDECREF((PyObject *)__pyx_v_m2);
  __Pyx_XDECREF((PyObject *)__pyx_v_cvertices);
  __Pyx_XDECREF(__pyx_v_pmask_shape);
  __Pyx_XDECREF((PyObject *)__pyx_v_strides);
  __Pyx_XDECREF((PyObject *)__pyx_v_dstrides);
  __Pyx_XDECREF(__pyx_v_ss0d);
  __Pyx_XDECREF(__pyx_v_ss1d);
  __Pyx_XDECREF(__pyx_v_ss2d);
  __Pyx_XDECREF(__pyx_v_verts);
  __Pyx_XDECREF(__pyx_v_union);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_coords);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_19_convert_stride3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_18_convert_stride3[] = "\n    Take a voxel, expressed as in index in stride1 and\n    re-express it as an index in stride2\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_19_convert_stride3 = {__Pyx_NAMESTR("_convert_stride3"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_19_convert_stride3, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_18_convert_stride3)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_19_convert_stride3(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_v_stride1 = 0;
  PyObject *__pyx_v_stride2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_stride3 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__v,&__pyx_n_s__stride1,&__pyx_n_s__stride2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__v)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__stride1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_stride3", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__stride2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_stride3", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_convert_stride3") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_v = values[0];
    __pyx_v_stride1 = values[1];
    __pyx_v_stride2 = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_stride3", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._convert_stride3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_18_convert_stride3(__pyx_self, __pyx_v_v, __pyx_v_stride1, __pyx_v_stride2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":699
 * 
 * 
 * def _convert_stride3(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_18_convert_stride3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v, PyObject *__pyx_v_stride1, PyObject *__pyx_v_stride2) {
  PyObject *__pyx_v_v0 = NULL;
  PyObject *__pyx_v_v1 = NULL;
  PyObject *__pyx_v_v2 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_stride3", 0);
  __Pyx_INCREF(__pyx_v_v);

  /* "nipy/algorithms/statistics/intvol.pyx":704
 *     re-express it as an index in stride2
 *     """
 *     v0 = v // stride1[0]             # <<<<<<<<<<<<<<
 *     v -= v0 * stride1[0]
 *     v1 = v // stride1[1]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_stride1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 704; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_FloorDivide(__pyx_v_v, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 704; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":705
 *     """
 *     v0 = v // stride1[0]
 *     v -= v0 * stride1[0]             # <<<<<<<<<<<<<<
 *     v1 = v // stride1[1]
 *     v2 = v - v1 * stride1[1]
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_stride1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 705; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_v0, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 705; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceSubtract(__pyx_v_v, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 705; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_v_v);
  __pyx_v_v = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":706
 *     v0 = v // stride1[0]
 *     v -= v0 * stride1[0]
 *     v1 = v // stride1[1]             # <<<<<<<<<<<<<<
 *     v2 = v - v1 * stride1[1]
 *     return v0*stride2[0] + v1*stride2[1] + v2*stride2[2]
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_stride1, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 706; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_FloorDivide(__pyx_v_v, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 706; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_v1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":707
 *     v -= v0 * stride1[0]
 *     v1 = v // stride1[1]
 *     v2 = v - v1 * stride1[1]             # <<<<<<<<<<<<<<
 *     return v0*stride2[0] + v1*stride2[1] + v2*stride2[2]
 * 
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_stride1, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 707; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_v1, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 707; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_v, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 707; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_v2 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":708
 *     v1 = v // stride1[1]
 *     v2 = v - v1 * stride1[1]
 *     return v0*stride2[0] + v1*stride2[1] + v2*stride2[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_stride2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_v0, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_stride2, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_v1, __pyx_t_1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_stride2, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_3) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_v2, __pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._convert_stride3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v0);
  __Pyx_XDECREF(__pyx_v_v1);
  __Pyx_XDECREF(__pyx_v_v2);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_21_convert_stride2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_20_convert_stride2[] = "\n    Take a voxel, expressed as in index in stride1 and\n    re-express it as an index in stride2\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_21_convert_stride2 = {__Pyx_NAMESTR("_convert_stride2"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_21_convert_stride2, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_20_convert_stride2)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_21_convert_stride2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_v_stride1 = 0;
  PyObject *__pyx_v_stride2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_stride2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__v,&__pyx_n_s__stride1,&__pyx_n_s__stride2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__v)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__stride1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_stride2", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__stride2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_stride2", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_convert_stride2") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_v = values[0];
    __pyx_v_stride1 = values[1];
    __pyx_v_stride2 = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_stride2", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._convert_stride2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_20_convert_stride2(__pyx_self, __pyx_v_v, __pyx_v_stride1, __pyx_v_stride2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":711
 * 
 * 
 * def _convert_stride2(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_20_convert_stride2(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v, PyObject *__pyx_v_stride1, PyObject *__pyx_v_stride2) {
  PyObject *__pyx_v_v0 = NULL;
  PyObject *__pyx_v_v1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_stride2", 0);

  /* "nipy/algorithms/statistics/intvol.pyx":716
 *     re-express it as an index in stride2
 *     """
 *     v0 = v // stride1[0]             # <<<<<<<<<<<<<<
 *     v1 = v - v0 * stride1[0]
 *     return v0*stride2[0] + v1*stride2[1]
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_stride1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_FloorDivide(__pyx_v_v, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 716; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":717
 *     """
 *     v0 = v // stride1[0]
 *     v1 = v - v0 * stride1[0]             # <<<<<<<<<<<<<<
 *     return v0*stride2[0] + v1*stride2[1]
 * 
 */
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_stride1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 717; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_v0, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 717; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_v, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 717; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v1 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":718
 *     v0 = v // stride1[0]
 *     v1 = v - v0 * stride1[0]
 *     return v0*stride2[0] + v1*stride2[1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_stride2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_v0, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_stride2, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_v_v1, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 718; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._convert_stride2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v0);
  __Pyx_XDECREF(__pyx_v_v1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_23_convert_stride1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_22_convert_stride1[] = "\n    Take a voxel, expressed as in index in stride1 and\n    re-express it as an index in stride2\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_23_convert_stride1 = {__Pyx_NAMESTR("_convert_stride1"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_23_convert_stride1, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_22_convert_stride1)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_23_convert_stride1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_v = 0;
  PyObject *__pyx_v_stride1 = 0;
  PyObject *__pyx_v_stride2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_convert_stride1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__v,&__pyx_n_s__stride1,&__pyx_n_s__stride2,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__v)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__stride1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_stride1", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__stride2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_convert_stride1", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_convert_stride1") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_v = values[0];
    __pyx_v_stride1 = values[1];
    __pyx_v_stride2 = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_convert_stride1", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._convert_stride1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_22_convert_stride1(__pyx_self, __pyx_v_v, __pyx_v_stride1, __pyx_v_stride2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":721
 * 
 * 
 * def _convert_stride1(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_22_convert_stride1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_v, PyObject *__pyx_v_stride1, PyObject *__pyx_v_stride2) {
  PyObject *__pyx_v_v0 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_convert_stride1", 0);

  /* "nipy/algorithms/statistics/intvol.pyx":726
 *     re-express it as an index in stride2
 *     """
 *     v0 = v // stride1[0]             # <<<<<<<<<<<<<<
 *     return v0 * stride2[0]
 * 
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_stride1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_FloorDivide(__pyx_v_v, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 726; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_v0 = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":727
 *     """
 *     v0 = v // stride1[0]
 *     return v0 * stride2[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_stride2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 727; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_v0, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 727; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol._convert_stride1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v0);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_25Lips2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_24Lips2d[] = " Estimate intrinsic volumes for 2d region in `mask` given `coords`\n\n    Given a 2d `mask` and coordinates `coords`, estimate the intrinsic volumes\n    of the masked region. The region is broken up into triangles / edges /\n    vertices, which are included based on whether all voxels in the triangle /\n    edge / vertex are in the mask or not.\n\n    Parameters\n    ----------\n    coords : ndarray((N,i,j,k))\n         Coordinates for the voxels in the mask. ``N`` will often be 2 (for 2\n         dimensional coordinates, but can be any integer > 0\n    mask : ndarray((i,j), np.int)\n         Binary mask determining whether or not a voxel is in the mask.\n\n    Returns\n    -------\n    mu : ndarray\n        Array of intrinsic volumes [mu0, mu1, mu2], being, respectively:\n        #. Euler characteristic\n        #. 2 * mean caliper diameter\n        #. Area.\n\n    Notes\n    -----\n    The array mask is assumed to be binary. At the time of writing, it\n    is not clear how to get cython to use np.bool arrays.\n\n    References\n    ----------\n    Taylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n      with an application to brain mapping.\"\n      Journal of the American Statistical Association, 102(479):913-928.\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_25Lips2d = {__Pyx_NAMESTR("Lips2d"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_25Lips2d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_24Lips2d)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_25Lips2d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_coords = 0;
  PyObject *__pyx_v_mask = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Lips2d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__coords,&__pyx_n_s__mask,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__coords)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("Lips2d", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "Lips2d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coords = values[0];
    __pyx_v_mask = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Lips2d", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.Lips2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_24Lips2d(__pyx_self, __pyx_v_coords, __pyx_v_mask);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":730
 * 
 * 
 * def Lips2d(coords, mask):             # <<<<<<<<<<<<<<
 *     """ Estimate intrinsic volumes for 2d region in `mask` given `coords`
 * 
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_24Lips2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords, PyObject *__pyx_v_mask) {
  PyObject *__pyx_v_value = NULL;
  CYTHON_UNUSED PyArrayObject *__pyx_v_coords_c = 0;
  PyArrayObject *__pyx_v_mask_c = 0;
  PyArrayObject *__pyx_v_fcoords = 0;
  PyArrayObject *__pyx_v_D = 0;
  PyArrayObject *__pyx_v_fmask = 0;
  PyArrayObject *__pyx_v_fpmask = 0;
  PyArrayObject *__pyx_v_pmask = 0;
  PyArrayObject *__pyx_v_d3 = 0;
  PyArrayObject *__pyx_v_d2 = 0;
  PyArrayObject *__pyx_v_cvertices = 0;
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_j;
  npy_intp __pyx_v_l;
  npy_intp __pyx_v_r;
  npy_intp __pyx_v_s;
  npy_intp __pyx_v_rr;
  npy_intp __pyx_v_ss;
  npy_intp __pyx_v_mr;
  npy_intp __pyx_v_ms;
  npy_intp __pyx_v_s0;
  npy_intp __pyx_v_s1;
  npy_intp __pyx_v_ds2;
  npy_intp __pyx_v_ds3;
  npy_intp __pyx_v_index;
  npy_intp __pyx_v_m;
  npy_intp __pyx_v_npix;
  npy_intp __pyx_v_ss0;
  npy_intp __pyx_v_ss1;
  npy_intp __pyx_v_ss0d;
  npy_intp __pyx_v_ss1d;
  npy_intp __pyx_v_v0;
  npy_intp __pyx_v_v1;
  npy_intp __pyx_v_v2;
  double __pyx_v_l0;
  double __pyx_v_l1;
  double __pyx_v_l2;
  double __pyx_v_res;
  PyObject *__pyx_v_pmask_shape = NULL;
  PyArrayObject *__pyx_v_strides = 0;
  PyArrayObject *__pyx_v_dstrides = 0;
  PyObject *__pyx_v_verts = NULL;
  PyObject *__pyx_v_union = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_m3 = NULL;
  PyObject *__pyx_v_m2 = NULL;
  PyObject *__pyx_v_pindex = NULL;
  PyObject *__pyx_v_w0 = NULL;
  PyObject *__pyx_v_w1 = NULL;
  PyObject *__pyx_v_w2 = NULL;
  PyObject *__pyx_v_v = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_D;
  __Pyx_Buffer __pyx_pybuffer_D;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_coords_c;
  __Pyx_Buffer __pyx_pybuffer_coords_c;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_cvertices;
  __Pyx_Buffer __pyx_pybuffer_cvertices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d2;
  __Pyx_Buffer __pyx_pybuffer_d2;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d3;
  __Pyx_Buffer __pyx_pybuffer_d3;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dstrides;
  __Pyx_Buffer __pyx_pybuffer_dstrides;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fcoords;
  __Pyx_Buffer __pyx_pybuffer_fcoords;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fmask;
  __Pyx_Buffer __pyx_pybuffer_fmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fpmask;
  __Pyx_Buffer __pyx_pybuffer_fpmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask_c;
  __Pyx_Buffer __pyx_pybuffer_mask_c;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pmask;
  __Pyx_Buffer __pyx_pybuffer_pmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_strides;
  __Pyx_Buffer __pyx_pybuffer_strides;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyArrayObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  PyArrayObject *__pyx_t_13 = NULL;
  npy_intp __pyx_t_14;
  npy_intp __pyx_t_15;
  PyArrayObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  PyArrayObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyArrayObject *__pyx_t_20 = NULL;
  PyArrayObject *__pyx_t_21 = NULL;
  long __pyx_t_22;
  __pyx_t_5numpy_intp_t __pyx_t_23;
  long __pyx_t_24;
  __pyx_t_5numpy_intp_t __pyx_t_25;
  long __pyx_t_26;
  long __pyx_t_27;
  int __pyx_t_28;
  PyArrayObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  long __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  PyObject *(*__pyx_t_33)(PyObject *);
  PyObject *__pyx_t_34 = NULL;
  PyArrayObject *__pyx_t_35 = NULL;
  PyArrayObject *__pyx_t_36 = NULL;
  PyArrayObject *__pyx_t_37 = NULL;
  long __pyx_t_38;
  npy_intp __pyx_t_39;
  npy_intp __pyx_t_40;
  npy_intp __pyx_t_41;
  long __pyx_t_42;
  npy_intp __pyx_t_43;
  npy_intp __pyx_t_44;
  npy_intp __pyx_t_45;
  npy_intp __pyx_t_46;
  npy_intp __pyx_t_47;
  npy_intp __pyx_t_48;
  npy_intp __pyx_t_49;
  npy_intp __pyx_t_50;
  npy_intp __pyx_t_51;
  npy_intp __pyx_t_52;
  npy_intp __pyx_t_53;
  npy_intp __pyx_t_54;
  npy_intp __pyx_t_55;
  npy_intp __pyx_t_56;
  npy_intp __pyx_t_57;
  npy_intp __pyx_t_58;
  npy_intp __pyx_t_59;
  long __pyx_t_60;
  npy_intp __pyx_t_61;
  npy_intp __pyx_t_62;
  long __pyx_t_63;
  npy_intp __pyx_t_64;
  long __pyx_t_65;
  npy_intp __pyx_t_66;
  long __pyx_t_67;
  npy_intp __pyx_t_68;
  long __pyx_t_69;
  npy_intp __pyx_t_70;
  npy_intp __pyx_t_71;
  double __pyx_t_72;
  double __pyx_t_73;
  double __pyx_t_74;
  double __pyx_t_75;
  double __pyx_t_76;
  double __pyx_t_77;
  npy_intp __pyx_t_78;
  long __pyx_t_79;
  npy_intp __pyx_t_80;
  long __pyx_t_81;
  npy_intp __pyx_t_82;
  npy_intp __pyx_t_83;
  long __pyx_t_84;
  npy_intp __pyx_t_85;
  long __pyx_t_86;
  npy_intp __pyx_t_87;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Lips2d", 0);
  __Pyx_INCREF(__pyx_v_coords);
  __Pyx_INCREF(__pyx_v_mask);
  __pyx_pybuffer_coords_c.pybuffer.buf = NULL;
  __pyx_pybuffer_coords_c.refcount = 0;
  __pyx_pybuffernd_coords_c.data = NULL;
  __pyx_pybuffernd_coords_c.rcbuffer = &__pyx_pybuffer_coords_c;
  __pyx_pybuffer_mask_c.pybuffer.buf = NULL;
  __pyx_pybuffer_mask_c.refcount = 0;
  __pyx_pybuffernd_mask_c.data = NULL;
  __pyx_pybuffernd_mask_c.rcbuffer = &__pyx_pybuffer_mask_c;
  __pyx_pybuffer_fcoords.pybuffer.buf = NULL;
  __pyx_pybuffer_fcoords.refcount = 0;
  __pyx_pybuffernd_fcoords.data = NULL;
  __pyx_pybuffernd_fcoords.rcbuffer = &__pyx_pybuffer_fcoords;
  __pyx_pybuffer_D.pybuffer.buf = NULL;
  __pyx_pybuffer_D.refcount = 0;
  __pyx_pybuffernd_D.data = NULL;
  __pyx_pybuffernd_D.rcbuffer = &__pyx_pybuffer_D;
  __pyx_pybuffer_fmask.pybuffer.buf = NULL;
  __pyx_pybuffer_fmask.refcount = 0;
  __pyx_pybuffernd_fmask.data = NULL;
  __pyx_pybuffernd_fmask.rcbuffer = &__pyx_pybuffer_fmask;
  __pyx_pybuffer_fpmask.pybuffer.buf = NULL;
  __pyx_pybuffer_fpmask.refcount = 0;
  __pyx_pybuffernd_fpmask.data = NULL;
  __pyx_pybuffernd_fpmask.rcbuffer = &__pyx_pybuffer_fpmask;
  __pyx_pybuffer_pmask.pybuffer.buf = NULL;
  __pyx_pybuffer_pmask.refcount = 0;
  __pyx_pybuffernd_pmask.data = NULL;
  __pyx_pybuffernd_pmask.rcbuffer = &__pyx_pybuffer_pmask;
  __pyx_pybuffer_d3.pybuffer.buf = NULL;
  __pyx_pybuffer_d3.refcount = 0;
  __pyx_pybuffernd_d3.data = NULL;
  __pyx_pybuffernd_d3.rcbuffer = &__pyx_pybuffer_d3;
  __pyx_pybuffer_d2.pybuffer.buf = NULL;
  __pyx_pybuffer_d2.refcount = 0;
  __pyx_pybuffernd_d2.data = NULL;
  __pyx_pybuffernd_d2.rcbuffer = &__pyx_pybuffer_d2;
  __pyx_pybuffer_cvertices.pybuffer.buf = NULL;
  __pyx_pybuffer_cvertices.refcount = 0;
  __pyx_pybuffernd_cvertices.data = NULL;
  __pyx_pybuffernd_cvertices.rcbuffer = &__pyx_pybuffer_cvertices;
  __pyx_pybuffer_strides.pybuffer.buf = NULL;
  __pyx_pybuffer_strides.refcount = 0;
  __pyx_pybuffernd_strides.data = NULL;
  __pyx_pybuffernd_strides.rcbuffer = &__pyx_pybuffer_strides;
  __pyx_pybuffer_dstrides.pybuffer.buf = NULL;
  __pyx_pybuffer_dstrides.refcount = 0;
  __pyx_pybuffernd_dstrides.data = NULL;
  __pyx_pybuffernd_dstrides.rcbuffer = &__pyx_pybuffer_dstrides;

  /* "nipy/algorithms/statistics/intvol.pyx":765
 *       Journal of the American Statistical Association, 102(479):913-928.
 *     """
 *     if mask.shape != coords.shape[1:]:             # <<<<<<<<<<<<<<
 *         raise ValueError('shape of mask does not match coordinates')
 *     # if the data can be squeezed, we must use the lower dimensional function
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PySequence_GetSlice(__pyx_t_2, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "nipy/algorithms/statistics/intvol.pyx":766
 *     """
 *     if mask.shape != coords.shape[1:]:
 *         raise ValueError('shape of mask does not match coordinates')             # <<<<<<<<<<<<<<
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_40), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 766; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 766; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":768
 *         raise ValueError('shape of mask does not match coordinates')
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)             # <<<<<<<<<<<<<<
 *     if mask.ndim == 1:
 *         value = np.zeros(3)
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 768; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__squeeze); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 768; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 768; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  __pyx_t_1 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 768; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_v_mask);
  __pyx_v_mask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":769
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)
 *     if mask.ndim == 1:             # <<<<<<<<<<<<<<
 *         value = np.zeros(3)
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__ndim); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "nipy/algorithms/statistics/intvol.pyx":770
 *     mask = np.squeeze(mask)
 *     if mask.ndim == 1:
 *         value = np.zeros(3)             # <<<<<<<<<<<<<<
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         value[:2] = Lips1d(coords, mask)
 */
    __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__zeros); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_k_tuple_41), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_value = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "nipy/algorithms/statistics/intvol.pyx":771
 *     if mask.ndim == 1:
 *         value = np.zeros(3)
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)             # <<<<<<<<<<<<<<
 *         value[:2] = Lips1d(coords, mask)
 *         return value
 */
    __pyx_t_2 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__shape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_3) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Add(((PyObject *)__pyx_t_1), __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 771; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_v_coords);
    __pyx_v_coords = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "nipy/algorithms/statistics/intvol.pyx":772
 *         value = np.zeros(3)
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         value[:2] = Lips1d(coords, mask)             # <<<<<<<<<<<<<<
 *         return value
 * 
 */
    __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__Lips1d); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_coords);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_coords);
    __Pyx_GIVEREF(__pyx_v_coords);
    __Pyx_INCREF(__pyx_v_mask);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_mask);
    __Pyx_GIVEREF(__pyx_v_mask);
    __pyx_t_2 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
    if (__Pyx_PySequence_SetSlice(__pyx_v_value, 0, 2, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "nipy/algorithms/statistics/intvol.pyx":773
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         value[:2] = Lips1d(coords, mask)
 *         return value             # <<<<<<<<<<<<<<
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = __pyx_v_value;
    goto __pyx_L0;
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "nipy/algorithms/statistics/intvol.pyx":775
 *         return value
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):             # <<<<<<<<<<<<<<
 *       raise ValueError('mask should be filled with 0/1 '
 *                        'values, but be of type np.int')
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__unique); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  __pyx_t_5 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(((PyObject *)((PyObject*)(&PySet_Type))), ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__issubset); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyList_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = (!__pyx_t_4);
  if (__pyx_t_6) {

    /* "nipy/algorithms/statistics/intvol.pyx":776
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 * 
 */
    __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_42), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 776; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 776; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "nipy/algorithms/statistics/intvol.pyx":803
 *         double res
 * 
 *     coords_c = coords             # <<<<<<<<<<<<<<
 *     mask_c = mask
 *     l0 = 0; l1 = 0; l2 = 0
 */
  if (!(likely(((__pyx_v_coords) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_coords, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_7 = ((PyArrayObject *)__pyx_v_coords);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer, (PyObject*)__pyx_v_coords_c, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_coords_c.diminfo[0].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_coords_c.diminfo[0].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_coords_c.diminfo[1].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_coords_c.diminfo[1].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_coords_c.diminfo[2].strides = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_coords_c.diminfo[2].shape = __pyx_pybuffernd_coords_c.rcbuffer->pybuffer.shape[2];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_7 = 0;
  __Pyx_INCREF(__pyx_v_coords);
  __pyx_v_coords_c = ((PyArrayObject *)__pyx_v_coords);

  /* "nipy/algorithms/statistics/intvol.pyx":804
 * 
 *     coords_c = coords
 *     mask_c = mask             # <<<<<<<<<<<<<<
 *     l0 = 0; l1 = 0; l2 = 0
 * 
 */
  if (!(likely(((__pyx_v_mask) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_mask, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_12 = ((PyArrayObject *)__pyx_v_mask);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask_c, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_mask_c.diminfo[0].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask_c.diminfo[0].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask_c.diminfo[1].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask_c.diminfo[1].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 804; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_12 = 0;
  __Pyx_INCREF(__pyx_v_mask);
  __pyx_v_mask_c = ((PyArrayObject *)__pyx_v_mask);

  /* "nipy/algorithms/statistics/intvol.pyx":805
 *     coords_c = coords
 *     mask_c = mask
 *     l0 = 0; l1 = 0; l2 = 0             # <<<<<<<<<<<<<<
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 */
  __pyx_v_l0 = 0.0;
  __pyx_v_l1 = 0.0;
  __pyx_v_l2 = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":807
 *     l0 = 0; l1 = 0; l2 = 0
 * 
 *     pmask_shape = np.array(mask.shape) + 1             # <<<<<<<<<<<<<<
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1] = mask_c
 */
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__array); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Add(__pyx_t_5, __pyx_int_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 807; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_pmask_shape = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":808
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)             # <<<<<<<<<<<<<<
 *     pmask[:-1,:-1] = mask_c
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__zeros); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__int); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_pmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_pmask.diminfo[0].strides = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pmask.diminfo[0].shape = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pmask.diminfo[1].strides = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pmask.diminfo[1].shape = __pyx_pybuffernd_pmask.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 808; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_13 = 0;
  __pyx_v_pmask = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":809
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1 = pmask.shape[0], pmask.shape[1]
 */
  if (PyObject_SetItem(((PyObject *)__pyx_v_pmask), ((PyObject *)__pyx_k_tuple_45), ((PyObject *)__pyx_v_mask_c)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":811
 *     pmask[:-1,:-1] = mask_c
 * 
 *     s0, s1 = pmask.shape[0], pmask.shape[1]             # <<<<<<<<<<<<<<
 * 
 *     fpmask = pmask.reshape(-1)
 */
  __pyx_t_14 = (__pyx_v_pmask->dimensions[0]);
  __pyx_t_15 = (__pyx_v_pmask->dimensions[1]);
  __pyx_v_s0 = __pyx_t_14;
  __pyx_v_s1 = __pyx_t_15;

  /* "nipy/algorithms/statistics/intvol.pyx":813
 *     s0, s1 = pmask.shape[0], pmask.shape[1]
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 *     fmask = mask_c.reshape(-1)
 *     fcoords = coords.reshape((coords.shape[0], -1))
 */
  __pyx_t_3 = PyObject_GetAttr(((PyObject *)__pyx_v_pmask), __pyx_n_s__reshape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_k_tuple_46), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_16 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_16, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_fpmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_fpmask.diminfo[0].strides = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fpmask.diminfo[0].shape = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_16 = 0;
  __pyx_v_fpmask = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":814
 * 
 *     fpmask = pmask.reshape(-1)
 *     fmask = mask_c.reshape(-1)             # <<<<<<<<<<<<<<
 *     fcoords = coords.reshape((coords.shape[0], -1))
 * 
 */
  __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_v_mask_c), __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_47), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_fmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_fmask.diminfo[0].strides = __pyx_pybuffernd_fmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fmask.diminfo[0].shape = __pyx_pybuffernd_fmask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_17 = 0;
  __pyx_v_fmask = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":815
 *     fpmask = pmask.reshape(-1)
 *     fmask = mask_c.reshape(-1)
 *     fcoords = coords.reshape((coords.shape[0], -1))             # <<<<<<<<<<<<<<
 * 
 *     # First do the interior contributions.
 */
  __pyx_t_3 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__reshape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_coords, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_18 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer, (PyObject*)__pyx_t_18, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer, (PyObject*)__pyx_v_fcoords, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_fcoords.diminfo[0].strides = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fcoords.diminfo[0].shape = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_fcoords.diminfo[1].strides = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_fcoords.diminfo[1].shape = __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 815; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_18 = 0;
  __pyx_v_fcoords = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":826
 *         np.ndarray[np.intp_t, ndim=1] strides
 *         np.ndarray[np.intp_t, ndim=1] dstrides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)             # <<<<<<<<<<<<<<
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
 *     ss0, ss1 = strides[0], strides[1]
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__array); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__strides_from); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__bool); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__intp); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, ((PyObject *)__pyx_n_s__dtype), __pyx_t_19) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_1)); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_19) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_19, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_20 = ((PyArrayObject *)__pyx_t_19);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_t_20, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_v_strides, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_strides.diminfo[0].strides = __pyx_pybuffernd_strides.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_strides.diminfo[0].shape = __pyx_pybuffernd_strides.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_20 = 0;
  __pyx_v_strides = ((PyArrayObject *)__pyx_t_19);
  __pyx_t_19 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":827
 *         np.ndarray[np.intp_t, ndim=1] dstrides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)             # <<<<<<<<<<<<<<
 *     ss0, ss1 = strides[0], strides[1]
 *     ss0d, ss1d = dstrides[0], dstrides[1]
 */
  __pyx_t_19 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_19, __pyx_n_s__array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_GetName(__pyx_m, __pyx_n_s__strides_from); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_3 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_5, __pyx_n_s__bool); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_19, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_2));
  __pyx_t_19 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_19, __pyx_n_s__intp); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyDict_SetItem(__pyx_t_2, ((PyObject *)__pyx_n_s__dtype), __pyx_t_3) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_5), ((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_21 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer, (PyObject*)__pyx_v_dstrides, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_dstrides.diminfo[0].strides = __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dstrides.diminfo[0].shape = __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_21 = 0;
  __pyx_v_dstrides = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":828
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
 *     ss0, ss1 = strides[0], strides[1]             # <<<<<<<<<<<<<<
 *     ss0d, ss1d = dstrides[0], dstrides[1]
 *     verts = []
 */
  __pyx_t_22 = 0;
  __pyx_t_8 = -1;
  if (__pyx_t_22 < 0) {
    __pyx_t_22 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_22 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_22 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_23 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_22, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_t_24 = 1;
  __pyx_t_8 = -1;
  if (__pyx_t_24 < 0) {
    __pyx_t_24 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_24 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_24 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_25 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_v_ss0 = __pyx_t_23;
  __pyx_v_ss1 = __pyx_t_25;

  /* "nipy/algorithms/statistics/intvol.pyx":829
 *     dstrides = np.array(strides_from(mask.shape, np.bool), dtype=np.intp)
 *     ss0, ss1 = strides[0], strides[1]
 *     ss0d, ss1d = dstrides[0], dstrides[1]             # <<<<<<<<<<<<<<
 *     verts = []
 *     for i in range(2):
 */
  __pyx_t_26 = 0;
  __pyx_t_8 = -1;
  if (__pyx_t_26 < 0) {
    __pyx_t_26 += __pyx_pybuffernd_dstrides.diminfo[0].shape;
    if (unlikely(__pyx_t_26 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_26 >= __pyx_pybuffernd_dstrides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_25 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_dstrides.diminfo[0].strides));
  __pyx_t_27 = 1;
  __pyx_t_8 = -1;
  if (__pyx_t_27 < 0) {
    __pyx_t_27 += __pyx_pybuffernd_dstrides.diminfo[0].shape;
    if (unlikely(__pyx_t_27 < 0)) __pyx_t_8 = 0;
  } else if (unlikely(__pyx_t_27 >= __pyx_pybuffernd_dstrides.diminfo[0].shape)) __pyx_t_8 = 0;
  if (unlikely(__pyx_t_8 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_8);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_23 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_dstrides.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_dstrides.diminfo[0].strides));
  __pyx_v_ss0d = __pyx_t_25;
  __pyx_v_ss1d = __pyx_t_23;

  /* "nipy/algorithms/statistics/intvol.pyx":830
 *     ss0, ss1 = strides[0], strides[1]
 *     ss0d, ss1d = dstrides[0], dstrides[1]
 *     verts = []             # <<<<<<<<<<<<<<
 *     for i in range(2):
 *         for j in range(2):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_verts = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":831
 *     ss0d, ss1d = dstrides[0], dstrides[1]
 *     verts = []
 *     for i in range(2):             # <<<<<<<<<<<<<<
 *         for j in range(2):
 *             verts.append(ss0d * i + ss1d * j)
 */
  for (__pyx_t_15 = 0; __pyx_t_15 < 2; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "nipy/algorithms/statistics/intvol.pyx":832
 *     verts = []
 *     for i in range(2):
 *         for j in range(2):             # <<<<<<<<<<<<<<
 *             verts.append(ss0d * i + ss1d * j)
 *     cvertices = np.array(sorted(verts), np.intp)
 */
    for (__pyx_t_14 = 0; __pyx_t_14 < 2; __pyx_t_14+=1) {
      __pyx_v_j = __pyx_t_14;

      /* "nipy/algorithms/statistics/intvol.pyx":833
 *     for i in range(2):
 *         for j in range(2):
 *             verts.append(ss0d * i + ss1d * j)             # <<<<<<<<<<<<<<
 *     cvertices = np.array(sorted(verts), np.intp)
 * 
 */
      __pyx_t_3 = PyInt_FromLong(((__pyx_v_ss0d * __pyx_v_i) + (__pyx_v_ss1d * __pyx_v_j))); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_28 = PyList_Append(__pyx_v_verts, __pyx_t_3); if (unlikely(__pyx_t_28 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "nipy/algorithms/statistics/intvol.pyx":834
 *         for j in range(2):
 *             verts.append(ss0d * i + ss1d * j)
 *     cvertices = np.array(sorted(verts), np.intp)             # <<<<<<<<<<<<<<
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_verts));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_verts));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_verts));
  __pyx_t_5 = PyObject_Call(__pyx_builtin_sorted, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__intp); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_5 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_29 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer, (PyObject*)__pyx_t_29, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer, (PyObject*)__pyx_v_cvertices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_cvertices.diminfo[0].strides = __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_cvertices.diminfo[0].shape = __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_29 = 0;
  __pyx_v_cvertices = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":836
 *     cvertices = np.array(sorted(verts), np.intp)
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__join_complexes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_48));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_k_tuple_48));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_48));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_5 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":837
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1), strides)])
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_49));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_49));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_49));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_19 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":838
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])             # <<<<<<<<<<<<<<
 * 
 *     c = cube_with_strides_center((0,0), strides)
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_50));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_k_tuple_50));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_50));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_30 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_19);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_30);
  __Pyx_GIVEREF(__pyx_t_30);
  __pyx_t_5 = 0;
  __pyx_t_19 = 0;
  __pyx_t_30 = 0;
  __pyx_t_30 = PySequence_Tuple(((PyObject *)__pyx_t_2)); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_30));
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_30), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_30)); __pyx_t_30 = 0;
  __pyx_v_union = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":840
 *                              cube_with_strides_center((1,1), strides)])
 * 
 *     c = cube_with_strides_center((0,0), strides)             # <<<<<<<<<<<<<<
 *     m3 = np.array(list(c[3].difference(union[3])))
 *     m2 = np.array(list(c[2].difference(union[2])))
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_30 = PyTuple_New(2); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_51));
  PyTuple_SET_ITEM(__pyx_t_30, 0, ((PyObject *)__pyx_k_tuple_51));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_51));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_30, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_30), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_30)); __pyx_t_30 = 0;
  __pyx_v_c = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":841
 * 
 *     c = cube_with_strides_center((0,0), strides)
 *     m3 = np.array(list(c[3].difference(union[3])))             # <<<<<<<<<<<<<<
 *     m2 = np.array(list(c[2].difference(union[2])))
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_c, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__difference); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_union, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_19 = PyTuple_New(1); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_19), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_19)); __pyx_t_19 = 0;
  __pyx_t_19 = PyTuple_New(1); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_19), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_19)); __pyx_t_19 = 0;
  __pyx_t_19 = PyTuple_New(1); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_30, ((PyObject *)__pyx_t_19), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_19)); __pyx_t_19 = 0;
  __pyx_v_m3 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":842
 *     c = cube_with_strides_center((0,0), strides)
 *     m3 = np.array(list(c[3].difference(union[3])))
 *     m2 = np.array(list(c[2].difference(union[2])))             # <<<<<<<<<<<<<<
 * 
 *     d3 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m3[i]] for i in range(m3.shape[0])])
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_19 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_c, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__difference); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_union, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_30, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_19, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_v_m2 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":844
 *     m2 = np.array(list(c[2].difference(union[2])))
 * 
 *     d3 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m3[i]] for i in range(m3.shape[0])])             # <<<<<<<<<<<<<<
 *     d3 = np.hstack([m3, d3])
 *     ds3 = d3.shape[0]
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_19 = PyObject_GetAttr(__pyx_v_m3, __pyx_n_s__shape); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_30 = __Pyx_GetItemInt(__pyx_t_19, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_30) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_31 = __Pyx_PyInt_AsLong(__pyx_t_30); if (unlikely((__pyx_t_31 == (long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_31; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_30 = PyList_New(0); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_19 = __Pyx_GetItemInt(__pyx_v_m3, __pyx_v_i, sizeof(npy_intp), __Pyx_PyInt_to_py_Py_intptr_t); if (!__pyx_t_19) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_19);
    if (PyList_CheckExact(__pyx_t_19) || PyTuple_CheckExact(__pyx_t_19)) {
      __pyx_t_5 = __pyx_t_19; __Pyx_INCREF(__pyx_t_5); __pyx_t_32 = 0;
      __pyx_t_33 = NULL;
    } else {
      __pyx_t_32 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_19); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_33 = Py_TYPE(__pyx_t_5)->tp_iternext;
    }
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    for (;;) {
      if (!__pyx_t_33 && PyList_CheckExact(__pyx_t_5)) {
        if (__pyx_t_32 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_19 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_32); __Pyx_INCREF(__pyx_t_19); __pyx_t_32++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_19 = PySequence_ITEM(__pyx_t_5, __pyx_t_32); __pyx_t_32++; if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_33 && PyTuple_CheckExact(__pyx_t_5)) {
        if (__pyx_t_32 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_19 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_32); __Pyx_INCREF(__pyx_t_19); __pyx_t_32++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_19 = PySequence_ITEM(__pyx_t_5, __pyx_t_32); __pyx_t_32++; if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_19 = __pyx_t_33(__pyx_t_5);
        if (unlikely(!__pyx_t_19)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_v_v);
      __pyx_v_v = __pyx_t_19;
      __pyx_t_19 = 0;
      __pyx_t_19 = __Pyx_GetName(__pyx_m, __pyx_n_s___convert_stride2); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_19);
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      __Pyx_INCREF(((PyObject *)__pyx_v_strides));
      PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
      __Pyx_INCREF(((PyObject *)__pyx_k_tuple_52));
      PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_k_tuple_52));
      __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_52));
      __pyx_t_34 = PyObject_Call(__pyx_t_19, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_34);
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      if (unlikely(__Pyx_PyList_Append(__pyx_t_30, (PyObject*)__pyx_t_34))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(__Pyx_PyList_Append(__pyx_t_1, (PyObject*)__pyx_t_30))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_t_30)); __pyx_t_30 = 0;
  }
  __pyx_t_30 = PyTuple_New(1); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  PyTuple_SET_ITEM(__pyx_t_30, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_30), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_30)); __pyx_t_30 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_35 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_t_35, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_v_d3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_d3.diminfo[0].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d3.diminfo[0].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d3.diminfo[1].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d3.diminfo[1].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_35 = 0;
  __pyx_v_d3 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":845
 * 
 *     d3 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m3[i]] for i in range(m3.shape[0])])
 *     d3 = np.hstack([m3, d3])             # <<<<<<<<<<<<<<
 *     ds3 = d3.shape[0]
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__hstack); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_m3);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_m3);
  __Pyx_GIVEREF(__pyx_v_m3);
  __Pyx_INCREF(((PyObject *)__pyx_v_d3));
  PyList_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_d3));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_d3));
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_30, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_35 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_t_35, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_v_d3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_d3.diminfo[0].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d3.diminfo[0].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d3.diminfo[1].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d3.diminfo[1].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_35 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_v_d3));
  __pyx_v_d3 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":846
 *     d3 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m3[i]] for i in range(m3.shape[0])])
 *     d3 = np.hstack([m3, d3])
 *     ds3 = d3.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     d2 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m2[i]] for i in range(m2.shape[0])])
 */
  __pyx_v_ds3 = (__pyx_v_d3->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":848
 *     ds3 = d3.shape[0]
 * 
 *     d2 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m2[i]] for i in range(m2.shape[0])])             # <<<<<<<<<<<<<<
 *     d2 = np.hstack([m2, d2])
 *     ds2 = d2.shape[0]
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_30 = PyObject_GetAttr(__pyx_v_m2, __pyx_n_s__shape); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_30, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_5) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __pyx_t_31 = __Pyx_PyInt_AsLong(__pyx_t_5); if (unlikely((__pyx_t_31 == (long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_31; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_30 = __Pyx_GetItemInt(__pyx_v_m2, __pyx_v_i, sizeof(npy_intp), __Pyx_PyInt_to_py_Py_intptr_t); if (!__pyx_t_30) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_30);
    if (PyList_CheckExact(__pyx_t_30) || PyTuple_CheckExact(__pyx_t_30)) {
      __pyx_t_34 = __pyx_t_30; __Pyx_INCREF(__pyx_t_34); __pyx_t_32 = 0;
      __pyx_t_33 = NULL;
    } else {
      __pyx_t_32 = -1; __pyx_t_34 = PyObject_GetIter(__pyx_t_30); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_34);
      __pyx_t_33 = Py_TYPE(__pyx_t_34)->tp_iternext;
    }
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    for (;;) {
      if (!__pyx_t_33 && PyList_CheckExact(__pyx_t_34)) {
        if (__pyx_t_32 >= PyList_GET_SIZE(__pyx_t_34)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_30 = PyList_GET_ITEM(__pyx_t_34, __pyx_t_32); __Pyx_INCREF(__pyx_t_30); __pyx_t_32++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_30 = PySequence_ITEM(__pyx_t_34, __pyx_t_32); __pyx_t_32++; if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else if (!__pyx_t_33 && PyTuple_CheckExact(__pyx_t_34)) {
        if (__pyx_t_32 >= PyTuple_GET_SIZE(__pyx_t_34)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_30 = PyTuple_GET_ITEM(__pyx_t_34, __pyx_t_32); __Pyx_INCREF(__pyx_t_30); __pyx_t_32++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_30 = PySequence_ITEM(__pyx_t_34, __pyx_t_32); __pyx_t_32++; if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #endif
      } else {
        __pyx_t_30 = __pyx_t_33(__pyx_t_34);
        if (unlikely(!__pyx_t_30)) {
          if (PyErr_Occurred()) {
            if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_30);
      }
      __Pyx_XDECREF(__pyx_v_v);
      __pyx_v_v = __pyx_t_30;
      __pyx_t_30 = 0;
      __pyx_t_30 = __Pyx_GetName(__pyx_m, __pyx_n_s___convert_stride2); if (unlikely(!__pyx_t_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_30);
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_v);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_v);
      __Pyx_GIVEREF(__pyx_v_v);
      __Pyx_INCREF(((PyObject *)__pyx_v_strides));
      PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
      __Pyx_INCREF(((PyObject *)__pyx_k_tuple_53));
      PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_k_tuple_53));
      __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_53));
      __pyx_t_19 = PyObject_Call(__pyx_t_30, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
      __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
      if (unlikely(__Pyx_PyList_Append(__pyx_t_5, (PyObject*)__pyx_t_19))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    }
    __Pyx_DECREF(__pyx_t_34); __pyx_t_34 = 0;
    if (unlikely(__Pyx_PyList_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_5), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_5)); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_36 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_t_36, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_v_d2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_d2.diminfo[0].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d2.diminfo[0].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d2.diminfo[1].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d2.diminfo[1].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_36 = 0;
  __pyx_v_d2 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":849
 * 
 *     d2 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m2[i]] for i in range(m2.shape[0])])
 *     d2 = np.hstack([m2, d2])             # <<<<<<<<<<<<<<
 *     ds2 = d2.shape[0]
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__hstack); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_m2);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_m2);
  __Pyx_GIVEREF(__pyx_v_m2);
  __Pyx_INCREF(((PyObject *)__pyx_v_d2));
  PyList_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_d2));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_d2));
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_1));
  __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_36 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_t_36, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_v_d2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
    }
    __pyx_pybuffernd_d2.diminfo[0].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d2.diminfo[0].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d2.diminfo[1].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d2.diminfo[1].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_36 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_v_d2));
  __pyx_v_d2 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":850
 *     d2 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m2[i]] for i in range(m2.shape[0])])
 *     d2 = np.hstack([m2, d2])
 *     ds2 = d2.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     D = np.zeros((4,4))
 */
  __pyx_v_ds2 = (__pyx_v_d2->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":852
 *     ds2 = d2.shape[0]
 * 
 *     D = np.zeros((4,4))             # <<<<<<<<<<<<<<
 * 
 *     npix = mask.size
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_55), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_37 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_D.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_D.rcbuffer->pybuffer, (PyObject*)__pyx_t_37, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_D.rcbuffer->pybuffer, (PyObject*)__pyx_v_D, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
    }
    __pyx_pybuffernd_D.diminfo[0].strides = __pyx_pybuffernd_D.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_D.diminfo[0].shape = __pyx_pybuffernd_D.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_D.diminfo[1].strides = __pyx_pybuffernd_D.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_D.diminfo[1].shape = __pyx_pybuffernd_D.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_8 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_37 = 0;
  __pyx_v_D = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":854
 *     D = np.zeros((4,4))
 * 
 *     npix = mask.size             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0-1):
 */
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__size); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 854; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_15 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 854; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_npix = __pyx_t_15;

  /* "nipy/algorithms/statistics/intvol.pyx":856
 *     npix = mask.size
 * 
 *     for i in range(s0-1):             # <<<<<<<<<<<<<<
 *         for j in range(s1-1):
 *             pindex = i*ss0+j*ss1
 */
  __pyx_t_31 = (__pyx_v_s0 - 1);
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_31; __pyx_t_15+=1) {
    __pyx_v_i = __pyx_t_15;

    /* "nipy/algorithms/statistics/intvol.pyx":857
 * 
 *     for i in range(s0-1):
 *         for j in range(s1-1):             # <<<<<<<<<<<<<<
 *             pindex = i*ss0+j*ss1
 *             index = i*ss0d+j*ss1d
 */
    __pyx_t_38 = (__pyx_v_s1 - 1);
    for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_38; __pyx_t_14+=1) {
      __pyx_v_j = __pyx_t_14;

      /* "nipy/algorithms/statistics/intvol.pyx":858
 *     for i in range(s0-1):
 *         for j in range(s1-1):
 *             pindex = i*ss0+j*ss1             # <<<<<<<<<<<<<<
 *             index = i*ss0d+j*ss1d
 *             for r in range(4):
 */
      __pyx_t_1 = PyInt_FromLong(((__pyx_v_i * __pyx_v_ss0) + (__pyx_v_j * __pyx_v_ss1))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 858; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF(__pyx_v_pindex);
      __pyx_v_pindex = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "nipy/algorithms/statistics/intvol.pyx":859
 *         for j in range(s1-1):
 *             pindex = i*ss0+j*ss1
 *             index = i*ss0d+j*ss1d             # <<<<<<<<<<<<<<
 *             for r in range(4):
 *                 rr = (index+cvertices[r]) % npix
 */
      __pyx_v_index = ((__pyx_v_i * __pyx_v_ss0d) + (__pyx_v_j * __pyx_v_ss1d));

      /* "nipy/algorithms/statistics/intvol.pyx":860
 *             pindex = i*ss0+j*ss1
 *             index = i*ss0d+j*ss1d
 *             for r in range(4):             # <<<<<<<<<<<<<<
 *                 rr = (index+cvertices[r]) % npix
 *                 mr = fmask[rr]
 */
      for (__pyx_t_39 = 0; __pyx_t_39 < 4; __pyx_t_39+=1) {
        __pyx_v_r = __pyx_t_39;

        /* "nipy/algorithms/statistics/intvol.pyx":861
 *             index = i*ss0d+j*ss1d
 *             for r in range(4):
 *                 rr = (index+cvertices[r]) % npix             # <<<<<<<<<<<<<<
 *                 mr = fmask[rr]
 *                 for s in range(r+1):
 */
        __pyx_t_40 = __pyx_v_r;
        __pyx_t_8 = -1;
        if (__pyx_t_40 < 0) {
          __pyx_t_40 += __pyx_pybuffernd_cvertices.diminfo[0].shape;
          if (unlikely(__pyx_t_40 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_40 >= __pyx_pybuffernd_cvertices.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_8 = (__pyx_v_index + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.buf, __pyx_t_40, __pyx_pybuffernd_cvertices.diminfo[0].strides)));
        if (unlikely(__pyx_v_npix == 0)) {
          PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_rr = __Pyx_mod_int(__pyx_t_8, __pyx_v_npix);

        /* "nipy/algorithms/statistics/intvol.pyx":862
 *             for r in range(4):
 *                 rr = (index+cvertices[r]) % npix
 *                 mr = fmask[rr]             # <<<<<<<<<<<<<<
 *                 for s in range(r+1):
 *                     res = 0
 */
        __pyx_t_41 = __pyx_v_rr;
        __pyx_t_8 = -1;
        if (__pyx_t_41 < 0) {
          __pyx_t_41 += __pyx_pybuffernd_fmask.diminfo[0].shape;
          if (unlikely(__pyx_t_41 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_41 >= __pyx_pybuffernd_fmask.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 862; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_mr = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fmask.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_fmask.diminfo[0].strides));

        /* "nipy/algorithms/statistics/intvol.pyx":863
 *                 rr = (index+cvertices[r]) % npix
 *                 mr = fmask[rr]
 *                 for s in range(r+1):             # <<<<<<<<<<<<<<
 *                     res = 0
 *                     ss = (index+cvertices[s]) % npix
 */
        __pyx_t_42 = (__pyx_v_r + 1);
        for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_42; __pyx_t_43+=1) {
          __pyx_v_s = __pyx_t_43;

          /* "nipy/algorithms/statistics/intvol.pyx":864
 *                 mr = fmask[rr]
 *                 for s in range(r+1):
 *                     res = 0             # <<<<<<<<<<<<<<
 *                     ss = (index+cvertices[s]) % npix
 *                     ms = fmask[ss]
 */
          __pyx_v_res = 0.0;

          /* "nipy/algorithms/statistics/intvol.pyx":865
 *                 for s in range(r+1):
 *                     res = 0
 *                     ss = (index+cvertices[s]) % npix             # <<<<<<<<<<<<<<
 *                     ms = fmask[ss]
 *                     if mr * ms:
 */
          __pyx_t_44 = __pyx_v_s;
          __pyx_t_8 = -1;
          if (__pyx_t_44 < 0) {
            __pyx_t_44 += __pyx_pybuffernd_cvertices.diminfo[0].shape;
            if (unlikely(__pyx_t_44 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_44 >= __pyx_pybuffernd_cvertices.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_8 = (__pyx_v_index + (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_cvertices.rcbuffer->pybuffer.buf, __pyx_t_44, __pyx_pybuffernd_cvertices.diminfo[0].strides)));
          if (unlikely(__pyx_v_npix == 0)) {
            PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 865; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_ss = __Pyx_mod_int(__pyx_t_8, __pyx_v_npix);

          /* "nipy/algorithms/statistics/intvol.pyx":866
 *                     res = 0
 *                     ss = (index+cvertices[s]) % npix
 *                     ms = fmask[ss]             # <<<<<<<<<<<<<<
 *                     if mr * ms:
 *                         for l in range(fcoords.shape[0]):
 */
          __pyx_t_45 = __pyx_v_ss;
          __pyx_t_8 = -1;
          if (__pyx_t_45 < 0) {
            __pyx_t_45 += __pyx_pybuffernd_fmask.diminfo[0].shape;
            if (unlikely(__pyx_t_45 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_45 >= __pyx_pybuffernd_fmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 866; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_ms = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fmask.rcbuffer->pybuffer.buf, __pyx_t_45, __pyx_pybuffernd_fmask.diminfo[0].strides));

          /* "nipy/algorithms/statistics/intvol.pyx":867
 *                     ss = (index+cvertices[s]) % npix
 *                     ms = fmask[ss]
 *                     if mr * ms:             # <<<<<<<<<<<<<<
 *                         for l in range(fcoords.shape[0]):
 *                             res += fcoords[l,ss] * fcoords[l,rr]
 */
          __pyx_t_8 = (__pyx_v_mr * __pyx_v_ms);
          if (__pyx_t_8) {

            /* "nipy/algorithms/statistics/intvol.pyx":868
 *                     ms = fmask[ss]
 *                     if mr * ms:
 *                         for l in range(fcoords.shape[0]):             # <<<<<<<<<<<<<<
 *                             res += fcoords[l,ss] * fcoords[l,rr]
 *                             D[r,s] = res
 */
            __pyx_t_46 = (__pyx_v_fcoords->dimensions[0]);
            for (__pyx_t_47 = 0; __pyx_t_47 < __pyx_t_46; __pyx_t_47+=1) {
              __pyx_v_l = __pyx_t_47;

              /* "nipy/algorithms/statistics/intvol.pyx":869
 *                     if mr * ms:
 *                         for l in range(fcoords.shape[0]):
 *                             res += fcoords[l,ss] * fcoords[l,rr]             # <<<<<<<<<<<<<<
 *                             D[r,s] = res
 *                             D[s,r] = res
 */
              __pyx_t_48 = __pyx_v_l;
              __pyx_t_49 = __pyx_v_ss;
              __pyx_t_8 = -1;
              if (__pyx_t_48 < 0) {
                __pyx_t_48 += __pyx_pybuffernd_fcoords.diminfo[0].shape;
                if (unlikely(__pyx_t_48 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_48 >= __pyx_pybuffernd_fcoords.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_49 < 0) {
                __pyx_t_49 += __pyx_pybuffernd_fcoords.diminfo[1].shape;
                if (unlikely(__pyx_t_49 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_49 >= __pyx_pybuffernd_fcoords.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 869; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              __pyx_t_50 = __pyx_v_l;
              __pyx_t_51 = __pyx_v_rr;
              __pyx_t_8 = -1;
              if (__pyx_t_50 < 0) {
                __pyx_t_50 += __pyx_pybuffernd_fcoords.diminfo[0].shape;
                if (unlikely(__pyx_t_50 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_50 >= __pyx_pybuffernd_fcoords.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_51 < 0) {
                __pyx_t_51 += __pyx_pybuffernd_fcoords.diminfo[1].shape;
                if (unlikely(__pyx_t_51 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_51 >= __pyx_pybuffernd_fcoords.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 869; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              __pyx_v_res = (__pyx_v_res + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.buf, __pyx_t_48, __pyx_pybuffernd_fcoords.diminfo[0].strides, __pyx_t_49, __pyx_pybuffernd_fcoords.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_fcoords.rcbuffer->pybuffer.buf, __pyx_t_50, __pyx_pybuffernd_fcoords.diminfo[0].strides, __pyx_t_51, __pyx_pybuffernd_fcoords.diminfo[1].strides))));

              /* "nipy/algorithms/statistics/intvol.pyx":870
 *                         for l in range(fcoords.shape[0]):
 *                             res += fcoords[l,ss] * fcoords[l,rr]
 *                             D[r,s] = res             # <<<<<<<<<<<<<<
 *                             D[s,r] = res
 *                     else:
 */
              __pyx_t_52 = __pyx_v_r;
              __pyx_t_53 = __pyx_v_s;
              __pyx_t_8 = -1;
              if (__pyx_t_52 < 0) {
                __pyx_t_52 += __pyx_pybuffernd_D.diminfo[0].shape;
                if (unlikely(__pyx_t_52 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_52 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_53 < 0) {
                __pyx_t_53 += __pyx_pybuffernd_D.diminfo[1].shape;
                if (unlikely(__pyx_t_53 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_53 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 870; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_53, __pyx_pybuffernd_D.diminfo[1].strides) = __pyx_v_res;

              /* "nipy/algorithms/statistics/intvol.pyx":871
 *                             res += fcoords[l,ss] * fcoords[l,rr]
 *                             D[r,s] = res
 *                             D[s,r] = res             # <<<<<<<<<<<<<<
 *                     else:
 *                         D[r,s] = 0
 */
              __pyx_t_54 = __pyx_v_s;
              __pyx_t_55 = __pyx_v_r;
              __pyx_t_8 = -1;
              if (__pyx_t_54 < 0) {
                __pyx_t_54 += __pyx_pybuffernd_D.diminfo[0].shape;
                if (unlikely(__pyx_t_54 < 0)) __pyx_t_8 = 0;
              } else if (unlikely(__pyx_t_54 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
              if (__pyx_t_55 < 0) {
                __pyx_t_55 += __pyx_pybuffernd_D.diminfo[1].shape;
                if (unlikely(__pyx_t_55 < 0)) __pyx_t_8 = 1;
              } else if (unlikely(__pyx_t_55 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
              if (unlikely(__pyx_t_8 != -1)) {
                __Pyx_RaiseBufferIndexError(__pyx_t_8);
                {__pyx_filename = __pyx_f[0]; __pyx_lineno = 871; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_54, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_55, __pyx_pybuffernd_D.diminfo[1].strides) = __pyx_v_res;
            }
            goto __pyx_L26;
          }
          /*else*/ {

            /* "nipy/algorithms/statistics/intvol.pyx":873
 *                             D[s,r] = res
 *                     else:
 *                         D[r,s] = 0             # <<<<<<<<<<<<<<
 *                         D[s,r] = 0
 * 
 */
            __pyx_t_46 = __pyx_v_r;
            __pyx_t_47 = __pyx_v_s;
            __pyx_t_8 = -1;
            if (__pyx_t_46 < 0) {
              __pyx_t_46 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_46 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_46 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_47 < 0) {
              __pyx_t_47 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_47 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_47 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 873; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_46, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_47, __pyx_pybuffernd_D.diminfo[1].strides) = 0.0;

            /* "nipy/algorithms/statistics/intvol.pyx":874
 *                     else:
 *                         D[r,s] = 0
 *                         D[s,r] = 0             # <<<<<<<<<<<<<<
 * 
 *             for l in range(ds3):
 */
            __pyx_t_56 = __pyx_v_s;
            __pyx_t_57 = __pyx_v_r;
            __pyx_t_8 = -1;
            if (__pyx_t_56 < 0) {
              __pyx_t_56 += __pyx_pybuffernd_D.diminfo[0].shape;
              if (unlikely(__pyx_t_56 < 0)) __pyx_t_8 = 0;
            } else if (unlikely(__pyx_t_56 >= __pyx_pybuffernd_D.diminfo[0].shape)) __pyx_t_8 = 0;
            if (__pyx_t_57 < 0) {
              __pyx_t_57 += __pyx_pybuffernd_D.diminfo[1].shape;
              if (unlikely(__pyx_t_57 < 0)) __pyx_t_8 = 1;
            } else if (unlikely(__pyx_t_57 >= __pyx_pybuffernd_D.diminfo[1].shape)) __pyx_t_8 = 1;
            if (unlikely(__pyx_t_8 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_8);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 874; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            *__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_D.rcbuffer->pybuffer.buf, __pyx_t_56, __pyx_pybuffernd_D.diminfo[0].strides, __pyx_t_57, __pyx_pybuffernd_D.diminfo[1].strides) = 0.0;
          }
          __pyx_L26:;
        }
      }

      /* "nipy/algorithms/statistics/intvol.pyx":876
 *                         D[s,r] = 0
 * 
 *             for l in range(ds3):             # <<<<<<<<<<<<<<
 *                 v0 = pindex + d3[l,0]
 *                 w0 = d3[l,3]
 */
      __pyx_t_39 = __pyx_v_ds3;
      for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_39; __pyx_t_43+=1) {
        __pyx_v_l = __pyx_t_43;

        /* "nipy/algorithms/statistics/intvol.pyx":877
 * 
 *             for l in range(ds3):
 *                 v0 = pindex + d3[l,0]             # <<<<<<<<<<<<<<
 *                 w0 = d3[l,3]
 *                 m = fpmask[v0]
 */
        __pyx_t_58 = __pyx_v_l;
        __pyx_t_42 = 0;
        __pyx_t_8 = -1;
        if (__pyx_t_58 < 0) {
          __pyx_t_58 += __pyx_pybuffernd_d3.diminfo[0].shape;
          if (unlikely(__pyx_t_58 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_58 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
        if (__pyx_t_42 < 0) {
          __pyx_t_42 += __pyx_pybuffernd_d3.diminfo[1].shape;
          if (unlikely(__pyx_t_42 < 0)) __pyx_t_8 = 1;
        } else if (unlikely(__pyx_t_42 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_58, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_d3.diminfo[1].strides))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PyNumber_Add(__pyx_v_pindex, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_59 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_59 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_v0 = __pyx_t_59;

        /* "nipy/algorithms/statistics/intvol.pyx":878
 *             for l in range(ds3):
 *                 v0 = pindex + d3[l,0]
 *                 w0 = d3[l,3]             # <<<<<<<<<<<<<<
 *                 m = fpmask[v0]
 *                 if m:
 */
        __pyx_t_59 = __pyx_v_l;
        __pyx_t_60 = 3;
        __pyx_t_8 = -1;
        if (__pyx_t_59 < 0) {
          __pyx_t_59 += __pyx_pybuffernd_d3.diminfo[0].shape;
          if (unlikely(__pyx_t_59 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_59 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
        if (__pyx_t_60 < 0) {
          __pyx_t_60 += __pyx_pybuffernd_d3.diminfo[1].shape;
          if (unlikely(__pyx_t_60 < 0)) __pyx_t_8 = 1;
        } else if (unlikely(__pyx_t_60 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_59, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_60, __pyx_pybuffernd_d3.diminfo[1].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 878; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_XDECREF(__pyx_v_w0);
        __pyx_v_w0 = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "nipy/algorithms/statistics/intvol.pyx":879
 *                 v0 = pindex + d3[l,0]
 *                 w0 = d3[l,3]
 *                 m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                 if m:
 *                     v1 = pindex + d3[l,1]
 */
        __pyx_t_61 = __pyx_v_v0;
        __pyx_t_8 = -1;
        if (__pyx_t_61 < 0) {
          __pyx_t_61 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
          if (unlikely(__pyx_t_61 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_61 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_61, __pyx_pybuffernd_fpmask.diminfo[0].strides));

        /* "nipy/algorithms/statistics/intvol.pyx":880
 *                 w0 = d3[l,3]
 *                 m = fpmask[v0]
 *                 if m:             # <<<<<<<<<<<<<<
 *                     v1 = pindex + d3[l,1]
 *                     v2 = pindex + d3[l,2]
 */
        if (__pyx_v_m) {

          /* "nipy/algorithms/statistics/intvol.pyx":881
 *                 m = fpmask[v0]
 *                 if m:
 *                     v1 = pindex + d3[l,1]             # <<<<<<<<<<<<<<
 *                     v2 = pindex + d3[l,2]
 *                     w1 = d3[l,4]
 */
          __pyx_t_62 = __pyx_v_l;
          __pyx_t_63 = 1;
          __pyx_t_8 = -1;
          if (__pyx_t_62 < 0) {
            __pyx_t_62 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_62 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_62 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_63 < 0) {
            __pyx_t_63 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_63 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_63 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_62, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_63, __pyx_pybuffernd_d3.diminfo[1].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = PyNumber_Add(__pyx_v_pindex, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_64 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_64 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 881; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_v1 = __pyx_t_64;

          /* "nipy/algorithms/statistics/intvol.pyx":882
 *                 if m:
 *                     v1 = pindex + d3[l,1]
 *                     v2 = pindex + d3[l,2]             # <<<<<<<<<<<<<<
 *                     w1 = d3[l,4]
 *                     w2 = d3[l,5]
 */
          __pyx_t_64 = __pyx_v_l;
          __pyx_t_65 = 2;
          __pyx_t_8 = -1;
          if (__pyx_t_64 < 0) {
            __pyx_t_64 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_64 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_64 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_65 < 0) {
            __pyx_t_65 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_65 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_65 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_64, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_65, __pyx_pybuffernd_d3.diminfo[1].strides))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyNumber_Add(__pyx_v_pindex, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_66 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_66 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 882; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_v2 = __pyx_t_66;

          /* "nipy/algorithms/statistics/intvol.pyx":883
 *                     v1 = pindex + d3[l,1]
 *                     v2 = pindex + d3[l,2]
 *                     w1 = d3[l,4]             # <<<<<<<<<<<<<<
 *                     w2 = d3[l,5]
 *                     m = m * fpmask[v1] * fpmask[v2]
 */
          __pyx_t_66 = __pyx_v_l;
          __pyx_t_67 = 4;
          __pyx_t_8 = -1;
          if (__pyx_t_66 < 0) {
            __pyx_t_66 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_66 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_66 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_67 < 0) {
            __pyx_t_67 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_67 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_67 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 883; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_67, __pyx_pybuffernd_d3.diminfo[1].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 883; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_XDECREF(__pyx_v_w1);
          __pyx_v_w1 = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":884
 *                     v2 = pindex + d3[l,2]
 *                     w1 = d3[l,4]
 *                     w2 = d3[l,5]             # <<<<<<<<<<<<<<
 *                     m = m * fpmask[v1] * fpmask[v2]
 *                     l2 = l2 + mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 */
          __pyx_t_68 = __pyx_v_l;
          __pyx_t_69 = 5;
          __pyx_t_8 = -1;
          if (__pyx_t_68 < 0) {
            __pyx_t_68 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_68 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_68 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_69 < 0) {
            __pyx_t_69 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_69 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_69 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 884; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_68, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_69, __pyx_pybuffernd_d3.diminfo[1].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 884; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_XDECREF(__pyx_v_w2);
          __pyx_v_w2 = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":885
 *                     w1 = d3[l,4]
 *                     w2 = d3[l,5]
 *                     m = m * fpmask[v1] * fpmask[v2]             # <<<<<<<<<<<<<<
 *                     l2 = l2 + mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 */
          __pyx_t_70 = __pyx_v_v1;
          __pyx_t_8 = -1;
          if (__pyx_t_70 < 0) {
            __pyx_t_70 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_70 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_70 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 885; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_71 = __pyx_v_v2;
          __pyx_t_8 = -1;
          if (__pyx_t_71 < 0) {
            __pyx_t_71 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_71 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_71 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 885; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = ((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_70, __pyx_pybuffernd_fpmask.diminfo[0].strides))) * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":886
 *                     w2 = d3[l,5]
 *                     m = m * fpmask[v1] * fpmask[v2]
 *                     l2 = l2 + mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 *                     l1 = l1 - mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 */
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_72 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_72 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_73 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_73 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_74 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_74 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 886; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":887
 *                     m = m * fpmask[v1] * fpmask[v2]
 *                     l2 = l2 + mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m             # <<<<<<<<<<<<<<
 *                     l1 = l1 - mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 */
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_75 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_75 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_76 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_76 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_77 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_77 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 887; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_l2 = (__pyx_v_l2 + (__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu2_tri(__pyx_t_72, __pyx_t_73, __pyx_t_74, __pyx_t_75, __pyx_t_76, __pyx_t_77, 0) * __pyx_v_m));

          /* "nipy/algorithms/statistics/intvol.pyx":888
 *                     l2 = l2 + mu2_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 *                     l1 = l1 - mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],             # <<<<<<<<<<<<<<
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 *                     l0 = l0 + m
 */
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_77 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_77 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_76 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_76 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_75 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_75 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 888; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":889
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 *                     l1 = l1 - mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m             # <<<<<<<<<<<<<<
 *                     l0 = l0 + m
 * 
 */
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_74 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_74 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_73 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_73 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __Pyx_INCREF(__pyx_v_w2);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w2);
          __Pyx_GIVEREF(__pyx_v_w2);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_72 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_72 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 889; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_l1 = (__pyx_v_l1 - (__pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_tri(__pyx_t_77, __pyx_t_76, __pyx_t_75, __pyx_t_74, __pyx_t_73, __pyx_t_72, 0) * __pyx_v_m));

          /* "nipy/algorithms/statistics/intvol.pyx":890
 *                     l1 = l1 - mu1_tri(D[w0,w0], D[w0,w1], D[w0,w2],
 *                                         D[w1,w1], D[w1,w2], D[w2,w2]) * m
 *                     l0 = l0 + m             # <<<<<<<<<<<<<<
 * 
 *             for l in range(ds2):
 */
          __pyx_v_l0 = (__pyx_v_l0 + __pyx_v_m);
          goto __pyx_L31;
        }
        __pyx_L31:;
      }

      /* "nipy/algorithms/statistics/intvol.pyx":892
 *                     l0 = l0 + m
 * 
 *             for l in range(ds2):             # <<<<<<<<<<<<<<
 *                 v0 = pindex + d2[l,0]
 *                 w0 = d2[l,2]
 */
      __pyx_t_39 = __pyx_v_ds2;
      for (__pyx_t_43 = 0; __pyx_t_43 < __pyx_t_39; __pyx_t_43+=1) {
        __pyx_v_l = __pyx_t_43;

        /* "nipy/algorithms/statistics/intvol.pyx":893
 * 
 *             for l in range(ds2):
 *                 v0 = pindex + d2[l,0]             # <<<<<<<<<<<<<<
 *                 w0 = d2[l,2]
 *                 m = fpmask[v0]
 */
        __pyx_t_78 = __pyx_v_l;
        __pyx_t_79 = 0;
        __pyx_t_8 = -1;
        if (__pyx_t_78 < 0) {
          __pyx_t_78 += __pyx_pybuffernd_d2.diminfo[0].shape;
          if (unlikely(__pyx_t_78 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_78 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
        if (__pyx_t_79 < 0) {
          __pyx_t_79 += __pyx_pybuffernd_d2.diminfo[1].shape;
          if (unlikely(__pyx_t_79 < 0)) __pyx_t_8 = 1;
        } else if (unlikely(__pyx_t_79 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 893; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_78, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_79, __pyx_pybuffernd_d2.diminfo[1].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 893; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = PyNumber_Add(__pyx_v_pindex, __pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 893; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_80 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_80 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 893; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_v0 = __pyx_t_80;

        /* "nipy/algorithms/statistics/intvol.pyx":894
 *             for l in range(ds2):
 *                 v0 = pindex + d2[l,0]
 *                 w0 = d2[l,2]             # <<<<<<<<<<<<<<
 *                 m = fpmask[v0]
 *                 if m:
 */
        __pyx_t_80 = __pyx_v_l;
        __pyx_t_81 = 2;
        __pyx_t_8 = -1;
        if (__pyx_t_80 < 0) {
          __pyx_t_80 += __pyx_pybuffernd_d2.diminfo[0].shape;
          if (unlikely(__pyx_t_80 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_80 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
        if (__pyx_t_81 < 0) {
          __pyx_t_81 += __pyx_pybuffernd_d2.diminfo[1].shape;
          if (unlikely(__pyx_t_81 < 0)) __pyx_t_8 = 1;
        } else if (unlikely(__pyx_t_81 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 894; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_80, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_81, __pyx_pybuffernd_d2.diminfo[1].strides))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 894; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF(__pyx_v_w0);
        __pyx_v_w0 = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "nipy/algorithms/statistics/intvol.pyx":895
 *                 v0 = pindex + d2[l,0]
 *                 w0 = d2[l,2]
 *                 m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                 if m:
 *                     v1 = pindex + d2[l,1]
 */
        __pyx_t_82 = __pyx_v_v0;
        __pyx_t_8 = -1;
        if (__pyx_t_82 < 0) {
          __pyx_t_82 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
          if (unlikely(__pyx_t_82 < 0)) __pyx_t_8 = 0;
        } else if (unlikely(__pyx_t_82 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
        if (unlikely(__pyx_t_8 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_8);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 895; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_82, __pyx_pybuffernd_fpmask.diminfo[0].strides));

        /* "nipy/algorithms/statistics/intvol.pyx":896
 *                 w0 = d2[l,2]
 *                 m = fpmask[v0]
 *                 if m:             # <<<<<<<<<<<<<<
 *                     v1 = pindex + d2[l,1]
 *                     w1 = d2[l,3]
 */
        if (__pyx_v_m) {

          /* "nipy/algorithms/statistics/intvol.pyx":897
 *                 m = fpmask[v0]
 *                 if m:
 *                     v1 = pindex + d2[l,1]             # <<<<<<<<<<<<<<
 *                     w1 = d2[l,3]
 *                     m = m * fpmask[v1]
 */
          __pyx_t_83 = __pyx_v_l;
          __pyx_t_84 = 1;
          __pyx_t_8 = -1;
          if (__pyx_t_83 < 0) {
            __pyx_t_83 += __pyx_pybuffernd_d2.diminfo[0].shape;
            if (unlikely(__pyx_t_83 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_83 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_84 < 0) {
            __pyx_t_84 += __pyx_pybuffernd_d2.diminfo[1].shape;
            if (unlikely(__pyx_t_84 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_84 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 897; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_1 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_83, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_84, __pyx_pybuffernd_d2.diminfo[1].strides))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 897; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = PyNumber_Add(__pyx_v_pindex, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 897; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_85 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_85 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 897; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_v1 = __pyx_t_85;

          /* "nipy/algorithms/statistics/intvol.pyx":898
 *                 if m:
 *                     v1 = pindex + d2[l,1]
 *                     w1 = d2[l,3]             # <<<<<<<<<<<<<<
 *                     m = m * fpmask[v1]
 *                     l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
 */
          __pyx_t_85 = __pyx_v_l;
          __pyx_t_86 = 3;
          __pyx_t_8 = -1;
          if (__pyx_t_85 < 0) {
            __pyx_t_85 += __pyx_pybuffernd_d2.diminfo[0].shape;
            if (unlikely(__pyx_t_85 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_85 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_8 = 0;
          if (__pyx_t_86 < 0) {
            __pyx_t_86 += __pyx_pybuffernd_d2.diminfo[1].shape;
            if (unlikely(__pyx_t_86 < 0)) __pyx_t_8 = 1;
          } else if (unlikely(__pyx_t_86 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_8 = 1;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_85, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_86, __pyx_pybuffernd_d2.diminfo[1].strides))); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 898; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_XDECREF(__pyx_v_w1);
          __pyx_v_w1 = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":899
 *                     v1 = pindex + d2[l,1]
 *                     w1 = d2[l,3]
 *                     m = m * fpmask[v1]             # <<<<<<<<<<<<<<
 *                     l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
 *                     l0 = l0 - m
 */
          __pyx_t_87 = __pyx_v_v1;
          __pyx_t_8 = -1;
          if (__pyx_t_87 < 0) {
            __pyx_t_87 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_87 < 0)) __pyx_t_8 = 0;
          } else if (unlikely(__pyx_t_87 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_8 = 0;
          if (unlikely(__pyx_t_8 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_8);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 899; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_87, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":900
 *                     w1 = d2[l,3]
 *                     m = m * fpmask[v1]
 *                     l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])             # <<<<<<<<<<<<<<
 *                     l0 = l0 - m
 * 
 */
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_72 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_72 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_w0);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_w0);
          __Pyx_GIVEREF(__pyx_v_w0);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __pyx_t_2 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_1)); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
          __pyx_t_73 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_73 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __Pyx_INCREF(__pyx_v_w1);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_w1);
          __Pyx_GIVEREF(__pyx_v_w1);
          __pyx_t_1 = PyObject_GetItem(((PyObject *)__pyx_v_D), ((PyObject *)__pyx_t_2)); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
          __pyx_t_74 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_74 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 900; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_l1 = (__pyx_v_l1 + (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(__pyx_t_72, __pyx_t_73, __pyx_t_74, 0)));

          /* "nipy/algorithms/statistics/intvol.pyx":901
 *                     m = m * fpmask[v1]
 *                     l1 = l1 + m * mu1_edge(D[w0,w0], D[w0,w1], D[w1,w1])
 *                     l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *     l0 += mask.sum()
 */
          __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
          goto __pyx_L34;
        }
        __pyx_L34:;
      }
    }
  }

  /* "nipy/algorithms/statistics/intvol.pyx":903
 *                     l0 = l0 - m
 * 
 *     l0 += mask.sum()             # <<<<<<<<<<<<<<
 *     return np.array([l0,l1,l2])
 * 
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__sum); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_74 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_74 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 903; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_l0 = __pyx_t_74;

  /* "nipy/algorithms/statistics/intvol.pyx":904
 * 
 *     l0 += mask.sum()
 *     return np.array([l0,l1,l2])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__array); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_l1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_34 = PyFloat_FromDouble(__pyx_v_l2); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  __pyx_t_19 = PyList_New(3); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  PyList_SET_ITEM(__pyx_t_19, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_19, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_19, 2, __pyx_t_34);
  __Pyx_GIVEREF(__pyx_t_34);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_34 = 0;
  __pyx_t_34 = PyTuple_New(1); if (unlikely(!__pyx_t_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_34);
  PyTuple_SET_ITEM(__pyx_t_34, 0, ((PyObject *)__pyx_t_19));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_19));
  __pyx_t_19 = 0;
  __pyx_t_19 = PyObject_Call(__pyx_t_5, ((PyObject *)__pyx_t_34), NULL); if (unlikely(!__pyx_t_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 904; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_34)); __pyx_t_34 = 0;
  __pyx_r = __pyx_t_19;
  __pyx_t_19 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_34);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_D.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.Lips2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_D.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords_c.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_cvertices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dstrides.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fcoords.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_coords_c);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask_c);
  __Pyx_XDECREF((PyObject *)__pyx_v_fcoords);
  __Pyx_XDECREF((PyObject *)__pyx_v_D);
  __Pyx_XDECREF((PyObject *)__pyx_v_fmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_fpmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_pmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_d3);
  __Pyx_XDECREF((PyObject *)__pyx_v_d2);
  __Pyx_XDECREF((PyObject *)__pyx_v_cvertices);
  __Pyx_XDECREF(__pyx_v_pmask_shape);
  __Pyx_XDECREF((PyObject *)__pyx_v_strides);
  __Pyx_XDECREF((PyObject *)__pyx_v_dstrides);
  __Pyx_XDECREF(__pyx_v_verts);
  __Pyx_XDECREF(__pyx_v_union);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_m3);
  __Pyx_XDECREF(__pyx_v_m2);
  __Pyx_XDECREF(__pyx_v_pindex);
  __Pyx_XDECREF(__pyx_v_w0);
  __Pyx_XDECREF(__pyx_v_w1);
  __Pyx_XDECREF(__pyx_v_w2);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_coords);
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_27EC2d(PyObject *__pyx_self, PyObject *__pyx_v_mask); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_26EC2d[] = " Compute Euler characteristic of 2D region in `mask`\n\n    Given a 2d `mask`, compute the 0th intrinsic volume (Euler characteristic)\n    of the masked region. The region is broken up into triangles / edges /\n    vertices, which are included based on whether all voxels in the triangle /\n    edge / vertex are in the mask or not.\n\n    Parameters\n    ----------\n    mask : ndarray((i,j), np.int)\n         Binary mask determining whether or not a voxel is in the mask.\n\n    Returns\n    -------\n    mu0 : int\n        Euler characteristic\n\n    Notes\n    -----\n    The array mask is assumed to be binary. At the time of writing, it\n    is not clear how to get cython to use np.bool arrays.\n\n    References\n    ----------\n    Taylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n      with an application to brain mapping.\"\n      Journal of the American Statistical Association, 102(479):913-928.\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_27EC2d = {__Pyx_NAMESTR("EC2d"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_27EC2d, METH_O, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_26EC2d)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_27EC2d(PyObject *__pyx_self, PyObject *__pyx_v_mask) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EC2d (wrapper)", 0);
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_26EC2d(__pyx_self, ((PyObject *)__pyx_v_mask));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":907
 * 
 * 
 * def EC2d(mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic of 2D region in `mask`
 * 
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_26EC2d(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_mask) {
  PyArrayObject *__pyx_v_mask_c = 0;
  PyArrayObject *__pyx_v_fpmask = 0;
  PyArrayObject *__pyx_v_d2 = 0;
  PyArrayObject *__pyx_v_d3 = 0;
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_j;
  npy_intp __pyx_v_l;
  npy_intp __pyx_v_s0;
  npy_intp __pyx_v_s1;
  npy_intp __pyx_v_ds2;
  npy_intp __pyx_v_ds3;
  npy_intp __pyx_v_index;
  npy_intp __pyx_v_m;
  npy_intp __pyx_v_ss0;
  npy_intp __pyx_v_ss1;
  npy_intp __pyx_v_v0;
  npy_intp __pyx_v_v1;
  long __pyx_v_l0;
  PyObject *__pyx_v_pmask_shape = NULL;
  PyObject *__pyx_v_pmask = NULL;
  PyArrayObject *__pyx_v_strides = 0;
  PyObject *__pyx_v_union = NULL;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_v2 = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d2;
  __Pyx_Buffer __pyx_pybuffer_d2;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_d3;
  __Pyx_Buffer __pyx_pybuffer_d3;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_fpmask;
  __Pyx_Buffer __pyx_pybuffer_fpmask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask_c;
  __Pyx_Buffer __pyx_pybuffer_mask_c;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_strides;
  __Pyx_Buffer __pyx_pybuffer_strides;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyArrayObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  npy_intp __pyx_t_11;
  npy_intp __pyx_t_12;
  PyArrayObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyArrayObject *__pyx_t_16 = NULL;
  long __pyx_t_17;
  __pyx_t_5numpy_intp_t __pyx_t_18;
  long __pyx_t_19;
  __pyx_t_5numpy_intp_t __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyArrayObject *__pyx_t_22 = NULL;
  PyArrayObject *__pyx_t_23 = NULL;
  long __pyx_t_24;
  long __pyx_t_25;
  npy_intp __pyx_t_26;
  npy_intp __pyx_t_27;
  npy_intp __pyx_t_28;
  long __pyx_t_29;
  npy_intp __pyx_t_30;
  npy_intp __pyx_t_31;
  long __pyx_t_32;
  npy_intp __pyx_t_33;
  long __pyx_t_34;
  npy_intp __pyx_t_35;
  npy_intp __pyx_t_36;
  long __pyx_t_37;
  npy_intp __pyx_t_38;
  npy_intp __pyx_t_39;
  long __pyx_t_40;
  npy_intp __pyx_t_41;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("EC2d", 0);
  __pyx_pybuffer_mask_c.pybuffer.buf = NULL;
  __pyx_pybuffer_mask_c.refcount = 0;
  __pyx_pybuffernd_mask_c.data = NULL;
  __pyx_pybuffernd_mask_c.rcbuffer = &__pyx_pybuffer_mask_c;
  __pyx_pybuffer_fpmask.pybuffer.buf = NULL;
  __pyx_pybuffer_fpmask.refcount = 0;
  __pyx_pybuffernd_fpmask.data = NULL;
  __pyx_pybuffernd_fpmask.rcbuffer = &__pyx_pybuffer_fpmask;
  __pyx_pybuffer_d2.pybuffer.buf = NULL;
  __pyx_pybuffer_d2.refcount = 0;
  __pyx_pybuffernd_d2.data = NULL;
  __pyx_pybuffernd_d2.rcbuffer = &__pyx_pybuffer_d2;
  __pyx_pybuffer_d3.pybuffer.buf = NULL;
  __pyx_pybuffer_d3.refcount = 0;
  __pyx_pybuffernd_d3.data = NULL;
  __pyx_pybuffernd_d3.rcbuffer = &__pyx_pybuffer_d3;
  __pyx_pybuffer_strides.pybuffer.buf = NULL;
  __pyx_pybuffer_strides.refcount = 0;
  __pyx_pybuffernd_strides.data = NULL;
  __pyx_pybuffernd_strides.rcbuffer = &__pyx_pybuffer_strides;

  /* "nipy/algorithms/statistics/intvol.pyx":936
 *       Journal of the American Statistical Association, 102(479):913-928.
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):             # <<<<<<<<<<<<<<
 *       raise ValueError('mask should be filled with 0/1 '
 *                        'values, but be of type np.int')
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__unique); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_mask);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_mask);
  __Pyx_GIVEREF(__pyx_v_mask);
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(((PyObject *)((PyObject*)(&PySet_Type))), ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__issubset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  if (__pyx_t_5) {

    /* "nipy/algorithms/statistics/intvol.pyx":937
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_56), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 937; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 937; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":952
 *         np.npy_intp ss0, ss1 # strides
 *         np.npy_intp v0, v1 # vertices
 *         long l0 = 0             # <<<<<<<<<<<<<<
 * 
 *     mask_c = mask
 */
  __pyx_v_l0 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":954
 *         long l0 = 0
 * 
 *     mask_c = mask             # <<<<<<<<<<<<<<
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 */
  if (!(likely(((__pyx_v_mask) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_mask, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 954; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_6 = ((PyArrayObject *)__pyx_v_mask);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask_c, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    __pyx_pybuffernd_mask_c.diminfo[0].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask_c.diminfo[0].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_mask_c.diminfo[1].strides = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_mask_c.diminfo[1].shape = __pyx_pybuffernd_mask_c.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 954; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_6 = 0;
  __Pyx_INCREF(__pyx_v_mask);
  __pyx_v_mask_c = ((PyArrayObject *)__pyx_v_mask);

  /* "nipy/algorithms/statistics/intvol.pyx":956
 *     mask_c = mask
 * 
 *     pmask_shape = np.array(mask.shape) + 1             # <<<<<<<<<<<<<<
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1] = mask_c
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__shape); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_int_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 956; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_pmask_shape = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":957
 * 
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)             # <<<<<<<<<<<<<<
 *     pmask[:-1,:-1] = mask_c
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__int); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_v_pmask = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":958
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1 = (pmask.shape[0], pmask.shape[1])
 */
  if (PyObject_SetItem(__pyx_v_pmask, ((PyObject *)__pyx_k_tuple_59), ((PyObject *)__pyx_v_mask_c)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 958; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":960
 *     pmask[:-1,:-1] = mask_c
 * 
 *     s0, s1 = (pmask.shape[0], pmask.shape[1])             # <<<<<<<<<<<<<<
 * 
 *     fpmask = pmask.reshape(-1)
 */
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__shape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, 0, sizeof(long), PyInt_FromLong); if (!__pyx_t_1) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_11 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__shape); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 1, sizeof(long), PyInt_FromLong); if (!__pyx_t_2) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_2); if (unlikely((__pyx_t_12 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 960; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_s0 = __pyx_t_11;
  __pyx_v_s1 = __pyx_t_12;

  /* "nipy/algorithms/statistics/intvol.pyx":962
 *     s0, s1 = (pmask.shape[0], pmask.shape[1])
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
  __pyx_t_2 = PyObject_GetAttr(__pyx_v_pmask, __pyx_n_s__reshape); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_k_tuple_60), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer, (PyObject*)__pyx_v_fpmask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
    }
    __pyx_pybuffernd_fpmask.diminfo[0].strides = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_fpmask.diminfo[0].shape = __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_13 = 0;
  __pyx_v_fpmask = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":966
 *     cdef:
 *         np.ndarray[np.intp_t, ndim=1] strides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)             # <<<<<<<<<<<<<<
 *     ss0, ss1 = strides[0], strides[1]
 * 
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__strides_from); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_14 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__bool); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_pmask_shape);
  __Pyx_GIVEREF(__pyx_v_pmask_shape);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_14);
  __pyx_t_14 = 0;
  __pyx_t_14 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_14);
  __pyx_t_14 = 0;
  __pyx_t_14 = PyDict_New(); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_14));
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__intp); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_14, ((PyObject *)__pyx_n_s__dtype), __pyx_t_15) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), ((PyObject *)__pyx_t_14)); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_14)); __pyx_t_14 = 0;
  if (!(likely(((__pyx_t_15) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_16 = ((PyArrayObject *)__pyx_t_15);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_t_16, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_strides.rcbuffer->pybuffer, (PyObject*)__pyx_v_strides, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    __pyx_pybuffernd_strides.diminfo[0].strides = __pyx_pybuffernd_strides.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_strides.diminfo[0].shape = __pyx_pybuffernd_strides.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 966; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_16 = 0;
  __pyx_v_strides = ((PyArrayObject *)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":967
 *         np.ndarray[np.intp_t, ndim=1] strides
 *     strides = np.array(strides_from(pmask_shape, np.bool), dtype=np.intp)
 *     ss0, ss1 = strides[0], strides[1]             # <<<<<<<<<<<<<<
 * 
 *     # First do the interior contributions.
 */
  __pyx_t_17 = 0;
  __pyx_t_7 = -1;
  if (__pyx_t_17 < 0) {
    __pyx_t_17 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_17 < 0)) __pyx_t_7 = 0;
  } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_7 = 0;
  if (unlikely(__pyx_t_7 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_7);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 967; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_18 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_17, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_t_19 = 1;
  __pyx_t_7 = -1;
  if (__pyx_t_19 < 0) {
    __pyx_t_19 += __pyx_pybuffernd_strides.diminfo[0].shape;
    if (unlikely(__pyx_t_19 < 0)) __pyx_t_7 = 0;
  } else if (unlikely(__pyx_t_19 >= __pyx_pybuffernd_strides.diminfo[0].shape)) __pyx_t_7 = 0;
  if (unlikely(__pyx_t_7 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_7);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 967; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_20 = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_strides.rcbuffer->pybuffer.buf, __pyx_t_19, __pyx_pybuffernd_strides.diminfo[0].strides));
  __pyx_v_ss0 = __pyx_t_18;
  __pyx_v_ss1 = __pyx_t_20;

  /* "nipy/algorithms/statistics/intvol.pyx":972
 *     # We first figure out which vertices, edges, triangles, tetrahedra
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])
 */
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s__join_complexes); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_61));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_61));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_61));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_2 = PyObject_Call(__pyx_t_14, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":973
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1), strides)])
 *     c = cube_with_strides_center((0,0), strides)
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 973; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 973; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_62));
  PyTuple_SET_ITEM(__pyx_t_14, 0, ((PyObject *)__pyx_k_tuple_62));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_62));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_14, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_1 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_14), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 973; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_14)); __pyx_t_14 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":974
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])             # <<<<<<<<<<<<<<
 *     c = cube_with_strides_center((0,0), strides)
 * 
 */
  __pyx_t_14 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 974; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 974; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_63));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_k_tuple_63));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_63));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_21 = PyObject_Call(__pyx_t_14, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 974; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_21);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_21 = 0;
  __pyx_t_21 = PySequence_Tuple(((PyObject *)__pyx_t_3)); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_21));
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_15, ((PyObject *)__pyx_t_21), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_21)); __pyx_t_21 = 0;
  __pyx_v_union = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":975
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])
 *     c = cube_with_strides_center((0,0), strides)             # <<<<<<<<<<<<<<
 * 
 *     d3 = np.array(list(c[3].difference(union[3])))
 */
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 975; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 975; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_64));
  PyTuple_SET_ITEM(__pyx_t_21, 0, ((PyObject *)__pyx_k_tuple_64));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_64));
  __Pyx_INCREF(((PyObject *)__pyx_v_strides));
  PyTuple_SET_ITEM(__pyx_t_21, 1, ((PyObject *)__pyx_v_strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_strides));
  __pyx_t_15 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_21), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 975; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_21)); __pyx_t_21 = 0;
  __pyx_v_c = __pyx_t_15;
  __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":977
 *     c = cube_with_strides_center((0,0), strides)
 * 
 *     d3 = np.array(list(c[3].difference(union[3])))             # <<<<<<<<<<<<<<
 *     d2 = np.array(list(c[2].difference(union[2])))
 * 
 */
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_21 = PyObject_GetAttr(__pyx_t_15, __pyx_n_s__array); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_c, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_15) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_15, __pyx_n_s__difference); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_union, 3, sizeof(long), PyInt_FromLong); if (!__pyx_t_15) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(__pyx_t_21, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_15) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_22 = ((PyArrayObject *)__pyx_t_15);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_t_22, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d3.rcbuffer->pybuffer, (PyObject*)__pyx_v_d3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
    }
    __pyx_pybuffernd_d3.diminfo[0].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d3.diminfo[0].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d3.diminfo[1].strides = __pyx_pybuffernd_d3.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d3.diminfo[1].shape = __pyx_pybuffernd_d3.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 977; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_22 = 0;
  __pyx_v_d3 = ((PyArrayObject *)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":978
 * 
 *     d3 = np.array(list(c[3].difference(union[3])))
 *     d2 = np.array(list(c[2].difference(union[2])))             # <<<<<<<<<<<<<<
 * 
 *     ds2 = d2.shape[0]
 */
  __pyx_t_15 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_15, __pyx_n_s__array); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_c, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_15) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_21 = PyObject_GetAttr(__pyx_t_15, __pyx_n_s__difference); if (unlikely(!__pyx_t_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_union, 2, sizeof(long), PyInt_FromLong); if (!__pyx_t_15) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(__pyx_t_21, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(((PyObject *)((PyObject*)(&PyList_Type))), ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_15) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_5numpy_ndarray))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_23 = ((PyArrayObject *)__pyx_t_15);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_t_23, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_d2.rcbuffer->pybuffer, (PyObject*)__pyx_v_d2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
    }
    __pyx_pybuffernd_d2.diminfo[0].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_d2.diminfo[0].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_d2.diminfo[1].strides = __pyx_pybuffernd_d2.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_d2.diminfo[1].shape = __pyx_pybuffernd_d2.rcbuffer->pybuffer.shape[1];
    if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 978; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_23 = 0;
  __pyx_v_d2 = ((PyArrayObject *)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":980
 *     d2 = np.array(list(c[2].difference(union[2])))
 * 
 *     ds2 = d2.shape[0]             # <<<<<<<<<<<<<<
 *     ds3 = d3.shape[0]
 * 
 */
  __pyx_v_ds2 = (__pyx_v_d2->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":981
 * 
 *     ds2 = d2.shape[0]
 *     ds3 = d3.shape[0]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0-1):
 */
  __pyx_v_ds3 = (__pyx_v_d3->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":983
 *     ds3 = d3.shape[0]
 * 
 *     for i in range(s0-1):             # <<<<<<<<<<<<<<
 *         for j in range(s1-1):
 *             index = i*ss0+j*ss1
 */
  __pyx_t_24 = (__pyx_v_s0 - 1);
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_24; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "nipy/algorithms/statistics/intvol.pyx":984
 * 
 *     for i in range(s0-1):
 *         for j in range(s1-1):             # <<<<<<<<<<<<<<
 *             index = i*ss0+j*ss1
 *             for l in range(ds3):
 */
    __pyx_t_25 = (__pyx_v_s1 - 1);
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_25; __pyx_t_11+=1) {
      __pyx_v_j = __pyx_t_11;

      /* "nipy/algorithms/statistics/intvol.pyx":985
 *     for i in range(s0-1):
 *         for j in range(s1-1):
 *             index = i*ss0+j*ss1             # <<<<<<<<<<<<<<
 *             for l in range(ds3):
 *                 v0 = index + d3[l,0]
 */
      __pyx_v_index = ((__pyx_v_i * __pyx_v_ss0) + (__pyx_v_j * __pyx_v_ss1));

      /* "nipy/algorithms/statistics/intvol.pyx":986
 *         for j in range(s1-1):
 *             index = i*ss0+j*ss1
 *             for l in range(ds3):             # <<<<<<<<<<<<<<
 *                 v0 = index + d3[l,0]
 *                 m = fpmask[v0]
 */
      __pyx_t_26 = __pyx_v_ds3;
      for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
        __pyx_v_l = __pyx_t_27;

        /* "nipy/algorithms/statistics/intvol.pyx":987
 *             index = i*ss0+j*ss1
 *             for l in range(ds3):
 *                 v0 = index + d3[l,0]             # <<<<<<<<<<<<<<
 *                 m = fpmask[v0]
 *                 if m and v0:
 */
        __pyx_t_28 = __pyx_v_l;
        __pyx_t_29 = 0;
        __pyx_t_7 = -1;
        if (__pyx_t_28 < 0) {
          __pyx_t_28 += __pyx_pybuffernd_d3.diminfo[0].shape;
          if (unlikely(__pyx_t_28 < 0)) __pyx_t_7 = 0;
        } else if (unlikely(__pyx_t_28 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_7 = 0;
        if (__pyx_t_29 < 0) {
          __pyx_t_29 += __pyx_pybuffernd_d3.diminfo[1].shape;
          if (unlikely(__pyx_t_29 < 0)) __pyx_t_7 = 1;
        } else if (unlikely(__pyx_t_29 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_7 = 1;
        if (unlikely(__pyx_t_7 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_7);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 987; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_v0 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_d3.diminfo[1].strides)));

        /* "nipy/algorithms/statistics/intvol.pyx":988
 *             for l in range(ds3):
 *                 v0 = index + d3[l,0]
 *                 m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                 if m and v0:
 *                     v1 = index + d3[l,1]
 */
        __pyx_t_30 = __pyx_v_v0;
        __pyx_t_7 = -1;
        if (__pyx_t_30 < 0) {
          __pyx_t_30 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
          if (unlikely(__pyx_t_30 < 0)) __pyx_t_7 = 0;
        } else if (unlikely(__pyx_t_30 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
        if (unlikely(__pyx_t_7 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_7);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 988; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_fpmask.diminfo[0].strides));

        /* "nipy/algorithms/statistics/intvol.pyx":989
 *                 v0 = index + d3[l,0]
 *                 m = fpmask[v0]
 *                 if m and v0:             # <<<<<<<<<<<<<<
 *                     v1 = index + d3[l,1]
 *                     v2 = index + d3[l,2]
 */
        if (__pyx_v_m) {
          __pyx_t_5 = __pyx_v_v0;
        } else {
          __pyx_t_5 = __pyx_v_m;
        }
        if (__pyx_t_5) {

          /* "nipy/algorithms/statistics/intvol.pyx":990
 *                 m = fpmask[v0]
 *                 if m and v0:
 *                     v1 = index + d3[l,1]             # <<<<<<<<<<<<<<
 *                     v2 = index + d3[l,2]
 *                     m = m * fpmask[v1] * fpmask[v2]
 */
          __pyx_t_31 = __pyx_v_l;
          __pyx_t_32 = 1;
          __pyx_t_7 = -1;
          if (__pyx_t_31 < 0) {
            __pyx_t_31 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_31 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_31 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_7 = 0;
          if (__pyx_t_32 < 0) {
            __pyx_t_32 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_32 < 0)) __pyx_t_7 = 1;
          } else if (unlikely(__pyx_t_32 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_7 = 1;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 990; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v1 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_32, __pyx_pybuffernd_d3.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":991
 *                 if m and v0:
 *                     v1 = index + d3[l,1]
 *                     v2 = index + d3[l,2]             # <<<<<<<<<<<<<<
 *                     m = m * fpmask[v1] * fpmask[v2]
 *                     l0 = l0 + m
 */
          __pyx_t_33 = __pyx_v_l;
          __pyx_t_34 = 2;
          __pyx_t_7 = -1;
          if (__pyx_t_33 < 0) {
            __pyx_t_33 += __pyx_pybuffernd_d3.diminfo[0].shape;
            if (unlikely(__pyx_t_33 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_33 >= __pyx_pybuffernd_d3.diminfo[0].shape)) __pyx_t_7 = 0;
          if (__pyx_t_34 < 0) {
            __pyx_t_34 += __pyx_pybuffernd_d3.diminfo[1].shape;
            if (unlikely(__pyx_t_34 < 0)) __pyx_t_7 = 1;
          } else if (unlikely(__pyx_t_34 >= __pyx_pybuffernd_d3.diminfo[1].shape)) __pyx_t_7 = 1;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 991; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_15 = PyInt_FromLong((__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d3.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_d3.diminfo[0].strides, __pyx_t_34, __pyx_pybuffernd_d3.diminfo[1].strides)))); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 991; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_XDECREF(__pyx_v_v2);
          __pyx_v_v2 = __pyx_t_15;
          __pyx_t_15 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":992
 *                     v1 = index + d3[l,1]
 *                     v2 = index + d3[l,2]
 *                     m = m * fpmask[v1] * fpmask[v2]             # <<<<<<<<<<<<<<
 *                     l0 = l0 + m
 * 
 */
          __pyx_t_35 = __pyx_v_v1;
          __pyx_t_7 = -1;
          if (__pyx_t_35 < 0) {
            __pyx_t_35 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_35 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_35 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_t_15 = PyInt_FromLong((__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_fpmask.diminfo[0].strides)))); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_3 = PyObject_GetItem(((PyObject *)__pyx_v_fpmask), __pyx_v_v2); if (!__pyx_t_3) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = PyNumber_Multiply(__pyx_t_15, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_36 = __Pyx_PyInt_from_py_Py_intptr_t(__pyx_t_1); if (unlikely((__pyx_t_36 == (npy_intp)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 992; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_m = __pyx_t_36;

          /* "nipy/algorithms/statistics/intvol.pyx":993
 *                     v2 = index + d3[l,2]
 *                     m = m * fpmask[v1] * fpmask[v2]
 *                     l0 = l0 + m             # <<<<<<<<<<<<<<
 * 
 *             for l in range(ds2):
 */
          __pyx_v_l0 = (__pyx_v_l0 + __pyx_v_m);
          goto __pyx_L10;
        }
        __pyx_L10:;
      }

      /* "nipy/algorithms/statistics/intvol.pyx":995
 *                     l0 = l0 + m
 * 
 *             for l in range(ds2):             # <<<<<<<<<<<<<<
 *                 v0 = index + d2[l,0]
 *                 m = fpmask[v0]
 */
      __pyx_t_26 = __pyx_v_ds2;
      for (__pyx_t_27 = 0; __pyx_t_27 < __pyx_t_26; __pyx_t_27+=1) {
        __pyx_v_l = __pyx_t_27;

        /* "nipy/algorithms/statistics/intvol.pyx":996
 * 
 *             for l in range(ds2):
 *                 v0 = index + d2[l,0]             # <<<<<<<<<<<<<<
 *                 m = fpmask[v0]
 *                 if m:
 */
        __pyx_t_36 = __pyx_v_l;
        __pyx_t_37 = 0;
        __pyx_t_7 = -1;
        if (__pyx_t_36 < 0) {
          __pyx_t_36 += __pyx_pybuffernd_d2.diminfo[0].shape;
          if (unlikely(__pyx_t_36 < 0)) __pyx_t_7 = 0;
        } else if (unlikely(__pyx_t_36 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_7 = 0;
        if (__pyx_t_37 < 0) {
          __pyx_t_37 += __pyx_pybuffernd_d2.diminfo[1].shape;
          if (unlikely(__pyx_t_37 < 0)) __pyx_t_7 = 1;
        } else if (unlikely(__pyx_t_37 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_7 = 1;
        if (unlikely(__pyx_t_7 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_7);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 996; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_v0 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_37, __pyx_pybuffernd_d2.diminfo[1].strides)));

        /* "nipy/algorithms/statistics/intvol.pyx":997
 *             for l in range(ds2):
 *                 v0 = index + d2[l,0]
 *                 m = fpmask[v0]             # <<<<<<<<<<<<<<
 *                 if m:
 *                     v1 = index + d2[l,1]
 */
        __pyx_t_38 = __pyx_v_v0;
        __pyx_t_7 = -1;
        if (__pyx_t_38 < 0) {
          __pyx_t_38 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
          if (unlikely(__pyx_t_38 < 0)) __pyx_t_7 = 0;
        } else if (unlikely(__pyx_t_38 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
        if (unlikely(__pyx_t_7 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_7);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 997; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_fpmask.diminfo[0].strides));

        /* "nipy/algorithms/statistics/intvol.pyx":998
 *                 v0 = index + d2[l,0]
 *                 m = fpmask[v0]
 *                 if m:             # <<<<<<<<<<<<<<
 *                     v1 = index + d2[l,1]
 *                     m = m * fpmask[v1]
 */
        if (__pyx_v_m) {

          /* "nipy/algorithms/statistics/intvol.pyx":999
 *                 m = fpmask[v0]
 *                 if m:
 *                     v1 = index + d2[l,1]             # <<<<<<<<<<<<<<
 *                     m = m * fpmask[v1]
 *                     l0 = l0 - m
 */
          __pyx_t_39 = __pyx_v_l;
          __pyx_t_40 = 1;
          __pyx_t_7 = -1;
          if (__pyx_t_39 < 0) {
            __pyx_t_39 += __pyx_pybuffernd_d2.diminfo[0].shape;
            if (unlikely(__pyx_t_39 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_39 >= __pyx_pybuffernd_d2.diminfo[0].shape)) __pyx_t_7 = 0;
          if (__pyx_t_40 < 0) {
            __pyx_t_40 += __pyx_pybuffernd_d2.diminfo[1].shape;
            if (unlikely(__pyx_t_40 < 0)) __pyx_t_7 = 1;
          } else if (unlikely(__pyx_t_40 >= __pyx_pybuffernd_d2.diminfo[1].shape)) __pyx_t_7 = 1;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 999; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_v1 = (__pyx_v_index + (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_d2.rcbuffer->pybuffer.buf, __pyx_t_39, __pyx_pybuffernd_d2.diminfo[0].strides, __pyx_t_40, __pyx_pybuffernd_d2.diminfo[1].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":1000
 *                 if m:
 *                     v1 = index + d2[l,1]
 *                     m = m * fpmask[v1]             # <<<<<<<<<<<<<<
 *                     l0 = l0 - m
 * 
 */
          __pyx_t_41 = __pyx_v_v1;
          __pyx_t_7 = -1;
          if (__pyx_t_41 < 0) {
            __pyx_t_41 += __pyx_pybuffernd_fpmask.diminfo[0].shape;
            if (unlikely(__pyx_t_41 < 0)) __pyx_t_7 = 0;
          } else if (unlikely(__pyx_t_41 >= __pyx_pybuffernd_fpmask.diminfo[0].shape)) __pyx_t_7 = 0;
          if (unlikely(__pyx_t_7 != -1)) {
            __Pyx_RaiseBufferIndexError(__pyx_t_7);
            {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1000; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          __pyx_v_m = (__pyx_v_m * (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_fpmask.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_fpmask.diminfo[0].strides)));

          /* "nipy/algorithms/statistics/intvol.pyx":1001
 *                     v1 = index + d2[l,1]
 *                     m = m * fpmask[v1]
 *                     l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *     l0 += mask.sum()
 */
          __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
          goto __pyx_L13;
        }
        __pyx_L13:;
      }
    }
  }

  /* "nipy/algorithms/statistics/intvol.pyx":1003
 *                     l0 = l0 - m
 * 
 *     l0 += mask.sum()             # <<<<<<<<<<<<<<
 *     return l0
 * 
 */
  __pyx_t_1 = PyInt_FromLong(__pyx_v_l0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1003; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_GetAttr(__pyx_v_mask, __pyx_n_s__sum); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1003; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1003; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_15); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1003; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_24 = __Pyx_PyInt_AsLong(__pyx_t_3); if (unlikely((__pyx_t_24 == (long)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1003; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_l0 = __pyx_t_24;

  /* "nipy/algorithms/statistics/intvol.pyx":1004
 * 
 *     l0 += mask.sum()
 *     return l0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyInt_FromLong(__pyx_v_l0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1004; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_21);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.EC2d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d2.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_d3.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_fpmask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask_c.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_strides.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mask_c);
  __Pyx_XDECREF((PyObject *)__pyx_v_fpmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_d2);
  __Pyx_XDECREF((PyObject *)__pyx_v_d3);
  __Pyx_XDECREF(__pyx_v_pmask_shape);
  __Pyx_XDECREF(__pyx_v_pmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_strides);
  __Pyx_XDECREF(__pyx_v_union);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_v2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_29Lips1d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_28Lips1d[] = " Estimate intrinsic volumes for 1D region in `mask` given `coords`\n\n    Given a 1d `mask` and coordinates `coords`, estimate the intrinsic volumes\n    of the masked region. The region is broken up into edges / vertices, which\n    are included based on whether all voxels in the edge / vertex are in the\n    mask or not.\n\n    Parameters\n    ----------\n    coords : ndarray((N,i,j,k))\n         Coordinates for the voxels in the mask. ``N`` will often be 1 (for 1\n         dimensional coordinates, but can be any integer > 0\n    mask : ndarray((i,), np.int)\n         Binary mask determining whether or not a voxel is in the mask.\n\n    Returns\n    -------\n    mu : ndarray\n        Array of intrinsic volumes [mu0, mu1], being, respectively:\n        #. Euler characteristic\n        #. Line segment length\n\n    Notes\n    -----\n    The array mask is assumed to be binary. At the time of writing, it\n    is not clear how to get cython to use np.bool arrays.\n\n    References\n    ----------\n    Taylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n      with an application to brain mapping.\"\n      Journal of the American Statistical Association, 102(479):913-928.\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_29Lips1d = {__Pyx_NAMESTR("Lips1d"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_29Lips1d, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_28Lips1d)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_29Lips1d(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_coords = 0;
  PyArrayObject *__pyx_v_mask = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Lips1d (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__coords,&__pyx_n_s__mask,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__coords)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s__mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("Lips1d", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "Lips1d") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_coords = ((PyArrayObject *)values[0]);
    __pyx_v_mask = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Lips1d", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.Lips1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coords), __pyx_ptype_5numpy_ndarray, 1, "coords", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1008; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_28Lips1d(__pyx_self, __pyx_v_coords, __pyx_v_mask);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":1007
 * 
 * 
 * def Lips1d(np.ndarray[np.float_t, ndim=2] coords,             # <<<<<<<<<<<<<<
 *            np.ndarray[np.intp_t, ndim=1] mask):
 *     """ Estimate intrinsic volumes for 1D region in `mask` given `coords`
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_28Lips1d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_coords, PyArrayObject *__pyx_v_mask) {
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_l;
  npy_intp __pyx_v_r;
  npy_intp __pyx_v_s;
  npy_intp __pyx_v_rr;
  npy_intp __pyx_v_ss;
  npy_intp __pyx_v_mr;
  npy_intp __pyx_v_ms;
  npy_intp __pyx_v_s0;
  npy_intp __pyx_v_m;
  double __pyx_v_l0;
  double __pyx_v_l1;
  double __pyx_v_res;
  PyObject *__pyx_v_D = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_coords;
  __Pyx_Buffer __pyx_pybuffer_coords;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  npy_intp __pyx_t_6;
  npy_intp __pyx_t_7;
  npy_intp __pyx_t_8;
  int __pyx_t_9;
  npy_intp __pyx_t_10;
  long __pyx_t_11;
  npy_intp __pyx_t_12;
  npy_intp __pyx_t_13;
  npy_intp __pyx_t_14;
  npy_intp __pyx_t_15;
  npy_intp __pyx_t_16;
  npy_intp __pyx_t_17;
  npy_intp __pyx_t_18;
  npy_intp __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  long __pyx_t_21;
  double __pyx_t_22;
  double __pyx_t_23;
  double __pyx_t_24;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Lips1d", 0);
  __pyx_pybuffer_coords.pybuffer.buf = NULL;
  __pyx_pybuffer_coords.refcount = 0;
  __pyx_pybuffernd_coords.data = NULL;
  __pyx_pybuffernd_coords.rcbuffer = &__pyx_pybuffer_coords;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_coords.rcbuffer->pybuffer, (PyObject*)__pyx_v_coords, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_coords.diminfo[0].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_coords.diminfo[0].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_coords.diminfo[1].strides = __pyx_pybuffernd_coords.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_coords.diminfo[1].shape = __pyx_pybuffernd_coords.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "nipy/algorithms/statistics/intvol.pyx":1042
 *       Journal of the American Statistical Association, 102(479):913-928.
 *     """
 *     if mask.shape[0] != coords.shape[1]:             # <<<<<<<<<<<<<<
 *         raise ValueError('shape of mask does not match coordinates')
 *     if not set(np.unique(mask)).issubset([0,1]):
 */
  __pyx_t_1 = ((__pyx_v_mask->dimensions[0]) != (__pyx_v_coords->dimensions[1]));
  if (__pyx_t_1) {

    /* "nipy/algorithms/statistics/intvol.pyx":1043
 *     """
 *     if mask.shape[0] != coords.shape[1]:
 *         raise ValueError('shape of mask does not match coordinates')             # <<<<<<<<<<<<<<
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '
 */
    __pyx_t_2 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_65), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1043; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1043; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":1044
 *     if mask.shape[0] != coords.shape[1]:
 *         raise ValueError('shape of mask does not match coordinates')
 *     if not set(np.unique(mask)).issubset([0,1]):             # <<<<<<<<<<<<<<
 *       raise ValueError('mask should be filled with 0/1 '
 *                        'values, but be of type np.int')
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__unique); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_mask));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_mask));
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_Call(((PyObject *)((PyObject*)(&PySet_Type))), ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__issubset); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_4));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_4));
  __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1044; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = (!__pyx_t_1);
  if (__pyx_t_5) {

    /* "nipy/algorithms/statistics/intvol.pyx":1045
 *         raise ValueError('shape of mask does not match coordinates')
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_66), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1045; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1045; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L4;
  }
  __pyx_L4:;

  /* "nipy/algorithms/statistics/intvol.pyx":1052
 *         double res
 * 
 *     l0 = 0; l1 = 0             # <<<<<<<<<<<<<<
 *     s0 = mask.shape[0]
 *     D = np.zeros((2,2))
 */
  __pyx_v_l0 = 0.0;
  __pyx_v_l1 = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":1053
 * 
 *     l0 = 0; l1 = 0
 *     s0 = mask.shape[0]             # <<<<<<<<<<<<<<
 *     D = np.zeros((2,2))
 * 
 */
  __pyx_v_s0 = (__pyx_v_mask->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":1054
 *     l0 = 0; l1 = 0
 *     s0 = mask.shape[0]
 *     D = np.zeros((2,2))             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0):
 */
  __pyx_t_4 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_GetAttr(__pyx_t_4, __pyx_n_s__zeros); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_k_tuple_68), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_D = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":1056
 *     D = np.zeros((2,2))
 * 
 *     for i in range(s0):             # <<<<<<<<<<<<<<
 *         for r in range(2):
 *             rr = (i+r) % s0
 */
  __pyx_t_6 = __pyx_v_s0;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "nipy/algorithms/statistics/intvol.pyx":1057
 * 
 *     for i in range(s0):
 *         for r in range(2):             # <<<<<<<<<<<<<<
 *             rr = (i+r) % s0
 *             mr = mask[rr]
 */
    for (__pyx_t_8 = 0; __pyx_t_8 < 2; __pyx_t_8+=1) {
      __pyx_v_r = __pyx_t_8;

      /* "nipy/algorithms/statistics/intvol.pyx":1058
 *     for i in range(s0):
 *         for r in range(2):
 *             rr = (i+r) % s0             # <<<<<<<<<<<<<<
 *             mr = mask[rr]
 *             for s in range(r+1):
 */
      __pyx_t_9 = (__pyx_v_i + __pyx_v_r);
      if (unlikely(__pyx_v_s0 == 0)) {
        PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1058; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_v_rr = __Pyx_mod_int(__pyx_t_9, __pyx_v_s0);

      /* "nipy/algorithms/statistics/intvol.pyx":1059
 *         for r in range(2):
 *             rr = (i+r) % s0
 *             mr = mask[rr]             # <<<<<<<<<<<<<<
 *             for s in range(r+1):
 *                 res = 0
 */
      __pyx_t_10 = __pyx_v_rr;
      __pyx_t_9 = -1;
      if (__pyx_t_10 < 0) {
        __pyx_t_10 += __pyx_pybuffernd_mask.diminfo[0].shape;
        if (unlikely(__pyx_t_10 < 0)) __pyx_t_9 = 0;
      } else if (unlikely(__pyx_t_10 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_9 = 0;
      if (unlikely(__pyx_t_9 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_9);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1059; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_v_mr = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_mask.diminfo[0].strides));

      /* "nipy/algorithms/statistics/intvol.pyx":1060
 *             rr = (i+r) % s0
 *             mr = mask[rr]
 *             for s in range(r+1):             # <<<<<<<<<<<<<<
 *                 res = 0
 *                 ss = (i+s) % s0
 */
      __pyx_t_11 = (__pyx_v_r + 1);
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_s = __pyx_t_12;

        /* "nipy/algorithms/statistics/intvol.pyx":1061
 *             mr = mask[rr]
 *             for s in range(r+1):
 *                 res = 0             # <<<<<<<<<<<<<<
 *                 ss = (i+s) % s0
 *                 ms = mask[ss]
 */
        __pyx_v_res = 0.0;

        /* "nipy/algorithms/statistics/intvol.pyx":1062
 *             for s in range(r+1):
 *                 res = 0
 *                 ss = (i+s) % s0             # <<<<<<<<<<<<<<
 *                 ms = mask[ss]
 *                 if mr * ms * ((i+r) < s0) * ((i+s) < s0):
 */
        __pyx_t_9 = (__pyx_v_i + __pyx_v_s);
        if (unlikely(__pyx_v_s0 == 0)) {
          PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1062; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_ss = __Pyx_mod_int(__pyx_t_9, __pyx_v_s0);

        /* "nipy/algorithms/statistics/intvol.pyx":1063
 *                 res = 0
 *                 ss = (i+s) % s0
 *                 ms = mask[ss]             # <<<<<<<<<<<<<<
 *                 if mr * ms * ((i+r) < s0) * ((i+s) < s0):
 *                     for l in range(coords.shape[0]):
 */
        __pyx_t_13 = __pyx_v_ss;
        __pyx_t_9 = -1;
        if (__pyx_t_13 < 0) {
          __pyx_t_13 += __pyx_pybuffernd_mask.diminfo[0].shape;
          if (unlikely(__pyx_t_13 < 0)) __pyx_t_9 = 0;
        } else if (unlikely(__pyx_t_13 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_9 = 0;
        if (unlikely(__pyx_t_9 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_9);
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1063; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        __pyx_v_ms = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_mask.diminfo[0].strides));

        /* "nipy/algorithms/statistics/intvol.pyx":1064
 *                 ss = (i+s) % s0
 *                 ms = mask[ss]
 *                 if mr * ms * ((i+r) < s0) * ((i+s) < s0):             # <<<<<<<<<<<<<<
 *                     for l in range(coords.shape[0]):
 *                         res += coords[l,ss] * coords[l,rr]
 */
        __pyx_t_9 = (((__pyx_v_mr * __pyx_v_ms) * ((__pyx_v_i + __pyx_v_r) < __pyx_v_s0)) * ((__pyx_v_i + __pyx_v_s) < __pyx_v_s0));
        if (__pyx_t_9) {

          /* "nipy/algorithms/statistics/intvol.pyx":1065
 *                 ms = mask[ss]
 *                 if mr * ms * ((i+r) < s0) * ((i+s) < s0):
 *                     for l in range(coords.shape[0]):             # <<<<<<<<<<<<<<
 *                         res += coords[l,ss] * coords[l,rr]
 *                         D[r,s] = res
 */
          __pyx_t_14 = (__pyx_v_coords->dimensions[0]);
          for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
            __pyx_v_l = __pyx_t_15;

            /* "nipy/algorithms/statistics/intvol.pyx":1066
 *                 if mr * ms * ((i+r) < s0) * ((i+s) < s0):
 *                     for l in range(coords.shape[0]):
 *                         res += coords[l,ss] * coords[l,rr]             # <<<<<<<<<<<<<<
 *                         D[r,s] = res
 *                         D[s,r] = res
 */
            __pyx_t_16 = __pyx_v_l;
            __pyx_t_17 = __pyx_v_ss;
            __pyx_t_9 = -1;
            if (__pyx_t_16 < 0) {
              __pyx_t_16 += __pyx_pybuffernd_coords.diminfo[0].shape;
              if (unlikely(__pyx_t_16 < 0)) __pyx_t_9 = 0;
            } else if (unlikely(__pyx_t_16 >= __pyx_pybuffernd_coords.diminfo[0].shape)) __pyx_t_9 = 0;
            if (__pyx_t_17 < 0) {
              __pyx_t_17 += __pyx_pybuffernd_coords.diminfo[1].shape;
              if (unlikely(__pyx_t_17 < 0)) __pyx_t_9 = 1;
            } else if (unlikely(__pyx_t_17 >= __pyx_pybuffernd_coords.diminfo[1].shape)) __pyx_t_9 = 1;
            if (unlikely(__pyx_t_9 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_9);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1066; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_t_18 = __pyx_v_l;
            __pyx_t_19 = __pyx_v_rr;
            __pyx_t_9 = -1;
            if (__pyx_t_18 < 0) {
              __pyx_t_18 += __pyx_pybuffernd_coords.diminfo[0].shape;
              if (unlikely(__pyx_t_18 < 0)) __pyx_t_9 = 0;
            } else if (unlikely(__pyx_t_18 >= __pyx_pybuffernd_coords.diminfo[0].shape)) __pyx_t_9 = 0;
            if (__pyx_t_19 < 0) {
              __pyx_t_19 += __pyx_pybuffernd_coords.diminfo[1].shape;
              if (unlikely(__pyx_t_19 < 0)) __pyx_t_9 = 1;
            } else if (unlikely(__pyx_t_19 >= __pyx_pybuffernd_coords.diminfo[1].shape)) __pyx_t_9 = 1;
            if (unlikely(__pyx_t_9 != -1)) {
              __Pyx_RaiseBufferIndexError(__pyx_t_9);
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1066; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            __pyx_v_res = (__pyx_v_res + ((*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_coords.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_coords.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_coords.diminfo[1].strides)) * (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float_t *, __pyx_pybuffernd_coords.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_coords.diminfo[0].strides, __pyx_t_19, __pyx_pybuffernd_coords.diminfo[1].strides))));

            /* "nipy/algorithms/statistics/intvol.pyx":1067
 *                     for l in range(coords.shape[0]):
 *                         res += coords[l,ss] * coords[l,rr]
 *                         D[r,s] = res             # <<<<<<<<<<<<<<
 *                         D[s,r] = res
 *                 else:
 */
            __pyx_t_4 = PyFloat_FromDouble(__pyx_v_res); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1067; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_r); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1067; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_s); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1067; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1067; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_20);
            PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_2);
            __pyx_t_3 = 0;
            __pyx_t_2 = 0;
            if (PyObject_SetItem(__pyx_v_D, ((PyObject *)__pyx_t_20), __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1067; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_DECREF(((PyObject *)__pyx_t_20)); __pyx_t_20 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "nipy/algorithms/statistics/intvol.pyx":1068
 *                         res += coords[l,ss] * coords[l,rr]
 *                         D[r,s] = res
 *                         D[s,r] = res             # <<<<<<<<<<<<<<
 *                 else:
 *                     D[r,s] = 0
 */
            __pyx_t_4 = PyFloat_FromDouble(__pyx_v_res); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1068; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_20 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_s); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1068; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_20);
            __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_r); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1068; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1068; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_20);
            __Pyx_GIVEREF(__pyx_t_20);
            PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
            __Pyx_GIVEREF(__pyx_t_2);
            __pyx_t_20 = 0;
            __pyx_t_2 = 0;
            if (PyObject_SetItem(__pyx_v_D, ((PyObject *)__pyx_t_3), __pyx_t_4) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1068; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          goto __pyx_L11;
        }
        /*else*/ {

          /* "nipy/algorithms/statistics/intvol.pyx":1070
 *                         D[s,r] = res
 *                 else:
 *                     D[r,s] = 0             # <<<<<<<<<<<<<<
 *                     D[s,r] = 0
 * 
 */
          __pyx_t_4 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_r); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1070; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_s); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1070; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1070; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_3);
          __pyx_t_4 = 0;
          __pyx_t_3 = 0;
          if (PyObject_SetItem(__pyx_v_D, ((PyObject *)__pyx_t_2), __pyx_int_0) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1070; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;

          /* "nipy/algorithms/statistics/intvol.pyx":1071
 *                 else:
 *                     D[r,s] = 0
 *                     D[s,r] = 0             # <<<<<<<<<<<<<<
 * 
 *         m = mask[i]
 */
          __pyx_t_2 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_s); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1071; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyInt_to_py_Py_intptr_t(__pyx_v_r); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1071; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1071; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_2);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_3);
          __pyx_t_2 = 0;
          __pyx_t_3 = 0;
          if (PyObject_SetItem(__pyx_v_D, ((PyObject *)__pyx_t_4), __pyx_int_0) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1071; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
        }
        __pyx_L11:;
      }
    }

    /* "nipy/algorithms/statistics/intvol.pyx":1073
 *                     D[s,r] = 0
 * 
 *         m = mask[i]             # <<<<<<<<<<<<<<
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 */
    __pyx_t_8 = __pyx_v_i;
    __pyx_t_9 = -1;
    if (__pyx_t_8 < 0) {
      __pyx_t_8 += __pyx_pybuffernd_mask.diminfo[0].shape;
      if (unlikely(__pyx_t_8 < 0)) __pyx_t_9 = 0;
    } else if (unlikely(__pyx_t_8 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_9 = 0;
    if (unlikely(__pyx_t_9 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_9);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1073; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_mask.diminfo[0].strides));

    /* "nipy/algorithms/statistics/intvol.pyx":1074
 * 
 *         m = mask[i]
 *         if m:             # <<<<<<<<<<<<<<
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 *             l1 = l1 + m * mu1_edge(D[0,0], D[0,1], D[1,1])
 */
    if (__pyx_v_m) {

      /* "nipy/algorithms/statistics/intvol.pyx":1075
 *         m = mask[i]
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))             # <<<<<<<<<<<<<<
 *             l1 = l1 + m * mu1_edge(D[0,0], D[0,1], D[1,1])
 *             l0 = l0 - m
 */
      __pyx_t_11 = (__pyx_v_i + 1);
      if (unlikely(__pyx_v_s0 == 0)) {
        PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1075; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_t_21 = __Pyx_mod_long(__pyx_t_11, __pyx_v_s0);
      __pyx_t_9 = -1;
      if (__pyx_t_21 < 0) {
        __pyx_t_21 += __pyx_pybuffernd_mask.diminfo[0].shape;
        if (unlikely(__pyx_t_21 < 0)) __pyx_t_9 = 0;
      } else if (unlikely(__pyx_t_21 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_9 = 0;
      if (unlikely(__pyx_t_9 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_9);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1075; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_v_m = (__pyx_v_m * ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_mask.diminfo[0].strides)) * ((__pyx_v_i + 1) < __pyx_v_s0)));

      /* "nipy/algorithms/statistics/intvol.pyx":1076
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 *             l1 = l1 + m * mu1_edge(D[0,0], D[0,1], D[1,1])             # <<<<<<<<<<<<<<
 *             l0 = l0 - m
 * 
 */
      __pyx_t_4 = PyObject_GetItem(__pyx_v_D, ((PyObject *)__pyx_k_tuple_69)); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_22 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_22 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_GetItem(__pyx_v_D, ((PyObject *)__pyx_k_tuple_70)); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_23 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_23 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_GetItem(__pyx_v_D, ((PyObject *)__pyx_k_tuple_71)); if (!__pyx_t_4) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_t_4); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_l1 = (__pyx_v_l1 + (__pyx_v_m * __pyx_f_4nipy_10algorithms_10statistics_6intvol_mu1_edge(__pyx_t_22, __pyx_t_23, __pyx_t_24, 0)));

      /* "nipy/algorithms/statistics/intvol.pyx":1077
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 *             l1 = l1 + m * mu1_edge(D[0,0], D[0,1], D[1,1])
 *             l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *     l0 += mask.sum()
 */
      __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
      goto __pyx_L14;
    }
    __pyx_L14:;
  }

  /* "nipy/algorithms/statistics/intvol.pyx":1079
 *             l0 = l0 - m
 * 
 *     l0 += mask.sum()             # <<<<<<<<<<<<<<
 *     return np.array([l0,l1])
 * 
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1079; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_GetAttr(((PyObject *)__pyx_v_mask), __pyx_n_s__sum); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1079; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_Call(__pyx_t_3, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1079; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1079; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_24 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_24 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1079; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_l0 = __pyx_t_24;

  /* "nipy/algorithms/statistics/intvol.pyx":1080
 * 
 *     l0 += mask.sum()
 *     return np.array([l0,l1])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__array); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_l1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_20 = PyList_New(2); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  PyList_SET_ITEM(__pyx_t_20, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_20, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_20));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_20));
  __pyx_t_20 = 0;
  __pyx_t_20 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1080; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_20;
  __pyx_t_20 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_20);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.Lips1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_coords.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_D);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_31EC1d(PyObject *__pyx_self, PyObject *__pyx_v_mask); /*proto*/
static char __pyx_doc_4nipy_10algorithms_10statistics_6intvol_30EC1d[] = " Compute Euler characteristic for 1d `mask`\n\n    Given a 1d mask `mask`, compute the 0th intrinsic volume (Euler\n    characteristic) of the masked region. The region is broken up into edges /\n    vertices, which are included based on whether all voxels in the edge /\n    vertex are in the mask or not.\n\n    Parameters\n    ----------\n    mask : ndarray((i,), np.int)\n         Binary mask determining whether or not a voxel is in the mask.\n\n    Returns\n    -------\n    mu0 : int\n        Euler characteristic\n\n    Notes\n    -----\n    The array mask is assumed to be binary. At the time of writing, it\n    is not clear how to get cython to use np.bool arrays.\n\n    The 3d cubes are triangulated into 6 tetrahedra of equal volume, as\n    described in the reference below.\n\n    References\n    ----------\n    Taylor, J.E. & Worsley, K.J. (2007). \"Detecting sparse signal in random fields,\n      with an application to brain mapping.\"\n      Journal of the American Statistical Association, 102(479):913-928.\n    ";
static PyMethodDef __pyx_mdef_4nipy_10algorithms_10statistics_6intvol_31EC1d = {__Pyx_NAMESTR("EC1d"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_31EC1d, METH_O, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_30EC1d)};
static PyObject *__pyx_pw_4nipy_10algorithms_10statistics_6intvol_31EC1d(PyObject *__pyx_self, PyObject *__pyx_v_mask) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("EC1d (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_r = __pyx_pf_4nipy_10algorithms_10statistics_6intvol_30EC1d(__pyx_self, ((PyArrayObject *)__pyx_v_mask));
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nipy/algorithms/statistics/intvol.pyx":1083
 * 
 * 
 * def EC1d(np.ndarray[np.intp_t, ndim=1] mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic for 1d `mask`
 * 
 */

static PyObject *__pyx_pf_4nipy_10algorithms_10statistics_6intvol_30EC1d(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_mask) {
  npy_intp __pyx_v_i;
  npy_intp __pyx_v_m;
  npy_intp __pyx_v_s0;
  double __pyx_v_l0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  npy_intp __pyx_t_6;
  npy_intp __pyx_t_7;
  npy_intp __pyx_t_8;
  int __pyx_t_9;
  long __pyx_t_10;
  long __pyx_t_11;
  double __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("EC1d", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_intp_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "nipy/algorithms/statistics/intvol.pyx":1115
 *       Journal of the American Statistical Association, 102(479):913-928.
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):             # <<<<<<<<<<<<<<
 *       raise ValueError('mask should be filled with 0/1 '
 *                        'values, but be of type np.int')
 */
  __pyx_t_1 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__unique); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_mask));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_mask));
  __pyx_t_3 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(((PyObject *)((PyObject*)(&PySet_Type))), ((PyObject *)__pyx_t_1), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_3, __pyx_n_s__issubset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_t_3));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_3));
  __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_Call(__pyx_t_1, ((PyObject *)__pyx_t_2), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (!__pyx_t_4);
  if (__pyx_t_5) {

    /* "nipy/algorithms/statistics/intvol.pyx":1116
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
    __pyx_t_3 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_72), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1116; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1116; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "nipy/algorithms/statistics/intvol.pyx":1120
 *     cdef:
 *         np.npy_intp i, m, s0
 *         double l0 = 0             # <<<<<<<<<<<<<<
 * 
 *     s0 = mask.shape[0]
 */
  __pyx_v_l0 = 0.0;

  /* "nipy/algorithms/statistics/intvol.pyx":1122
 *         double l0 = 0
 * 
 *     s0 = mask.shape[0]             # <<<<<<<<<<<<<<
 *     for i in range(s0):
 *         m = mask[i]
 */
  __pyx_v_s0 = (__pyx_v_mask->dimensions[0]);

  /* "nipy/algorithms/statistics/intvol.pyx":1123
 * 
 *     s0 = mask.shape[0]
 *     for i in range(s0):             # <<<<<<<<<<<<<<
 *         m = mask[i]
 *         if m:
 */
  __pyx_t_6 = __pyx_v_s0;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "nipy/algorithms/statistics/intvol.pyx":1124
 *     s0 = mask.shape[0]
 *     for i in range(s0):
 *         m = mask[i]             # <<<<<<<<<<<<<<
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 */
    __pyx_t_8 = __pyx_v_i;
    __pyx_t_9 = -1;
    if (__pyx_t_8 < 0) {
      __pyx_t_8 += __pyx_pybuffernd_mask.diminfo[0].shape;
      if (unlikely(__pyx_t_8 < 0)) __pyx_t_9 = 0;
    } else if (unlikely(__pyx_t_8 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_9 = 0;
    if (unlikely(__pyx_t_9 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_9);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1124; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_v_m = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_mask.diminfo[0].strides));

    /* "nipy/algorithms/statistics/intvol.pyx":1125
 *     for i in range(s0):
 *         m = mask[i]
 *         if m:             # <<<<<<<<<<<<<<
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 *             l0 = l0 - m
 */
    if (__pyx_v_m) {

      /* "nipy/algorithms/statistics/intvol.pyx":1126
 *         m = mask[i]
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))             # <<<<<<<<<<<<<<
 *             l0 = l0 - m
 * 
 */
      __pyx_t_10 = (__pyx_v_i + 1);
      if (unlikely(__pyx_v_s0 == 0)) {
        PyErr_Format(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_t_11 = __Pyx_mod_long(__pyx_t_10, __pyx_v_s0);
      __pyx_t_9 = -1;
      if (__pyx_t_11 < 0) {
        __pyx_t_11 += __pyx_pybuffernd_mask.diminfo[0].shape;
        if (unlikely(__pyx_t_11 < 0)) __pyx_t_9 = 0;
      } else if (unlikely(__pyx_t_11 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_9 = 0;
      if (unlikely(__pyx_t_9 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_9);
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1126; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_v_m = (__pyx_v_m * ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_intp_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_11, __pyx_pybuffernd_mask.diminfo[0].strides)) * ((__pyx_v_i + 1) < __pyx_v_s0)));

      /* "nipy/algorithms/statistics/intvol.pyx":1127
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 *             l0 = l0 - m             # <<<<<<<<<<<<<<
 * 
 *     l0 += mask.sum()
 */
      __pyx_v_l0 = (__pyx_v_l0 - __pyx_v_m);
      goto __pyx_L6;
    }
    __pyx_L6:;
  }

  /* "nipy/algorithms/statistics/intvol.pyx":1129
 *             l0 = l0 - m
 * 
 *     l0 += mask.sum()             # <<<<<<<<<<<<<<
 *     return l0
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1129; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_GetAttr(((PyObject *)__pyx_v_mask), __pyx_n_s__sum); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1129; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Call(__pyx_t_2, ((PyObject *)__pyx_empty_tuple), NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1129; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1129; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_12 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1129; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_l0 = __pyx_t_12;

  /* "nipy/algorithms/statistics/intvol.pyx":1130
 * 
 *     l0 += mask.sum()
 *     return l0             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_l0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1130; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("nipy.algorithms.statistics.intvol.EC1d", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":194
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "numpy.pxd":200
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = (__pyx_v_info == NULL);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":203
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":204
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":206
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "numpy.pxd":208
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = ((sizeof(npy_intp)) != (sizeof(Py_ssize_t)));
  if (__pyx_t_1) {

    /* "numpy.pxd":209
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;
    goto __pyx_L4;
  }
  /*else*/ {

    /* "numpy.pxd":211
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "numpy.pxd":213
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS);
  if (__pyx_t_1) {

    /* "numpy.pxd":214
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
    __pyx_t_2 = (!PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS));
    __pyx_t_3 = __pyx_t_2;
  } else {
    __pyx_t_3 = __pyx_t_1;
  }
  if (__pyx_t_3) {

    /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_74), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L5;
  }
  __pyx_L5:;

  /* "numpy.pxd":217
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_3 = ((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS);
  if (__pyx_t_3) {

    /* "numpy.pxd":218
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
    __pyx_t_1 = (!PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS));
    __pyx_t_2 = __pyx_t_1;
  } else {
    __pyx_t_2 = __pyx_t_3;
  }
  if (__pyx_t_2) {

    /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_76), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    goto __pyx_L6;
  }
  __pyx_L6:;

  /* "numpy.pxd":221
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "numpy.pxd":222
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "numpy.pxd":223
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  if (__pyx_v_copy_shape) {

    /* "numpy.pxd":226
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "numpy.pxd":227
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "numpy.pxd":228
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_5 = __pyx_v_ndim;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "numpy.pxd":229
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "numpy.pxd":230
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }
    goto __pyx_L7;
  }
  /*else*/ {

    /* "numpy.pxd":232
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "numpy.pxd":233
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L7:;

  /* "numpy.pxd":234
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "numpy.pxd":235
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "numpy.pxd":236
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!PyArray_ISWRITEABLE(__pyx_v_self));

  /* "numpy.pxd":239
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef list stack
 */
  __pyx_v_f = NULL;

  /* "numpy.pxd":240
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef list stack
 *             cdef int offset
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "numpy.pxd":244
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "numpy.pxd":246
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = (!__pyx_v_hasfields);
  if (__pyx_t_2) {
    __pyx_t_3 = (!__pyx_v_copy_shape);
    __pyx_t_1 = __pyx_t_3;
  } else {
    __pyx_t_1 = __pyx_t_2;
  }
  if (__pyx_t_1) {

    /* "numpy.pxd":248
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;
    goto __pyx_L10;
  }
  /*else*/ {

    /* "numpy.pxd":251
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L10:;

  /* "numpy.pxd":253
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = (!__pyx_v_hasfields);
  if (__pyx_t_1) {

    /* "numpy.pxd":254
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_5 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_5;

    /* "numpy.pxd":255
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_1 = (__pyx_v_descr->byteorder == '>');
    if (__pyx_t_1) {
      __pyx_t_2 = __pyx_v_little_endian;
    } else {
      __pyx_t_2 = __pyx_t_1;
    }
    if (!__pyx_t_2) {

      /* "numpy.pxd":256
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
      __pyx_t_1 = (__pyx_v_descr->byteorder == '<');
      if (__pyx_t_1) {
        __pyx_t_3 = (!__pyx_v_little_endian);
        __pyx_t_7 = __pyx_t_3;
      } else {
        __pyx_t_7 = __pyx_t_1;
      }
      __pyx_t_1 = __pyx_t_7;
    } else {
      __pyx_t_1 = __pyx_t_2;
    }
    if (__pyx_t_1) {

      /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_4 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_78), NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L12;
    }
    __pyx_L12:;

    /* "numpy.pxd":258
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_BYTE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__b;
      goto __pyx_L13;
    }

    /* "numpy.pxd":259
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_UBYTE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__B;
      goto __pyx_L13;
    }

    /* "numpy.pxd":260
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_SHORT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__h;
      goto __pyx_L13;
    }

    /* "numpy.pxd":261
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_USHORT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__H;
      goto __pyx_L13;
    }

    /* "numpy.pxd":262
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_INT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__i;
      goto __pyx_L13;
    }

    /* "numpy.pxd":263
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_UINT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__I;
      goto __pyx_L13;
    }

    /* "numpy.pxd":264
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__l;
      goto __pyx_L13;
    }

    /* "numpy.pxd":265
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_ULONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__L;
      goto __pyx_L13;
    }

    /* "numpy.pxd":266
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONGLONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__q;
      goto __pyx_L13;
    }

    /* "numpy.pxd":267
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_ULONGLONG);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Q;
      goto __pyx_L13;
    }

    /* "numpy.pxd":268
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_FLOAT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__f;
      goto __pyx_L13;
    }

    /* "numpy.pxd":269
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_DOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__d;
      goto __pyx_L13;
    }

    /* "numpy.pxd":270
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_LONGDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__g;
      goto __pyx_L13;
    }

    /* "numpy.pxd":271
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CFLOAT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zf;
      goto __pyx_L13;
    }

    /* "numpy.pxd":272
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zd;
      goto __pyx_L13;
    }

    /* "numpy.pxd":273
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
    __pyx_t_1 = (__pyx_v_t == NPY_CLONGDOUBLE);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__Zg;
      goto __pyx_L13;
    }

    /* "numpy.pxd":274
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
    __pyx_t_1 = (__pyx_v_t == NPY_OBJECT);
    if (__pyx_t_1) {
      __pyx_v_f = __pyx_k__O;
      goto __pyx_L13;
    }
    /*else*/ {

      /* "numpy.pxd":276
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_4 = PyInt_FromLong(__pyx_v_t); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_79), __pyx_t_4); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(((PyObject *)__pyx_t_8));
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_t_8));
      __Pyx_GIVEREF(((PyObject *)__pyx_t_8));
      __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_4), NULL); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 276; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_L13:;

    /* "numpy.pxd":277
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "numpy.pxd":278
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;
    goto __pyx_L11;
  }
  /*else*/ {

    /* "numpy.pxd":280
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
    __pyx_v_info->format = ((char *)malloc(255));

    /* "numpy.pxd":281
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "numpy.pxd":282
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "numpy.pxd":285
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)             # <<<<<<<<<<<<<<
 *                 f[0] = c'\0' # Terminate format string
 * 
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 255), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 283; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_f = __pyx_t_9;

    /* "numpy.pxd":286
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }
  __pyx_L11:;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));
  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":288
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "numpy.pxd":289
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = PyArray_HASFIELDS(__pyx_v_self);
  if (__pyx_t_1) {

    /* "numpy.pxd":290
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);
    goto __pyx_L3;
  }
  __pyx_L3:;

  /* "numpy.pxd":291
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = ((sizeof(npy_intp)) != (sizeof(Py_ssize_t)));
  if (__pyx_t_1) {

    /* "numpy.pxd":292
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);
    goto __pyx_L4;
  }
  __pyx_L4:;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":768
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "numpy.pxd":769
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":771
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "numpy.pxd":772
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 772; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":774
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "numpy.pxd":775
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 775; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":777
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "numpy.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 778; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":780
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "numpy.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 781; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":783
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  long __pyx_t_11;
  char *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "numpy.pxd":790
 *     cdef int delta_offset
 *     cdef tuple i
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "numpy.pxd":791
 *     cdef tuple i
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "numpy.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(((PyObject *)__pyx_v_descr->names) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __pyx_t_1 = ((PyObject *)__pyx_v_descr->names); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 794; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    #endif
    __Pyx_XDECREF(__pyx_v_childname);
    __pyx_v_childname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "numpy.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (!__pyx_t_3) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected tuple, got %.200s", Py_TYPE(__pyx_t_3)->tp_name), 0))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 795; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_fields));
    __pyx_v_fields = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "numpy.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 */
    if (likely(PyTuple_CheckExact(((PyObject *)__pyx_v_fields)))) {
      PyObject* sequence = ((PyObject *)__pyx_v_fields);
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      #endif
    } else if (1) {
      __Pyx_RaiseNoneNotIterableError(); {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    } else
    {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(((PyObject *)__pyx_v_fields)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_L6_unpacking_done:;
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 796; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_XDECREF(((PyObject *)__pyx_v_child));
    __pyx_v_child = ((PyArray_Descr *)__pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_v_new_offset);
    __pyx_v_new_offset = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "numpy.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = PyInt_FromLong((__pyx_v_end - __pyx_v_f)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_3); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_int_15, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 798; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_81), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L7;
    }
    __pyx_L7:;

    /* "numpy.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = (__pyx_v_child->byteorder == '>');
    if (__pyx_t_7) {
      __pyx_t_8 = __pyx_v_little_endian;
    } else {
      __pyx_t_8 = __pyx_t_7;
    }
    if (!__pyx_t_8) {

      /* "numpy.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
      __pyx_t_7 = (__pyx_v_child->byteorder == '<');
      if (__pyx_t_7) {
        __pyx_t_9 = (!__pyx_v_little_endian);
        __pyx_t_10 = __pyx_t_9;
      } else {
        __pyx_t_10 = __pyx_t_7;
      }
      __pyx_t_7 = __pyx_t_10;
    } else {
      __pyx_t_7 = __pyx_t_8;
    }
    if (__pyx_t_7) {

      /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_k_tuple_82), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      goto __pyx_L8;
    }
    __pyx_L8:;

    /* "numpy.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_5 = PyInt_FromLong((__pyx_v_offset[0])); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!__pyx_t_7) break;

      /* "numpy.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 120;

      /* "numpy.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "numpy.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_11 = 0;
      (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + 1);
    }

    /* "numpy.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_11 = 0;
    (__pyx_v_offset[__pyx_t_11]) = ((__pyx_v_offset[__pyx_t_11]) + __pyx_v_child->elsize);

    /* "numpy.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_7 = (!PyDataType_HASFIELDS(__pyx_v_child));
    if (__pyx_t_7) {

      /* "numpy.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_3 = PyInt_FromLong(__pyx_v_child->type_num); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 821; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF(__pyx_v_t);
      __pyx_v_t = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "numpy.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_7 = ((__pyx_v_end - __pyx_v_f) < 5);
      if (__pyx_t_7) {

        /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_3 = PyObject_Call(__pyx_builtin_RuntimeError, ((PyObject *)__pyx_k_tuple_84), NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        goto __pyx_L12;
      }
      __pyx_L12:;

      /* "numpy.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_BYTE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 826; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L13;
      }

      /* "numpy.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UBYTE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 827; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L13;
      }

      /* "numpy.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_SHORT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 828; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 104;
        goto __pyx_L13;
      }

      /* "numpy.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_USHORT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 829; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L13;
      }

      /* "numpy.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_INT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 830; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 105;
        goto __pyx_L13;
      }

      /* "numpy.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_UINT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 831; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L13;
      }

      /* "numpy.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 108;
        goto __pyx_L13;
      }

      /* "numpy.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 833; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L13;
      }

      /* "numpy.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGLONG); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 113;
        goto __pyx_L13;
      }

      /* "numpy.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_5 = PyInt_FromLong(NPY_ULONGLONG); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 835; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L13;
      }

      /* "numpy.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_FLOAT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 102;
        goto __pyx_L13;
      }

      /* "numpy.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_5 = PyInt_FromLong(NPY_DOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 100;
        goto __pyx_L13;
      }

      /* "numpy.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_3 = PyInt_FromLong(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 103;
        goto __pyx_L13;
      }

      /* "numpy.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CFLOAT); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 839; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 102;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_3 = PyInt_FromLong(NPY_CDOUBLE); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 100;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_5 = PyInt_FromLong(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 103;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L13;
      }

      /* "numpy.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_3 = PyInt_FromLong(NPY_OBJECT); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_7) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L13;
      }
      /*else*/ {

        /* "numpy.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
        __pyx_t_5 = PyNumber_Remainder(((PyObject *)__pyx_kp_u_79), __pyx_v_t); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(((PyObject *)__pyx_t_5));
        __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_t_5));
        __Pyx_GIVEREF(((PyObject *)__pyx_t_5));
        __pyx_t_5 = 0;
        __pyx_t_5 = PyObject_Call(__pyx_builtin_ValueError, ((PyObject *)__pyx_t_3), NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(((PyObject *)__pyx_t_3)); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        {__pyx_filename = __pyx_f[1]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_L13:;

      /* "numpy.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);
      goto __pyx_L11;
    }
    /*else*/ {

      /* "numpy.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
      __pyx_t_12 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_12 == NULL)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 849; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_v_f = __pyx_t_12;
    }
    __pyx_L11:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "numpy.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "numpy.pxd":965
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "numpy.pxd":967
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  if (__pyx_t_1) {

    /* "numpy.pxd":968
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":970
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
    Py_INCREF(__pyx_v_base);

    /* "numpy.pxd":971
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "numpy.pxd":972
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "numpy.pxd":973
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  __Pyx_RefNannyFinishContext();
}

/* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "numpy.pxd":976
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = (__pyx_v_arr->base == NULL);
  if (__pyx_t_1) {

    /* "numpy.pxd":977
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
    goto __pyx_L3;
  }
  /*else*/ {

    /* "numpy.pxd":979
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }
  __pyx_L3:;

  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {__Pyx_NAMESTR("mu3_tet"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_1mu3_tet, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_mu3_tet)},
  {__Pyx_NAMESTR("mu2_tet"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_3mu2_tet, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_2mu2_tet)},
  {__Pyx_NAMESTR("mu1_tet"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_5mu1_tet, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_4mu1_tet)},
  {__Pyx_NAMESTR("_mu1_tetface"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_7_mu1_tetface, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(0)},
  {__Pyx_NAMESTR("mu2_tri"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_9mu2_tri, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_8mu2_tri)},
  {__Pyx_NAMESTR("mu1_tri"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_11mu1_tri, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_10mu1_tri)},
  {__Pyx_NAMESTR("mu1_edge"), (PyCFunction)__pyx_pw_4nipy_10algorithms_10statistics_6intvol_13mu1_edge, METH_VARARGS|METH_KEYWORDS, __Pyx_DOCSTR(__pyx_doc_4nipy_10algorithms_10statistics_6intvol_12mu1_edge)},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    __Pyx_NAMESTR("intvol"),
    __Pyx_DOCSTR(__pyx_k_85), /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 0, 1, 0},
  {&__pyx_kp_s_17, __pyx_k_17, sizeof(__pyx_k_17), 0, 0, 1, 0},
  {&__pyx_kp_u_73, __pyx_k_73, sizeof(__pyx_k_73), 0, 1, 0, 0},
  {&__pyx_kp_u_75, __pyx_k_75, sizeof(__pyx_k_75), 0, 1, 0, 0},
  {&__pyx_kp_u_77, __pyx_k_77, sizeof(__pyx_k_77), 0, 1, 0, 0},
  {&__pyx_kp_u_79, __pyx_k_79, sizeof(__pyx_k_79), 0, 1, 0, 0},
  {&__pyx_n_s_8, __pyx_k_8, sizeof(__pyx_k_8), 0, 0, 1, 1},
  {&__pyx_kp_u_80, __pyx_k_80, sizeof(__pyx_k_80), 0, 1, 0, 0},
  {&__pyx_kp_u_83, __pyx_k_83, sizeof(__pyx_k_83), 0, 1, 0, 0},
  {&__pyx_n_s_86, __pyx_k_86, sizeof(__pyx_k_86), 0, 0, 1, 1},
  {&__pyx_n_s_87, __pyx_k_87, sizeof(__pyx_k_87), 0, 0, 1, 1},
  {&__pyx_kp_s_90, __pyx_k_90, sizeof(__pyx_k_90), 0, 0, 1, 0},
  {&__pyx_n_s_91, __pyx_k_91, sizeof(__pyx_k_91), 0, 0, 1, 1},
  {&__pyx_n_s__D, __pyx_k__D, sizeof(__pyx_k__D), 0, 0, 1, 1},
  {&__pyx_n_s__D00, __pyx_k__D00, sizeof(__pyx_k__D00), 0, 0, 1, 1},
  {&__pyx_n_s__D01, __pyx_k__D01, sizeof(__pyx_k__D01), 0, 0, 1, 1},
  {&__pyx_n_s__D02, __pyx_k__D02, sizeof(__pyx_k__D02), 0, 0, 1, 1},
  {&__pyx_n_s__D03, __pyx_k__D03, sizeof(__pyx_k__D03), 0, 0, 1, 1},
  {&__pyx_n_s__D11, __pyx_k__D11, sizeof(__pyx_k__D11), 0, 0, 1, 1},
  {&__pyx_n_s__D12, __pyx_k__D12, sizeof(__pyx_k__D12), 0, 0, 1, 1},
  {&__pyx_n_s__D13, __pyx_k__D13, sizeof(__pyx_k__D13), 0, 0, 1, 1},
  {&__pyx_n_s__D22, __pyx_k__D22, sizeof(__pyx_k__D22), 0, 0, 1, 1},
  {&__pyx_n_s__D23, __pyx_k__D23, sizeof(__pyx_k__D23), 0, 0, 1, 1},
  {&__pyx_n_s__D33, __pyx_k__D33, sizeof(__pyx_k__D33), 0, 0, 1, 1},
  {&__pyx_n_s__Ds0s0, __pyx_k__Ds0s0, sizeof(__pyx_k__Ds0s0), 0, 0, 1, 1},
  {&__pyx_n_s__Ds0s1, __pyx_k__Ds0s1, sizeof(__pyx_k__Ds0s1), 0, 0, 1, 1},
  {&__pyx_n_s__Ds0t0, __pyx_k__Ds0t0, sizeof(__pyx_k__Ds0t0), 0, 0, 1, 1},
  {&__pyx_n_s__Ds0t1, __pyx_k__Ds0t1, sizeof(__pyx_k__Ds0t1), 0, 0, 1, 1},
  {&__pyx_n_s__Ds1s1, __pyx_k__Ds1s1, sizeof(__pyx_k__Ds1s1), 0, 0, 1, 1},
  {&__pyx_n_s__Ds1t0, __pyx_k__Ds1t0, sizeof(__pyx_k__Ds1t0), 0, 0, 1, 1},
  {&__pyx_n_s__Ds1t1, __pyx_k__Ds1t1, sizeof(__pyx_k__Ds1t1), 0, 0, 1, 1},
  {&__pyx_n_s__Dt0t0, __pyx_k__Dt0t0, sizeof(__pyx_k__Dt0t0), 0, 0, 1, 1},
  {&__pyx_n_s__Dt0t1, __pyx_k__Dt0t1, sizeof(__pyx_k__Dt0t1), 0, 0, 1, 1},
  {&__pyx_n_s__Dt1t1, __pyx_k__Dt1t1, sizeof(__pyx_k__Dt1t1), 0, 0, 1, 1},
  {&__pyx_n_s__EC1d, __pyx_k__EC1d, sizeof(__pyx_k__EC1d), 0, 0, 1, 1},
  {&__pyx_n_s__EC2d, __pyx_k__EC2d, sizeof(__pyx_k__EC2d), 0, 0, 1, 1},
  {&__pyx_n_s__EC3d, __pyx_k__EC3d, sizeof(__pyx_k__EC3d), 0, 0, 1, 1},
  {&__pyx_n_s__Lips1d, __pyx_k__Lips1d, sizeof(__pyx_k__Lips1d), 0, 0, 1, 1},
  {&__pyx_n_s__Lips2d, __pyx_k__Lips2d, sizeof(__pyx_k__Lips2d), 0, 0, 1, 1},
  {&__pyx_n_s__Lips3d, __pyx_k__Lips3d, sizeof(__pyx_k__Lips3d), 0, 0, 1, 1},
  {&__pyx_n_s__RuntimeError, __pyx_k__RuntimeError, sizeof(__pyx_k__RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s__ValueError, __pyx_k__ValueError, sizeof(__pyx_k__ValueError), 0, 0, 1, 1},
  {&__pyx_n_s____main__, __pyx_k____main__, sizeof(__pyx_k____main__), 0, 0, 1, 1},
  {&__pyx_n_s____test__, __pyx_k____test__, sizeof(__pyx_k____test__), 0, 0, 1, 1},
  {&__pyx_n_s___convert_stride1, __pyx_k___convert_stride1, sizeof(__pyx_k___convert_stride1), 0, 0, 1, 1},
  {&__pyx_n_s___convert_stride2, __pyx_k___convert_stride2, sizeof(__pyx_k___convert_stride2), 0, 0, 1, 1},
  {&__pyx_n_s___convert_stride3, __pyx_k___convert_stride3, sizeof(__pyx_k___convert_stride3), 0, 0, 1, 1},
  {&__pyx_n_s__array, __pyx_k__array, sizeof(__pyx_k__array), 0, 0, 1, 1},
  {&__pyx_n_s__bool, __pyx_k__bool, sizeof(__pyx_k__bool), 0, 0, 1, 1},
  {&__pyx_n_s__c, __pyx_k__c, sizeof(__pyx_k__c), 0, 0, 1, 1},
  {&__pyx_n_s__coords, __pyx_k__coords, sizeof(__pyx_k__coords), 0, 0, 1, 1},
  {&__pyx_n_s__coords_c, __pyx_k__coords_c, sizeof(__pyx_k__coords_c), 0, 0, 1, 1},
  {&__pyx_n_s__cvertices, __pyx_k__cvertices, sizeof(__pyx_k__cvertices), 0, 0, 1, 1},
  {&__pyx_n_s__d2, __pyx_k__d2, sizeof(__pyx_k__d2), 0, 0, 1, 1},
  {&__pyx_n_s__d3, __pyx_k__d3, sizeof(__pyx_k__d3), 0, 0, 1, 1},
  {&__pyx_n_s__d4, __pyx_k__d4, sizeof(__pyx_k__d4), 0, 0, 1, 1},
  {&__pyx_n_s__difference, __pyx_k__difference, sizeof(__pyx_k__difference), 0, 0, 1, 1},
  {&__pyx_n_s__dok_matrix, __pyx_k__dok_matrix, sizeof(__pyx_k__dok_matrix), 0, 0, 1, 1},
  {&__pyx_n_s__ds2, __pyx_k__ds2, sizeof(__pyx_k__ds2), 0, 0, 1, 1},
  {&__pyx_n_s__ds3, __pyx_k__ds3, sizeof(__pyx_k__ds3), 0, 0, 1, 1},
  {&__pyx_n_s__ds4, __pyx_k__ds4, sizeof(__pyx_k__ds4), 0, 0, 1, 1},
  {&__pyx_n_s__dstrides, __pyx_k__dstrides, sizeof(__pyx_k__dstrides), 0, 0, 1, 1},
  {&__pyx_n_s__dtype, __pyx_k__dtype, sizeof(__pyx_k__dtype), 0, 0, 1, 1},
  {&__pyx_n_s__fcoords, __pyx_k__fcoords, sizeof(__pyx_k__fcoords), 0, 0, 1, 1},
  {&__pyx_n_s__fmask, __pyx_k__fmask, sizeof(__pyx_k__fmask), 0, 0, 1, 1},
  {&__pyx_n_s__fpmask, __pyx_k__fpmask, sizeof(__pyx_k__fpmask), 0, 0, 1, 1},
  {&__pyx_n_s__hstack, __pyx_k__hstack, sizeof(__pyx_k__hstack), 0, 0, 1, 1},
  {&__pyx_n_s__i, __pyx_k__i, sizeof(__pyx_k__i), 0, 0, 1, 1},
  {&__pyx_n_s__index, __pyx_k__index, sizeof(__pyx_k__index), 0, 0, 1, 1},
  {&__pyx_n_s__int, __pyx_k__int, sizeof(__pyx_k__int), 0, 0, 1, 1},
  {&__pyx_n_s__intp, __pyx_k__intp, sizeof(__pyx_k__intp), 0, 0, 1, 1},
  {&__pyx_n_s__issubset, __pyx_k__issubset, sizeof(__pyx_k__issubset), 0, 0, 1, 1},
  {&__pyx_n_s__j, __pyx_k__j, sizeof(__pyx_k__j), 0, 0, 1, 1},
  {&__pyx_n_s__join_complexes, __pyx_k__join_complexes, sizeof(__pyx_k__join_complexes), 0, 0, 1, 1},
  {&__pyx_n_s__k, __pyx_k__k, sizeof(__pyx_k__k), 0, 0, 1, 1},
  {&__pyx_n_s__l, __pyx_k__l, sizeof(__pyx_k__l), 0, 0, 1, 1},
  {&__pyx_n_s__l0, __pyx_k__l0, sizeof(__pyx_k__l0), 0, 0, 1, 1},
  {&__pyx_n_s__l1, __pyx_k__l1, sizeof(__pyx_k__l1), 0, 0, 1, 1},
  {&__pyx_n_s__l2, __pyx_k__l2, sizeof(__pyx_k__l2), 0, 0, 1, 1},
  {&__pyx_n_s__l3, __pyx_k__l3, sizeof(__pyx_k__l3), 0, 0, 1, 1},
  {&__pyx_n_s__m, __pyx_k__m, sizeof(__pyx_k__m), 0, 0, 1, 1},
  {&__pyx_n_s__m2, __pyx_k__m2, sizeof(__pyx_k__m2), 0, 0, 1, 1},
  {&__pyx_n_s__m3, __pyx_k__m3, sizeof(__pyx_k__m3), 0, 0, 1, 1},
  {&__pyx_n_s__m4, __pyx_k__m4, sizeof(__pyx_k__m4), 0, 0, 1, 1},
  {&__pyx_n_s__mask, __pyx_k__mask, sizeof(__pyx_k__mask), 0, 0, 1, 1},
  {&__pyx_n_s__mask_c, __pyx_k__mask_c, sizeof(__pyx_k__mask_c), 0, 0, 1, 1},
  {&__pyx_n_s__mr, __pyx_k__mr, sizeof(__pyx_k__mr), 0, 0, 1, 1},
  {&__pyx_n_s__ms, __pyx_k__ms, sizeof(__pyx_k__ms), 0, 0, 1, 1},
  {&__pyx_n_s__ndim, __pyx_k__ndim, sizeof(__pyx_k__ndim), 0, 0, 1, 1},
  {&__pyx_n_s__np, __pyx_k__np, sizeof(__pyx_k__np), 0, 0, 1, 1},
  {&__pyx_n_s__npix, __pyx_k__npix, sizeof(__pyx_k__npix), 0, 0, 1, 1},
  {&__pyx_n_s__numpy, __pyx_k__numpy, sizeof(__pyx_k__numpy), 0, 0, 1, 1},
  {&__pyx_n_s__nvox, __pyx_k__nvox, sizeof(__pyx_k__nvox), 0, 0, 1, 1},
  {&__pyx_n_s__pi, __pyx_k__pi, sizeof(__pyx_k__pi), 0, 0, 1, 1},
  {&__pyx_n_s__pindex, __pyx_k__pindex, sizeof(__pyx_k__pindex), 0, 0, 1, 1},
  {&__pyx_n_s__pmask, __pyx_k__pmask, sizeof(__pyx_k__pmask), 0, 0, 1, 1},
  {&__pyx_n_s__pmask_shape, __pyx_k__pmask_shape, sizeof(__pyx_k__pmask_shape), 0, 0, 1, 1},
  {&__pyx_n_s__r, __pyx_k__r, sizeof(__pyx_k__r), 0, 0, 1, 1},
  {&__pyx_n_s__range, __pyx_k__range, sizeof(__pyx_k__range), 0, 0, 1, 1},
  {&__pyx_n_s__res, __pyx_k__res, sizeof(__pyx_k__res), 0, 0, 1, 1},
  {&__pyx_n_s__reshape, __pyx_k__reshape, sizeof(__pyx_k__reshape), 0, 0, 1, 1},
  {&__pyx_n_s__rr, __pyx_k__rr, sizeof(__pyx_k__rr), 0, 0, 1, 1},
  {&__pyx_n_s__s, __pyx_k__s, sizeof(__pyx_k__s), 0, 0, 1, 1},
  {&__pyx_n_s__s0, __pyx_k__s0, sizeof(__pyx_k__s0), 0, 0, 1, 1},
  {&__pyx_n_s__s1, __pyx_k__s1, sizeof(__pyx_k__s1), 0, 0, 1, 1},
  {&__pyx_n_s__s2, __pyx_k__s2, sizeof(__pyx_k__s2), 0, 0, 1, 1},
  {&__pyx_n_s__shape, __pyx_k__shape, sizeof(__pyx_k__shape), 0, 0, 1, 1},
  {&__pyx_n_s__size, __pyx_k__size, sizeof(__pyx_k__size), 0, 0, 1, 1},
  {&__pyx_n_s__sorted, __pyx_k__sorted, sizeof(__pyx_k__sorted), 0, 0, 1, 1},
  {&__pyx_n_s__squeeze, __pyx_k__squeeze, sizeof(__pyx_k__squeeze), 0, 0, 1, 1},
  {&__pyx_n_s__ss, __pyx_k__ss, sizeof(__pyx_k__ss), 0, 0, 1, 1},
  {&__pyx_n_s__ss0, __pyx_k__ss0, sizeof(__pyx_k__ss0), 0, 0, 1, 1},
  {&__pyx_n_s__ss0d, __pyx_k__ss0d, sizeof(__pyx_k__ss0d), 0, 0, 1, 1},
  {&__pyx_n_s__ss1, __pyx_k__ss1, sizeof(__pyx_k__ss1), 0, 0, 1, 1},
  {&__pyx_n_s__ss1d, __pyx_k__ss1d, sizeof(__pyx_k__ss1d), 0, 0, 1, 1},
  {&__pyx_n_s__ss2, __pyx_k__ss2, sizeof(__pyx_k__ss2), 0, 0, 1, 1},
  {&__pyx_n_s__ss2d, __pyx_k__ss2d, sizeof(__pyx_k__ss2d), 0, 0, 1, 1},
  {&__pyx_n_s__stride1, __pyx_k__stride1, sizeof(__pyx_k__stride1), 0, 0, 1, 1},
  {&__pyx_n_s__stride2, __pyx_k__stride2, sizeof(__pyx_k__stride2), 0, 0, 1, 1},
  {&__pyx_n_s__strides, __pyx_k__strides, sizeof(__pyx_k__strides), 0, 0, 1, 1},
  {&__pyx_n_s__strides_from, __pyx_k__strides_from, sizeof(__pyx_k__strides_from), 0, 0, 1, 1},
  {&__pyx_n_s__sum, __pyx_k__sum, sizeof(__pyx_k__sum), 0, 0, 1, 1},
  {&__pyx_n_s__union, __pyx_k__union, sizeof(__pyx_k__union), 0, 0, 1, 1},
  {&__pyx_n_s__unique, __pyx_k__unique, sizeof(__pyx_k__unique), 0, 0, 1, 1},
  {&__pyx_n_s__utils, __pyx_k__utils, sizeof(__pyx_k__utils), 0, 0, 1, 1},
  {&__pyx_n_s__v, __pyx_k__v, sizeof(__pyx_k__v), 0, 0, 1, 1},
  {&__pyx_n_s__v0, __pyx_k__v0, sizeof(__pyx_k__v0), 0, 0, 1, 1},
  {&__pyx_n_s__v1, __pyx_k__v1, sizeof(__pyx_k__v1), 0, 0, 1, 1},
  {&__pyx_n_s__v2, __pyx_k__v2, sizeof(__pyx_k__v2), 0, 0, 1, 1},
  {&__pyx_n_s__v3, __pyx_k__v3, sizeof(__pyx_k__v3), 0, 0, 1, 1},
  {&__pyx_n_s__value, __pyx_k__value, sizeof(__pyx_k__value), 0, 0, 1, 1},
  {&__pyx_n_s__verts, __pyx_k__verts, sizeof(__pyx_k__verts), 0, 0, 1, 1},
  {&__pyx_n_s__w0, __pyx_k__w0, sizeof(__pyx_k__w0), 0, 0, 1, 1},
  {&__pyx_n_s__w1, __pyx_k__w1, sizeof(__pyx_k__w1), 0, 0, 1, 1},
  {&__pyx_n_s__w2, __pyx_k__w2, sizeof(__pyx_k__w2), 0, 0, 1, 1},
  {&__pyx_n_s__w3, __pyx_k__w3, sizeof(__pyx_k__w3), 0, 0, 1, 1},
  {&__pyx_n_s__zeros, __pyx_k__zeros, sizeof(__pyx_k__zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ValueError = __Pyx_GetName(__pyx_b, __pyx_n_s__ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_range = __Pyx_GetName(__pyx_b, __pyx_n_s__range); if (!__pyx_builtin_range) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 430; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_sorted = __Pyx_GetName(__pyx_b, __pyx_n_s__sorted); if (!__pyx_builtin_sorted) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 581; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetName(__pyx_b, __pyx_n_s__RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "nipy/algorithms/statistics/intvol.pyx":373
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
  __pyx_k_tuple_2 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_2);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_1));
  PyTuple_SET_ITEM(__pyx_k_tuple_2, 0, ((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_2));

  /* "nipy/algorithms/statistics/intvol.pyx":395
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 */
  __pyx_k_slice_3 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 395; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_3);
  __Pyx_GIVEREF(__pyx_k_slice_3);
  __pyx_k_slice_4 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 395; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_4);
  __Pyx_GIVEREF(__pyx_k_slice_4);
  __pyx_k_slice_5 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 395; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_5);
  __Pyx_GIVEREF(__pyx_k_slice_5);
  __pyx_k_tuple_6 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 395; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_6);
  __Pyx_INCREF(__pyx_k_slice_3);
  PyTuple_SET_ITEM(__pyx_k_tuple_6, 0, __pyx_k_slice_3);
  __Pyx_GIVEREF(__pyx_k_slice_3);
  __Pyx_INCREF(__pyx_k_slice_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_6, 1, __pyx_k_slice_4);
  __Pyx_GIVEREF(__pyx_k_slice_4);
  __Pyx_INCREF(__pyx_k_slice_5);
  PyTuple_SET_ITEM(__pyx_k_tuple_6, 2, __pyx_k_slice_5);
  __Pyx_GIVEREF(__pyx_k_slice_5);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_6));

  /* "nipy/algorithms/statistics/intvol.pyx":399
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 *     cdef:
 *         np.ndarray[np.intp_t, ndim=1] strides
 */
  __pyx_k_tuple_7 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_7);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_7, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_7));

  /* "nipy/algorithms/statistics/intvol.pyx":407
 *     # We first figure out which vertices, edges, triangles, tetrahedra
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 */
  __pyx_k_tuple_9 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 407; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_9);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_9, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_9, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_9, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_9));

  /* "nipy/algorithms/statistics/intvol.pyx":408
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 */
  __pyx_k_tuple_10 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 408; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_10);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_10, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_10, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_10, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_10));

  /* "nipy/algorithms/statistics/intvol.pyx":409
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 */
  __pyx_k_tuple_11 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_11);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_11, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_11, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_11, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_11));

  /* "nipy/algorithms/statistics/intvol.pyx":410
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 */
  __pyx_k_tuple_12 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 410; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_12);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_12, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_12, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_12, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_12));

  /* "nipy/algorithms/statistics/intvol.pyx":411
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 */
  __pyx_k_tuple_13 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_13);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_13, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_13, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_13, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_13));

  /* "nipy/algorithms/statistics/intvol.pyx":412
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)
 */
  __pyx_k_tuple_14 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_14);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_14, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_14, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_14, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_14));

  /* "nipy/algorithms/statistics/intvol.pyx":413
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])             # <<<<<<<<<<<<<<
 *     c = cube_with_strides_center((0,0,0), strides)
 * 
 */
  __pyx_k_tuple_15 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_15);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_15, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_15, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_15, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_15));

  /* "nipy/algorithms/statistics/intvol.pyx":414
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)             # <<<<<<<<<<<<<<
 * 
 *     d4 = np.array(list(c[4].difference(union[4])))
 */
  __pyx_k_tuple_16 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 414; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_16);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_16, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_16, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_16, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_16));

  /* "nipy/algorithms/statistics/intvol.pyx":505
 *     """
 *     if mask.shape != coords.shape[1:]:
 *         raise ValueError('shape of mask does not match coordinates')             # <<<<<<<<<<<<<<
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)
 */
  __pyx_k_tuple_18 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 505; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_18);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_17));
  PyTuple_SET_ITEM(__pyx_k_tuple_18, 0, ((PyObject *)__pyx_kp_s_17));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_17));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_18));

  /* "nipy/algorithms/statistics/intvol.pyx":509
 *     mask = np.squeeze(mask)
 *     if mask.ndim < 3:
 *         value = np.zeros(4)             # <<<<<<<<<<<<<<
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         if mask.ndim == 2:
 */
  __pyx_k_tuple_19 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 509; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_19);
  __Pyx_INCREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_19, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_19));

  /* "nipy/algorithms/statistics/intvol.pyx":518
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
  __pyx_k_tuple_20 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 518; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_20);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_1));
  PyTuple_SET_ITEM(__pyx_k_tuple_20, 0, ((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_20));

  /* "nipy/algorithms/statistics/intvol.pyx":555
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 */
  __pyx_k_slice_21 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 555; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_21);
  __Pyx_GIVEREF(__pyx_k_slice_21);
  __pyx_k_slice_22 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 555; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_22);
  __Pyx_GIVEREF(__pyx_k_slice_22);
  __pyx_k_slice_23 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 555; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_23);
  __Pyx_GIVEREF(__pyx_k_slice_23);
  __pyx_k_tuple_24 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 555; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_24);
  __Pyx_INCREF(__pyx_k_slice_21);
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 0, __pyx_k_slice_21);
  __Pyx_GIVEREF(__pyx_k_slice_21);
  __Pyx_INCREF(__pyx_k_slice_22);
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 1, __pyx_k_slice_22);
  __Pyx_GIVEREF(__pyx_k_slice_22);
  __Pyx_INCREF(__pyx_k_slice_23);
  PyTuple_SET_ITEM(__pyx_k_tuple_24, 2, __pyx_k_slice_23);
  __Pyx_GIVEREF(__pyx_k_slice_23);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_24));

  /* "nipy/algorithms/statistics/intvol.pyx":559
 *     s0, s1, s2 = (pmask.shape[0], pmask.shape[1], pmask.shape[2])
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 *     fmask = mask_c.reshape(-1)
 *     fcoords = coords_c.reshape((coords_c.shape[0], -1))
 */
  __pyx_k_tuple_25 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 559; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_25);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_25, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_25));

  /* "nipy/algorithms/statistics/intvol.pyx":560
 * 
 *     fpmask = pmask.reshape(-1)
 *     fmask = mask_c.reshape(-1)             # <<<<<<<<<<<<<<
 *     fcoords = coords_c.reshape((coords_c.shape[0], -1))
 * 
 */
  __pyx_k_tuple_26 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 560; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_26);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_26, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_26));

  /* "nipy/algorithms/statistics/intvol.pyx":583
 *     cvertices = np.array(sorted(verts), np.intp)
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 */
  __pyx_k_tuple_27 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 583; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_27);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_27, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_27, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_27, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_27));

  /* "nipy/algorithms/statistics/intvol.pyx":584
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 */
  __pyx_k_tuple_28 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_28);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_28, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_28));

  /* "nipy/algorithms/statistics/intvol.pyx":585
 *     union = join_complexes(*[cube_with_strides_center((0,0,1), strides),
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 */
  __pyx_k_tuple_29 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 585; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_29);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_29, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_29, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_29, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_29));

  /* "nipy/algorithms/statistics/intvol.pyx":586
 *                              cube_with_strides_center((0,1,0), strides),
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 */
  __pyx_k_tuple_30 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 586; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_30);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_30, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_30));

  /* "nipy/algorithms/statistics/intvol.pyx":587
 *                              cube_with_strides_center((0,1,1), strides),
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 */
  __pyx_k_tuple_31 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 587; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_31);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_31, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_31, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_31, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_31));

  /* "nipy/algorithms/statistics/intvol.pyx":588
 *                              cube_with_strides_center((1,0,0), strides),
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)
 */
  __pyx_k_tuple_32 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 588; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_32);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_32, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_32));

  /* "nipy/algorithms/statistics/intvol.pyx":589
 *                              cube_with_strides_center((1,0,1), strides),
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])             # <<<<<<<<<<<<<<
 *     c = cube_with_strides_center((0,0,0), strides)
 *     m4 = np.array(list(c[4].difference(union[4])))
 */
  __pyx_k_tuple_33 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 589; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_33);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_33, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_33, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_33, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_33));

  /* "nipy/algorithms/statistics/intvol.pyx":590
 *                              cube_with_strides_center((1,1,0), strides),
 *                              cube_with_strides_center((1,1,1), strides)])
 *     c = cube_with_strides_center((0,0,0), strides)             # <<<<<<<<<<<<<<
 *     m4 = np.array(list(c[4].difference(union[4])))
 *     m3 = np.array(list(c[3].difference(union[3])))
 */
  __pyx_k_tuple_34 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 590; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_34);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_34, 2, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_34));

  /* "nipy/algorithms/statistics/intvol.pyx":595
 *     m2 = np.array(list(c[2].difference(union[2])))
 * 
 *     d4 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m4[i]] for i in range(m4.shape[0])])             # <<<<<<<<<<<<<<
 *     d4 = np.hstack([m4, d4])
 *     ds4 = d4.shape[0]
 */
  __pyx_k_tuple_35 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_35)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 595; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_35);
  __Pyx_INCREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_35, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_35, 1, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_35, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_35));

  /* "nipy/algorithms/statistics/intvol.pyx":599
 *     ds4 = d4.shape[0]
 * 
 *     d3 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m3[i]] for i in range(m3.shape[0])])             # <<<<<<<<<<<<<<
 *     d3 = np.hstack([m3, d3])
 *     ds3 = d3.shape[0]
 */
  __pyx_k_tuple_36 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_36)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 599; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_36);
  __Pyx_INCREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_36, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_36, 1, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_36, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_36));

  /* "nipy/algorithms/statistics/intvol.pyx":603
 *     ds3 = d3.shape[0]
 * 
 *     d2 = np.array([[_convert_stride3(v, strides, (4,2,1)) for v in m2[i]] for i in range(m2.shape[0])])             # <<<<<<<<<<<<<<
 *     d2 = np.hstack([m2, d2])
 *     ds2 = d2.shape[0]
 */
  __pyx_k_tuple_37 = PyTuple_New(3); if (unlikely(!__pyx_k_tuple_37)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 603; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_37);
  __Pyx_INCREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 1, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_37, 2, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_37));

  /* "nipy/algorithms/statistics/intvol.pyx":609
 *     nvox = mask.size
 * 
 *     D = np.zeros((8,8))             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0-1):
 */
  __pyx_k_tuple_38 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_38)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_38);
  __Pyx_INCREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_k_tuple_38, 0, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  __Pyx_INCREF(__pyx_int_8);
  PyTuple_SET_ITEM(__pyx_k_tuple_38, 1, __pyx_int_8);
  __Pyx_GIVEREF(__pyx_int_8);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_38));
  __pyx_k_tuple_39 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_39)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 609; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_39);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_38));
  PyTuple_SET_ITEM(__pyx_k_tuple_39, 0, ((PyObject *)__pyx_k_tuple_38));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_38));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_39));

  /* "nipy/algorithms/statistics/intvol.pyx":766
 *     """
 *     if mask.shape != coords.shape[1:]:
 *         raise ValueError('shape of mask does not match coordinates')             # <<<<<<<<<<<<<<
 *     # if the data can be squeezed, we must use the lower dimensional function
 *     mask = np.squeeze(mask)
 */
  __pyx_k_tuple_40 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_40)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 766; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_40);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_17));
  PyTuple_SET_ITEM(__pyx_k_tuple_40, 0, ((PyObject *)__pyx_kp_s_17));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_17));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_40));

  /* "nipy/algorithms/statistics/intvol.pyx":770
 *     mask = np.squeeze(mask)
 *     if mask.ndim == 1:
 *         value = np.zeros(3)             # <<<<<<<<<<<<<<
 *         coords = coords.reshape((coords.shape[0],) + mask.shape)
 *         value[:2] = Lips1d(coords, mask)
 */
  __pyx_k_tuple_41 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_41)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 770; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_41);
  __Pyx_INCREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_k_tuple_41, 0, __pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_41));

  /* "nipy/algorithms/statistics/intvol.pyx":776
 * 
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 * 
 */
  __pyx_k_tuple_42 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_42)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 776; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_42);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_1));
  PyTuple_SET_ITEM(__pyx_k_tuple_42, 0, ((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_42));

  /* "nipy/algorithms/statistics/intvol.pyx":809
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1 = pmask.shape[0], pmask.shape[1]
 */
  __pyx_k_slice_43 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_43)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_43);
  __Pyx_GIVEREF(__pyx_k_slice_43);
  __pyx_k_slice_44 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_44)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_44);
  __Pyx_GIVEREF(__pyx_k_slice_44);
  __pyx_k_tuple_45 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_45)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_45);
  __Pyx_INCREF(__pyx_k_slice_43);
  PyTuple_SET_ITEM(__pyx_k_tuple_45, 0, __pyx_k_slice_43);
  __Pyx_GIVEREF(__pyx_k_slice_43);
  __Pyx_INCREF(__pyx_k_slice_44);
  PyTuple_SET_ITEM(__pyx_k_tuple_45, 1, __pyx_k_slice_44);
  __Pyx_GIVEREF(__pyx_k_slice_44);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_45));

  /* "nipy/algorithms/statistics/intvol.pyx":813
 *     s0, s1 = pmask.shape[0], pmask.shape[1]
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 *     fmask = mask_c.reshape(-1)
 *     fcoords = coords.reshape((coords.shape[0], -1))
 */
  __pyx_k_tuple_46 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_46)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 813; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_46);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_46, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_46));

  /* "nipy/algorithms/statistics/intvol.pyx":814
 * 
 *     fpmask = pmask.reshape(-1)
 *     fmask = mask_c.reshape(-1)             # <<<<<<<<<<<<<<
 *     fcoords = coords.reshape((coords.shape[0], -1))
 * 
 */
  __pyx_k_tuple_47 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_47)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 814; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_47);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_47, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_47));

  /* "nipy/algorithms/statistics/intvol.pyx":836
 *     cvertices = np.array(sorted(verts), np.intp)
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])
 */
  __pyx_k_tuple_48 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_48)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 836; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_48);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_48, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_48, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_48));

  /* "nipy/algorithms/statistics/intvol.pyx":837
 * 
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1), strides)])
 * 
 */
  __pyx_k_tuple_49 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_49)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 837; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_49);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_49, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_49, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_49));

  /* "nipy/algorithms/statistics/intvol.pyx":838
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])             # <<<<<<<<<<<<<<
 * 
 *     c = cube_with_strides_center((0,0), strides)
 */
  __pyx_k_tuple_50 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_50)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_50);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_50, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_50, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_50));

  /* "nipy/algorithms/statistics/intvol.pyx":840
 *                              cube_with_strides_center((1,1), strides)])
 * 
 *     c = cube_with_strides_center((0,0), strides)             # <<<<<<<<<<<<<<
 *     m3 = np.array(list(c[3].difference(union[3])))
 *     m2 = np.array(list(c[2].difference(union[2])))
 */
  __pyx_k_tuple_51 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_51)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_51);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_51, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_51, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_51));

  /* "nipy/algorithms/statistics/intvol.pyx":844
 *     m2 = np.array(list(c[2].difference(union[2])))
 * 
 *     d3 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m3[i]] for i in range(m3.shape[0])])             # <<<<<<<<<<<<<<
 *     d3 = np.hstack([m3, d3])
 *     ds3 = d3.shape[0]
 */
  __pyx_k_tuple_52 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_52)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 844; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_52);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_52, 0, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_52, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_52));

  /* "nipy/algorithms/statistics/intvol.pyx":848
 *     ds3 = d3.shape[0]
 * 
 *     d2 = np.array([[_convert_stride2(v, strides, (2,1)) for v in m2[i]] for i in range(m2.shape[0])])             # <<<<<<<<<<<<<<
 *     d2 = np.hstack([m2, d2])
 *     ds2 = d2.shape[0]
 */
  __pyx_k_tuple_53 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_53)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 848; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_53);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_53, 0, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_53, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_53));

  /* "nipy/algorithms/statistics/intvol.pyx":852
 *     ds2 = d2.shape[0]
 * 
 *     D = np.zeros((4,4))             # <<<<<<<<<<<<<<
 * 
 *     npix = mask.size
 */
  __pyx_k_tuple_54 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_54)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_54);
  __Pyx_INCREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_54, 0, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  __Pyx_INCREF(__pyx_int_4);
  PyTuple_SET_ITEM(__pyx_k_tuple_54, 1, __pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_54));
  __pyx_k_tuple_55 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_55)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 852; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_55);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_54));
  PyTuple_SET_ITEM(__pyx_k_tuple_55, 0, ((PyObject *)__pyx_k_tuple_54));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_54));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_55));

  /* "nipy/algorithms/statistics/intvol.pyx":937
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
  __pyx_k_tuple_56 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_56)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 937; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_56);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_1));
  PyTuple_SET_ITEM(__pyx_k_tuple_56, 0, ((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_56));

  /* "nipy/algorithms/statistics/intvol.pyx":958
 *     pmask_shape = np.array(mask.shape) + 1
 *     pmask = np.zeros(pmask_shape, np.int)
 *     pmask[:-1,:-1] = mask_c             # <<<<<<<<<<<<<<
 * 
 *     s0, s1 = (pmask.shape[0], pmask.shape[1])
 */
  __pyx_k_slice_57 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_57)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 958; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_57);
  __Pyx_GIVEREF(__pyx_k_slice_57);
  __pyx_k_slice_58 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_k_slice_58)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 958; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_slice_58);
  __Pyx_GIVEREF(__pyx_k_slice_58);
  __pyx_k_tuple_59 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_59)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 958; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_59);
  __Pyx_INCREF(__pyx_k_slice_57);
  PyTuple_SET_ITEM(__pyx_k_tuple_59, 0, __pyx_k_slice_57);
  __Pyx_GIVEREF(__pyx_k_slice_57);
  __Pyx_INCREF(__pyx_k_slice_58);
  PyTuple_SET_ITEM(__pyx_k_tuple_59, 1, __pyx_k_slice_58);
  __Pyx_GIVEREF(__pyx_k_slice_58);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_59));

  /* "nipy/algorithms/statistics/intvol.pyx":962
 *     s0, s1 = (pmask.shape[0], pmask.shape[1])
 * 
 *     fpmask = pmask.reshape(-1)             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
  __pyx_k_tuple_60 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_60)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 962; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_60);
  __Pyx_INCREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_60, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_60));

  /* "nipy/algorithms/statistics/intvol.pyx":972
 *     # We first figure out which vertices, edges, triangles, tetrahedra
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])
 */
  __pyx_k_tuple_61 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_61)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 972; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_61);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_61, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_61, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_61));

  /* "nipy/algorithms/statistics/intvol.pyx":973
 *     # are uniquely associated with an interior voxel
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),             # <<<<<<<<<<<<<<
 *                              cube_with_strides_center((1,1), strides)])
 *     c = cube_with_strides_center((0,0), strides)
 */
  __pyx_k_tuple_62 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_62)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 973; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_62);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_62, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_62, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_62));

  /* "nipy/algorithms/statistics/intvol.pyx":974
 *     union = join_complexes(*[cube_with_strides_center((0,1), strides),
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])             # <<<<<<<<<<<<<<
 *     c = cube_with_strides_center((0,0), strides)
 * 
 */
  __pyx_k_tuple_63 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_63)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 974; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_63);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_63, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_63, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_63));

  /* "nipy/algorithms/statistics/intvol.pyx":975
 *                              cube_with_strides_center((1,0), strides),
 *                              cube_with_strides_center((1,1), strides)])
 *     c = cube_with_strides_center((0,0), strides)             # <<<<<<<<<<<<<<
 * 
 *     d3 = np.array(list(c[3].difference(union[3])))
 */
  __pyx_k_tuple_64 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_64)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 975; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_64);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_64, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_64, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_64));

  /* "nipy/algorithms/statistics/intvol.pyx":1043
 *     """
 *     if mask.shape[0] != coords.shape[1]:
 *         raise ValueError('shape of mask does not match coordinates')             # <<<<<<<<<<<<<<
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '
 */
  __pyx_k_tuple_65 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_65)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1043; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_65);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_17));
  PyTuple_SET_ITEM(__pyx_k_tuple_65, 0, ((PyObject *)__pyx_kp_s_17));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_17));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_65));

  /* "nipy/algorithms/statistics/intvol.pyx":1045
 *         raise ValueError('shape of mask does not match coordinates')
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
  __pyx_k_tuple_66 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_66)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1045; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_66);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_1));
  PyTuple_SET_ITEM(__pyx_k_tuple_66, 0, ((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_66));

  /* "nipy/algorithms/statistics/intvol.pyx":1054
 *     l0 = 0; l1 = 0
 *     s0 = mask.shape[0]
 *     D = np.zeros((2,2))             # <<<<<<<<<<<<<<
 * 
 *     for i in range(s0):
 */
  __pyx_k_tuple_67 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_67)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_67);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_67, 0, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_INCREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_k_tuple_67, 1, __pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_67));
  __pyx_k_tuple_68 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_68)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1054; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_68);
  __Pyx_INCREF(((PyObject *)__pyx_k_tuple_67));
  PyTuple_SET_ITEM(__pyx_k_tuple_68, 0, ((PyObject *)__pyx_k_tuple_67));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_67));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_68));

  /* "nipy/algorithms/statistics/intvol.pyx":1076
 *         if m:
 *             m = m * (mask[(i+1) % s0] * ((i+1) < s0))
 *             l1 = l1 + m * mu1_edge(D[0,0], D[0,1], D[1,1])             # <<<<<<<<<<<<<<
 *             l0 = l0 - m
 * 
 */
  __pyx_k_tuple_69 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_69)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_69);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_69, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_69, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_69));
  __pyx_k_tuple_70 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_70)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_70);
  __Pyx_INCREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_k_tuple_70, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_70, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_70));
  __pyx_k_tuple_71 = PyTuple_New(2); if (unlikely(!__pyx_k_tuple_71)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1076; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_71);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_71, 0, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_INCREF(__pyx_int_1);
  PyTuple_SET_ITEM(__pyx_k_tuple_71, 1, __pyx_int_1);
  __Pyx_GIVEREF(__pyx_int_1);
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_71));

  /* "nipy/algorithms/statistics/intvol.pyx":1116
 *     """
 *     if not set(np.unique(mask)).issubset([0,1]):
 *       raise ValueError('mask should be filled with 0/1 '             # <<<<<<<<<<<<<<
 *                        'values, but be of type np.int')
 *     cdef:
 */
  __pyx_k_tuple_72 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_72)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1116; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_72);
  __Pyx_INCREF(((PyObject *)__pyx_kp_s_1));
  PyTuple_SET_ITEM(__pyx_k_tuple_72, 0, ((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_s_1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_72));

  /* "numpy.pxd":215
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_k_tuple_74 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_74)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 215; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_74);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_73));
  PyTuple_SET_ITEM(__pyx_k_tuple_74, 0, ((PyObject *)__pyx_kp_u_73));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_73));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_74));

  /* "numpy.pxd":219
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_k_tuple_76 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_76)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 219; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_76);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_75));
  PyTuple_SET_ITEM(__pyx_k_tuple_76, 0, ((PyObject *)__pyx_kp_u_75));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_75));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_76));

  /* "numpy.pxd":257
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_k_tuple_78 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_78)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_78);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_77));
  PyTuple_SET_ITEM(__pyx_k_tuple_78, 0, ((PyObject *)__pyx_kp_u_77));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_77));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_78));

  /* "numpy.pxd":799
 * 
 *         if (end - f) - (new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_k_tuple_81 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_81)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 799; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_81);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_80));
  PyTuple_SET_ITEM(__pyx_k_tuple_81, 0, ((PyObject *)__pyx_kp_u_80));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_80));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_81));

  /* "numpy.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_k_tuple_82 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_82)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 803; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_82);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_77));
  PyTuple_SET_ITEM(__pyx_k_tuple_82, 0, ((PyObject *)__pyx_kp_u_77));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_77));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_82));

  /* "numpy.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_k_tuple_84 = PyTuple_New(1); if (unlikely(!__pyx_k_tuple_84)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 823; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_84);
  __Pyx_INCREF(((PyObject *)__pyx_kp_u_83));
  PyTuple_SET_ITEM(__pyx_k_tuple_84, 0, ((PyObject *)__pyx_kp_u_83));
  __Pyx_GIVEREF(((PyObject *)__pyx_kp_u_83));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_84));

  /* "nipy/algorithms/statistics/intvol.pyx":340
 * 
 * 
 * def EC3d(mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic of region within `mask`
 * 
 */
  __pyx_k_tuple_88 = PyTuple_New(32); if (unlikely(!__pyx_k_tuple_88)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 340; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_88);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 0, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask_c));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 1, ((PyObject *)__pyx_n_s__mask_c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask_c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fpmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 2, ((PyObject *)__pyx_n_s__fpmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fpmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d2));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 3, ((PyObject *)__pyx_n_s__d2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d3));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 4, ((PyObject *)__pyx_n_s__d3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d4));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 5, ((PyObject *)__pyx_n_s__d4));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d4));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 6, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 7, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 8, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 9, ((PyObject *)__pyx_n_s__l));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s0));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 10, ((PyObject *)__pyx_n_s__s0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s1));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 11, ((PyObject *)__pyx_n_s__s1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s2));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 12, ((PyObject *)__pyx_n_s__s2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds2));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 13, ((PyObject *)__pyx_n_s__ds2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds3));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 14, ((PyObject *)__pyx_n_s__ds3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds4));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 15, ((PyObject *)__pyx_n_s__ds4));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds4));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__index));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 16, ((PyObject *)__pyx_n_s__index));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__index));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 17, ((PyObject *)__pyx_n_s__m));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nvox));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 18, ((PyObject *)__pyx_n_s__nvox));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nvox));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss0));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 19, ((PyObject *)__pyx_n_s__ss0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss1));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 20, ((PyObject *)__pyx_n_s__ss1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss2));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 21, ((PyObject *)__pyx_n_s__ss2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 22, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v1));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 23, ((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v2));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 24, ((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v3));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 25, ((PyObject *)__pyx_n_s__v3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l0));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 26, ((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask_shape));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 27, ((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 28, ((PyObject *)__pyx_n_s__pmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__strides));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 29, ((PyObject *)__pyx_n_s__strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__strides));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__union));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 30, ((PyObject *)__pyx_n_s__union));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__union));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__c));
  PyTuple_SET_ITEM(__pyx_k_tuple_88, 31, ((PyObject *)__pyx_n_s__c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__c));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_88));
  __pyx_k_codeobj_89 = (PyObject*)__Pyx_PyCode_New(1, 0, 32, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_88, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s__EC3d, 340, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_89)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 340; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":464
 * 
 * 
 * def Lips3d(coords, mask):             # <<<<<<<<<<<<<<
 *     """ Estimated intrinsic volumes within masked region given coordinates
 * 
 */
  __pyx_k_tuple_92 = PyTuple_New(63); if (unlikely(!__pyx_k_tuple_92)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_92);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__coords));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 0, ((PyObject *)__pyx_n_s__coords));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__coords));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 1, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__value));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 2, ((PyObject *)__pyx_n_s__value));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__value));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__coords_c));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 3, ((PyObject *)__pyx_n_s__coords_c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__coords_c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask_c));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 4, ((PyObject *)__pyx_n_s__mask_c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask_c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fcoords));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 5, ((PyObject *)__pyx_n_s__fcoords));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fcoords));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__D));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 6, ((PyObject *)__pyx_n_s__D));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__D));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 7, ((PyObject *)__pyx_n_s__fmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fpmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 8, ((PyObject *)__pyx_n_s__fpmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fpmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 9, ((PyObject *)__pyx_n_s__pmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d4));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 10, ((PyObject *)__pyx_n_s__d4));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d4));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m4));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 11, ((PyObject *)__pyx_n_s__m4));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m4));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d3));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 12, ((PyObject *)__pyx_n_s__d3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m3));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 13, ((PyObject *)__pyx_n_s__m3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 14, ((PyObject *)__pyx_n_s__d2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 15, ((PyObject *)__pyx_n_s__m2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__cvertices));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 16, ((PyObject *)__pyx_n_s__cvertices));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__cvertices));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 17, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 18, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 19, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 20, ((PyObject *)__pyx_n_s__l));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s0));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 21, ((PyObject *)__pyx_n_s__s0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s1));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 22, ((PyObject *)__pyx_n_s__s1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 23, ((PyObject *)__pyx_n_s__s2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds4));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 24, ((PyObject *)__pyx_n_s__ds4));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds4));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds3));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 25, ((PyObject *)__pyx_n_s__ds3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 26, ((PyObject *)__pyx_n_s__ds2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__index));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 27, ((PyObject *)__pyx_n_s__index));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__index));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pindex));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 28, ((PyObject *)__pyx_n_s__pindex));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pindex));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 29, ((PyObject *)__pyx_n_s__m));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__nvox));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 30, ((PyObject *)__pyx_n_s__nvox));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__nvox));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__r));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 31, ((PyObject *)__pyx_n_s__r));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__r));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 32, ((PyObject *)__pyx_n_s__s));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__rr));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 33, ((PyObject *)__pyx_n_s__rr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__rr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 34, ((PyObject *)__pyx_n_s__ss));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mr));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 35, ((PyObject *)__pyx_n_s__mr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ms));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 36, ((PyObject *)__pyx_n_s__ms));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ms));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss0));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 37, ((PyObject *)__pyx_n_s__ss0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss1));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 38, ((PyObject *)__pyx_n_s__ss1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 39, ((PyObject *)__pyx_n_s__ss2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 40, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v1));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 41, ((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 42, ((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v3));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 43, ((PyObject *)__pyx_n_s__v3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w0));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 44, ((PyObject *)__pyx_n_s__w0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w1));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 45, ((PyObject *)__pyx_n_s__w1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 46, ((PyObject *)__pyx_n_s__w2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w3));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 47, ((PyObject *)__pyx_n_s__w3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l0));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 48, ((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l1));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 49, ((PyObject *)__pyx_n_s__l1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l2));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 50, ((PyObject *)__pyx_n_s__l2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l3));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 51, ((PyObject *)__pyx_n_s__l3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__res));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 52, ((PyObject *)__pyx_n_s__res));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__res));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask_shape));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 53, ((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__strides));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 54, ((PyObject *)__pyx_n_s__strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__strides));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dstrides));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 55, ((PyObject *)__pyx_n_s__dstrides));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dstrides));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss0d));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 56, ((PyObject *)__pyx_n_s__ss0d));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss0d));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss1d));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 57, ((PyObject *)__pyx_n_s__ss1d));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss1d));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss2d));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 58, ((PyObject *)__pyx_n_s__ss2d));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss2d));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__verts));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 59, ((PyObject *)__pyx_n_s__verts));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__verts));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__union));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 60, ((PyObject *)__pyx_n_s__union));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__union));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__c));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 61, ((PyObject *)__pyx_n_s__c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v));
  PyTuple_SET_ITEM(__pyx_k_tuple_92, 62, ((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_92));
  __pyx_k_codeobj_93 = (PyObject*)__Pyx_PyCode_New(2, 0, 63, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_92, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s__Lips3d, 464, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_93)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":699
 * 
 * 
 * def _convert_stride3(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */
  __pyx_k_tuple_94 = PyTuple_New(6); if (unlikely(!__pyx_k_tuple_94)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_94);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v));
  PyTuple_SET_ITEM(__pyx_k_tuple_94, 0, ((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__stride1));
  PyTuple_SET_ITEM(__pyx_k_tuple_94, 1, ((PyObject *)__pyx_n_s__stride1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__stride1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__stride2));
  PyTuple_SET_ITEM(__pyx_k_tuple_94, 2, ((PyObject *)__pyx_n_s__stride2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__stride2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_94, 3, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v1));
  PyTuple_SET_ITEM(__pyx_k_tuple_94, 4, ((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v2));
  PyTuple_SET_ITEM(__pyx_k_tuple_94, 5, ((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_94));
  __pyx_k_codeobj_95 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s___convert_stride3, 699, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_95)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":711
 * 
 * 
 * def _convert_stride2(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */
  __pyx_k_tuple_96 = PyTuple_New(5); if (unlikely(!__pyx_k_tuple_96)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_96);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v));
  PyTuple_SET_ITEM(__pyx_k_tuple_96, 0, ((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__stride1));
  PyTuple_SET_ITEM(__pyx_k_tuple_96, 1, ((PyObject *)__pyx_n_s__stride1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__stride1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__stride2));
  PyTuple_SET_ITEM(__pyx_k_tuple_96, 2, ((PyObject *)__pyx_n_s__stride2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__stride2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_96, 3, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v1));
  PyTuple_SET_ITEM(__pyx_k_tuple_96, 4, ((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_96));
  __pyx_k_codeobj_97 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s___convert_stride2, 711, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_97)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":721
 * 
 * 
 * def _convert_stride1(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */
  __pyx_k_tuple_98 = PyTuple_New(4); if (unlikely(!__pyx_k_tuple_98)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_98);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v));
  PyTuple_SET_ITEM(__pyx_k_tuple_98, 0, ((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__stride1));
  PyTuple_SET_ITEM(__pyx_k_tuple_98, 1, ((PyObject *)__pyx_n_s__stride1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__stride1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__stride2));
  PyTuple_SET_ITEM(__pyx_k_tuple_98, 2, ((PyObject *)__pyx_n_s__stride2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__stride2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_98, 3, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_98));
  __pyx_k_codeobj_99 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s___convert_stride1, 721, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_99)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":730
 * 
 * 
 * def Lips2d(coords, mask):             # <<<<<<<<<<<<<<
 *     """ Estimate intrinsic volumes for 2d region in `mask` given `coords`
 * 
 */
  __pyx_k_tuple_100 = PyTuple_New(54); if (unlikely(!__pyx_k_tuple_100)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_100);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__coords));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 0, ((PyObject *)__pyx_n_s__coords));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__coords));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 1, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__value));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 2, ((PyObject *)__pyx_n_s__value));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__value));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__coords_c));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 3, ((PyObject *)__pyx_n_s__coords_c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__coords_c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask_c));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 4, ((PyObject *)__pyx_n_s__mask_c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask_c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fcoords));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 5, ((PyObject *)__pyx_n_s__fcoords));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fcoords));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__D));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 6, ((PyObject *)__pyx_n_s__D));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__D));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 7, ((PyObject *)__pyx_n_s__fmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fpmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 8, ((PyObject *)__pyx_n_s__fpmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fpmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 9, ((PyObject *)__pyx_n_s__pmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d3));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 10, ((PyObject *)__pyx_n_s__d3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d2));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 11, ((PyObject *)__pyx_n_s__d2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__cvertices));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 12, ((PyObject *)__pyx_n_s__cvertices));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__cvertices));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 13, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 14, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 15, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 16, ((PyObject *)__pyx_n_s__l));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__r));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 17, ((PyObject *)__pyx_n_s__r));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__r));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 18, ((PyObject *)__pyx_n_s__s));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__rr));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 19, ((PyObject *)__pyx_n_s__rr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__rr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 20, ((PyObject *)__pyx_n_s__ss));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mr));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 21, ((PyObject *)__pyx_n_s__mr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ms));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 22, ((PyObject *)__pyx_n_s__ms));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ms));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s0));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 23, ((PyObject *)__pyx_n_s__s0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s1));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 24, ((PyObject *)__pyx_n_s__s1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds2));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 25, ((PyObject *)__pyx_n_s__ds2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds3));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 26, ((PyObject *)__pyx_n_s__ds3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__index));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 27, ((PyObject *)__pyx_n_s__index));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__index));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 28, ((PyObject *)__pyx_n_s__m));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__npix));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 29, ((PyObject *)__pyx_n_s__npix));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__npix));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss0));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 30, ((PyObject *)__pyx_n_s__ss0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss1));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 31, ((PyObject *)__pyx_n_s__ss1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss0d));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 32, ((PyObject *)__pyx_n_s__ss0d));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss0d));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss1d));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 33, ((PyObject *)__pyx_n_s__ss1d));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss1d));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 34, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v1));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 35, ((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v2));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 36, ((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l0));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 37, ((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l1));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 38, ((PyObject *)__pyx_n_s__l1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l2));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 39, ((PyObject *)__pyx_n_s__l2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__res));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 40, ((PyObject *)__pyx_n_s__res));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__res));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask_shape));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 41, ((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__strides));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 42, ((PyObject *)__pyx_n_s__strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__strides));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dstrides));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 43, ((PyObject *)__pyx_n_s__dstrides));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dstrides));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__verts));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 44, ((PyObject *)__pyx_n_s__verts));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__verts));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__union));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 45, ((PyObject *)__pyx_n_s__union));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__union));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__c));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 46, ((PyObject *)__pyx_n_s__c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m3));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 47, ((PyObject *)__pyx_n_s__m3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m2));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 48, ((PyObject *)__pyx_n_s__m2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pindex));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 49, ((PyObject *)__pyx_n_s__pindex));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pindex));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w0));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 50, ((PyObject *)__pyx_n_s__w0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w1));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 51, ((PyObject *)__pyx_n_s__w1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__w2));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 52, ((PyObject *)__pyx_n_s__w2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__w2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v));
  PyTuple_SET_ITEM(__pyx_k_tuple_100, 53, ((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_100));
  __pyx_k_codeobj_101 = (PyObject*)__Pyx_PyCode_New(2, 0, 54, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_100, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s__Lips2d, 730, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_101)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":907
 * 
 * 
 * def EC2d(mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic of 2D region in `mask`
 * 
 */
  __pyx_k_tuple_102 = PyTuple_New(26); if (unlikely(!__pyx_k_tuple_102)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 907; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_102);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 0, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask_c));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 1, ((PyObject *)__pyx_n_s__mask_c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask_c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__fpmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 2, ((PyObject *)__pyx_n_s__fpmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__fpmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d2));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 3, ((PyObject *)__pyx_n_s__d2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__d3));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 4, ((PyObject *)__pyx_n_s__d3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__d3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 5, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__j));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 6, ((PyObject *)__pyx_n_s__j));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__j));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__k));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 7, ((PyObject *)__pyx_n_s__k));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__k));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 8, ((PyObject *)__pyx_n_s__l));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s0));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 9, ((PyObject *)__pyx_n_s__s0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s1));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 10, ((PyObject *)__pyx_n_s__s1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds2));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 11, ((PyObject *)__pyx_n_s__ds2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds2));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ds3));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 12, ((PyObject *)__pyx_n_s__ds3));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ds3));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__index));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 13, ((PyObject *)__pyx_n_s__index));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__index));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 14, ((PyObject *)__pyx_n_s__m));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss0));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 15, ((PyObject *)__pyx_n_s__ss0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss1));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 16, ((PyObject *)__pyx_n_s__ss1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v0));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 17, ((PyObject *)__pyx_n_s__v0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v1));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 18, ((PyObject *)__pyx_n_s__v1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l0));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 19, ((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask_shape));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 20, ((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask_shape));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__pmask));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 21, ((PyObject *)__pyx_n_s__pmask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__pmask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__strides));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 22, ((PyObject *)__pyx_n_s__strides));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__strides));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__union));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 23, ((PyObject *)__pyx_n_s__union));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__union));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__c));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 24, ((PyObject *)__pyx_n_s__c));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__c));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__v2));
  PyTuple_SET_ITEM(__pyx_k_tuple_102, 25, ((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__v2));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_102));
  __pyx_k_codeobj_103 = (PyObject*)__Pyx_PyCode_New(1, 0, 26, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_102, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s__EC2d, 907, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_103)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 907; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":1007
 * 
 * 
 * def Lips1d(np.ndarray[np.float_t, ndim=2] coords,             # <<<<<<<<<<<<<<
 *            np.ndarray[np.intp_t, ndim=1] mask):
 *     """ Estimate intrinsic volumes for 1D region in `mask` given `coords`
 */
  __pyx_k_tuple_104 = PyTuple_New(17); if (unlikely(!__pyx_k_tuple_104)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_104);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__coords));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 0, ((PyObject *)__pyx_n_s__coords));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__coords));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 1, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 2, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 3, ((PyObject *)__pyx_n_s__l));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__r));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 4, ((PyObject *)__pyx_n_s__r));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__r));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 5, ((PyObject *)__pyx_n_s__s));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__rr));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 6, ((PyObject *)__pyx_n_s__rr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__rr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ss));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 7, ((PyObject *)__pyx_n_s__ss));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ss));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mr));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 8, ((PyObject *)__pyx_n_s__mr));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mr));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__ms));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 9, ((PyObject *)__pyx_n_s__ms));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__ms));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s0));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 10, ((PyObject *)__pyx_n_s__s0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__index));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 11, ((PyObject *)__pyx_n_s__index));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__index));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 12, ((PyObject *)__pyx_n_s__m));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l0));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 13, ((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l1));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 14, ((PyObject *)__pyx_n_s__l1));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l1));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__res));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 15, ((PyObject *)__pyx_n_s__res));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__res));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__D));
  PyTuple_SET_ITEM(__pyx_k_tuple_104, 16, ((PyObject *)__pyx_n_s__D));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__D));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_104));
  __pyx_k_codeobj_105 = (PyObject*)__Pyx_PyCode_New(2, 0, 17, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_104, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s__Lips1d, 1007, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_105)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "nipy/algorithms/statistics/intvol.pyx":1083
 * 
 * 
 * def EC1d(np.ndarray[np.intp_t, ndim=1] mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic for 1d `mask`
 * 
 */
  __pyx_k_tuple_106 = PyTuple_New(5); if (unlikely(!__pyx_k_tuple_106)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_k_tuple_106);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__mask));
  PyTuple_SET_ITEM(__pyx_k_tuple_106, 0, ((PyObject *)__pyx_n_s__mask));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__mask));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__i));
  PyTuple_SET_ITEM(__pyx_k_tuple_106, 1, ((PyObject *)__pyx_n_s__i));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__i));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__m));
  PyTuple_SET_ITEM(__pyx_k_tuple_106, 2, ((PyObject *)__pyx_n_s__m));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__m));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__s0));
  PyTuple_SET_ITEM(__pyx_k_tuple_106, 3, ((PyObject *)__pyx_n_s__s0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__s0));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__l0));
  PyTuple_SET_ITEM(__pyx_k_tuple_106, 4, ((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__l0));
  __Pyx_GIVEREF(((PyObject *)__pyx_k_tuple_106));
  __pyx_k_codeobj_107 = (PyObject*)__Pyx_PyCode_New(1, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_k_tuple_106, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_90, __pyx_n_s__EC1d, 1083, __pyx_empty_bytes); if (unlikely(!__pyx_k_codeobj_107)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initintvol(void); /*proto*/
PyMODINIT_FUNC initintvol(void)
#else
PyMODINIT_FUNC PyInit_intvol(void); /*proto*/
PyMODINIT_FUNC PyInit_intvol(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_intvol(void)", 0);
  if ( __Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__Pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4(__Pyx_NAMESTR("intvol"), __pyx_methods, __Pyx_DOCSTR(__pyx_k_85), 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "nipy.algorithms.statistics.intvol")) {
      if (unlikely(PyDict_SetItemString(modules, "nipy.algorithms.statistics.intvol", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  __pyx_b = PyImport_AddModule(__Pyx_NAMESTR(__Pyx_BUILTIN_MODULE_NAME)); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (__Pyx_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (unlikely(__Pyx_InitGlobals() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_module_is_main_nipy__algorithms__statistics__intvol) {
    if (__Pyx_SetAttrString(__pyx_m, "__name__", __pyx_n_s____main__) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  }
  /*--- Builtin init code ---*/
  if (unlikely(__Pyx_InitCachedBuiltins() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (unlikely(__Pyx_InitCachedConstants() < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) {__pyx_filename = __pyx_f[2]; __pyx_lineno = 9; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 165; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 169; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 178; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 861; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/

  /* "nipy/algorithms/statistics/intvol.pyx":12
 * cimport cython
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * cimport numpy as np
 * 
 */
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s__numpy), 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 12; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__np, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 12; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":15
 * cimport numpy as np
 * 
 * from scipy.sparse import dok_matrix             # <<<<<<<<<<<<<<
 * 
 * # Array helper
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__dok_matrix));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_n_s__dok_matrix));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__dok_matrix));
  __pyx_t_2 = __Pyx_Import(((PyObject *)__pyx_n_s_86), ((PyObject *)__pyx_t_1), -1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__dok_matrix);
  if (__pyx_t_1 == NULL) {
    if (PyErr_ExceptionMatches(PyExc_AttributeError)) __Pyx_RaiseImportError(__pyx_n_s__dok_matrix);
    if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__dok_matrix, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 15; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":18
 * 
 * # Array helper
 * from nipy.utils.arrays import strides_from             # <<<<<<<<<<<<<<
 * 
 * # local imports
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_n_s__strides_from));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_n_s__strides_from));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__strides_from));
  __pyx_t_1 = __Pyx_Import(((PyObject *)__pyx_n_s_87), ((PyObject *)__pyx_t_2), -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(((PyObject *)__pyx_t_2)); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_GetAttr(__pyx_t_1, __pyx_n_s__strides_from);
  if (__pyx_t_2 == NULL) {
    if (PyErr_ExceptionMatches(PyExc_AttributeError)) __Pyx_RaiseImportError(__pyx_n_s__strides_from);
    if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_2);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__strides_from, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 18; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":21
 * 
 * # local imports
 * from utils import cube_with_strides_center, join_complexes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 21; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_n_s_8));
  PyList_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_n_s_8));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s_8));
  __Pyx_INCREF(((PyObject *)__pyx_n_s__join_complexes));
  PyList_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_n_s__join_complexes));
  __Pyx_GIVEREF(((PyObject *)__pyx_n_s__join_complexes));
  __pyx_t_2 = __Pyx_Import(((PyObject *)__pyx_n_s__utils), ((PyObject *)__pyx_t_1), -1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 21; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s_8);
  if (__pyx_t_1 == NULL) {
    if (PyErr_ExceptionMatches(PyExc_AttributeError)) __Pyx_RaiseImportError(__pyx_n_s_8);
    if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 21; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_8, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 21; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__join_complexes);
  if (__pyx_t_1 == NULL) {
    if (PyErr_ExceptionMatches(PyExc_AttributeError)) __Pyx_RaiseImportError(__pyx_n_s__join_complexes);
    if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 21; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__join_complexes, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 21; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":24
 * 
 * 
 * cdef double PI = np.pi             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_GetName(__pyx_m, __pyx_n_s__np); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 24; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_GetAttr(__pyx_t_2, __pyx_n_s__pi); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 24; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 24; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_4nipy_10algorithms_10statistics_6intvol_PI = __pyx_t_3;

  /* "nipy/algorithms/statistics/intvol.pyx":340
 * 
 * 
 * def EC3d(mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic of region within `mask`
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_15EC3d, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 340; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__EC3d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 340; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":464
 * 
 * 
 * def Lips3d(coords, mask):             # <<<<<<<<<<<<<<
 *     """ Estimated intrinsic volumes within masked region given coordinates
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_17Lips3d, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__Lips3d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 464; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":699
 * 
 * 
 * def _convert_stride3(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_19_convert_stride3, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s___convert_stride3, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 699; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":711
 * 
 * 
 * def _convert_stride2(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_21_convert_stride2, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s___convert_stride2, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 711; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":721
 * 
 * 
 * def _convert_stride1(v, stride1, stride2):             # <<<<<<<<<<<<<<
 *     """
 *     Take a voxel, expressed as in index in stride1 and
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_23_convert_stride1, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s___convert_stride1, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 721; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":730
 * 
 * 
 * def Lips2d(coords, mask):             # <<<<<<<<<<<<<<
 *     """ Estimate intrinsic volumes for 2d region in `mask` given `coords`
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_25Lips2d, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__Lips2d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 730; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":907
 * 
 * 
 * def EC2d(mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic of 2D region in `mask`
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_27EC2d, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 907; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__EC2d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 907; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":1007
 * 
 * 
 * def Lips1d(np.ndarray[np.float_t, ndim=2] coords,             # <<<<<<<<<<<<<<
 *            np.ndarray[np.intp_t, ndim=1] mask):
 *     """ Estimate intrinsic volumes for 1D region in `mask` given `coords`
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_29Lips1d, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__Lips1d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1007; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":1083
 * 
 * 
 * def EC1d(np.ndarray[np.intp_t, ndim=1] mask):             # <<<<<<<<<<<<<<
 *     """ Compute Euler characteristic for 1d `mask`
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4nipy_10algorithms_10statistics_6intvol_31EC1d, NULL, __pyx_n_s_91); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s__EC1d, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1083; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nipy/algorithms/statistics/intvol.pyx":1
 * """             # <<<<<<<<<<<<<<
 * The estimators for the intrinsic volumes appearing in this module
 * were partially supported by NSF grant DMS-0405970.
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s____test__, ((PyObject *)__pyx_t_1)) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;

  /* "numpy.pxd":975
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    __Pyx_AddTraceback("init nipy.algorithms.statistics.intvol", __pyx_clineno, __pyx_lineno, __pyx_filename);
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init nipy.algorithms.statistics.intvol");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* Runtime support code */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif /* CYTHON_REFNANNY */

static PyObject *__Pyx_GetName(PyObject *dict, PyObject *name) {
    PyObject *result;
    result = PyObject_GetAttr(dict, name);
    if (!result) {
        if (dict != __pyx_b) {
            PyErr_Clear();
            result = PyObject_GetAttr(__pyx_b, name);
        }
        if (!result) {
            PyErr_SetObject(PyExc_NameError, name);
        }
    }
    return result;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%s() takes %s %" CYTHON_FORMAT_SSIZE_T "d positional argument%s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%s() got an unexpected keyword argument '%s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    #if PY_VERSION_HEX < 0x02050000
    if (PyClass_Check(type)) {
    #else
    if (PyType_Check(type)) {
    #endif
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        #if PY_VERSION_HEX < 0x02050000
            if (PyInstance_Check(type)) {
                type = (PyObject*) ((PyInstanceObject*)type)->in_class;
                Py_INCREF(type);
            }
            else {
                type = 0;
                PyErr_SetString(PyExc_TypeError,
                    "raise: exception must be an old-style class or instance");
                goto raise_error;
            }
        #else
            type = (PyObject*) Py_TYPE(type);
            Py_INCREF(type);
            if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
                PyErr_SetString(PyExc_TypeError,
                    "raise: exception class must be a subclass of BaseException");
                goto raise_error;
            }
        #endif
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else /* Python 3+ */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *args;
        if (!value)
            args = PyTuple_New(0);
        else if (PyTuple_Check(value)) {
            Py_INCREF(value);
            args = value;
        }
        else
            args = PyTuple_Pack(1, value);
        if (!args)
            goto bad;
        owned_instance = PyEval_CallObject(type, args);
        Py_DECREF(args);
        if (!owned_instance)
            goto bad;
        value = owned_instance;
        if (!PyExceptionInstance_Check(value)) {
            PyErr_Format(PyExc_TypeError,
                         "calling %R should have returned an instance of "
                         "BaseException, not %R",
                         type, Py_TYPE(value));
            goto bad;
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause && cause != Py_None) {
        PyObject *fixed_cause;
        if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        }
        else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        }
        else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static CYTHON_INLINE int __Pyx_IsLittleEndian(void) {
  unsigned int n = 1;
  return *(unsigned char*)(&n) != 0;
}
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t < '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1) /* First char was not a digit */
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count; /* Consume from buffer string */
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break; /* breaks both loops as ctx->enc_count == 0 */
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue; /* empty struct */
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static CYTHON_INLINE PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        if (isspace(*ts))
            continue;
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
                return ts;
      case ' ':
      case 10:
      case 13:
        ++ts;
        break;
      case '<':
        if (!__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T': /* substruct */
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0; /* Erase processed last struct element */
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}': /* end of substruct; either repeat or move on */
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0; /* Erase processed last struct element */
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }        /* fall through */
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 's': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
        } else {
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_count = ctx->new_count;
          ctx->enc_packmode = ctx->new_packmode;
          ctx->enc_type = *ts;
          ctx->is_complex = got_Z;
        }
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}
static CYTHON_INLINE void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static CYTHON_INLINE int __Pyx_GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  if (obj == Py_None || obj == NULL) {
    __Pyx_ZeroBuffer(buf);
    return 0;
  }
  buf->buf = NULL;
  if (__Pyx_GetBuffer(obj, buf, flags) == -1) goto fail;
  if (buf->ndim != nd) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if ((unsigned)buf->itemsize != dtype->size) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_ZeroBuffer(buf);
  return -1;
}
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (info->buf == NULL) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}

static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_Format(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

static CYTHON_INLINE int __Pyx_mod_int(int a, int b) {
    int r = a % b;
    r += ((r != 0) & ((r ^ b) < 0)) * b;
    return r;
}

static int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (!type) {
        PyErr_Format(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (Py_TYPE(obj) == type) return 1;
    }
    else {
        if (PyObject_TypeCheck(obj, type)) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%s' has incorrect type (expected %s, got %s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

static CYTHON_INLINE long __Pyx_mod_long(long a, long b) {
    long r = a % b;
    r += ((r != 0) & ((r ^ b) < 0)) * b;
    return r;
}

static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%s to unpack",
                 index, (index == 1) ? "" : "s");
}

static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    CYTHON_UNUSED PyObject *getbuffer_cobj;
  #if PY_VERSION_HEX >= 0x02060000
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
  #endif
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
  #if PY_VERSION_HEX < 0x02060000
    if (obj->ob_type->tp_dict &&
        (getbuffer_cobj = PyMapping_GetItemString(obj->ob_type->tp_dict,
                                             "__pyx_getbuffer"))) {
        getbufferproc func;
      #if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION == 0)
        func = (getbufferproc) PyCapsule_GetPointer(getbuffer_cobj, "getbuffer(obj, view, flags)");
      #else
        func = (getbufferproc) PyCObject_AsVoidPtr(getbuffer_cobj);
      #endif
        Py_DECREF(getbuffer_cobj);
        if (!func)
            goto fail;
        return func(obj, view, flags);
    } else {
        PyErr_Clear();
    }
  #endif
    PyErr_Format(PyExc_TypeError, "'%100s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
#if PY_VERSION_HEX < 0x02060000
fail:
#endif
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    CYTHON_UNUSED PyObject *releasebuffer_cobj;
    if (!obj) return;
  #if PY_VERSION_HEX >= 0x02060000
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
  #endif
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) { __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view); return; }
  #if PY_VERSION_HEX < 0x02060000
    if (obj->ob_type->tp_dict &&
        (releasebuffer_cobj = PyMapping_GetItemString(obj->ob_type->tp_dict,
                                                      "__pyx_releasebuffer"))) {
        releasebufferproc func;
      #if PY_VERSION_HEX >= 0x02070000 && !(PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION == 0)
        func = (releasebufferproc) PyCapsule_GetPointer(releasebuffer_cobj, "releasebuffer(obj, view)");
      #else
        func = (releasebufferproc) PyCObject_AsVoidPtr(releasebuffer_cobj);
      #endif
        Py_DECREF(releasebuffer_cobj);
        if (!func)
            goto fail;
        func(obj, view);
        return;
    } else {
        PyErr_Clear();
    }
  #endif
    goto nofail;
#if PY_VERSION_HEX < 0x02060000
fail:
#endif
    PyErr_WriteUnraisable(obj);
nofail:
    Py_DECREF(obj);
    view->obj = NULL;
}
#endif /*  PY_MAJOR_VERSION < 3 */


    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, long level) {
    PyObject *py_import = 0;
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    py_import = __Pyx_GetAttrString(__pyx_b, "__import__");
    if (!py_import)
        goto bad;
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    #if PY_VERSION_HEX >= 0x02050000
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                /* try package relative import first */
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0; /* try absolute import on failure */
        }
        #endif
        if (!module) {
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
        }
    }
    #else
    if (level>0) {
        PyErr_SetString(PyExc_RuntimeError, "Relative import is not supported for Python <=2.4.");
        goto bad;
    }
    module = PyObject_CallFunctionObjArgs(py_import,
        name, global_dict, empty_dict, list, NULL);
    #endif
bad:
    Py_XDECREF(empty_list);
    Py_XDECREF(py_import);
    Py_XDECREF(empty_dict);
    return module;
}

static CYTHON_INLINE void __Pyx_RaiseImportError(PyObject *name) {
#if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_ImportError, "cannot import name %.230s",
                 PyString_AsString(name));
#else
    PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
#endif
}

static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_from_py_Py_intptr_t(PyObject* x) {
    const Py_intptr_t neg_one = (Py_intptr_t)-1, const_zero = (Py_intptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if (sizeof(Py_intptr_t) == sizeof(char)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedChar(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedChar(x);
    } else if (sizeof(Py_intptr_t) == sizeof(short)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedShort(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedShort(x);
    } else if (sizeof(Py_intptr_t) == sizeof(int)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedInt(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedInt(x);
    } else if (sizeof(Py_intptr_t) == sizeof(long)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedLong(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedLong(x);
    } else if (sizeof(Py_intptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return (Py_intptr_t)__Pyx_PyInt_AsUnsignedLongLong(x);
        else
            return (Py_intptr_t)__Pyx_PyInt_AsSignedLongLong(x);
    }  else {
        #if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
        #else
        Py_intptr_t val;
        PyObject *v = __Pyx_PyNumber_Int(x);
        #if PY_VERSION_HEX < 0x03000000
        if (likely(v) && !PyLong_Check(v)) {
            PyObject *tmp = v;
            v = PyNumber_Long(tmp);
            Py_DECREF(tmp);
        }
        #endif
        if (likely(v)) {
            int one = 1; int is_little = (int)*(unsigned char *)&one;
            unsigned char *bytes = (unsigned char *)&val;
            int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                          bytes, sizeof(val),
                                          is_little, !is_unsigned);
            Py_DECREF(v);
            if (likely(!ret))
                return val;
        }
        #endif
        return (Py_intptr_t)-1;
    }
}

static CYTHON_INLINE PyObject *__Pyx_PyInt_to_py_Py_intptr_t(Py_intptr_t val) {
    const Py_intptr_t neg_one = (Py_intptr_t)-1, const_zero = (Py_intptr_t)0;
    const int is_unsigned = const_zero < neg_one;
    if ((sizeof(Py_intptr_t) == sizeof(char))  ||
        (sizeof(Py_intptr_t) == sizeof(short))) {
        return PyInt_FromLong((long)val);
    } else if ((sizeof(Py_intptr_t) == sizeof(int)) ||
               (sizeof(Py_intptr_t) == sizeof(long))) {
        if (is_unsigned)
            return PyLong_FromUnsignedLong((unsigned long)val);
        else
            return PyInt_FromLong((long)val);
    } else if (sizeof(Py_intptr_t) == sizeof(PY_LONG_LONG)) {
        if (is_unsigned)
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG)val);
        else
            return PyLong_FromLongLong((PY_LONG_LONG)val);
    } else {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eqf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sumf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_difff(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prodf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quotf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        float denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_negf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zerof(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conjf(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_absf(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_powf(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(a, a);
                    case 3:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, a);
                    case 4:
                        z = __Pyx_c_prodf(a, a);
                        return __Pyx_c_prodf(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_absf(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        double denom = b.real * b.real + b.imag * b.imag;
        z.real = (a.real * b.real + a.imag * b.imag) / denom;
        z.imag = (a.imag * b.real - a.real * b.imag) / denom;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(a, a);
                    case 3:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, a);
                    case 4:
                        z = __Pyx_c_prod(a, a);
                        return __Pyx_c_prod(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                }
                r = a.real;
                theta = 0;
            } else {
                r = __Pyx_c_abs(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

static CYTHON_INLINE unsigned char __Pyx_PyInt_AsUnsignedChar(PyObject* x) {
    const unsigned char neg_one = (unsigned char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned char" :
                    "value too large to convert to unsigned char");
            }
            return (unsigned char)-1;
        }
        return (unsigned char)val;
    }
    return (unsigned char)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned short __Pyx_PyInt_AsUnsignedShort(PyObject* x) {
    const unsigned short neg_one = (unsigned short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned short" :
                    "value too large to convert to unsigned short");
            }
            return (unsigned short)-1;
        }
        return (unsigned short)val;
    }
    return (unsigned short)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE unsigned int __Pyx_PyInt_AsUnsignedInt(PyObject* x) {
    const unsigned int neg_one = (unsigned int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(unsigned int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(unsigned int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to unsigned int" :
                    "value too large to convert to unsigned int");
            }
            return (unsigned int)-1;
        }
        return (unsigned int)val;
    }
    return (unsigned int)__Pyx_PyInt_AsUnsignedLong(x);
}

static CYTHON_INLINE char __Pyx_PyInt_AsChar(PyObject* x) {
    const char neg_one = (char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to char" :
                    "value too large to convert to char");
            }
            return (char)-1;
        }
        return (char)val;
    }
    return (char)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE short __Pyx_PyInt_AsShort(PyObject* x) {
    const short neg_one = (short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to short" :
                    "value too large to convert to short");
            }
            return (short)-1;
        }
        return (short)val;
    }
    return (short)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsInt(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE signed char __Pyx_PyInt_AsSignedChar(PyObject* x) {
    const signed char neg_one = (signed char)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed char) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed char)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed char" :
                    "value too large to convert to signed char");
            }
            return (signed char)-1;
        }
        return (signed char)val;
    }
    return (signed char)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed short __Pyx_PyInt_AsSignedShort(PyObject* x) {
    const signed short neg_one = (signed short)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed short) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed short)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed short" :
                    "value too large to convert to signed short");
            }
            return (signed short)-1;
        }
        return (signed short)val;
    }
    return (signed short)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE signed int __Pyx_PyInt_AsSignedInt(PyObject* x) {
    const signed int neg_one = (signed int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(signed int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(signed int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to signed int" :
                    "value too large to convert to signed int");
            }
            return (signed int)-1;
        }
        return (signed int)val;
    }
    return (signed int)__Pyx_PyInt_AsSignedLong(x);
}

static CYTHON_INLINE int __Pyx_PyInt_AsLongDouble(PyObject* x) {
    const int neg_one = (int)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(int) < sizeof(long)) {
        long val = __Pyx_PyInt_AsLong(x);
        if (unlikely(val != (long)(int)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to int" :
                    "value too large to convert to int");
            }
            return (int)-1;
        }
        return (int)val;
    }
    return (int)__Pyx_PyInt_AsLong(x);
}

static CYTHON_INLINE unsigned long __Pyx_PyInt_AsUnsignedLong(PyObject* x) {
    const unsigned long neg_one = (unsigned long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned long");
            return (unsigned long)-1;
        }
        return (unsigned long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned long");
                return (unsigned long)-1;
            }
            return (unsigned long)PyLong_AsUnsignedLong(x);
        } else {
            return (unsigned long)PyLong_AsLong(x);
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned long)-1;
        val = __Pyx_PyInt_AsUnsignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to unsigned PY_LONG_LONG");
            return (unsigned PY_LONG_LONG)-1;
        }
        return (unsigned PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to unsigned PY_LONG_LONG");
                return (unsigned PY_LONG_LONG)-1;
            }
            return (unsigned PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (unsigned PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsUnsignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE long __Pyx_PyInt_AsLong(PyObject* x) {
    const long neg_one = (long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to long");
            return (long)-1;
        }
        return (long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to long");
                return (long)-1;
            }
            return (long)PyLong_AsUnsignedLong(x);
        } else {
            return (long)PyLong_AsLong(x);
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long)-1;
        val = __Pyx_PyInt_AsLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_AsLongLong(PyObject* x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to PY_LONG_LONG");
            return (PY_LONG_LONG)-1;
        }
        return (PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to PY_LONG_LONG");
                return (PY_LONG_LONG)-1;
            }
            return (PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed long __Pyx_PyInt_AsSignedLong(PyObject* x) {
    const signed long neg_one = (signed long)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed long");
            return (signed long)-1;
        }
        return (signed long)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed long");
                return (signed long)-1;
            }
            return (signed long)PyLong_AsUnsignedLong(x);
        } else {
            return (signed long)PyLong_AsLong(x);
        }
    } else {
        signed long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed long)-1;
        val = __Pyx_PyInt_AsSignedLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static CYTHON_INLINE signed PY_LONG_LONG __Pyx_PyInt_AsSignedLongLong(PyObject* x) {
    const signed PY_LONG_LONG neg_one = (signed PY_LONG_LONG)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to signed PY_LONG_LONG");
            return (signed PY_LONG_LONG)-1;
        }
        return (signed PY_LONG_LONG)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to signed PY_LONG_LONG");
                return (signed PY_LONG_LONG)-1;
            }
            return (signed PY_LONG_LONG)PyLong_AsUnsignedLongLong(x);
        } else {
            return (signed PY_LONG_LONG)PyLong_AsLongLong(x);
        }
    } else {
        signed PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (signed PY_LONG_LONG)-1;
        val = __Pyx_PyInt_AsSignedLongLong(tmp);
        Py_DECREF(tmp);
        return val;
    }
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        #if PY_VERSION_HEX < 0x02050000
        return PyErr_Warn(NULL, message);
        #else
        return PyErr_WarnEx(NULL, message, 1);
        #endif
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%s.%s is not a type object",
            module_name, class_name);
        goto bad;
    }
    if (!strict && (size_t)((PyTypeObject *)result)->tp_basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        #if PY_VERSION_HEX < 0x02050000
        if (PyErr_Warn(NULL, warning) < 0) goto bad;
        #else
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
        #endif
    }
    else if ((size_t)((PyTypeObject *)result)->tp_basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%s.%s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = (start + end) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,            /*int argcount,*/
        0,            /*int kwonlyargcount,*/
        0,            /*int nlocals,*/
        0,            /*int stacksize,*/
        0,            /*int flags,*/
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,      /*int firstlineno,*/
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_globals = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_globals = PyModule_GetDict(__pyx_m);
    if (!py_globals) goto bad;
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        py_globals,          /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else  /* Python 3+ has unicode identifiers */
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}


/* Type Conversion Functions */

static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}

static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_VERSION_HEX < 0x03000000
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return Py_INCREF(x), x;
  m = Py_TYPE(x)->tp_as_number;
#if PY_VERSION_HEX < 0x03000000
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_VERSION_HEX < 0x03000000
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%s__ returned non-%s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}

static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject* x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}

static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
#if PY_VERSION_HEX < 0x02050000
   if (ival <= LONG_MAX)
       return PyInt_FromLong((long)ival);
   else {
       unsigned char *bytes = (unsigned char *) &ival;
       int one = 1; int little = (int)*(unsigned char*)&one;
       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
   }
#else
   return PyInt_FromSize_t(ival);
#endif
}

static CYTHON_INLINE size_t __Pyx_PyInt_AsSize_t(PyObject* x) {
   unsigned PY_LONG_LONG val = __Pyx_PyInt_AsUnsignedLongLong(x);
   if (unlikely(val == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred())) {
       return (size_t)-1;
   } else if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
       PyErr_SetString(PyExc_OverflowError,
                       "value too large to convert to size_t");
       return (size_t)-1;
   }
   return (size_t)val;
}


#endif /* Py_PYTHON_H */
